#include "alignCurvesDialog.h"
#include "parser.h"
#include "graph.h"
#include "application.h"
#include "nrutil.h"

#include <qapplication.h>
#include <qvariant.h>
#include <qpushbutton.h>
#include <qcheckbox.h>
#include <qlabel.h>
#include <qcombobox.h>
#include <qlayout.h>
#include <qbuttongroup.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qmessagebox.h>
#include <qhbuttongroup.h>

alignCurvesDialog::alignCurvesDialog(QWidget * parent, const char *name, 
				     bool modal, WFlags fl):
  QDialog (parent, name, modal, fl)
{
  if (!name)
    setName ("alignCurvesDialog");
  setCaption (tr ("QtiPlot - Align Curves"));
  setSizeGripEnabled (true);
  setSizePolicy (QSizePolicy (QSizePolicy::Expanding, QSizePolicy::Fixed));

  GroupBox1 =
    new QButtonGroup (2, QGroupBox::Horizontal, tr (""), this, "GroupBox1");

  new QLabel (tr ("Method"), GroupBox1, "TextLabel2", 0);
  boxMethod = new QComboBox (GroupBox1, "boxMethod");
  QStringList methods;
  methods << "Maximum" << "Minimum" << "Rising Edge" << "Falling Edge";
  boxMethod->insertStringList(methods,-1);

  // for Edge only
  new QLabel( tr("Edge Level"), GroupBox1 );
  boxLevel = new QLineEdit(GroupBox1,"boxLevel");
  boxLevel->setText("0");
  boxLevel->setEnabled(false);

  boxSmooth = new QCheckBox( tr("Smoothing Points"), GroupBox1 );
  boxSmooth->setChecked(false);
  boxPoints = new QSpinBox(3,101,2,GroupBox1,"boxPoints");
  boxPoints->setEnabled(false);

  new QLabel( tr("Move to Origin"), GroupBox1 );
  boxOrigin = new QCheckBox(GroupBox1);
  boxOrigin->setChecked(false);

  GroupBox2 = new QHButtonGroup (this, "GroupBox2");
  GroupBox2->setFlat (TRUE);
  GroupBox2->setLineWidth (0);

  buttonOk = new QPushButton (GroupBox2, "buttonOk");
  buttonOk->setText (tr ("Align"));
  buttonOk->setAutoDefault (TRUE);
  buttonOk->setDefault (TRUE);

  buttonDone = new QPushButton (GroupBox2, "buttonDone");
  buttonDone->setText(tr("Close"));
  buttonDone->setAutoDefault (TRUE);

  QVBoxLayout *vlayout = new QVBoxLayout (this, 5, 5, "vlayout");
  vlayout->addWidget (GroupBox1);
  vlayout->addWidget (GroupBox2);

  // signals and slots connections
  connect (buttonOk, SIGNAL (clicked ()), this, SLOT (accept ()));
  connect (buttonDone, SIGNAL (clicked ()), this, SLOT (close ()));
  connect (boxMethod, SIGNAL(activated(int)), this, SLOT(methodChanged(int)));
  connect (boxSmooth, SIGNAL(toggled(bool)), this, SLOT(smoothChanged(bool)));
  
  firsttime = true;
}

alignCurvesDialog::~alignCurvesDialog ()
{
}

void
alignCurvesDialog::setGraph (Graph * g, const QString &name)
{
  graph = g;
  graphname = name;
  connect (graph, SIGNAL (closedGraph ()), this, SLOT (close ()));
}

void
alignCurvesDialog::methodChanged(int item)
{
  if(item <= 1)
    boxLevel->setEnabled(false);
  else
    boxLevel->setEnabled(true);
}

void
alignCurvesDialog::smoothChanged(bool on)
{
  boxPoints->setEnabled(on);
}

void
alignCurvesDialog::accept ()
{
  ApplicationWindow *app = (ApplicationWindow *)this->parent();
  int prec = app->tbl_output_precision;

  if(firsttime)
    {
      int maxrows = 0;
      for (int i = 0; i < graph->curves(); i++)
        {
          QwtPlotCurve *c = graph->curve(i);
          if (c && c->rtti () != QwtPlotItem::Rtti_PlotSpectrogram
              && graph->curveType(i) != Graph::XErrorBars
              && graph->curveType(i) != Graph::YErrorBars)
            {
              curves << i;
              int ds=c->dataSize();
              if (maxrows < ds) maxrows = ds;
            }
        }
      QString tname = app->generateUniqueName("Aligned");
      QString legend = tr("Aligned data of ") + graphname;
      alignedTable = app->newHiddenTable( tname, legend, maxrows,
                                          curves.count()*2, prec);
      for(int i=0;i<(int)curves.count();i++)
        {
          alignedTable->setColPlotDesignation(i*2,  Table::X);
          alignedTable->setColPlotDesignation(i*2+1,Table::Y);
        }
      alignedTable->setHeaderColType();
    }

  QApplication::setOverrideCursor (waitCursor);

  int nc = (int)curves.count();
  double xorigin = 0;
  for (int i = 0; i < nc; i++)
    {
      QwtPlotCurve *c = graph->curve(curves[i]);
      int n = c->dataSize();
      QMemArray <double> s(n);
      if(boxSmooth->isChecked())
	{
	  QMemArray <double> y(n);
	  int np = boxPoints->value();
	  for (int j = 0; j<n; j++) y[j] = c->y(j);
	  smooth_vect( y.data(), n, np/2, np/2, 2, s.data());
	}
      else
	{
	  for(int j=0; j<n; j++)
	    s[j] = c->y(j);
	}

      int j,dj;
      double v,xp;
      double level = boxLevel->text().toDouble();
      bool inv = ( c->x(0) > c->x(n-1) );
      if (inv)
	{
	  j = n - 1;
	  dj = -1;
	}
      else
	{
	  j = 0;
	  dj = 1;
	}
      v = s[j];
      xp = c->x(j);
      while (j+dj>=0 && j+dj<=n-1)
	{
	  j += dj;
	  switch (boxMethod->currentItem()) 
	    {
	    case 0 : //Maximum 
	      if ( v < s[j] )
		{
		  v = s[j];
		  xp = c->x(j);
		}
	      break;
	    case 1 : //Minimum
	      if ( v > s[j] )
		{
		  v = s[j];
		  xp = c->x(j);
		}
	      break;
	    case 2 : //Rising Edge
	      if (s[j-dj] <= level && s[j] > level )
		{
		  xp = c->x(j-dj) + (c->x(j) - c->x(j-dj)) * (level - s[j-dj]) 
		    / (s[j] - s[j-dj]);
		  goto loopend;
		}
	      break;
	    case 3 : //Falling Edge
	      if (s[j-dj] >= level && s[j] < level )
		{
		  xp = c->x(j-dj) + (c->x(j) - c->x(j-dj)) * (level - s[j-dj]) 
		    / (s[j] - s[j-dj]);
		  goto loopend;
		}
	      break;
	    }
	}
    loopend:
      double xoff = 0;
      if (i==0 && !boxOrigin->isChecked())
	xorigin = xp;
      else
	xoff = xorigin - xp;

      for (int r = 0; r < c->dataSize(); r++)
        {
          double xv = c->x(r) + xoff;
          double yv = c->y(r);
          alignedTable->setText( r, i*2, QString::number(xv,'g',prec) );
          alignedTable->setText( r, i*2+1, QString::number(yv,'g',prec) );
        }
    }
  
  if(firsttime)
    {
      alignedPlot = app->multilayerPlot(alignedTable,
					alignedTable->YColumns(),
					Graph::Line );
      Graph *ag = app->activeGraph;
      for(int i = 0; i < nc; i++)
        {
          QwtPlotCurve *c = graph->curve(curves[i]);
          QwtPlotCurve *c2 = ag->curve(i);
          c2->setPen(c->pen());
          c2->setSymbol(c->symbol());
          int style = graph->curveType(curves[i]);
          if(style <= Graph::LineSymbols)
            ag->setCurveType(i,style);
          if(style==Graph::Scatter)
            ag->setCurveStyle(i,QwtPlotCurve::NoCurve);
        }
      app->activeGraph->replot();
      firsttime = false;
    }
  else
    alignedTable->notifyChanges();
      
  QApplication::restoreOverrideCursor ();
}

