/***************************************************************************
*                                                                         *
*  This program is free software; you can redistribute it and/or modify   *
*  it under the terms of the GNU General Public License as published by   *
*  the Free Software Foundation; either version 2 of the License, or      *
*  (at your option) any later version.                                    *
*                                                                         *
*  This program is distributed in the hope that it will be useful,        *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
*  GNU General Public License for more details.                           *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the Free Software           *
*   Foundation, Inc., 51 Franklin Street, Fifth Floor,                    *
*   Boston, MA  02110-1301  USA                                           *
*                                                                         *
***************************************************************************/

#include "graph.h"
#include "plot.h"
#include "colorBox.h"
#include "ImageMarker.h"
#include "Histogram.h"
#include "nrutil.h"
#include "application.h"
#include "LineMarker.h"
#include "FunctionCurve.h"

#include <qapplication.h>
#include <qdatetime.h>
#include <qimage.h>
#include <qmessagebox.h>

#define GSL_DISABLE_DEPRECATED

#include <gsl/gsl_sort.h>
#include <gsl/gsl_interp.h>
#include <gsl/gsl_fft_halfcomplex.h>
#include <gsl/gsl_fft_complex.h>
#include <gsl/gsl_histogram.h>
#include <gsl/gsl_multifit_nlin.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_fit.h>

#define SQR(x) (x)*(x)

void
Graph::updateHistogram (Table * w, const QString & curveName, int index,
                        bool automatic, double binSize, double begin,
                        double end)
{
  int i;
  QwtHistogram *cv = (QwtHistogram *) curve (index);
  if (!cv)
    return;
  if (cv->autoBinning () && automatic)
    return;

  int ycol = w->colIndex (curveName);
  int r = w->tableRows ();

  QMemArray < double > Y(r);
  int it = 0;
  for (i = 0; i < r; i++)
  {
    QString yval = w->text (i, ycol);
    if (!yval.isEmpty ())
    {
      it++;
      Y[it - 1] = yval.toDouble ();
    }
  }

  if (it < 2 || (it == 2 && Y[0] == Y[1]))
  {				//non valid histogram
    QMemArray < double > X (2);
    Y.resize (2);
    for (i = 0; i < 2; i++)
    {
      Y[i] = 0;
      X[i] = 0;
    }
    cv->setData (X, Y, 2);
    d_plot->replot ();
    updateScale ();
    return;
  }

  int n;
  gsl_histogram *h;
  if (automatic)
  {
    n = 10;
    h = gsl_histogram_alloc (n);
    if (!h)
      return;

    gsl_vector *v;
    v = gsl_vector_alloc (it);
    for (i = 0; i < it; i++)
      gsl_vector_set (v, i, Y[i]);

    double min, max;
    gsl_vector_minmax (v, &min, &max);
    gsl_vector_free (v);

    double from = floor (min);
    double to = ceil (max);

    gsl_histogram_set_ranges_uniform (h, floor (min), ceil (max));
    cv->setBinning (true, from, to, (double) (to - from) / (double) n);
  }
  else
  {
    n = int ((end - begin) / binSize + 1);
    h = gsl_histogram_alloc (n);
    if (!h)
      return;

    //double *range;
    //range = vector (0, n + 1);
    QMemArray <double> range(n+2);

    for (i = 0; i <= n + 1; i++)
      range[i] = begin + i * binSize;

    gsl_histogram_set_ranges (h, range.data(), n + 1);
    //free_vector (range, 0, n + 1);
    cv->setBinning (false, begin, end, binSize);
  }

  for (i = 0; i < it; i++)
    gsl_histogram_increment (h, Y[i]);

  QMemArray < double > X(n);	//stores ranges (x) and bins (y)
  Y.resize (n);
  for (i = 0; i < n; i++)
  {
    Y[i] = gsl_histogram_get (h, i);
    double lower, upper;
    gsl_histogram_get_range (h, i, &lower, &upper);
    X[i] = lower;
  }
  cv->setData (X, Y, n);
  gsl_histogram_free (h);
  d_plot->replot ();
  updateScale ();
  emit modifiedGraph ();
}

void
Graph::updateHistogram (Table * w, const QString & curveName, int index)
{
  int i;
  QwtHistogram *cv = (QwtHistogram *)curve (index);
  if (!cv)
    return;

  int ycol = w->colIndex (curveName);
  int r = w->tableRows ();

  QMemArray < double >Y (r);
  int it = 0;
  for (i = 0; i < r; i++)
  {
    QString yval = w->text (i, ycol);
    if (!yval.isEmpty ())
    {
      it++;
      Y[it - 1] = yval.toDouble ();
    }
  }

  if (!it)
    removeCurve (index, true);

  if (it < 2 || (it == 2 && Y[0] == Y[1]))
  {				//non valid histogram data
    QMemArray < double >X (2);
    Y.resize (2);
    for (i = 0; i < 2; i++)
    {
      Y[i] = 0;
      X[i] = 0;
    }

    cv->setData (X, Y, 2);
    d_plot->replot ();
    updateScale ();
    return;
  }

  double binSize = cv->binSize ();
  double begin = cv->begin ();
  double end = cv->end ();
  int n;
  gsl_histogram *h;

  if (cv->autoBinning ())
  {
    n = 10;
    h = gsl_histogram_alloc (n);
    if (!h)
      return;

    gsl_vector *v;
    v = gsl_vector_alloc (it);
    for (i = 0; i < it; i++)
      gsl_vector_set (v, i, Y[i]);

    double min, max;
    gsl_vector_minmax (v, &min, &max);
    gsl_vector_free (v);

    double from = floor (min);
    double to = ceil (max);

    gsl_histogram_set_ranges_uniform (h, floor (min), ceil (max));
    cv->setBinning (true, from, to, (double) (to - from) / (double) n);
  }
  else
  {
    n = int ((end - begin) / binSize + 1);
    h = gsl_histogram_alloc (n);
    if (!h)
      return;

    //double *range;
    //range = vector (0, n + 1);
    QMemArray <double> range(n+2);

    for (i = 0; i <= n + 1; i++)
      range[i] = begin + i * binSize;

    gsl_histogram_set_ranges (h, range.data(), n + 1);
    //free_vector (range, 0, n + 1);
  }

  for (i = 0; i < it; i++)
    gsl_histogram_increment (h, Y[i]);

  QMemArray < double > X(n);	//stores ranges (x) and bins (y)
  Y.resize (n);
  for (i = 0; i < n; i++)
  {
    Y[i] = gsl_histogram_get (h, i);
    double lower, upper;
    gsl_histogram_get_range (h, i, &lower, &upper);
    X[i] = lower;
  }

  cv->setData (X, Y, n);
  gsl_histogram_free (h);
  d_plot->replot ();
  updateScale ();
}

QString
Graph::showHistogramStats (Table * w, const QString & curveName, int index)
{
  int i;
  QwtHistogram *cv = (QwtHistogram *)curve (index);
  if (!cv)
    return "";

  int ycol = w->colIndex (curveName);
  int r = w->tableRows ();

  QMemArray < double >Y (r);
  int it = 0;
  for (i = 0; i < r; i++)
  {
    QString yval = w->text (i, ycol);
    if (!yval.isEmpty ())
    {
      it++;
      Y[it - 1] = yval.toDouble ();
    }
  }

  if (it < 2 || (it == 2 && Y[0] == Y[1]))
  {				//non valid histogram data
    QMessageBox::critical (0, tr ("Error - QtiPlot"),
      tr
      ("Your data is not valid. You need at least two different points for a histogram!"));
    return "";
  }

  bool automatic = cv->autoBinning ();
  double binSize = cv->binSize ();
  double begin = cv->begin ();
  double end = cv->end ();
  int n;
  gsl_histogram *h;

  if (automatic)
  {
    n = 10;
    h = gsl_histogram_alloc (n);
    if (!h)
      return "";

    gsl_vector *v;
    v = gsl_vector_alloc (it);
    for (i = 0; i < it; i++)
      gsl_vector_set (v, i, Y[i]);

    double min, max;
    gsl_vector_minmax (v, &min, &max);
    gsl_vector_free (v);

    gsl_histogram_set_ranges_uniform (h, floor (min), ceil (max));
  }
  else
  {
    n = int ((end - begin) / binSize + 1);
    h = gsl_histogram_alloc (n);
    if (!h)
      return "";

    //double *range;
    //range = vector (0, n + 1);
    QMemArray <double> range(n+2);

    for (i = 0; i <= n + 1; i++)
      range[i] = begin + i * binSize;

    gsl_histogram_set_ranges (h, range.data(), n + 1);
    //free_vector (range, 0, n + 1);
  }

  for (i = 0; i < it; i++)
    gsl_histogram_increment (h, Y[i]);

  QMemArray < double >X (n);	//stock ranges (x) and bins (y)
  Y.resize (n);
  for (i = 0; i < n; i++)
  {
    Y[i] = gsl_histogram_get (h, i);
    double lower, upper;
    gsl_histogram_get_range (h, i, &lower, &upper);
    X[i] = lower;
  }

  double h_sum = 0.0;
  for (i = 0; i < n; i++)
    h_sum += Y[i];

  double sum = 0.0;
  QString text =
    tr ("Bins") + "\t" + tr ("Quantity") + "\t" + tr ("Sum") + "\t" +
    tr ("Percent") + "\n";
  for (i = 0; i < n; i++)
  {
    sum += Y[i];
    text += QString::number (X[i]) + "\t";
    text += QString::number (Y[i]) + "\t";
    text += QString::number (sum) + "\t";
    text += QString::number (sum / h_sum * 100) + "\n";
  }

  QDateTime dt = QDateTime::currentDateTime ();
  QString info =
    dt.toString (Qt::LocalDate) + "\t" +
    tr ("Histogram and Probabilities for") + ": " + curveName + "\n";
  info +=
    tr ("Mean") + " = " + QString::number (gsl_histogram_mean (h)) + "\t";
  info +=
    tr ("Standard Deviation") + " = " +
    QString::number (gsl_histogram_sigma (h)) + "\n";
  info +=
    tr ("Minimum") + " = " + QString::number (gsl_histogram_min_val (h)) +
    "\t";
  info +=
    tr ("Maximum") + " = " + QString::number (gsl_histogram_max_val (h)) +
    "\t";
  info +=
    tr ("Bins") + " = " + QString::number (gsl_histogram_bins (h)) + "\n";
  info += "-------------------------------------------------------------\n";
  gsl_histogram_free (h);

  QString caption = "Bins-" + curveName;
  emit createHistogramTable (caption, n, 4, text);
  return info;
}

void
Graph::initHistogram (int index, const QMemArray < double >&Y, int it)
{
  QwtHistogram *hc = (QwtHistogram *)curve(index);
  if (!hc)
    return;

  int i;
  if (it < 2 || (it == 2 && Y[0] == Y[1]))
  {				//non valid histogram data
    QMemArray < double >x (2), y (2);
    for (i = 0; i < 2; i++)
    {
      y[i] = 0;
      x[i] = 0;
    }
    hc->setData (x, y, 2);
    d_plot->replot ();
    updateScale ();
    return;
  }

  int n = 10;			//default value
  QMemArray < double >x (n), y (n);	//store ranges (x) and bins (y)
  gsl_histogram *h = gsl_histogram_alloc (n);
  if (!h)
    return;

  gsl_vector *v;
  v = gsl_vector_alloc (it);
  for (i = 0; i < it; i++)
    gsl_vector_set (v, i, Y[i]);

  double min, max;
  gsl_vector_minmax (v, &min, &max);
  gsl_vector_free (v);

  double from = floor (min);
  double to = ceil (max);

  gsl_histogram_set_ranges_uniform (h, floor (min), ceil (max));

  for (i = 0; i < it; i++)
    gsl_histogram_increment (h, Y[i]);

  for (i = 0; i < n; i++)
  {
    y[i] = gsl_histogram_get (h, i);
    double lower, upper;
    gsl_histogram_get_range (h, i, &lower, &upper);
    x[i] = lower;
  }

  hc->setBinning (true, (double) from, (double) to,
    (double) (to - from) / (double) n);
  hc->setData (x, y, n);
  gsl_histogram_free (h);
}

bool
Graph::stepCheck( int cindex )
{
  QwtPlotCurve *c = (QwtPlotCurve *)this->curve(cindex);
  int n = c->dataSize();

  double sep0 = c->x(1) - c->x(0);
  if(sep0==0) return false;
  for(int i=2;i<n;i++)
  {
    double sep = c->x(i) - c->x(i-1);
    if( fabs((sep-sep0)/sep0) > 0.01 )
    {
      if(QMessageBox::warning (0, "QtiPlot",
        tr("X steps are not constant.\nDo you want to continue the process?"),
        QMessageBox::Ignore,QMessageBox::Cancel)
        == QMessageBox::Ignore)
        return true;
      else
        return false;
    }
  }
  return true;
}

bool
Graph::fft (ApplicationWindow *app, int index, bool forward, double sampling,
            bool resample, bool normalizeAmp, bool order, FFTWindow win)
{
  QwtPlotCurve *c = (QwtPlotCurve *)this->curve(index);
  if (!c)
    return false;

  if(!resample && !stepCheck( index ))
    return false;

  QApplication::setOverrideCursor (waitCursor);

  double start,end;
  range( index, &start, &end );
    
  int n;
  if(resample)
    n = qRound(end - start) / sampling + 1;  
  else
    n = numPoints( index, start, end );

  QMemArray <double> x(n), y(n), amp(n), result(n*2);

  if(resample)
  {
    for(int i=0;i<n;i++)
      x[i] = start + sampling * i;
    if(!interpolate(index, 1, x, y)) // cubic spline
      return false;
  }
  else
  {    
    int top = 0;
    while ( c->x(top) < start ) top++;
    for (int i = 0; i < n; i++)
      y[i] = c->y(top+i);
  }

  if(win==Hanning){
    for(int i=0;i<n;i++)
      y[i] = y[i] * 0.5 * ( 1 - cos( M_PI * 2 * i / n ) );
  } else if(win==Hamming){
    for(int i=0;i<n;i++)
      y[i] = y[i] * 0.54 - 0.46 * cos( M_PI * 2 * i / n );
  }

  double df = 1.0 / (double) (n * sampling);	//frequency sampling
  double aMax = 0.0;		//max amplitude
  QString label, wlabel, text;
  if (forward)
  {
    //label = "ForwardFFT" + QString::number (++fitID);
    label = "FFT";
    wlabel = tr("ForwardFFT of ") + c->title().text();
    text = tr ("Frequency");

    gsl_fft_real_workspace *work = gsl_fft_real_workspace_alloc (n);
    gsl_fft_real_wavetable *real = gsl_fft_real_wavetable_alloc (n);

    if (!work || !real)
    {
      QMessageBox::warning (0, "QtiPlot",
        tr
        ("Could not allocate memory, operation aborted!"));
      return false;
    }

    gsl_fft_real_transform (y.data(), 1, n, real, work);
    gsl_fft_halfcomplex_unpack (y.data(), result.data(), 1, n);

    gsl_fft_real_wavetable_free (real);
    gsl_fft_real_workspace_free (work);
  }
  else
  {
    //label = "InverseFFT" + QString::number (++fitID);
    label = "IFFT";
    wlabel = tr ("InverseFFT of ") + c->title ().text ();	
    text = tr ("Time");

    gsl_fft_real_unpack (y.data(), result.data(), 1, n);
    gsl_fft_complex_wavetable *wavetable = gsl_fft_complex_wavetable_alloc(n);
    gsl_fft_complex_workspace *workspace = gsl_fft_complex_workspace_alloc(n);

    if (!workspace || !wavetable)
    {
      QMessageBox::warning (0, "QtiPlot",
        tr
        ("Could not allocate memory, operation aborted!"));
      return false;
    }

    gsl_fft_complex_inverse (result.data(), 1, n, wavetable, workspace);

    gsl_fft_complex_wavetable_free (wavetable);
    gsl_fft_complex_workspace_free (workspace);
  }

  QMemArray <double> yr(n), yi(n);

  if(order)
  {
    int n2 = n / 2;
    int d = (n+1) / 2;
    for (int i = 0; i < n; i++)
    {
      int i2 = (i+d) % n;
      x[i] = (i-n2) * df;
      yr[i] = result[i2*2];
      yi[i] = result[i2*2+1];
    }
  }
  else
  {
    for (int i = 0; i < n; i++)
    {
      x[i] = i * df;
      yr[i] = result[i*2];
      yi[i] = result[i*2+1];
    }
  }
  aMax = 0;
  for(int i=0;i<n;i++)
  {
    amp[i] = sqrt (yr[i] * yr[i] + yi[i] * yi[i]);
    if (amp[i] > aMax)
      aMax = amp[i];
  }
  text +=
    "\t" + tr ("Real") + "\t" + tr ("Imaginary") + "\t" + tr ("Amplitude") +
    "\t" + tr ("Angle") + "\n";
  int prec = app->tbl_output_precision;
  for (int i = 0; i < n; i++)
  {
    text += QString::number (x[i],'g',prec) + "\t";
    text += QString::number (yr[i],'g',prec) + "\t";
    text += QString::number (yi[i],'g',prec) + "\t";
    if (normalizeAmp)
      text += QString::number (amp[i] / aMax,'g',prec) + "\t";
    else
      text += QString::number (amp[i],'g',prec) + "\t";
    text += QString::number (atan2(yi[i],yr[i]),'g',prec) + "\n";
  }

  QString title = app->generateUniqueName( label );

  emit createHiddenTable (title+"\t"+wlabel, n, 5, prec, text);
  QApplication::restoreOverrideCursor ();
  return true;
}

// store interpolated data correspoiding to given x array into y array
bool
Graph::interpolate( int cindex, int spline, 
		    QMemArray <double> &x, QMemArray <double> &y )
{
  int points = x.size();
  if (points<1) return false;
  if (y.size() != points)
    y.resize(points);

  QwtPlotCurve *c = curve( cindex );
  double start = c->minXValue();
  double end = c->maxXValue();

  QMemArray <double> x1, y1;
  int n1 = sortedCurveData (cindex, start, end, x1, y1);
  if (n1==-1)
  {
    QMessageBox::critical (this, tr ("QtiPlot - Error"),
      tr("Several data points have the same x value causing divisions by zero, operation aborted!"));
    return false;
  }
  else if (n1==0)
  {
    QMessageBox::critical (this, tr ("QtiPlot - Error"),
      tr("Failed to obtain sorted curve data!"));
    return false;
  }
  else if ((int) n1 < spline + 3)
  {
    QMessageBox::critical (this, tr ("QtiPlot - Error"),
      tr("You need at least %1 points to perform this operation! Operation aborted!").
      arg (QString::number (spline + 3)));
    return false;
  }

  const gsl_interp_type *method;
  switch (spline)
  {
  case 0:
    method = gsl_interp_linear;
    break;
  case 1:
    method = gsl_interp_cspline;
    break;
  case 2:
    method = gsl_interp_akima;
    break;
  }

  gsl_interp_accel *acc = gsl_interp_accel_alloc ();
  gsl_interp *sp = gsl_interp_alloc (method, n1);
  gsl_interp_init (sp, x1.data(), y1.data(), n1);

  for (int i = 0; i < points; i++)
    y[i] = gsl_interp_eval (sp, x1.data(), y1.data(), x[i], acc);

  gsl_interp_free (sp);
  gsl_interp_accel_free (acc);

  return true;
}

bool
Graph::diffCurve (ApplicationWindow *app, const QString & curveTitle)
{
  int cindex = curveIndex (curveTitle);
  if (cindex < 0)
    return false;

  QwtPlotCurve *c = (QwtPlotCurve *)curve(cindex);
  double start = c->minXValue();
  double end = c->maxXValue();
  QMemArray <double> x,y;
  int n = sortedCurveData (cindex, start, end, x, y);

  if( n==-1 )
  {
    QMessageBox::critical (this, tr ("QtiPlot - Error"),
      tr("Several data points have the same x value causing divisions by zero, operation aborted!"));
    return false;
  }
  else if( n==0 )
  {
    QMessageBox::critical (this, tr ("QtiPlot - Error"),
      tr("Failed to prepare sorted curve data!"));
    return false;
  }
  else if (n < 4)
  {
    QMessageBox::critical (this, tr ("QtiPlot - Error"),
      tr("You need at least %1 points to perform this operation! Operation aborted!").arg (QString::number (4)));
    return false;
  }

  int prec = app->tbl_output_precision;
  QString text = "x\tderivative\n";
  for (int i=0; i<n; i++)
  {
    double result;
    if(i==0)
      result = (y[1] - y[0]) / (x[1] - x[0]);
    else if(i==n-1)
      result = (y[n-1] - y[n-2]) / (x[n-1] - x[n-2]);
    else
      result = 0.5 * ((y[i + 1] - y[i]) / (x[i + 1] - x[i]) +
      (y[i] - y[i - 1]) / (x[i] - x[i - 1]));
    text += QString::number (x[i],'g',prec);
    text += "\t";
    text += QString::number (result,'g',prec);
    text += "\n";
  }

  QString title = app->generateUniqueName( tr("Derivative") );
  emit createHiddenTable (title + "\t" + "Derivative of " +
    curve (cindex)->title ().text (),
    n, 2, prec, text);

  return true;
}


QString
Graph::integrateCurve (QString &curveTitle, double low, double up)
{
  int index = curveIndex( curveTitle );
  if (index<0) return "";

  QMemArray <double> x,y;
  int d_n = sortedCurveData(index, low, up, x, y);
  if (d_n == -1)
  {
    QMessageBox::critical (this, tr ("QtiPlot - Integration error"),
      tr("Several data points have the same x value causing divisions by zero, operation aborted!"));
    return "";
  }
  else if (d_n == 0)
  {
    QMessageBox::critical (this, tr ("QtiPlot - Integration error"),
      tr("Failed to prepare sorted curve data!"));
    return "";
  }
  else if (d_n < 4)
  {
    QMessageBox::critical (this, tr ("QtiPlot - Integration error"),
      tr("You need at least 4 points to integrate! Integration aborted!"));
    return "";
  }

  // data check
  int maxID = 0;
  double value = y[0];
  for(int i=1;i<d_n;i++)
    if( y[i] > value )
    {
      value = y[i];
      maxID = i;
    }

    // Do the interpolation, use GSL libraries for that
    gsl_interp_accel *acc = gsl_interp_accel_alloc ();
    const gsl_interp_type *method;
    // The method for interpolation is chosen based on the number of points
    if (d_n > 5)
      method = gsl_interp_akima;
    else if (d_n > 4)
      method = gsl_interp_cspline;
    else
      method = gsl_interp_linear;

    gsl_interp *interp;
    interp = gsl_interp_alloc (method, d_n);
    gsl_interp_init (interp, x.data(), y.data(), d_n);

    double result =
      gsl_interp_eval_integ (interp, x.data(), y.data(), low, up, acc);

    QDateTime dt = QDateTime::currentDateTime ();
    QString date = dt.toString (Qt::LocalDate);
    QString info = tr("[") + date
      + "\t" + this->caption () + "]\nNumerical integration of: " +
      curveTitle + " with a spline integration method \n";
    info +=
      "Points: " + QString::number (d_n) + " from x=" + QString::number (low) +
      " to x=" + QString::number (up) + "\n";
    info += "Peak at x=";
    info += QString::number (x[maxID]) + "\t";
    info += "y=";
    info += QString::number (y[maxID]) + "\n";
    info += "Area=";
    info += QString::number (result);
    info += "\n-------------------------------------------------------------\n";

    gsl_interp_free (interp);
    gsl_interp_accel_free (acc);

    return info;
}

QString
Graph::curveFWHM(QString &curveTitle, double low, double up, int peakfrac, int halffrac, int prec)
{
  int index = curveIndex( curveTitle );
  if (index<0) return "";

  QMemArray <double> x,y;
  int d_n = sortedCurveData(index, low, up, x, y);
  if (d_n == -1)
  {
    QMessageBox::critical (this, tr ("QtiPlot - FHWM Error"),
      tr("Several data points have the same x value causing divisions by zero, operation aborted!"));
    return "";
  }
  else if(d_n==0)
  {
    QMessageBox::critical (this, tr ("QtiPlot - FWHM Error"),
      tr("The dataset has no data points between %1 and %2, operation aborted!").arg(QString::number(low)).arg(QString::number(up)));
    return "";
  }
  // find maximum
  int imax = 0;
  double value = y[0];
  for(int i=1;i<d_n;i++)
  {
    if(y[i] > value)
    {
      imax = i;
      value = y[i];
    }
  }

  // determine peak range
  double thres = y[imax] * ( 100 - peakfrac ) * 0.01;
  int i1,i2;
  for(i1=imax-1;i1>=0;i1--)
  {
    if(y[i1]<thres) break;
  }
  if (i1==0)
  {
    QMessageBox::critical (this, tr ("QtiPlot - FWHM error"),
      tr("There is no half maximum in this range!"));
    return "";
  }
  for(i2=imax+1;i2<d_n;i2++)
  {
    if(y[i2]<thres) break;
  }
  if (i2==d_n-1)
  {
    QMessageBox::critical (this, tr ("QtiPlot - FWHM error"),
      tr("There is no half maximum in this range!"));
    return "";
  }

  // find exact peak
  int n = i2 - i1 + 1;
  gsl_matrix *X = gsl_matrix_alloc(n, 3);
  gsl_vector *s = gsl_vector_alloc(n);
  gsl_vector *cc = gsl_vector_alloc(3);
  gsl_matrix *covar = gsl_matrix_alloc (3, 3);
  gsl_multifit_linear_workspace *work = gsl_multifit_linear_alloc (n, 3);
  double chisqr;
  for(int j=0;j<n;j++)
  {
    gsl_vector_set(s, j, y[i1+j]);
    for (int k = 0; k < 3; k++)
      gsl_matrix_set (X, j, k, pow (x[i1+j], k));
  }

  gsl_multifit_linear (X, s, cc, covar, &chisqr, work);
  double pa = gsl_vector_get (cc, 2);
  double pb = gsl_vector_get (cc, 1);
  double pc = gsl_vector_get (cc, 0);

  gsl_multifit_linear_free (work);
  gsl_matrix_free (covar);
  gsl_vector_free (cc);
  gsl_vector_free (s);
  gsl_matrix_free (X);

  double peak_x = - pb / pa / 2;
  double height = pc - pa*peak_x*peak_x;

  QString formula = QString::number(pa,'g',prec)+"*(x";
  if(peak_x<0) formula += "+";
  formula += QString::number(-peak_x,'g',prec);
  formula += ")^2";
  if(height>=0) formula += "+";
  formula += QString::number(height,'g',prec);

  QStringList formulas;
  QValueList < double >ranges;
  formulas += formula;
  ranges += x[i1];
  ranges += x[i2];
  addFunctionCurve(FunctionCurve::Normal,formulas,"x",ranges,20,tr("Peak"));
  QPen pen = QPen(red,1,Qt::SolidLine,Qt::FlatCap,Qt::BevelJoin);
  setCurvePen( n_curves-1, pen ); // most recent curve

  double c0, c1, cov00, cov01, cov11;
  double half_x1,half_x2;

  // search lower half point
  for(i2=imax;i2>=0;i2--)
  {
    // first point inside
    if(y[i2] < height * ( 0.5 + (double)halffrac / 200 ) )
      break;
  }
  if((i2==0) || (y[i2] <= height * 0.5))
  {
    QMessageBox::critical (this, tr ("QtiPlot - FWHM error"),
      tr("There is no half maximum in this range!"));
    return "";
  }

  for(i1=i2;i1>=0;i1--)
  {
    // first point outside
    if(y[i1] < height * ( 0.5 - (double)halffrac / 200 ) )
      break;
  }
  i1++; // last point inside
  if(y[i1] >= height * 0.5)
  {
    QMessageBox::critical (this, tr ("QtiPlot - FWHM error"),
      tr("There is no half maximum in this range!"));
    return "";
  }

  n = i2 - i1 + 1;
  QMemArray <double> xx(n), yy(n);
  //fit around lower half point
  for(int j=0;j<n;j++)
  {
    xx[j] = x[i1+j];
    yy[j] = y[i1+j];
  }
  gsl_fit_linear (xx.data(), 1, yy.data(), 1, n,
		  &c0, &c1, &cov00, &cov01, &cov11, &chisqr);
  half_x1 = (height/2-c0)/c1;
  //fprintf(stderr,"half_x1 = %f\n",half_x1);
  insertLineMarker( xx[0], xx[0]*c1+c0, xx[n-1], xx[n-1]*c1+c0,
    false, false, QColor(red) );

  // search upper half point
  for(i1=imax;i1<d_n;i1++)
  {
    // first point inside
    if(y[i1] < height * ( 0.5 + (double)halffrac / 200 ) )
      break;
  }
  if((i1==d_n-1) || (y[i1] <= height * 0.5))
  {
    QMessageBox::critical (this, tr ("QtiPlot - FWHM error"),
      tr("There is no half maximum in this range!"));
    return "";
  }

  for(i2=i1;i2<d_n;i2++)
  {
    // first point outside
    if(y[i2] < height * ( 0.5 - (double)halffrac / 200 ) )
      break;
  }
  i2--; // last point inside
  if(y[i2] >= height * 0.5)
  {
    QMessageBox::critical (this, tr ("QtiPlot - FWHM error"),
      tr("There is no half maximum in this range!"));
    return "";
  }

  n = i2 - i1 + 1;
  xx.resize(n);
  yy.resize(n);
  //fit around upper half point
  for(int j=0;j<n;j++)
  {
    xx[j] = x[i1+j];
    yy[j] = y[i1+j];
  }
  gsl_fit_linear (xx.data(), 1, yy.data(), 1, n, 
		  &c0, &c1, &cov00, &cov01, &cov11, &chisqr);
  half_x2 = (height/2-c0)/c1;
  //fprintf(stderr,"half_x2 = %f\n",half_x2);
  insertLineMarker( xx[0], xx[0]*c1+c0, xx[n-1], xx[n-1]*c1+c0,
    false, false, QColor(red) );

  insertLineMarker( peak_x, 0, peak_x, height, false, false, QColor(green));
  insertLineMarker( half_x1, height/2, half_x2, height/2, false, false, QColor(green));

  d_plot->replot();

  QString text = tr("Name")+"\t"+tr("Values")+"\n";
  text += tr("Peak") + "\t" +QString::number(peak_x,'g',prec) + "\n";
  text += tr("Height") + "\t" + QString::number(height,'g',prec) + "\n";
  text += tr("LowerX") + "\t" + QString::number(half_x1,'g',prec) + "\n";
  text += tr("UpperX") + "\t" + QString::number(half_x2,'g',prec) + "\n";
  text += tr("FWHM") + "\t" + QString::number(fabs(half_x2-half_x1),'g',prec) + "\n";
  return text;
}

void
Graph::calculateLineProfile (const QPoint & start, const QPoint & end)
{
  ImageMarker *mrk = (ImageMarker *) d_plot->marker (selectedMarker);
  if (!mrk)
  {
    QMessageBox::warning (0, tr ("QtiPlot - Pixel selection warning"),
      tr
      ("Please select the start line point inside the image rectangle!"));
    int linesOnPlot = (int) d_lines.size ();
    d_plot->removeMarker (d_lines[--linesOnPlot]);
    d_lines.resize (linesOnPlot);
    return;
  }

  QRect rect = mrk->rect ();
  if (!rect.contains (start) || !rect.contains (end))
  {
    QMessageBox::warning (0, tr ("QtiPlot - Pixel selection warning"),
      tr
      ("Please select the end line point inside the image rectangle!"));
    int linesOnPlot = (int) d_lines.size ();
    d_plot->removeMarker (d_lines[--linesOnPlot]);
    d_lines.resize (linesOnPlot);
    return;
  }

  QPoint o = mrk->getOrigin ();
  QPixmap pic = mrk->image ();
  QImage image = pic.convertToImage ();
  lineProfileOn = FALSE;

  int x1 = start.x () - o.x ();
  int x2 = end.x () - o.x ();
  int y1 = start.y () - o.y ();
  int y2 = end.y () - o.y ();

  QSize realSize = pic.size ();
  QSize actualSize = mrk->size ();

  if (realSize != actualSize)
  {
    double ratioX =
      (double) realSize.width () / (double) actualSize.width ();
    double ratioY =
      (double) realSize.height () / (double) actualSize.height ();
    x1 = int (x1 * ratioX);
    x2 = int (x2 * ratioX);
    y1 = int (y1 * ratioY);
    y2 = int (y2 * ratioY);
  }

  QString text = "pixel\tx\ty\tintensity\n";
  //uses the fast Bresenham's line-drawing algorithm
#define sgn(x) ((x<0)?-1:((x>0)?1:0))
  int i, dx, dy, sdx, sdy, dxabs, dyabs, x, y, px, py, n;

  dx = x2 - x1;			//the horizontal distance of the line 
  dy = y2 - y1;			//the vertical distance of the line 
  dxabs = QABS(dx);
  dyabs = QABS(dy);
  sdx = sgn (dx);
  sdy = sgn (dy);
  x = dyabs >> 1;
  y = dxabs >> 1;
  px = x1;
  py = y1;

  if (dxabs >= dyabs)		//the line is more horizontal than vertical 
  {
    for (i = 0; i < dxabs; i++)
    {
      y += dyabs;
      if (y >= dxabs)
      {
        y -= dxabs;
        py += sdy;
      }
      px += sdx;

      n = dxabs;
      text += QString::number (i) + "\t";
      text += QString::number (px) + "\t";
      text += QString::number (py) + "\t";
      text +=
        QString::
        number (averageImagePixel (image, px, py, averagePixels, TRUE)) +
        "\n";
    }
  }
  else				// the line is more vertical than horizontal 
  {
    for (i = 0; i < dyabs; i++)
    {
      x += dxabs;
      if (x >= dyabs)
      {
        x -= dyabs;
        px += sdx;
      }
      py += sdy;

      n = dyabs;
      text += QString::number (i) + "\t";
      text += QString::number (px) + "\t";
      text += QString::number (py) + "\t";
      text +=
        QString::
        number (averageImagePixel (image, px, py, averagePixels, FALSE)) +
        "\n";
    }
  }
  QString caption = "Table1";
  emit createTablePlot (caption, n, 4, text);
}

int
Graph::averageImagePixel (const QImage & image, int px, int py, int average,
                          bool moreHorizontal)
{
  QRgb pixel;
  int sum = 0, start, i;
  int middle = int (0.5 * (average - 1));
  if (moreHorizontal)
  {
    start = py - middle;
    for (i = 0; i < average; i++)
    {
      pixel = image.pixel (px, start + i);
      sum += qGray (pixel);
    }
  }
  else
  {
    start = px - middle;
    for (i = 0; i < average; i++)
    {
      pixel = image.pixel (start + i, py);
      sum += qGray (pixel);
    }
  }
  return sum / average;
}

void
Graph::calculateProfile (int average, bool ok)
{
  lineProfileOn = ok;
  if (average % 2 == 0)
    averagePixels = average + 1;
  else
    averagePixels = average;
}

void
Graph::addResultCurve (int n, QMemArray <double> &x, QMemArray <double> &y,
		       int colorIndex,
                       const QString & tableName, const QString & legend,
                       int prec, QwtPlotCurve * orig_curve)
{
  QString label = tableName + "_2";
  QwtPlotCurve *c = new QwtPlotCurve (label);
  c->setPen (QPen (ColorBox::color (colorIndex), 1));
  c->setAxis( orig_curve->xAxis(), orig_curve->yAxis() );
  int curveID = d_plot->insertCurve (c);

  n_curves++;
  c_attr.append( new CurveAttr( curveID, Line ) );

  addLegendItem (label);
  label = tableName + "_1," + label;
  associations << label;

  c->setData (x.data(), y.data(), n);
  QString text = "1\t2\n";

  for (int i = 0; i < n; i++)
  {
    text += QString::number (x[i], 'g', prec);
    text += "\t";
    text += QString::number (y[i], 'g', prec);
    text += "\n";
  }

  emit createHiddenTable (tableName + "\t" + legend, n, 2, prec, text);
  updatePlot ();
}

bool
Graph::smoothSavGol (ApplicationWindow *app, int cindex, int order, int nl, int nr, int colIndex)
{
  QwtPlotCurve *curve = (QwtPlotCurve *)this->curve(cindex);
  if (!curve || curve->rtti () != QwtPlotItem::Rtti_PlotCurve)
    return false;

  if(!stepCheck( cindex ))
    return false;

  QApplication::setOverrideCursor (waitCursor);

  int n = curve->dataSize ();

  QMemArray <double> x(n), y(n), s(n);

  for (int i = 0; i < n; i++)
  {				// The data to be smoothed 
    x[i] = curve->x (i);
    y[i] = curve->y (i);
  }

  smooth_vect( y.data(), n, nl, nr, order, s.data() );

  QString title = app->generateUniqueName( tr("Smoothed") );
  QString label = tr ("Savitzky-Golay smoothing of ") + curve->title().text();
  addResultCurve (n, x, s, colIndex, title, label,
		  app->tbl_output_precision, curve);

  QApplication::restoreOverrideCursor ();
  return true;
}

bool
Graph::PeakConv(ApplicationWindow *app, int cindex, int points, int colIndex,
                int peakType)
{//peakType = 0 : Gaussian
  //          1 : Lorentzian
  if (peakType != 0 && peakType != 1)
    return false;

  QwtPlotCurve *curve = (QwtPlotCurve *)this->curve (cindex);
  if (!curve || curve->rtti () != QwtPlotItem::Rtti_PlotCurve)
    return false;

  if(!stepCheck( cindex ))
    return false;

  QApplication::setOverrideCursor (waitCursor);

  int i, n = curve->dataSize ();
  QMemArray <double> x(n), y(n);

  for (i = 0; i < n; i++)
  {				// The data to be filtered 
    x[i] = curve->x(i);
    y[i] = curve->y(i);
  }

  gsl_fft_real_workspace *work = gsl_fft_real_workspace_alloc (n);
  gsl_fft_real_wavetable *real = gsl_fft_real_wavetable_alloc (n);
  // FFT forward (real to halfcomplex sequence)
  gsl_fft_real_transform (y.data(), 1, n, real, work);
  gsl_fft_real_wavetable_free (real);

  double w = (double) points / (double) n;
  QString filtername;
  switch (peakType) 
  {
  case 0 : //Gaussian Filter
    filtername = "Gaussian";
    for (i = 0; i < n; i++)
    {
      int j = (i+1)/2;
      y[i] *= exp( - (double) j * j * w * w);
    }
    break;
  case 1 : //Lorentzian Filter
    filtername = "Lorentzian";
    for (i = 0; i < n; i++)
    {
      int j = (i+1)/2;
      y[i] *= exp( - (double) 2 * j * w );
    }
    break;
  }

  gsl_fft_halfcomplex_wavetable *hc = gsl_fft_halfcomplex_wavetable_alloc (n);
  // FFT backward (halfcomplex to real sequence)
  gsl_fft_halfcomplex_inverse (y.data(), 1, n, hc, work);	//FFT inverse
  gsl_fft_halfcomplex_wavetable_free (hc);
  gsl_fft_real_workspace_free (work);

  QString title = app->generateUniqueName( tr("Smoothed") );
  QString label = QString::number (points) + " points " + filtername
    + " Convolution of " + curve->title ().text ();

  addResultCurve (n, x, y, colIndex, title, label,
    app->tbl_output_precision, curve);

  QApplication::restoreOverrideCursor ();
  return true;
}

bool
Graph::smoothAverage (ApplicationWindow *app, int index, int points, int colIndex)
{
  QwtPlotCurve *curve = (QwtPlotCurve *)this->curve(index);
  if (!curve || curve->rtti () != QwtPlotItem::Rtti_PlotCurve)
    return false; 

  QApplication::setOverrideCursor (waitCursor);

  int i, n = curve->dataSize ();

  QMemArray <double> x(n), y(n), s(n);

  for (i = 0; i < n; i++)
  {				// The data to be smoothed 
    x[i] = curve->x (i);
    y[i] = curve->y (i);
  }

  smooth_ave( y.data(), n, points, s.data() );

  QString title = app->generateUniqueName( tr("Smoothed") );

  addResultCurve (n, x, s, colIndex,
		  // tr ("Smoothed") + QString::number (++fitID),
		  title,
		  QString::number (points) + " " + tr ("points") + " " +
		  tr ("average smoothing of") + " " +
		  curve->title ().text (),
		  app->tbl_output_precision,
		  curve);
  
  QApplication::restoreOverrideCursor ();
  return true;
}

void fft_filter_interpolate(int j1, int j2, int j3, int j4,
			    QMemArray <double> &dat)
{
  double chisq;
  int d_n = j2 - j1 + 1 + j4 - j3 + 1;
  gsl_matrix *X = gsl_matrix_alloc (d_n, 4);
  gsl_vector *y = gsl_vector_alloc (d_n);
  gsl_vector *c = gsl_vector_alloc (4);
  gsl_matrix *cov = gsl_matrix_alloc (4, 4);

  int k=0;
  for (int j = j1; j <= j4; j++)
    if(j<=j2 || j>=j3)
    {
      gsl_matrix_set(X, k, 0, 1);
      gsl_matrix_set(X, k, 1, j);
      gsl_matrix_set(X, k, 2, j*j);
      gsl_matrix_set(X, k, 3, j*j*j);
      gsl_vector_set(y, k, dat[j]);
      k++;
    }

  if(k!=d_n)
  {
    qDebug("k is not equal to d_n!");
  }

  gsl_multifit_linear_workspace *work = gsl_multifit_linear_alloc (d_n, 4);
  gsl_multifit_linear (X, y, c, cov, &chisq, work);

  for(int j=j2;j<=j3;j++)
  {
    dat[j] = 0;
    for (int p = 3; p >= 0; p--)
      dat[j] = dat[j] * j + gsl_vector_get (c, p);
  }

  gsl_multifit_linear_free (work);
  gsl_matrix_free(cov);
  gsl_vector_free(c);
  gsl_vector_free(y);
  gsl_matrix_free(X);
}

bool
Graph::filterFFT (ApplicationWindow *app, int cindex, int filter_type, 
                  double lf, double hf, bool resample, bool DCOffset, int colIndex)
{
  QwtPlotCurve *curve = (QwtPlotCurve *)this->curve (cindex);
  if (!curve || curve->rtti () != QwtPlotItem::Rtti_PlotCurve)
    return false;

  if(!resample && !stepCheck( cindex ))
    return false;

  QApplication::setOverrideCursor (waitCursor);

  int i, n = curve->dataSize ();

  QMemArray <double> x(n), y(n);
  
  double sampling = ( curve->maxXValue() - curve->minXValue() ) / (n-1);
 
  if(resample)
  {
    for(int i=0; i<n; i++)
      x[i] = curve->minXValue() + sampling * i;
    if(!interpolate(cindex,1,x,y))
      return false;
  }
  else
  {
    for (i = 0; i < n; i++)
    { 
      x[i] = curve->x (i);
      y[i] = curve->y (i);
    }
  }

  double df = 1.0 / ( n * sampling );

  gsl_fft_real_workspace *work = gsl_fft_real_workspace_alloc (n);
  gsl_fft_real_wavetable *real = gsl_fft_real_wavetable_alloc (n);

  gsl_fft_real_transform (y.data(), 1, n, real, work);
  gsl_fft_real_wavetable_free (real);

  QString label;
  if (filter_type <= 2)
    label = QString::number (lf) + " Hz ";
  else
    label = QString::number (lf) + " to " + QString::number (hf) + " Hz ";

  switch (filter_type)
  {
  case 1:			//low pass
    label += tr ("Low Pass FFT Filter of ");
    for (i = 0; i < n; i++)
    {
      int j = (i+1)/2;
      y[i] = j * df > lf ? 0 : y[i];
    }
    break;

  case 2:			//high pass
    label += tr ("High Pass FFT Filter of ");
    for (i = 0; i < n; i++)
    {
      int j = (i+1)/2;
      y[i] = j * df < lf ? 0 : y[i];
    }
    break;

  case 3:			//band pass
    label += tr ("Band Pass FFT Filter of ");
    if (DCOffset)
    {
      for (i = 1; i < n; i++)
      {
        int j = (i+1)/2;
        y[i] = ((j * df > lf) && (j * df < hf)) ? y[i] : 0;
      }
    }
    else
    {
      for (i = 0; i < n; i++)
      {
        int j = (i+1)/2;
        y[i] = ((j * df > lf) && (j * df < hf)) ? y[i] : 0;
      }
    }
    break;

  case 4:			//band block
    label += tr ("Band Block FFT Filter of ");

    if (DCOffset)
      y[0] = 0;		//substract DC offset

    for (i = 1; i < n; i++)
    {
      int j = (i+1)/2;
      y[i] = ((j * df > lf) && (j * df < hf)) ? 0 : y[i];
    }

    break;

  case 5:			//band block (interpolate)
    label += tr ("Band Block (interpolate) FFT Filter of ");

    if (DCOffset)
      y[0] = 0;		//substract DC offset

    int j2 = qRound(lf/df);
    int j3 = qRound(hf/df);
    if( j2>0 && j3<n/2 && j2 < j3 )
      {
	int j1 = QMAX(j2 - (j3-j2), 0);
	int j4 = QMIN(j3 + (j3-j2), n/2);
	QMemArray <double> yr(n/2+1), yi(n/2+1);

	yr[0] = y[0];     yi[0] = 0;
	yr[n/2] = y[n-1]; yi[n/2] = 0;
	for(int j=1;j<n/2;j++)
	  {
	    yr[j] = y[j*2-1];
	    yi[j] = y[j*2];
	  }
	fft_filter_interpolate(j1,j2,j3,j4,yr);
	fft_filter_interpolate(j1,j2,j3,j4,yi);
	y[0] = yr[0]; y[n-1] = yr[n/2];
	for(int j=1;j<n/2;j++)
	  {
	    y[j*2-1] = yr[j];
	    y[j*2]   = yi[j];
	  }
      }
    break;
  }
 
  gsl_fft_halfcomplex_wavetable *hc = gsl_fft_halfcomplex_wavetable_alloc (n);
  gsl_fft_halfcomplex_inverse (y.data(), 1, n, hc, work);
  gsl_fft_halfcomplex_wavetable_free (hc);
  gsl_fft_real_workspace_free (work);

  QString title = app->generateUniqueName( tr("FilteredFFT") );

  addResultCurve (n, x, y, colIndex,
		  title,
		  label + curve->title ().text (),
		  app->tbl_output_precision,
		  curve);

  QApplication::restoreOverrideCursor ();
  return true;
}

bool
Graph::validCurvesDataSize ()
{
  if (!n_curves)
  {
    QMessageBox::warning (this, tr ("QtiPlot - Warning"),
      tr
      ("There are no curves available on this plot!"));
    return false;
  }
  else
  {
    for (int i = 0; i < n_curves; i++)
    {
      QwtPlotCurve *c = curve (i);
      if (c && c->rtti () == QwtPlotItem::Rtti_PlotCurve
        && c->dataSize () > 1)
        return true;
    }
    QMessageBox::warning (this, tr ("QtiPlot - Error"),
      tr
      ("There are no curves with more than one points on this plot. Operation aborted!"));
    return false;
  }
}

bool
Graph::convoluteFFT(ApplicationWindow *app, int method,
                    const QString &sigTitle, const QString &resTitle,
                    bool normRes, double thres, double cutoff, int colorIndex )
                    // method = 0 : FFT Convolution
                    // method = 1 : FFT Deconvolution
                    // method = 2 : FFT Deconvolution for decay curves (Andre1979)
{
  int sig_index = curveIndex( sigTitle );
  int res_index = curveIndex( resTitle );

  if(!stepCheck( sig_index )) return false;

  QwtPlotCurve *sigc = curve(sig_index);
  QwtPlotCurve *resc = curve(res_index);

  double start = sigc->minXValue();
  double end = sigc->maxXValue();
  if(method==2)
  {
    if(start>0 || end<0)
    {
      QMessageBox::critical (this, tr ("QtiPlot - Error"),
        tr("Signal curve does not include origin in its X range, operation aborted!"));
      return false;
    }
    start = 0;
  }

  QMemArray <double>xs,ys;
  int n = sortedCurveData( sig_index, start, end, xs, ys );
  if (n==-1)
  {
    QMessageBox::critical (this, tr ("QtiPlot - Error"),
      tr("Several data points have the same x value causing divisions by zero, operation aborted!"));
    return false;
  }
  else if (n==0)
  {
    QMessageBox::critical (this, tr ("QtiPlot - Error"),
      tr("Failed to prepare sorted curve data!"));
    return false;
  }

  double step = xs[1] - xs[0];
  int mn = QMIN( n/2, int( fabs( resc->minXValue() ) / step ) );
  int mp = QMIN( n/2, int( fabs( resc->maxXValue() ) / step ) );

  QMemArray <double> xr(mn+mp+1), yr(mn+mp+1);

  for(int i = 0; i < mn+mp+1; i++)
    xr[i] = step * (i-mn); // zero at i=mn

  if (!interpolate(res_index, 1, xr, yr))
    return false;
  if(normRes)
  {
    // normalize response function
    double sum = 0;
    for(int i=0; i < mn+mp+1; i++)
      sum += yr[i];
    if(sum!=0)
      for(int i=0; i < mn+mp+1; i++)
        yr[i] /= sum;
  }

  int n2 = 16;
  int nmin = n + QMAX(mn,mp+1);
  while ( n2 < nmin ) n2 *= 2; //force radix2

  QMemArray <double> sig(n2), res(n2);
  sig.fill( 0 );
  res.fill( 0 );

  //res must be constracted in wrap around order
  for(int i = 0; i <= mp; i++)
    res[i] = yr[mn+i];
  for(int i = 0; i < mn; i++)
    res[n2-1-i] = yr[mn-1-i];

  for(int i = 0; i < n; i++)
    sig[i] = ys[i];

  gsl_fft_real_radix2_transform (sig.data(), 1, n2);
  gsl_fft_real_radix2_transform (res.data(), 1, n2);

  int cut_n = n2/2;
  double df = 1.0 / (step * n2);
  if(cutoff>0)
    cut_n = qRound( cutoff / df );
  double cut_f = cut_n * df;

  double a2max = 0;
  a2max = QMAX( res[0]*res[0], res[n2/2]*res[n2/2] );
  a2max *= thres * thres;

  double sig_r,sig_i,res_r,res_i;
  double re1,im1;
  for (int i=0;i<=n2/2;i++)
  {
    sig_r = sig[i];
    res_r = res[i];
    if(i==0 || i==n2/2)
    {
      sig_i = 0;
      res_i = 0;
    }
    else
    {
      sig_i = sig[n2-i];
      res_i = res[n2-i];       
    }
    double re,im;
    if(method==0) // convolute
    {
      re = (sig_r * res_r - sig_i * res_i) * step;
      im = (sig_r * res_i + res_r * sig_i) * step;
    }
    else // deconvolute
    {
      if(i<=cut_n)
      {
        double size = res_r * res_r + res_i * res_i;
        if ( method == 1 && size < a2max )
          size *= sqrt(a2max/size);
        re = ( sig_r * res_r + sig_i * res_i ) / size / step;
        im = ( res_r * sig_i - sig_r * res_i ) / size / step;
        if(i==cut_n)
        {
          re1 = re;
          im1 = im;
        }
      }
      else
      {
        if(method==1) //Simple cut off
        {
          re = 0;
          im = 0;
        }
        else // Extrapolation with transformed exponential decay
        {
          double A = - 2 * M_PI * cut_f * re1 / im1;
          double B = re1 * ( A*A + SQR(2*M_PI*cut_f) ) / A;
          double C = A*A + SQR(2*M_PI*i*df);
          re = B * A / C / step;
          im = - 2 * M_PI * B * i * df / C / step;
        }
      }
    }
    sig[i] = re;
    if(i>0 && i<n2/2)
      sig[n2-i] = im;
  }
  gsl_fft_halfcomplex_radix2_inverse (sig.data(), 1, n2);

  QString label, wlabel;
  if(method==0)
  {
    label = "Convoluted";
    wlabel = tr("FFT Convolution of ")
      + sigc->title().text() + " and " + resc->title().text();
  }
  else
  {
    label = "Deconvoluted";
    wlabel = tr("FFT Deconvolution of ")
      + sigc->title().text() + " by " + resc->title().text();
  }

  QString title = app->generateUniqueName( label );
  addResultCurve (n, xs, sig, colorIndex, title, wlabel,
    app->tbl_output_precision,
    sigc );

  return true;
}

bool
Graph::deconvoluteGS(ApplicationWindow *app,
                     const QString &sigTitle, const QString &resTitle,
                     bool normRes, int smooth, double damping, int iteration,
                     int constraint, int colorIndex )
{
  int sig_index = curveIndex( sigTitle );
  int res_index = curveIndex( resTitle );

  if(!stepCheck( sig_index )) return false;

  QwtPlotCurve *sigc = curve(sig_index);
  QwtPlotCurve *resc = curve(res_index);

  double start = sigc->minXValue();
  double end = sigc->maxXValue();

  QMemArray <double> xs,ys;
  int n = sortedCurveData( sig_index, start, end, xs, ys );
  if (n==-1)
  {
    QMessageBox::critical (this, tr ("QtiPlot - Error"),
      tr("Several data points have the same X value causing divisions by zero, operation aborted!"));
    return false;
  }
  else if (n==0)
  {
    QMessageBox::critical (this, tr ("QtiPlot - Error"),
      tr("Failed to prepare sorted curve data!"));
    return false;
  }
  QMemArray <double> ys2(n);
  if(smooth>3)
    smooth_vect( ys.data(), n, smooth/2, smooth/2, 2, ys2.data() );
  else
    for(int i=0;i<n;i++)
      ys2[i] = ys[i];

  double step = xs[1] - xs[0];
  int mn = QMIN( n/2, int( fabs( resc->minXValue() ) / step ) );
  int mp = QMIN( n/2, int( fabs( resc->maxXValue() ) / step ) );

  QMemArray <double> xr(mn+mp+1), yr(mn+mp+1);

  for(int i = 0; i < mn+mp+1; i++)
    xr[i] = step * (i-mn); // zero at i=mn

  if (!interpolate(res_index, 1, xr, yr))
    return false;

  if(normRes)
  {
    // normalize response function
    double sum = 0;
    for(int i=0; i < mn+mp+1; i++)
      sum += yr[i];
    if(sum!=0)
      for(int i=0; i < mn+mp+1; i++)
        yr[i] /= sum;
  }

  QMemArray <double> x(n+mn+mp); // solution curve

  for(int i=0; i<n; i++)
    x[mn+i] = ys2[i];

  for(int j=0;j<iteration;j++)
  {
    int i,k;
    for(i=0; i<mn; i++)
      x[i] = x[mn];
    for(i=0; i<mp; i++)
      x[mn+n+i] = x[mn+n-1];
    for(i=0; i<n; i++)
    {
      double sum = 0;
      for(k=0; k<mn+mp+1; k++)
        sum += yr[k] * x[i+k];
      x[mn+i] += (ys2[i] - sum) * damping / yr[mn];
      if((constraint==0 && x[mn+i]<0) || (constraint==1 && x[mn+i]>0))
        x[mn+i] = 0;
    }
  }

  for(int i=0;i<n;i++)
    ys[i] = x[mn+i];

  QString label, wlabel;
  label = "Deconvoluted";
  wlabel = tr("GS Deconvolution of ")
    + sigc->title().text() + " by " + resc->title().text();

  QString title = app->generateUniqueName( label );
  addResultCurve (n, xs, ys, colorIndex, title, wlabel,
    app->tbl_output_precision,
    sigc );

  return true;
}
