/***************************************************************************
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor,                    *
 *   Boston, MA  02110-1301  USA                                           *
 *                                                                         *
 ***************************************************************************/
#include "canvaspicker.h"
#include "ImageMarker.h"
#include "LegendMarker.h"
#include "LineMarker.h"

#include <qpainter.h>
#include <qpaintdevicemetrics.h>
#include <qpixmapcache.h>
#include <qcursor.h>
#include <qapplication.h>
#include <qmessagebox.h>

#include <qwt_text_label.h>
#include <qwt_plot_canvas.h>

CanvasPicker::CanvasPicker (Graph * plot):
QObject (plot)
{
  moved = false;
  movedGraph = false;
  pointSelected = false;
  resizeLineFromStart = false;
  resizeLineFromEnd = false;

  plotWidget = plot->plotWidget ();

  QwtPlotCanvas *canvas = plotWidget->canvas ();
  canvas->installEventFilter (this);
}

bool
CanvasPicker::eventFilter (QObject * object, QEvent * e)
{
  QMemArray < int >images = plot ()->imageMarkerKeys ();
  QMemArray < int >texts = plot ()->textMarkerKeys ();
  QMemArray < int >lines = plot ()->lineMarkerKeys ();

  if (object != (QObject *) plot ()->plotWidget ()->canvas ())
    return FALSE;

  bool moveRangeSelector = plot ()->selectorsEnabled ();
  bool pickerActivated = plot ()->pickerActivated ();
  bool dataCursorEnabled = plot ()->enabledCursor ();
  bool removePoint = plot ()->removePointActivated ();
  bool movePoint = plot ()->movePointsActivated ();

  switch (e->type ())
    {
    case QEvent::FocusIn:
      {
	if (plot ()->enabledCursor ())
	  plot ()->showCursor (TRUE);
	return TRUE;
      }
      break;

    case QEvent::FocusOut:
      {
	if (plot ()->enabledCursor ())
	  plot ()->showCursor (TRUE);
	// commented out by S.Yamamoto
	//plotWidget->titleLabel ()->repaint ();
	//plotWidget->replot ();
	return TRUE;
      }
      break;

    case QEvent::Paint:
      {
	// The inPaint guard protecs from producing endless paint events.
	static bool inPaint = FALSE;
	if (plotWidget->canvas ()->hasFocus () && !inPaint)
	  {
	    const QPaintEvent *pe = (const QPaintEvent *) e;
	    inPaint = TRUE;
	    plotWidget->canvas ()->repaint (pe->rect ().x (),
					    pe->rect ().y (),
					    pe->rect ().width (),
					    pe->rect ().height (),
					    pe->erased ());
	    inPaint = FALSE;
	    return TRUE;
	  }
	break;
      }

    case QEvent::MouseButtonPress:
      {
	const QMouseEvent *me = (const QMouseEvent *) e;
	presspos = me->pos ();

	bool allAxisDisabled = true;
	for (int i = 0; i < QwtPlot::axisCnt; i++)
	  {
	    if (plotWidget->axisEnabled (i))
	      {
		allAxisDisabled = FALSE;
		break;
	      }
	  }
	if (allAxisDisabled && plotWidget->margin () < 2
	    && plotWidget->lineWidth () < 2)
	  {
	    QRect r = plotWidget->canvas ()->rect ();
	    r.addCoords (2, 2, -2, -2);
	    if (!r.contains (presspos))
	      emit highlightGraph ();
	  }

	emit selectPlot ();

	xMouse = presspos.x ();
	yMouse = presspos.y ();

	bool drawText = plot ()->drawTextActive ();
	bool select = false;
	//first perform all other operations than marker selection
	if (!drawText && !movePoint &&
	    !dataCursorEnabled && !pickerActivated &&
	    !moveRangeSelector && !removePoint &&
	    !plot()->drawLineActive() &&
	    !plot ()->subtractLineActive() )
	  select = selectMarker (presspos);

	if (!select)
	  plot ()->deselectMarker();

	if (me->button () == QMouseEvent::RightButton && select)
	  emit showMarkerPopupMenu ();
	else if (removePoint || movePoint ||
	    (plot ()->translationInProgress () && !pickerActivated))
	  pointSelected = plot ()->selectPoint (presspos);
	else if (dataCursorEnabled)
	  plot ()->selectCurve (presspos);
	else if (plot ()->pickerActivated ())
	  {
	    plot ()->movedPicker (presspos, true);
	    if (plot ()->selectPeaksOn ())
	      pointSelected = plot ()->selectPoint (presspos);
	  }
 	else if (me->button () == QMouseEvent::LeftButton
	    && (plot ()->drawLineActive () 
		|| plot ()->lineProfile () 
		|| plot ()->subtractLineActive()))
	  {
	    startLinePoint = presspos;
	    plot ()->copyCanvas (TRUE);
	  }
	else if (me->button () == QMouseEvent::LeftButton && drawText)
	  drawTextMarker (presspos);
	else if (moveRangeSelector)
	  { //moves quickly the active selector at the mouse selected point
	    plot ()->selectPoint (presspos);
	    plot ()->moveRangeSelector(0);
	  }

	return TRUE;
      }
      break;

    case QEvent::MouseButtonDblClick:
      {
	const QMouseEvent *me = (const QMouseEvent *) e;
	presspos = me->pos ();

	if (plot ()->translationInProgress () && pointSelected)
	  {
	    if (!pickerActivated)
	      plot ()->startCurveTranslation ();
	    else
	      plot ()->translateCurveTo (presspos);
	    return true;
	  }

	if (plot ()->selectPeaksOn () && pointSelected && pickerActivated)
	  {
	    plot ()->selectPeak (presspos);
	    return TRUE;
	  }

	if (movePoint || moveRangeSelector || pickerActivated
	    || dataCursorEnabled)
	  return false;

	if (removePoint)
	  {
	    plot ()->removePoint ();
	    return TRUE;
	  }

	int selectedMarker = plot ()->selectedMarkerKey ();
	if (selectedMarker < 0)
	  {
	    if (plot ()->isPiePlot ())
	      emit showPieDialog ();
	    else
	    {
	      const QMouseEvent *me = (const QMouseEvent *) e;
	      int dist, point;
	      int curveKey =
		plotWidget->closestCurve (presspos.x (), presspos.y (),
					  dist, point);
	      int index = plot()->curveIndex(curveKey);
	      if (dist < 10)
		emit showPlotDialog (index);
	      else
	    if (plot ()->curves () > 0)
	      emit showPlotDialog (0);
	    }
	    return TRUE;
	  }
	else
	  {
	    if (texts.contains (selectedMarker) > 0)
	      {
		emit viewTextDialog ();
		return TRUE;
	      }
	    if (lines.contains (selectedMarker) > 0)
	      {
		emit viewLineDialog ();
		return TRUE;
	      }
	    if (images.contains (selectedMarker) > 0)
	      {
		emit viewImageDialog ();
		return TRUE;
	      }
	  }
      }
      break;

    case QEvent::MouseMove:
      {
	const QMouseEvent *me = (const QMouseEvent *) e;
	if (me->state () != QMouseEvent::LeftButton)
	  return true;

	QPoint pos = me->pos ();

	if (plot ()->zoomOn () || removePoint || moveRangeSelector
	    || dataCursorEnabled )
	  //  || (presspos - pos).manhattanLength () <=
       	  //  QApplication::startDragDistance ())
	  return true;

	int selectedMarker = plot ()->selectedMarkerKey ();

	if (plot ()->drawLineActive ())
	  drawLineMarker (pos, plot ()->drawArrow ());
	else if (plot ()->lineProfile () || resizeLineFromStart
		 || resizeLineFromEnd)
	  drawLineMarker (pos, false);
	else if (plot ()->subtractLineActive())
	  drawLineMarker (pos, false);
	else if (plot ()->movePointsActivated ())
	  plot ()->move (pos);
	else if (plot ()->pickerActivated ())
	  plot ()->movedPicker (pos, true);
	else if (selectedMarker >= 0)
	  moveMarker (pos);
	else
	  {
	    //plotWidget->canvas ()->setCursor (Qt::PointingHandCursor);
	    movedGraph = true;
	    emit moveGraph (pos);
	  }
	return TRUE;
      }
      break;

    case QEvent::MouseButtonRelease:
      {
	const QMouseEvent *me = (const QMouseEvent *) e;
	presspos = me->pos ();

	Graph *g = plot ();
	if (moved && !g->drawLineActive () && !g->lineProfile ())
	  releaseMarker ();
	else if (resizeLineFromStart || resizeLineFromEnd)
	  resizeLineMarker (presspos);
	else if (g->drawLineActive ())
	  {
	    emit showInfo("");
	    LineMarker mrk;
	    mrk.attach (g->plotWidget ());
	    mrk.setStartPoint (startLinePoint);
	    mrk.setEndPoint (QPoint (presspos.x (), presspos.y ()));
	    mrk.setColor (g->arrowDefaultColor ());
	    mrk.setWidth (g->arrowDefaultWidth ());
	    mrk.setStyle (g->arrowLineDefaultStyle ());
	    mrk.setHeadLength (g->arrowHeadDefaultLength ());
	    mrk.setHeadAngle (g->arrowHeadDefaultAngle ());
	    mrk.fillArrowHead (g->arrowHeadDefaultFill ());
	    mrk.drawEndArrow (g->drawArrow ());
	    mrk.drawStartArrow (false);

	    g->insertLineMarker (&mrk);
	    g->drawLine (false);

	    mrk.detach ();
	    plotWidget->replot ();
	  }
	else if (plot ()->lineProfile ())
	  {
	    emit showInfo("");
	    QPoint endLinePoint = QPoint (presspos.x (), presspos.y ());
	    if (endLinePoint == startLinePoint)
	      return FALSE;
	    else
	      {
		LineMarker mrk;
		mrk.attach (g->plotWidget ());
		mrk.setStartPoint (startLinePoint);
		mrk.setEndPoint (endLinePoint);
		mrk.setColor (Qt::red);
		mrk.setWidth (1);
		Qt::PenStyle style = Qt::SolidLine;
		mrk.setStyle (style);
		mrk.drawEndArrow (false);
		mrk.drawStartArrow (false);
		mrk.detach ();
		plot ()->insertLineMarker (&mrk);
		plotWidget->replot ();
		emit calculateProfile (startLinePoint, endLinePoint);
	      }

	  }
	else if (plot()->subtractLineActive())
	  {
	    QPoint endLinePoint = QPoint (presspos.x(), presspos.y());
	    if (endLinePoint == startLinePoint)
	      return FALSE;
	    else
	      plot ()->subtractLine(startLinePoint,endLinePoint);
	    emit showInfo("");
	  }
	else if (movedGraph)
	  {
	    plotWidget->canvas ()->setCursor (Qt::arrowCursor);
	    emit releasedGraph ();
	    movedGraph = false;
	  }

	return TRUE;
      }
      break;

    case QEvent::KeyPress:
      {
	const int delta = 5;
	int key = ((const QKeyEvent *) e)->key ();

	if ((key == Qt::Key_Enter || key == Qt::Key_Return) &&
	    plot ()->translationInProgress () && pointSelected)
	  {
	    if (!pickerActivated)
	      plot ()->startCurveTranslation ();
	    else
	      plot ()->translateCurveTo (plotWidget->canvas ()->
					 mapFromGlobal (QCursor::pos ()));
	    return true;
	  }

	if ((key == Qt::Key_Enter || key == Qt::Key_Return) &&
	    plot ()->selectPeaksOn () && pointSelected && pickerActivated)
	  {
	    plot ()->selectPeak (plotWidget->canvas ()->
				 mapFromGlobal (QCursor::pos ()));
	    return TRUE;
	  }

	if (key == Qt::Key_Tab)
	  {
	    plot ()->selectNextMarker ();
	    return true;
	  }

	if (plot ()->enabledCursor ())
	  {
	    switch (key)
	      {
	      case Qt::Key_Up:
		plot ()->shiftCurveCursor (1);
		return TRUE;

	      case Qt::Key_Down:
		plot ()->shiftCurveCursor (-1);
		return TRUE;

	      case Qt::Key_Right:
	      case Qt::Key_Plus:
		if (plot ()->selectedCurveID () < 0)
		  plot ()->shiftCurveCursor (1);
		else
		  plot ()->shiftPointCursor (1);
		return TRUE;

	      case Qt::Key_Left:
	      case Qt::Key_Minus:
		if (plot ()->selectedCurveID () < 0)
		  plot ()->shiftCurveCursor (1);
		else
		  plot ()->shiftPointCursor (-1);

		return TRUE;
	      }
	  }

	//moving range selector                         
	if (plot ()->selectorsEnabled ())
	  {
	    switch (key)
	      {
	      case Qt::Key_Up:
		plot ()->shiftCurveSelector (1);
		return TRUE;

	      case Qt::Key_Down:
		plot ()->shiftCurveSelector (-1);
		return TRUE;

	      case Qt::Key_Right:
	      case Qt::Key_Plus:
		if (((const QKeyEvent *) e)->state () == Qt::ControlButton)
		  plot ()->moveRangeSelector (1);
		else
		  plot ()->shiftRangeSelector (1);
		return TRUE;

	      case Qt::Key_Left:
	      case Qt::Key_Minus:
		if (((const QKeyEvent *) e)->state () == Qt::ControlButton)
		  plot ()->moveRangeSelector (-1);
		else
		  plot ()->shiftRangeSelector (1);

		return TRUE;
	      }
	  }

	// The following keys represent a direction, they are
	// organized on the keyboard.

	if (plot ()->movePointsActivated ())
	  {
	    switch (key)
	      {
	      case Key_1:
		plot ()->moveBy (-delta, delta);
		break;
	      case Key_2:
		plot ()->moveBy (0, delta);
		break;
	      case Key_3:
		plot ()->moveBy (delta, delta);
		break;
	      case Key_4:
		plot ()->moveBy (-delta, 0);
		break;
	      case Key_6:
		plot ()->moveBy (delta, 0);
		break;
	      case Key_7:
		plot ()->moveBy (-delta, -delta);
		break;
	      case Key_8:
		plot ()->moveBy (0, -delta);
		break;
	      case Key_9:
		plot ()->moveBy (delta, -delta);
		break;
	      }
	  }

	int selectedMarker = plot ()->selectedMarkerKey ();
	if (selectedMarker >= 0)
	  {
	    int delta = 1;
	    switch (key)
	      {
	      case Key_Left:
		plot ()->moveMarkerBy (-delta, 0);
		break;
	      case Key_Right:
		plot ()->moveMarkerBy (delta, 0);
		break;
	      case Key_Up:
		plot ()->moveMarkerBy (0, -delta);
		break;
	      case Key_Down:
		plot ()->moveMarkerBy (0, delta);
		break;
	      }
	  }

	if (texts.contains (selectedMarker) > 0 &&
	    (key == Qt::Key_Enter || key == Qt::Key_Return))
	  {
	    emit viewTextDialog ();
	    return TRUE;
	  }
	if (lines.contains (selectedMarker) > 0 &&
	    (key == Qt::Key_Enter || key == Qt::Key_Return))
	  {
	    emit viewLineDialog ();
	    return TRUE;
	  }
	if (images.contains (selectedMarker) > 0 &&
	    (key == Qt::Key_Enter || key == Qt::Key_Return))
	  {
	    emit viewImageDialog ();
	    return TRUE;
	  }
      }
      break;

    default:
      break;
    }
  return QObject::eventFilter (object, e);
}

void
CanvasPicker::releaseMarker ()
{
  QMemArray < int >images = plot ()->imageMarkerKeys ();
  QMemArray < int >texts = plot ()->textMarkerKeys ();
  QMemArray < int >lines = plot ()->lineMarkerKeys ();

  bool line = false, image = false;

  int selectedMarker = plot ()->selectedMarkerKey ();
  if (texts.contains (selectedMarker))
    {
      LegendMarker *mrkT =
	(LegendMarker *) plotWidget->marker (selectedMarker);
      mrkT->setOrigin (QPoint (xMrk, yMrk));
    }
  else if (images.contains (selectedMarker))
    {
      image = true;

      ImageMarker *mrk = (ImageMarker *) plotWidget->marker (selectedMarker);
      mrk->setOrigin (QPoint (xMrk, yMrk));
    }
  else if (lines.contains (selectedMarker))
    {
      line = true;

      LineMarker *mrk = (LineMarker *) plotWidget->marker (selectedMarker);
      //int clw = plotWidget->canvas ()->lineWidth ();
      int clw = 0;
      mrk->
	setStartPoint (QPoint
		       (startLinePoint.x () - clw,
			startLinePoint.y () - clw));
      mrk->
	setEndPoint (QPoint
		     (endLinePoint.x () - clw, endLinePoint.y () - clw));
    }

  plotWidget->replot ();

  if (line)
    plot ()->highlightLineMarker (selectedMarker);
  else if (image)
    plot ()->highlightImageMarker (selectedMarker);
  else
    plot ()->highlightTextMarker (selectedMarker);

  moved = FALSE;
  emit modified ();

  QApplication::restoreOverrideCursor ();
}

void
CanvasPicker::moveMarker (QPoint & point)
{
  static QRect oldRect=QRect(0,0,0,0);

  //fprintf(stderr,"CanvasPicker::moveMarker\n");
  QApplication::setOverrideCursor (QCursor (Qt::PointingHandCursor), true);

  QMemArray < int >images = plot ()->imageMarkerKeys ();
  QMemArray < int >texts = plot ()->textMarkerKeys ();
  QMemArray < int >lines = plot ()->lineMarkerKeys ();

  QPainter painter (plotWidget->canvas ());

  int w = plotWidget->canvas ()->width ();
  int h = plotWidget->canvas ()->height ();
  //fprintf(stderr, "moveMarker w,h = %d, %d\n",w,h);
  
  QPixmap pix (w, h, -1);
  pix.fill (QColor (white));
  if(!QPixmapCache::find ("field", pix)){
    fprintf(stderr,"failed to find Pixmap cache\n");
  }

  //painter.drawPixmap (0, 0, pix, 0, 0, -1, -1);
  painter.drawPixmap( oldRect.topLeft(), pix, oldRect );

  int selectedMarker = plot ()->selectedMarkerKey ();
  if (lines.contains (selectedMarker))
    {
      LineMarker *mrk = (LineMarker *) plotWidget->marker (selectedMarker);

      //int clw = plotWidget->canvas ()->lineWidth ();
      int clw = 0;
      point.setX (point.x () + clw);
      point.setY (point.y () + clw);

      QPoint aux = mrk->startPoint ();
      int x1 = aux.x () + point.x () - xMouse;
      int y1 = aux.y () + point.y () - yMouse;

      mrk->setStartPoint (QPoint (x1, y1));
      startLinePoint = QPoint (x1, y1);

      aux = mrk->endPoint ();
      int x2 = aux.x () + point.x () - xMouse;
      int y2 = aux.y () + point.y () - yMouse;
      mrk->setEndPoint (QPoint (x2, y2));
      endLinePoint = QPoint (x2, y2);

      mrk->draw (&painter, plotWidget->canvasMap (QwtPlot::xBottom),
		 plotWidget->canvasMap (QwtPlot::yLeft), QRect (0, 0, 0, 0));
      int lw = mrk->width();
      int left = QMIN(x1,x2)-lw;
      int top  = QMIN(y1,y2)-lw;
      int width  = abs(x1-x2)+lw*2+1;
      int height = abs(y1-y2)+lw*2+1;
      if(mrk->hasStartArrow() || mrk->hasEndArrow())
	{
	  left -= mrk->headLength();
	  top  -= mrk->headLength();
	  width += mrk->headLength()*2;
	  height += mrk->headLength()*2;
	}
      oldRect=QRect(left,top,width,height);
      //painter.drawRect(oldRect);
    }
  else if (images.contains (selectedMarker))
    {
      ImageMarker *mrk = (ImageMarker *) plotWidget->marker (selectedMarker);

      xMrk += point.x () - xMouse;
      yMrk += point.y () - yMouse;

      painter.setRasterOp (Qt::NotROP);
      painter.drawRect (QRect (QPoint (xMrk, yMrk), mrk->size ()));
      oldRect = QRect (QPoint (xMrk, yMrk), mrk->rect ().size ());
    }
  else if (texts.contains (selectedMarker))
    {
      xMrk += point.x () - xMouse;
      yMrk += point.y () - yMouse;

      LegendMarker *mrk =
	(LegendMarker *) plotWidget->marker (selectedMarker);
      painter.setRasterOp (Qt::NotROP);
      //painter.drawRect (QRect (QPoint (xMrk, yMrk), mrk->rect ().size ()));
      oldRect = QRect (QPoint (xMrk, yMrk), mrk->rect ().size ());
      painter.drawRect( oldRect );
    }
  xMouse = point.x ();
  yMouse = point.y ();
  moved = TRUE;

  //emit modified ();
}

void
CanvasPicker::drawTextMarker (const QPoint & point)
{
  LegendMarker mrkT (plotWidget);
  mrkT.setOrigin (point);
  mrkT.setBackground (plot ()->textMarkerDefaultFrame ());
  mrkT.setFont (plot()->defaultTextMarkerFont(), plot()->frameGeometry());
  mrkT.setTextColor (plot ()->textMarkerDefaultColor ());
  mrkT.setBackgroundColor (plot ()->textMarkerDefaultBackground ());
  mrkT.setText (tr ("enter your text here"));
  plot ()->insertTextMarker (&mrkT);
  plot ()->drawText (FALSE);
  emit drawTextOff ();
}

void
CanvasPicker::resizeLineMarker (const QPoint & point)
{
  int selectedMarker = plot ()->selectedMarkerKey ();
  LineMarker *mrk = (LineMarker *) plotWidget->marker (selectedMarker);
  if (!mrk)
    return;

  //int clw = plotWidget->canvas ()->lineWidth ();
  int clw = 0;

  if (resizeLineFromStart)
    mrk->setStartPoint (QPoint (point.x () - clw, point.y () - clw));
  else if (resizeLineFromEnd)
    mrk->setEndPoint (QPoint (point.x () - clw, point.y () - clw));

  plotWidget->replot ();
  plot ()->highlightLineMarker (selectedMarker);

  emit modified ();

  resizeLineFromStart = false;
  resizeLineFromEnd = false;

  emit showInfo("");

  QApplication::restoreOverrideCursor ();
}

void
CanvasPicker::drawLineMarker (const QPoint & point, bool endArrow)
{
  plotWidget->replot ();
  LineMarker mrk;
  mrk.attach (plotWidget);

  //int clw = plotWidget->canvas ()->lineWidth ();
  int clw = 0;

  mrk.
    setStartPoint (QPoint
		   (startLinePoint.x () + clw, startLinePoint.y () + clw));
  mrk.setEndPoint (QPoint (point.x () + clw, point.y () + clw));

  mrk.setWidth (1);
  mrk.setStyle (Qt::SolidLine);
  mrk.drawEndArrow (endArrow);
  mrk.drawStartArrow (false);

  if (plot ()->drawLineActive ())
    mrk.setColor (Qt::black);
  else
    mrk.setColor (Qt::red);

  QPainter painter (plot ()->plotWidget ()->canvas ());
  painter.save ();
  painter.setRasterOp (Qt::NotXorROP);
  mrk.draw (&painter, plotWidget->canvasMap (QwtPlot::xBottom),
	    plotWidget->canvasMap (QwtPlot::yLeft), QRect (0, 0, 0, 0));

  const QwtScaleMap & xMap = plotWidget->canvasMap (mrk.xAxis ());
  const QwtScaleMap & yMap = plotWidget->canvasMap (mrk.yAxis ());
  double x = xMap.invTransform( point.x()+clw );
  double y = yMap.invTransform( point.y()+clw );
  QString info = tr("x = ") + QString::number( x )
    + "; y = " + QString::number( y ) 
    + "; ( dx = " + QString::number( mrk.dx() )
    + "; dy = " + QString::number( mrk.dy() )
    + "; slope = "; 
  bool inf;
  double slope = mrk.slope(&inf);
  if (inf)
    info += "inf";
  else
    info += QString::number(slope);

  info += " )";

  emit showInfo( info );

  painter.restore ();
  mrk.detach ();

}

// Selects and highlights the marker 
bool
CanvasPicker::selectMarker (const QPoint & point)
{
  if (plot ()->zoomOn ())
    return false;

  QMemArray < int >images = plot ()->imageMarkerKeys ();
  QMemArray < int >texts = plot ()->textMarkerKeys ();
  QMemArray < int >lines = plot ()->lineMarkerKeys ();

  plotWidget->replot ();
  plot ()->copyCanvas (true);

  //
  // point is on the selected Marker?
  //
  int selected = plot()->selectedMarkerKey();
  //qDebug("selected marker = %d",selected);
  if(selected>=0)
    {
      if (lines.contains(selected))
	{
	  LineMarker *mrkL = (LineMarker *) plotWidget->marker (selected);
	  if (mrkL)
	    {
	      int d = mrkL->width () +
		(int) floor (mrkL->headLength () *
			     tan (M_PI * mrkL->headAngle () / 180.0) + 0.5);
	      double dist = mrkL->dist (point.x (), point.y ());
	      if (dist <= d)
		{
		  plot ()->highlightLineMarker (selected);
      int lw = mrkL->width();
		  QRect sr = QRect (QPoint (0, 0), QSize (7+lw, 7+lw));
		  sr.moveCenter (mrkL->startPoint ());
		  if (sr.contains (point))
		    {
		      //QApplication::setOverrideCursor (QCursor (Qt::SizeAllCursor), true);
		      QApplication::setOverrideCursor (QCursor (Qt::BlankCursor), true);
		      resizeLineFromStart = true;
		      startLinePoint = mrkL->endPoint ();
		      //emit showDisplayBar();
		      return TRUE;
		    }
		  QRect er = sr;
		  er.moveCenter (mrkL->endPoint ());
		  if (er.contains (point))
		    {
		      //QApplication::setOverrideCursor (QCursor (Qt::SizeAllCursor), true);
		      QApplication::setOverrideCursor (QCursor (Qt::BlankCursor), true);
		      resizeLineFromEnd = true;
		      startLinePoint = mrkL->startPoint ();
		      //emit showDisplayBar();
		    }
		  return true;
		}
	    }
	}
      else if (texts.contains(selected))
	{
	  LegendMarker *mrk = (LegendMarker *) plotWidget->marker (selected);
	  QRect mRect = mrk->rect ();
	  if (mRect.contains (point))
	    {
	      plot ()->highlightTextMarker (selected);
	      QPoint origin = mRect.topLeft ();
	      xMrk = origin.x ();
	      yMrk = origin.y ();
	      return true;
	    }
	}
      else if (images.contains(selected))
	{
	  ImageMarker *mrkI = (ImageMarker *) plotWidget->marker (selected);
	  if (mrkI->rect ().contains (point))
	    {
	      plot ()->highlightImageMarker (selected);
	      QPoint origin = mrkI->getOrigin ();
	      xMrk = origin.x ();
	      yMrk = origin.y ();
	      return true ;
	    }
	}
    }
  //
  // otherwise select another marker
  //
  int i;
  for (i = (int)lines.size()-1; i>=0; i--)
    { //line marker selected(first)? 
      LineMarker *mrkL = (LineMarker *) plotWidget->marker (lines[i]);
      if (mrkL)
	{
	  int d =
	    mrkL->width () +
	    (int) floor (mrkL->headLength () *
			 tan (M_PI * mrkL->headAngle () / 180.0) + 0.5);
	  double dist = mrkL->dist (point.x (), point.y ());
	  if (dist <= d)
	    {
	      plot ()->highlightLineMarker (lines[i]);
	      return TRUE;
	    }
	}
    }

  for (i = (int)texts.size()-1; i>=0; i--)
    { //text marker selected(second)?        
      LegendMarker *mrk = (LegendMarker *) plotWidget->marker (texts[i]);
      QRect mRect = mrk->rect ();
      if (mRect.contains (point))
	{
	  plot ()->highlightTextMarker (texts[i]);
	  QPoint origin = mRect.topLeft ();
	  xMrk = origin.x ();
	  yMrk = origin.y ();
	  return TRUE;
	}
    }

  for (i = (int)images.size()-1; i>= 0; i--)
    { //image marker selected(last)? 
      ImageMarker *mrkI = (ImageMarker *) plotWidget->marker (images[i]);
      if (mrkI->rect ().contains (point))
	{
	  plot ()->highlightImageMarker (images[i]);
	  QPoint origin = mrkI->getOrigin ();
	  xMrk = origin.x ();
	  yMrk = origin.y ();
	  return TRUE;
	}
    }

  return FALSE;
}
