/***************************************************************************
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor,                    *
 *   Boston, MA  02110-1301  USA                                           *
 *                                                                         *
 ***************************************************************************/
#include "cascadeDialog.h"
#include "plot.h"
#include "graph.h"
#include "parser.h"
#include "application.h"
#include "pixmaps.h"

#include <qpushbutton.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qbuttongroup.h>
#include <qlineedit.h>
#include <qmessagebox.h>
#include <qcombobox.h>
#include <qapplication.h>

#include <qwt_plot.h>
#include <qwt_scale_engine.h>

cascadeDialog::cascadeDialog(QWidget *parent, const char *name, bool modal,
			     WFlags fl):
  QDialog(parent, name, modal, fl)
{
  if (!name)
    setName ("cascadeDialog");
  setCaption (tr ("QtiPlot - Cascade Graph"));

  QButtonGroup *GroupBox1 =
    new QButtonGroup (2, QGroupBox::Horizontal, tr (""), this, "GroupBox1");
  
  new QLabel (tr ("X Step"), GroupBox1, "TextLabel1", 0);
  boxXstep = new QLineEdit (GroupBox1, "boxXstep");
  boxXstep->setText (tr ("0"));

  new QLabel (tr ("Y Step"), GroupBox1, "TextLabel2", 0);
  boxYstep = new QLineEdit (GroupBox1, "boxYstep");
  boxYstep->setText (tr ("0"));

  boxInverse = new QCheckBox (tr ("Inverse order"), GroupBox1, 0);

  QButtonGroup *GroupBox2 =
    new QButtonGroup (1, QGroupBox::Horizontal, tr (""), this, "GroupBox2");
  GroupBox2->setFlat (TRUE);
  GroupBox2->setLineWidth (0);

  buttonApply = new QPushButton (GroupBox2, "buttonApply");
  buttonApply->setAutoDefault (TRUE);
  buttonApply->setDefault (TRUE);
  buttonApply->setText (tr ("&Apply"));

  buttonClose = new QPushButton (GroupBox2, "buttonClose");
  buttonClose->setAutoDefault (TRUE);
  buttonClose->setText (tr ("&Close"));

  QHBoxLayout *hlayout = new QHBoxLayout (this, 5, 5, "hlayout");
  hlayout->addWidget (GroupBox1);
  hlayout->addWidget (GroupBox2);

  // signals and slots connections
  connect (buttonApply, SIGNAL (clicked ()), this, SLOT (accept()));
  connect (buttonClose, SIGNAL (clicked ()), this, SLOT (close()));

  firsttime = true;
  xlogscale = false;
  ylogscale = false;
  d_graph = 0;
  cascadeTable = 0;
  cascadePlot = 0;
  curves.clear();
}

cascadeDialog::~cascadeDialog()
{
}

void
cascadeDialog::setGraph(Graph *g, const QString &name)
{
  d_graph = g;
  graphname = name;

  connect (d_graph, SIGNAL (closedGraph ()), this, SLOT (close ()));
}

void
cascadeDialog::accept()
{
  double xstep,ystep;

  try
  {
    myParser parser;
    parser.SetExpr (boxXstep->text().ascii());
    xstep = parser.Eval ();
  }
  catch (mu::ParserError & e)
  {
    QMessageBox::critical (this, tr ("QtiPlot - X step value error"),
                           e.GetMsg ().c_str());
    boxXstep->setFocus ();
    return;
  }

  try
  {
    myParser parser;
    parser.SetExpr (boxYstep->text().ascii());
    ystep = parser.Eval ();
  }
  catch (mu::ParserError & e)
  {
    QMessageBox::critical (this, tr ("QtiPlot - Y step value error"),
                           e.GetMsg ().c_str());
    boxYstep->setFocus ();
    return;
  }

  ApplicationWindow *app = (ApplicationWindow *)this->parent();

  int prec = app->tbl_output_precision;

  QApplication::setOverrideCursor (waitCursor);

  if(firsttime)
    {
      int maxrows = 0;
      for (int i = 0; i < d_graph->curves(); i++)
	{
	  QwtPlotCurve *c = d_graph->curve(i);
	  if (c && c->rtti () != QwtPlotItem::Rtti_PlotSpectrogram
	      && d_graph->curveType(i) != Graph::XErrorBars
	      && d_graph->curveType(i) != Graph::YErrorBars)
	    {
	      curves << i;
	      if (maxrows < c->dataSize()) maxrows = c->dataSize();
	      if(curves.count()==1)
		{
		  Plot *plt=d_graph->plotWidget();
		  QwtScaleEngine *xse = plt->axisScaleEngine(c->xAxis());
		  xlogscale = (xse->transformation()->type()
			       == QwtScaleTransformation::Log10);
		  QwtScaleEngine *yse = plt->axisScaleEngine(c->yAxis());
		  ylogscale = (yse->transformation()->type()
			       == QwtScaleTransformation::Log10);	  
		}
	    }
	}
      QString tname = app->generateUniqueName("Cascade");
      QString legend = tr("Cascade data of ") + graphname;
      cascadeTable = app->newHiddenTable( tname, legend, maxrows,
					  curves.count()*2, prec);
      for(int i=0;i<(int)curves.count();i++)
	{
	  cascadeTable->setColPlotDesignation(i*2,  Table::X);
	  cascadeTable->setColPlotDesignation(i*2+1,Table::Y);
	}
      cascadeTable->setHeaderColType();
    }
  int nc = (int)curves.count();
  for (int i = 0; i < nc; i++)
    {
      QwtPlotCurve *c = d_graph->curve(curves[i]);
      for (int r = 0; r < c->dataSize(); r++)
	{
	  int order = i;
	  if(boxInverse->isChecked())
	    order = nc - 1 - i;

	  double xv,yv;
	  if(!xlogscale)
	    xv = c->x(r) + xstep * order;
	  else if(xstep>0)
	    xv = c->x(r) * pow( xstep, order );
	  else
	    xv = c->x(r);

	  if(!ylogscale)
	    yv = c->y(r) + ystep * order;
	  else if(ystep>0)
	    yv = c->y(r) * pow( ystep, order );
	  else
	    yv = c->y(r);

	  cascadeTable->setText( r, i*2, QString::number(xv,'g',prec) );
	  cascadeTable->setText( r, i*2+1, QString::number(yv,'g',prec) );
	}
    }
  if(firsttime)
    {
      cascadePlot = app->multilayerPlot(cascadeTable,
					cascadeTable->YColumns(),
					Graph::Line );
      ((myWidget *)cascadePlot)->setWindowLabel(tr("Cascad curves of ")+graphname);
      Graph *ag = app->activeGraph;
      for(int i = 0; i < nc; i++)
	{
	  QwtPlotCurve *c = d_graph->curve(curves[i]);
	  if(xlogscale)
	    ag->setScale( QwtPlot::xBottom, 1, 10, 1, 10, 10, 1, false );
	  if(ylogscale)
	    ag->setScale( QwtPlot::yLeft, 1, 10, 1, 10, 10, 1, false );
	  ag->setCurvePen(i, c->pen());
	  ag->setCurveSymbol(i, c->symbol());
          int style = d_graph->curveType(curves[i]);
	  if(style <= Graph::LineSymbols)
	    ag->setCurveType(i,style);
          if(style == Graph::Scatter)
            ag->setCurveStyle(i,QwtPlotCurve::NoCurve);
	}
      ag->setAutoScale();
      ag->replot();
      firsttime = false;
    }
  else
    cascadeTable->notifyChanges();

  QApplication::restoreOverrideCursor ();

}
