/***************************************************************************
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor,                    *
 *   Boston, MA  02110-1301  USA                                           *
 *                                                                         *
 ***************************************************************************/
#include <math.h>
#include <stddef.h>

#include <qmessagebox.h>

#include <gsl/gsl_blas.h>
#include <gsl/gsl_math.h>
#include "fit.h"
#include "parser.h"

int
expd3_f (const gsl_vector * x, void *params, gsl_vector * f)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  double A1 = gsl_vector_get (x, 0);
  double t1 = gsl_vector_get (x, 1);
  double A2 = gsl_vector_get (x, 2);
  double t2 = gsl_vector_get (x, 3);
  double A3 = gsl_vector_get (x, 4);
  double t3 = gsl_vector_get (x, 5);

  size_t i;
  for (i = 0; i < n; i++)
    {
      double Yi =
	A1 * exp (-X[i] * t1) + A2 * exp (-X[i] * t2) + A3 * exp (-X[i] * t3);
      gsl_vector_set (f, i, (Yi - Y[i]) * sqrt(weight[i]));
    }

  return GSL_SUCCESS;
}

double
expd3_d (const gsl_vector * x, void *params)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  double A1 = gsl_vector_get (x, 0);
  double t1 = gsl_vector_get (x, 1);
  double A2 = gsl_vector_get (x, 2);
  double t2 = gsl_vector_get (x, 3);
  double A3 = gsl_vector_get (x, 4);
  double t3 = gsl_vector_get (x, 5);

  size_t i;
  double val = 0;
  for (i = 0; i < n; i++)
    {
      double dYi =
	A1 * exp (-X[i] * t1) +	A2 * exp (-X[i] * t2) +	A3 * exp (-X[i] * t3)
	- Y[i];
      val += dYi * dYi * weight[i];
    }

  return val;
}

int
expd3_df (const gsl_vector * x, void *params, gsl_matrix * J)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *weight = ((struct fitData *) params)->weight;

  double A1 = gsl_vector_get (x, 0);
  double l1 = gsl_vector_get (x, 1);
  double A2 = gsl_vector_get (x, 2);
  double l2 = gsl_vector_get (x, 3);
  double A3 = gsl_vector_get (x, 4);
  double l3 = gsl_vector_get (x, 5);

  size_t i;
  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sigma[i],      */
      /*       Yi = A1 * exp(-xi*l1) + A2 * exp(-xi*l2) */
      /* and the xj are the parameters (A1,l1,A2,l2) */
      double t = X[i];
      double wt = sqrt(weight[i]);
      double e1 = exp (-t * l1) * wt;
      double e2 = exp (-t * l2) * wt;
      double e3 = exp (-t * l3) * wt;

      gsl_matrix_set (J, i, 0, e1);
      gsl_matrix_set (J, i, 1, -t * A1 * e1);
      gsl_matrix_set (J, i, 2, e2);
      gsl_matrix_set (J, i, 3, -t * A2 * e2);
      gsl_matrix_set (J, i, 4, e3);
      gsl_matrix_set (J, i, 5, -t * A3 * e3);
    }
  return GSL_SUCCESS;
}

int
expd3_fdf (const gsl_vector * x, void *params, gsl_vector * f, gsl_matrix * J)
{
  expd3_f (x, params, f);
  expd3_df (x, params, J);

  return GSL_SUCCESS;
}

int
expd2_f (const gsl_vector * x, void *params, gsl_vector * f)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  double A1 = gsl_vector_get (x, 0);
  double t1 = gsl_vector_get (x, 1);
  double A2 = gsl_vector_get (x, 2);
  double t2 = gsl_vector_get (x, 3);

  size_t i;
  for (i = 0; i < n; i++)
    {
      double Yi = A1 * exp (-X[i] * t1) + A2 * exp (-X[i] * t2);
      gsl_vector_set (f, i, (Yi - Y[i]) * sqrt( weight[i] ) );
    }

  return GSL_SUCCESS;
}

double
expd2_d (const gsl_vector * x, void *params)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  double A1 = gsl_vector_get (x, 0);
  double t1 = gsl_vector_get (x, 1);
  double A2 = gsl_vector_get (x, 2);
  double t2 = gsl_vector_get (x, 3);

  size_t i;
  double val = 0;
  for (i = 0; i < n; i++)
    {
      double dYi = A1 * exp (-X[i] * t1) + A2 * exp (-X[i] * t2) - Y[i];
      val += dYi * dYi * weight[i]; 
    }

  return val;
}

int
expd2_df (const gsl_vector * x, void *params, gsl_matrix * J)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *weight = ((struct fitData *) params)->weight;

  double A1 = gsl_vector_get (x, 0);
  double l1 = gsl_vector_get (x, 1);
  double A2 = gsl_vector_get (x, 2);
  double l2 = gsl_vector_get (x, 3);

  size_t i;

  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sigma[i],      */
      /*       Yi = A1 * exp(-xi*l1) + A2 * exp(-xi*l2) +y0  */
      /* and the xj are the parameters (A1,l1,A2,l2,y0) */
      double wt = sqrt( weight[i] );
      double t = X[i];
      double e1 = exp (-t * l1) * wt;
      double e2 = exp (-t * l2) * wt;

      gsl_matrix_set (J, i, 0, e1);
      gsl_matrix_set (J, i, 1, -t * A1 * e1);
      gsl_matrix_set (J, i, 2, e2);
      gsl_matrix_set (J, i, 3, -t * A2 * e2);
    }
  return GSL_SUCCESS;
}

int
expd2_fdf (const gsl_vector * x, void *params, gsl_vector * f, gsl_matrix * J)
{
  expd2_f (x, params, f);
  expd2_df (x, params, J);
  return GSL_SUCCESS;
}

int
exp_f (const gsl_vector * x, void *params, gsl_vector * f)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  double A = gsl_vector_get (x, 0);
  double lambda = gsl_vector_get (x, 1);
  size_t i;
  for (i = 0; i < n; i++)
    {
      double Yi = A * exp (-lambda * X[i]);
      gsl_vector_set (f, i, (Yi - Y[i]) * sqrt( weight[i] ) );
    }
  return GSL_SUCCESS;
}

double
exp_d (const gsl_vector * x, void *params)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  double A = gsl_vector_get (x, 0);
  double lambda = gsl_vector_get (x, 1);
  size_t i;
  double val = 0;
  for (i = 0; i < n; i++)
    {
      double dYi = A * exp (-lambda * X[i]) - Y[i];
      val += dYi * dYi * weight[i];
    }
  return val;
}

int
exp_df (const gsl_vector * x, void *params, gsl_matrix * J)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *weight = ((struct fitData *) params)->weight;

  double A = gsl_vector_get (x, 0);
  double lambda = gsl_vector_get (x, 1);
  size_t i;
  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sigma[i],      */
      /*       Yi = A * exp(-lambda * x[i]) + b  */
      /* and the xj are the parameters (A,lambda,b) */

      double t = X[i];
      double wt = sqrt( weight[i] );
      double e = exp (-lambda * t) * wt;
      gsl_matrix_set (J, i, 0, e );
      gsl_matrix_set (J, i, 1, -t * A * e );
    }
  return GSL_SUCCESS;
}

int
exp_fdf (const gsl_vector * x, void *params, gsl_vector * f, gsl_matrix * J)
{
  exp_f (x, params, f);
  exp_df (x, params, J);
  return GSL_SUCCESS;
}

int
gauss_f (const gsl_vector * x, void *params, gsl_vector * f)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  double A = gsl_vector_get (x, 0);
  double C = gsl_vector_get (x, 1);
  double w = gsl_vector_get (x, 2);

  size_t i;

  for (i = 0; i < n; i++)
    {
      double diff = X[i] - C;
      double Yi = A * exp (-0.5 * diff * diff / (w * w));
      gsl_vector_set (f, i, (Yi - Y[i]) * sqrt( weight[i] ) );
    }
  return GSL_SUCCESS;
}

double
gauss_d (const gsl_vector * x, void *params)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  double A = gsl_vector_get (x, 0);
  double C = gsl_vector_get (x, 1);
  double w = gsl_vector_get (x, 2);

  size_t i;
  double val = 0;

  for (i = 0; i < n; i++)
    {
      double diff = X[i] - C;
      double dYi = A * exp (-0.5 * diff * diff / (w * w)) - Y[i];
      val += dYi * dYi * weight[i];
    }
  return val;
}

int
gauss_df (const gsl_vector * x, void *params, gsl_matrix * J)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *weight = ((struct fitData *) params)->weight;

  double A = gsl_vector_get (x, 0);
  double C = gsl_vector_get (x, 1);
  double w = gsl_vector_get (x, 2);

  size_t i;
  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj,     */
      /* where fi = Yi - yi,                                    */
      /* Yi = y=A*exp[-(Xi-xc)^2/(2*w*w)]+B             */
      /* and the xj are the parameters (B,A,C,w) */

      double wt = sqrt( weight[i] );
      double diff = X[i] - C;
      double e = exp (-0.5 * diff * diff / (w * w)) * wt;

      gsl_matrix_set (J, i, 0, e);
      gsl_matrix_set (J, i, 1, diff * A * e / (w * w));
      gsl_matrix_set (J, i, 2, diff * diff * A * e / (w * w * w));
    }
  return GSL_SUCCESS;
}

int
gauss_fdf (const gsl_vector * x, void *params, gsl_vector * f, gsl_matrix * J)
{
  gauss_f (x, params, f);
  gauss_df (x, params, J);

  return GSL_SUCCESS;
}

int
gauss_multi_peak_f (const gsl_vector * x, void *params, gsl_vector * f)
{
  size_t n = ((struct fitData *) params)->n;
  size_t p = ((struct fitData *) params)->p;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  size_t peaks = p / 3;
  QMemArray <double> a(peaks), xc(peaks), w2(peaks);

  size_t i, j;
  for (i = 0; i < peaks; i++)
    {
      xc[i] = gsl_vector_get (x, 3 * i + 1);
      double wi = gsl_vector_get (x, 3 * i + 2);
      a[i] = sqrt (M_2_PI) * gsl_vector_get (x, 3 * i) / wi;
      w2[i] = wi * wi;
    }
  for (i = 0; i < n; i++)
    {
      double res = 0;
      for (j = 0; j < peaks; j++)
	{
	  double diff = X[i] - xc[j];
	  res += a[j] * exp (-2 * diff * diff / w2[j]);
	}
      gsl_vector_set (f, i, (res - Y[i]) * sqrt( weight[i] ) );
    }

  return GSL_SUCCESS;
}

double
gauss_multi_peak_d (const gsl_vector * x, void *params)
{
  size_t n = ((struct fitData *) params)->n;
  size_t p = ((struct fitData *) params)->p;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  size_t peaks = p / 3;
  QMemArray <double> a(peaks), xc(peaks), w2(peaks);

  size_t i, j;
  double val = 0;
  for (i = 0; i < peaks; i++)
    {
      xc[i] = gsl_vector_get (x, 3 * i + 1);
      double wi = gsl_vector_get (x, 3 * i + 2);
      a[i] = sqrt (M_2_PI) * gsl_vector_get (x, 3 * i) / wi;
      w2[i] = wi * wi;
    }
  double t;
  for (i = 0; i < n; i++)
    {
      double res = 0;
      for (j = 0; j < peaks; j++)
	{
	  double diff = X[i] - xc[j];
	  res += a[j] * exp (-2 * diff * diff / w2[j]);
	}
      t = res - Y[i];
      val += t * t * weight[i];
    }

  return val;
}

int
gauss_multi_peak_df (const gsl_vector * x, void *params, gsl_matrix * J)
{
  size_t n = ((struct fitData *) params)->n;
  size_t p = ((struct fitData *) params)->p;
  double *X = ((struct fitData *) params)->X;
  double *weight = ((struct fitData *) params)->weight;

  size_t peaks = p / 3;
  QMemArray <double> a(peaks), xc(peaks), w(peaks);

  size_t i, j;
  for (i = 0; i < peaks; i++)
    {
      a[i] = gsl_vector_get (x, 3 * i);
      xc[i] = gsl_vector_get (x, 3 * i + 1);
      w[i] = gsl_vector_get (x, 3 * i + 2);
    }
  for (i = 0; i < n; i++)
    {
      double wt = sqrt( weight[i] );
      for (j = 0; j < peaks; j++)
	{
	  double diff = X[i] - xc[j];
	  double w2 = w[j] * w[j];
	  double e = sqrt (M_2_PI) * wt * exp (-2 * diff * diff / w2);

	  gsl_matrix_set (J, i, 3 * j, e / w[j]);
	  gsl_matrix_set (J, i, 3 * j + 1, 4 * diff * a[j] * e / (w2 * w[j]));
	  gsl_matrix_set (J, i, 3 * j + 2,
			  a[j] / w2 * e * (4 * diff * diff / w2 - 1));
	}
    }

  return GSL_SUCCESS;
}

int
gauss_multi_peak_fdf (const gsl_vector * x, void *params, gsl_vector * f,
		      gsl_matrix * J)
{
  gauss_multi_peak_f (x, params, f);
  gauss_multi_peak_df (x, params, J);
  return GSL_SUCCESS;
}

int
lorentz_multi_peak_f (const gsl_vector * x, void *params, gsl_vector * f)
{
  size_t n = ((struct fitData *) params)->n;
  size_t p = ((struct fitData *) params)->p;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  size_t peaks = p / 3;
  QMemArray <double> a(peaks), xc(peaks), w(peaks);

  size_t i, j;
  for (i = 0; i < peaks; i++)
    {
      a[i] = gsl_vector_get (x, 3 * i);
      xc[i] = gsl_vector_get (x, 3 * i + 1);
      w[i] = gsl_vector_get (x, 3 * i + 2);
    }
  for (i = 0; i < n; i++)
    {
      double res = 0;
      for (j = 0; j < peaks; j++)
	{
	  double diff = X[i] - xc[j];
	  res += a[j] * w[j] / (4 * diff * diff + w[j] * w[j]);
	}
      gsl_vector_set (f, i, (res - Y[i]) * sqrt ( weight[i] ) );
    }

  return GSL_SUCCESS;
}

double
lorentz_multi_peak_d (const gsl_vector * x, void *params)
{
  size_t n = ((struct fitData *) params)->n;
  size_t p = ((struct fitData *) params)->p;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  size_t peaks = p / 3;
  QMemArray <double> a(peaks), xc(peaks), w(peaks);

  size_t i, j;
  double val = 0, t;
  for (i = 0; i < peaks; i++)
    {
      a[i] = gsl_vector_get (x, 3 * i);
      xc[i] = gsl_vector_get (x, 3 * i + 1);
      w[i] = gsl_vector_get (x, 3 * i + 2);
    }
  for (i = 0; i < n; i++)
    {
      double res = 0;
      for (j = 0; j < peaks; j++)
	{
	  double diff = X[i] - xc[j];
	  res += a[j] * w[j] / (4 * diff * diff + w[j] * w[j]);
	}
      t = res - Y[i];
      val += t * t * weight[i];
    }

  return GSL_SUCCESS;
}

int
lorentz_multi_peak_df (const gsl_vector * x, void *params, gsl_matrix * J)
{
  size_t n = ((struct fitData *) params)->n;
  size_t p = ((struct fitData *) params)->p;
  double *X = ((struct fitData *) params)->X;
  double *weight = ((struct fitData *) params)->weight;

  size_t peaks = p / 3;
  QMemArray <double> a(peaks), xc(peaks), w(peaks);

  size_t i, j;
  for (i = 0; i < peaks; i++)
    {
      a[i] = gsl_vector_get (x, 3 * i);
      xc[i] = gsl_vector_get (x, 3 * i + 1);
      w[i] = gsl_vector_get (x, 3 * i + 2);
    }
  for (i = 0; i < n; i++)
    {
      double wt = sqrt( weight[i] );
      for (j = 0; j < peaks; j++)
	{
	  double diff = X[i] - xc[j];
	  double w2 = w[j] * w[j];
	  double num = 1.0 / (4 * diff * diff + w2);
	  double den = 4 * diff * diff - w2;

	  gsl_matrix_set (J, i, 3 * j, w[j] * num * wt);
	  gsl_matrix_set (J, i, 3 * j + 1,
			  8 * diff * a[j] * w[j] * num * sqrt (num) * wt);
	  gsl_matrix_set (J, i, 3 * j + 2, den * a[j] * num * num * wt);
	}
    }

  return GSL_SUCCESS;
}

int
lorentz_multi_peak_fdf (const gsl_vector * x, void *params, gsl_vector * f,
			gsl_matrix * J)
{
  lorentz_multi_peak_f (x, params, f);
  lorentz_multi_peak_df (x, params, J);
  return GSL_SUCCESS;
}

int
user_f (const gsl_vector * x, void *params, gsl_vector * f)
{
  size_t n = ((struct fitData *) params)->n;
  size_t p = ((struct fitData *) params)->p;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  const char *function = ((struct fitData *) params)->function;

  QString names = (QString) ((struct fitData *) params)->names;
  QStringList parNames = QStringList::split (',', names, FALSE);

  myParser parser;
  try
  {
    QMemArray <double> parameters(p);
    double xvar;
    parser.DefineVar ("x", &xvar);
    for (int i = 0; i < (int) p; i++)
      {
	parameters[i] = gsl_vector_get (x, i);
	parser.DefineVar (parNames[i].ascii (), &parameters[i]);
      }
    parser.SetExpr (function);
    for (int j = 0; j < (int) n; j++)
      {
	xvar = X[j];
	gsl_vector_set (f, j, (parser.Eval () - Y[j]) * sqrt( weight[j] ) );
      }
  }
  catch (mu::ParserError & e)
  {
    QMessageBox::critical (0, "QtiPlot - Input function error", e.GetMsg ().c_str());
    return GSL_EINVAL;
  }
  return GSL_SUCCESS;
}

double
user_d (const gsl_vector * x, void *params)
{
  size_t n = ((struct fitData *) params)->n;
  size_t p = ((struct fitData *) params)->p;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  const char *function = ((struct fitData *) params)->function;
  QString names = (QString) ((struct fitData *) params)->names;
  QStringList parNames = QStringList::split (',', names, FALSE);

  double val = 0;
  myParser parser;
  try
  {
    QMemArray <double> parameters(p);
    double xvar;
    parser.DefineVar ("x", &xvar);
    for (int i = 0; i < (int) p; i++)
      {
	parameters[i] = gsl_vector_get (x, i);
	parser.DefineVar (parNames[i].ascii (), &parameters[i]);
      }
    parser.SetExpr (function);
    for (int j = 0; j < (int) n; j++)
      {
	xvar = X[j];
	double t = parser.Eval () - Y[j];
	val += t * t * weight[j];
      }
  }
  catch (mu::ParserError & e)
  {
    QMessageBox::critical (0, "QtiPlot - Input function error", e.GetMsg ().c_str());
    return GSL_EINVAL;
  }
  return val;
}

int
user_df (const gsl_vector * x, void *params, gsl_matrix * J)
{
  size_t n = ((struct fitData *) params)->n;
  size_t p = ((struct fitData *) params)->p;
  double *X = ((struct fitData *) params)->X;
  double *weight = ((struct fitData *) params)->weight;

  const char *function = ((struct fitData *) params)->function;
  QString names = (QString) ((struct fitData *) params)->names;
  QStringList parNames = QStringList::split (',', names, false);

  try
  {
    QMemArray <double> param(p);
    myParser parser;
    double xvar;
    parser.DefineVar ("x", &xvar);
    for (int k = 0; k < (int) p; k++)
      {
	param[k] = gsl_vector_get (x, k);
	parser.DefineVar (parNames[k].ascii(), &param[k]);
      }

    parser.SetExpr (function);

    for (int i = 0; i < (int) n; i++)
      {
	xvar = X[i];
	for (int j = 0; j < (int) p; j++)
	  gsl_matrix_set (J, i, j, 
			  parser.Diff(&param[j], param[j]) * sqrt(weight[i]) );
      }
  }
  catch (mu::ParserError &)
  {
    return GSL_EINVAL;
  }
  return GSL_SUCCESS;
}

int
user_fdf (const gsl_vector * x, void *params, gsl_vector * f, gsl_matrix * J)
{
  user_f (x, params, f);
  user_df (x, params, J);
  return GSL_SUCCESS;
}

int
boltzmann_f (const gsl_vector * x, void *params, gsl_vector * f)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  double A1 = gsl_vector_get (x, 0);
  double A2 = gsl_vector_get (x, 1);
  double x0 = gsl_vector_get (x, 2);
  double dx = gsl_vector_get (x, 3);
  size_t i;
  for (i = 0; i < n; i++)
    {
      double Yi = (A1 - A2) / (1 + exp ((X[i] - x0) / dx)) + A2;
      gsl_vector_set (f, i, (Yi - Y[i]) * sqrt( weight[i] ) );
    }

  return GSL_SUCCESS;
}

double
boltzmann_d (const gsl_vector * x, void *params)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  double A1 = gsl_vector_get (x, 0);
  double A2 = gsl_vector_get (x, 1);
  double x0 = gsl_vector_get (x, 2);
  double dx = gsl_vector_get (x, 3);
  size_t i;
  double val = 0;
  for (i = 0; i < n; i++)
    {
      double dYi = (A1 - A2) / (1 + exp ((X[i] - x0) / dx)) + A2 - Y[i];
      val += dYi * dYi * weight[i];
    }
  return val;
}

int
boltzmann_df (const gsl_vector * x, void *params, gsl_matrix * J)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *weight = ((struct fitData *) params)->weight;

  double A1 = gsl_vector_get (x, 0);
  double A2 = gsl_vector_get (x, 1);
  double x0 = gsl_vector_get (x, 2);
  double dx = gsl_vector_get (x, 3);
  size_t i;
  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj,            */
      /* where fi = Yi - yi,                                            */
      /* Yi = (A1-A2)/(1+exp((X[i]-x0)/dx)) + A2        */
      /* and the xj are the parameters (A1,A2,x0,dx) */
      double wt = sqrt( weight[i] );
      double diff = X[i] - x0;
      double e = exp (diff / dx);
      double r = 1 / (1 + e);
      double aux = wt * (A1 - A2) * e * r * r / dx;
      gsl_matrix_set (J, i, 0, r * wt);
      gsl_matrix_set (J, i, 1, (1 - r) * wt);
      gsl_matrix_set (J, i, 2, aux);
      gsl_matrix_set (J, i, 3, aux * diff / dx);
    }
  return GSL_SUCCESS;
}

int
boltzmann_fdf (const gsl_vector * x, void *params, gsl_vector * f,
	       gsl_matrix * J)
{
  boltzmann_f (x, params, f);
  boltzmann_df (x, params, J);
  return GSL_SUCCESS;
}


int sine_f(const gsl_vector * x, void *params, gsl_vector *f)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  double A = gsl_vector_get (x, 0); // amplitude
  double fr = gsl_vector_get (x, 1); // frequency
  double th = gsl_vector_get (x, 2); // phase
  double y0 = gsl_vector_get (x, 3); // offset

  size_t i;
  for (i = 0; i < n; i++)
    {
      double Yi = A * sin( 2 * M_PI * fr * X[i] + th * M_PI / 180 ) + y0;
      gsl_vector_set (f, i, (Yi - Y[i]) * sqrt( weight[i] ) );
    }

  return GSL_SUCCESS;
}

double sine_d (const gsl_vector * x, void *params)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *Y = ((struct fitData *) params)->Y;
  double *weight = ((struct fitData *) params)->weight;

  double A = gsl_vector_get (x, 0); // amplitude
  double fr = gsl_vector_get (x, 1); // frequency
  double th = gsl_vector_get (x, 2); // phase
  double y0 = gsl_vector_get (x, 3); // offset

  size_t i;
  double val = 0;
  for (i = 0; i < n; i++)
    {
      double dYi =  A * sin( 2 * M_PI * fr * X[i] + th * M_PI / 180 ) + y0 - Y[i];
      val += dYi * dYi * weight[i];
    }
  return val;
}

int sine_df (const gsl_vector * x, void *params, gsl_matrix * J)
{
  size_t n = ((struct fitData *) params)->n;
  double *X = ((struct fitData *) params)->X;
  double *weight = ((struct fitData *) params)->weight;

  double A = gsl_vector_get (x, 0); // amplitude
  double fr = gsl_vector_get (x, 1); // frequency
  double th = gsl_vector_get (x, 2); // phase
  double y0 = gsl_vector_get (x, 3); // offset

  size_t i;
  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj,            */
      /* where fi = Yi - yi,                                            */
      /* Yi = A * sin( 2 * M_PI * fr * X[i] + th * M_PI / 180 ) + y0 */
      /* and the xj are the parameters (A, fr, th, y0) */
      double wt = sqrt( weight[i] );
      double q = 2 * M_PI * fr * X[i] + th * M_PI / 180;
      double s = sin( q );
      double c = cos( q );
      gsl_matrix_set (J, i, 0, s * wt );
      gsl_matrix_set (J, i, 1, A * 2 * M_PI * X[i] * c * wt);
      gsl_matrix_set (J, i, 2, A * M_PI / 180 * c * wt);
      gsl_matrix_set (J, i, 3, wt);
    }
  return GSL_SUCCESS;
}

int sine_fdf (const gsl_vector * x, void *params, gsl_vector * f,
	       gsl_matrix * J)
{
  sine_f (x, params, f);
  sine_df (x, params, J);
  return GSL_SUCCESS;
}

