#include "fwhmDialog.h"
#include "parser.h"
#include "graph.h"
#include "application.h"

#include <qvariant.h>
#include <qpushbutton.h>
#include <qcheckbox.h>
#include <qlabel.h>
#include <qcombobox.h>
#include <qlayout.h>
#include <qbuttongroup.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qmessagebox.h>
#include <qhbuttongroup.h>

fwhmDialog::fwhmDialog (QWidget * parent, const char *name, bool modal,
		      WFlags fl):
QDialog (parent, name, modal, fl)
{
  if (!name)
    setName ("fwhmDialog");
  setCaption (tr ("QtiPlot - FWHM Calculation Options"));
  setSizeGripEnabled (true);
  setSizePolicy (QSizePolicy (QSizePolicy::Expanding, QSizePolicy::Fixed));

  ApplicationWindow *app = (ApplicationWindow *)parent;

  GroupBox1 =
    new QButtonGroup (2, QGroupBox::Horizontal, tr (""), this, "GroupBox1");

  new QLabel (tr ("Calculate FWHM of"), GroupBox1, "TextLabel1", 0);
  boxName = new QComboBox (GroupBox1, "boxShow");

  new QLabel (tr ("Lower limit"), GroupBox1, "TextLabel5", 0);
  boxStart = new QLineEdit (GroupBox1, "boxStart");

  new QLabel (tr ("Upper limit"), GroupBox1, "TextLabel6", 0);
  boxEnd = new QLineEdit (GroupBox1, "boxEnd");

  new QLabel (tr ("Peak Fraction(%)"), GroupBox1, "TextLabel7", 0);
  boxPeak = new QSpinBox (1,99,1,GroupBox1, "boxPeak");
  boxPeak->setValue(app->fwhmPeakFrac);

  new QLabel (tr ("Halfway Fraction(%)"), GroupBox1, "TextLabel8", 0);
  boxHalf = new QSpinBox (1,99,1,GroupBox1, "boxHalf");
  boxHalf->setValue(app->fwhmHalfFrac);

  GroupBox2 = new QHButtonGroup (this, "GroupBox2");
  GroupBox2->setFlat (TRUE);
  GroupBox2->setLineWidth (0);

  buttonOk = new QPushButton (GroupBox2, "buttonOk");
  buttonOk->setAutoDefault (TRUE);
  buttonOk->setDefault (TRUE);

  buttonCancel = new QPushButton (GroupBox2, "buttonCancel");
  buttonCancel->setAutoDefault (TRUE);

  buttonHelp = new QPushButton (GroupBox2, "buttonHelp");
  buttonHelp->setAutoDefault (TRUE);

  QVBoxLayout *vlayout = new QVBoxLayout (this, 5, 5, "vlayout");
  vlayout->addWidget (GroupBox1);
  vlayout->addWidget (GroupBox2);

  languageChange ();

  // signals and slots connections
  connect (buttonOk, SIGNAL (clicked ()), this, SLOT (accept ()));
  connect (buttonCancel, SIGNAL (clicked ()), this, SLOT (reject ()));
  connect (buttonHelp, SIGNAL (clicked ()), this, SLOT (help ()));
  connect (boxName, SIGNAL (activated (int)), this,
	   SLOT (activateCurve (int)));
}

fwhmDialog::~fwhmDialog ()
{
}

void
fwhmDialog::languageChange ()
{
  buttonOk->setText (tr ("&Calculate"));
  buttonCancel->setText (tr ("&Close"));
  buttonHelp->setText (tr ("&Help"));
}

void
fwhmDialog::accept ()
{
  QString curveTitle = boxName->currentText ();
  QStringList curvesList = d_graph->analyzableCurvesList ();
  if (curvesList.contains (curveTitle) <= 0)
    {
      QMessageBox::critical (this, tr ("QtiPlot - Warning"),
			     tr
			     ("The curve <b> %1 </b> doesn't exist anymore! Operation aborted!").
			     arg (curveTitle));
      boxName->clear ();
      boxName->insertStringList (curvesList);
      return;
    }

  //int index = boxName->currentItem ();
  int index = d_graph->curveIndex(curveTitle);
  QwtPlotCurve *c = d_graph->curve(index);
  if (!c || c->rtti () != QwtPlotItem::Rtti_PlotCurve || c->dataSize () < 2)
    {
      QString s = tr ("You cannot fit index:");
      s += "<p><b>'" + curveTitle + "'</b><p>";
      s += tr ("because it has less than 2 points!");
      QMessageBox::warning (0, tr ("QtiPlot - Warning"), s);
      return;
    }

  double start, stop;
  try
    {
      myParser parser;
      parser.SetExpr ((boxStart->text ()).ascii ());
      start = parser.Eval ();  
    }
  catch (mu::ParserError & e)
    {
      QMessageBox::critical (0, tr ("QtiPlot - Start limit error"),
			     e.GetMsg ().c_str());
      boxStart->setFocus ();
      return;
    }

  try
    {
      myParser parser;
      parser.SetExpr ((boxEnd->text ()).ascii ());
      stop = parser.Eval ();
    }
  catch (mu::ParserError & e)
    {
      QMessageBox::critical (0, tr ("QtiPlot - End limit error"),
			     e.GetMsg ().c_str());
      boxEnd->setFocus ();
      return;
    }

  if(start>=stop)
    {
      QMessageBox::critical (0, tr ("QtiPlot - Start or End limit error"),
			     tr("Upper limit must be greater than Lower limit."));
      return;
    }
  
  ApplicationWindow *app=(ApplicationWindow *)parent();
  int peakfrac = boxPeak->value();
  app->fwhmPeakFrac = peakfrac;
  int halffrac = boxHalf->value();
  app->fwhmHalfFrac = halffrac;

  QString res = d_graph->curveFWHM(curveTitle, start, stop, peakfrac, halffrac, app->fit_output_precision);
  if (!res.isEmpty ())
    {
      QDateTime dt = QDateTime::currentDateTime ();
      QString info = tr("[")+ dt.toString (Qt::LocalDate) + "]\n";
      info += tr("FWHM Analysis of ") + c->title().text() + "\n"
	+ res + 
	"-------------------------------------------------------------\n";
      app->logInfo += info;
      app->showResults (true);
      QString caption = app->generateUniqueName("FWHM") + "\t" +
        tr("FWHM Analysis of ")+c->title().text();
      Table *w = app->newTable( caption, 5, 2, res );
      QStringList ct;
      ct << tr("1") << tr("0;0/")+QString::number(app->fit_output_precision);
      w->setColumnTypes( ct );
      app->modifiedProject ();
    }
  close();
}

void
fwhmDialog::setGraph (Graph * g)
{
  d_graph = g;
  QStringList clist = g->analyzableCurvesList();
  boxName->insertStringList (clist, -1);
  if (d_graph->selectorsEnabled ())
    {
      int index = 0;
      QString selected = d_graph->selectedCurveTitle();
      if((index = clist.findIndex( selected )) >=0 )
        boxName->setCurrentItem( index );
      activateCurve(index);
    }
  else
    activateCurve(0);
 
  connect (d_graph, SIGNAL (closedGraph ()), this, SLOT (close ()));
  connect (d_graph, SIGNAL (dataRangeChanged ()), this,
	   SLOT (changeDataRange ()));
}

void
fwhmDialog::activateCurve (int index)
{
  QString ctitle = boxName->text(index);
  int cindex = d_graph->curveIndex(ctitle);
  QwtPlotCurve *c = d_graph->curve(cindex);
  if (!c || c->rtti () != QwtPlotItem::Rtti_PlotCurve)
    return;

  double start,end;
  d_graph->range(cindex, &start, &end);
  boxStart->setText (QString::number (QMIN (start, end), 'g', 15));
  boxEnd->setText (QString::number (QMAX (start, end), 'g', 15));
}

void
fwhmDialog::changeDataRange ()
{
  double start = d_graph->selectedXStartValue ();
  double end = d_graph->selectedXEndValue ();
  boxStart->setText (QString::number (QMIN (start, end), 'g', 15));
  boxEnd->setText (QString::number (QMAX (start, end), 'g', 15));
}

void
fwhmDialog::help ()
{
  QMessageBox::about (0, tr ("QtiPlot - Help for FWHM Analysis"),
		      tr(
"FWHM Analysis calculate the full width of half maximum.\n\
It first finds the peak of the curve by fitting a parabolic function\n\
around the maximum point. Then it calculates the FWHM by fitting\n\
linear functions around the half height. The ranges used for\n\
the fitting are specified as ratios to the height."));
}
