/***************************************************************************
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor,                    *
 *   Boston, MA  02110-1301  USA                                           *
 *                                                                         *
 ***************************************************************************/
#ifndef GRAPH_H
#define GRAPH_H

#include <qprinter.h>
#include <qpainter.h>
#include <qpointarray.h>

#include <gsl/gsl_multifit_nlin.h>
#include <gsl/gsl_multimin.h>

#include <qwt_plot.h>
#include <qwt_plot_marker.h>
#include <qwt_plot_curve.h>

#include "plotDialog.h"
#include "worksheet.h"
#include "axesDialog.h"

class QwtPlotCurve;
class QwtPlotZoomer;
class QwtPieCurve;
class Table;
class LegendMarker;
class LineMarker;
class ImageMarker;
class TitlePicker;
class ScalePicker;
class CanvasPicker;
class Plot;
class MultiPeakFit;
class ApplicationWindow;
class Matrix;

class Graph:public QWidget
{
Q_OBJECT public:
  Graph (QWidget * parent = 0, const char *name = 0, WFlags f = 0);
   ~Graph ();

  enum AxisType
  { Numeric = 0, Txt = 1, Day = 2, Month = 3, Time = 4, Date = 5, ColHeader =
      6 };
  enum MarkerType
  { None = -1, Text = 0, Arrow = 1, Image = 2 };

  enum FFTWindow
  { NoWindow = 0, Hanning = 1, Hamming = 2 }; 

  // _ErrorBars is for backward compatibility

  enum CurveType
  { Line, Scatter, LineSymbols, VerticalBars, Area, Pie, VerticalDropLines,
    Spline, HorizontalSteps, Histogram, HorizontalBars, VectXYXY, _ErrorBars,
    Box, VectXYAM, VerticalSteps, ColorMap, GrayMap, ContourMap,
    XErrorBars, YErrorBars
  };

  class CurveAttr
  {
  public:

    CurveAttr( int ckey=-1, int ctype=0, int lwidth=1, int ssize=5,
	       int sewidth=1 )
      {
	c_key = ckey;
	c_type = ctype;
	line_width = lwidth;
	symbol_size = ssize;
	symbol_edge_width = sewidth;
      };
    CurveAttr( CurveAttr *src )
      {
	c_key = src->c_key;
	c_type = src->c_type;
	line_width = src->line_width;
	symbol_size = src->symbol_size;
	symbol_edge_width = src->symbol_edge_width;
      };
    ~CurveAttr(){};
    
    int c_key;
    int c_type;
    int line_width;
    int symbol_size;
    int symbol_edge_width;
  };

  Plot *d_plot;
  QwtPlotZoomer *d_zoomer[2];
  TitlePicker *titlePicker;
  ScalePicker *scalePicker;
  CanvasPicker *cp;

  //! Returns the name of the parent multilayer plot
  const QString & parentPlotName ();

  public slots:Plot * plotWidget ()
  {
    return d_plot;
  };
  void copy (Graph * g);

  //
  void setRelativeGeometry( double x, double y, double w, double h )
    {
      gm_x = x;
      gm_y = y;
      gm_width = w;
      gm_height = h;
    }

  void getRelativeGeometry( double *x, double *y, double *w, double *h )
    {
      *x = gm_x;
      *y = gm_y;
      *w = gm_width;
      *h = gm_height;
    }

  // pie curves   
  bool isPiePlot ()
  {
    return piePlot;
  };
  void plotPie (QwtPieCurve * curve);
  void plotPie (Table * w, const QString & name);
  void plotPie (Table * w, const QString & name, const QPen & pen, int brush,
		int size, int firstColor);
  void updatePie (const QPen & pen, const Qt::BrushStyle & brushStyle,
		  int size, int firstColor);
  void updatePieCurveData (Table * w, const QString & yColName, int curve);
  void removePie ();
  QString pieLegendText ();
  QString savePieCurveLayout ();

  bool insertCurvesList (Table * w, const QStringList & names, int style,
			 int lWidth, int sSize);
  bool insertCurve (Table * w, const QString & name, int style);
  bool insertCurve (Table * w, int xcol, const QString & name, int style);
  bool insertCurve (Table * w, const QString & xColName,
		    const QString & yColName, int style);
  void insertPlotItem (QwtPlotItem * i, const QString & plotAssociation,
		       int type);

  //! Removes a curve defined by its index.
  void removeCurve (int index, bool update);

  //! Removes a curve defined by the title/plot association string s.
  void removeCurves (const QStringList &);

  void updateData (Table * w, int curve);
  void updateCurveData (Table * w, const QString & yColName, int curve);
  void updateBoxData (Table * w, const QString & yColName, int curve);

  int curves ()
  {
    return n_curves;
  };
  bool validCurvesDataSize ();
  double selectedXStartValue ();
  double selectedXEndValue ();

  int curveKey (int curve);
  int curveIndex (int key);
  //! map curve title to index
  int curveIndex (const QString &);
  //! get curve by index
  QwtPlotCurve *curve (int index);
  //! get curve by name
  QwtPlotCurve *curve (const QString & title)
  {
    return curve (curveIndex (title));
  }

  QString curveXColName (const QString & curveTitle);

  void insertPlottedList (const QStringList & names);

  //! Returns the names of all the QwtPlotCurve items on the plot, as a string list 
  QStringList curvesList ();
  QStringList analyzableCurvesList ();
  QStringList xyCurvesList ();
  //! Returns the names of all plot items, including spectrograms, as a string list 
  QStringList plotItemsList ();
  QStringList plotAssociations ();

  void updateTableName(const QString &, const QString &);
  void updateColName(const QString &, const QString &);

  void setPlotAssociations (const QStringList & newList);
  void changePlotAssociation (Table * t, int curve, int ctype, 
			      const QString & text);

  int curveType (int curveIndex);
  void setCurveType (int curve, int style);

  void print (const QString &);
  void copyImage ();
  void exportToSVG (const QString & fname);
  void exportToEPS (const QString & fname);
  void exportToEPS (const QString & fname, int res, QPrinter::Orientation o,
		    QPrinter::PageSize size, QPrinter::ColorMode col);

  void replot ();
  void updatePlot ();
  void activateLayout();

  QPixmap graphPixmap ();
  void exportImage (const QString & fileName, const QString & fileType,
		    int quality = 100, bool transparent = false);

  // error bars
  bool addErrorBars (Table * w, const QString & xColName,
		     const QString & yColName, Table * errTable,
		     const QString & errColName, int type = 1, int width =
		     1, int cap = 8, const QColor & color =
		     QColor (black), bool through = true, bool minus =
		     true, bool plus = true, double xOffset =
		     0, double yOffset = 0);

  bool addErrorBars (Table * w, const QString & yColName, Table * errTable,
		     const QString & errColName, int type = 1, int width =
		     1, int cap = 8, const QColor & color =
		     QColor (black), bool through = true, bool minus =
		     true, bool plus = true);

  void updateErrorBarsData (Table * w, int curve);
  void updateErrorBars (int curve, bool xErr, int width, int cap,
			const QColor & c, bool plus, bool minus,
			bool through);

  //! Called when a bar curve associated to an error bars curve curve is deleted
  void resetErrorBarsOffset (int index);

  // event handlers 
  void contextMenuEvent (QContextMenuEvent *);
  void closeEvent (QCloseEvent * e);

  //! Set axis scale
  void setScale (int axis, double start, double end, double step = 0.0,
		 int majorTicks = 5, int minorTicks = 5, int type =
		 0, bool inverted = false);

  double axisStep (int axis)
  {
    return d_user_step[axis];
  };

  //curves layout
  curveLayout initCurveLayout (int i, int curves, int style);
  static curveLayout initCurveLayout ();
  void updateCurveLayout (int index, const curveLayout * cL);
  //! Tries to guess not already used curve color and symbol style
  void guessUniqueCurveLayout (int &colorIndex, int &symbolIndex);

  gridOptions getGridOptions ();
  void setGridOptions (const gridOptions & options);

  // zoom
  void zoomed (const QwtDoubleRect &);
  void zoom (bool on);
  void zoomOut ();
  bool zoomOn ();

  void movedPicker (const QPoint & pos, bool mark);
  void setAutoScale ();
  void updateScale ();

  // saving to file 
  QString saveToString (bool saveAsTemplate = false);
  QString saveGridOptions ();
  QString saveScale ();
  QString saveScaleTitles ();
  QString saveFonts ();
  QString saveMarkers ();
  QString saveCurveLayout (int index);
  QString saveAxesTitleColors ();
  QString saveAxesColors ();
  QString saveEnabledAxes ();
  QString saveCanvas ();
  QString saveTitle ();
  QString saveAxesTitleAlignement ();
  QString saveEnabledTickLabels ();
  QString saveTicksType ();
  QString saveCurves ();
  QString saveLabelsFormat ();
  QString saveLabelsRotation ();
  QString saveAxesLabelsType ();
  QString saveAxesBaseline ();
  QString saveAxesFormulas ();

  // text markers 
  void drawText (bool on);
  bool drawTextActive ()
  {
    return drawTextOn;
  };
  int insertTextMarker (LegendMarker * mrk);

  //! Used when opening a project file
  int insertTextMarker (const QStringList & list, int fileVersion);
  void updateTextMarker (const QString & text, int angle, int bkg,
			 const QFont & fnt, const QColor & textColor,
			 const QColor & backgroundColor);

  void removeMarker ();
  void cutMarker ();
  void copyMarker ();
  void pasteMarker ();
  void selectNextMarker ();
  void highlightLineMarker (int markerID);
  void highlightTextMarker (int markerID);
  void highlightImageMarker (int markerID);
  void moveMarkerBy (int dx, int dy);

  QFont defaultTextMarkerFont ()
  {
    return defaultMarkerFont;
  };
  QColor textMarkerDefaultColor ()
  {
    return defaultTextMarkerColor;
  };
  QColor textMarkerDefaultBackground ()
  {
    return defaultTextMarkerBackground;
  };
  int textMarkerDefaultFrame ()
  {
    return defaultMarkerFrame;
  };
  void setTextMarkerDefaults (int f, const QFont & font,
			      const QColor & textCol,
			      const QColor & backgroundCol);

  void updateMarkerPos();

  Qt::PenStyle arrowLineDefaultStyle ()
  {
    return defaultArrowLineStyle;
  };
  bool arrowHeadDefaultFill ()
  {
    return defaultArrowHeadFill;
  };
  int arrowDefaultWidth ()
  {
    return defaultArrowLineWidth;
  };
  int arrowHeadDefaultLength ()
  {
    return defaultArrowHeadLength;
  };
  int arrowHeadDefaultAngle ()
  {
    return defaultArrowHeadAngle;
  };
  QColor arrowDefaultColor ()
  {
    return defaultArrowColor;
  };

  void setArrowDefaults (int lineWidth, const QColor & c, Qt::PenStyle style,
			 int headLength, int headAngle, bool fillHead);

  MarkerType copiedMarkerType ()
  {
    return selectedMarkerType;
  };
  void setCopiedMarkerType (Graph::MarkerType type)
  {
    selectedMarkerType = type;
  };
  void setCopiedMarkerEnds (const QPoint & start, const QPoint & end);
  void setCopiedTextOptions (int bkg, const QString & text,
			     const QFont & font, const QColor & color,
			     const QColor & bkgColor);
  void setCopiedArrowOptions (int width, Qt::PenStyle style,
			      const QColor & color, bool start, bool end,
			      int headLength, int headAngle, bool filledHead);
  void setCopiedImageName (const QString & fn)
  {
    auxMrkFileName = fn;
  };
  QRect copiedMarkerRect ()
  {
    return QRect (auxMrkStart, auxMrkEnd);
  };

  //legendMarker
  QwtArray < int >textMarkerKeys ()
  {
    return d_texts;
  };
  LegendMarker *textMarker (int id);

  void addTimeStamp ();

  // legend  
  void customLegend ();
  void removeLegend ();
  void removeLegendItem (int index);
  void addLegendItem (const QString & colName);
  void insertLegend (const QStringList & lst, int fileVersion);
  void newLegend ();

  LegendMarker *legend ();
  LegendMarker *newLegend (const QString & text);
  bool hasLegend ()
  {
    return legendMarkerID >= -0;
  };

  //! Creates a new legend text using the curves titles
  QString legendText ();

  // line markers 
  LineMarker *lineMarker (int id);
  void insertLineMarker (double,double,double,double,bool,bool,const QColor &);
  void insertLineMarker (LineMarker * mrk);

  //! Used when opening a project file
  void insertLineMarker (const QStringList & list, int fileVersion);
  QwtArray < int >lineMarkerKeys ()
  {
    return d_lines;
  };

  //!Draws a line/arrow depending on the value of "arrow"
  void drawLine (bool on, bool arrow = FALSE);
  bool drawArrow ()
  {
    return drawArrowOn;
  };
  bool drawLineActive ()
  {
    return drawLineOn;
  };
  void setSubtractLine(const QString &);
  void setSubtractLine(bool on);
  bool subtractLineActive ()
  {
    return subtractLineOn;
  };
  void subtractLine(const QPoint&, const QPoint&);
  

  //image markers
  ImageMarker *imageMarker (int id);
  QwtArray < int >imageMarkerKeys ()
  {
    return d_images;
  };
  void insertImageMarker (ImageMarker * mrk);
  void insertImageMarker (const QPixmap & photo, const QString & fileName);

  void insertImageMarker (const QStringList & lst, int fileVersion,
			  const QString & imageDir);
  bool imageMarkerSelected ();
  void updateImageMarker (int x, int y, int width, int height);

  //! Keep the markers on screen each time the scales are modified by adding/removing curves
  void updateMarkersBoundingRect ();

  int selectedMarkerKey ();
  void setSelectedMarker (int mrk);
  QwtPlotMarker *selectedMarkerPtr ();
  bool markerSelected ();
  bool arrowMarkerSelected ();
  void deselectMarker ();

  // axes 
  QValueList < int >axesType ();
  void setAxesType (const QValueList < int >tl);

  QStringList scalesTitles ();
  void setXAxisTitle (const QString & text);
  void setYAxisTitle (const QString & text);
  void setRightAxisTitle (const QString & text);
  void setTopAxisTitle (const QString & text);
  void setAxisTitle (int axis, const QString & text);

  QFont axisTitleFont (int axis);
  void setXAxisTitleFont (const QFont & fnt);
  void setYAxisTitleFont (const QFont & fnt);
  void setRightAxisTitleFont (const QFont & fnt);
  void setTopAxisTitleFont (const QFont & fnt);
  void setAxisTitleFont (int axis, const QFont & fnt);

  void setAxisFont (int axis, const QFont & fnt);
  QFont axisFont (int axis);
  void initFonts (const QFont & scaleTitleFnt, const QFont & numbersFnt);

  QColor axisTitleColor (int axis);
  void setXAxisTitleColor (const QColor & c);
  void setYAxisTitleColor (const QColor & c);
  void setRightAxisTitleColor (const QColor & c);
  void setTopAxisTitleColor (const QColor & c);
  void setAxesTitleColor (const QStringList &l);

  int axisTitleAlignment (int axis);
  void setXAxisTitleAlignment (int align);
  void setYAxisTitleAlignment (int align);
  void setTopAxisTitleAlignment (int align);
  void setRightAxisTitleAlignment (int align);
  void setAxesTitlesAlignment (const QStringList & align);

  QColor axisColor (int axis);
  QStringList axesColors ();
  void setAxesColors (const QStringList & colors);

  QColor axisNumbersColor (int axis);
  QStringList axesNumColors ();
  void setAxesNumColors (const QStringList & colors);

  void showAxis (int axis, int type, const QString & formatInfo,
		 Table * table, bool axisOn, int majTicksType,
		 int minTicksType, bool labelsOn, const QColor & c,
		 int format, int prec, int rotation, int baselineDist,
		 const QString & formula, const QColor & labelsColor);

  QMemArray < bool > enabledAxes ();
  void enableAxes (QMemArray < bool > axesOn);
  void enableAxes (const QStringList & list);

  int labelsRotation (int axis);
  void setAxisLabelRotation (int axis, int rotation);

  QStringList enabledTickLabels ();
  void setEnabledTickLabels (const QStringList & list);

  void setAxesLinewidth (int width);
  void loadAxesLinewidth (int width);	//used when opening a project file

  void drawAxesBackbones (bool yes);
  bool axesBackbones ()
  {
    return drawAxesBackbone;
  };
  void loadAxesOptions (const QString & s);	//used when opening a project file

  QValueList < int >axesBaseline ();
  void setAxesBaseline (const QValueList < int >&lst);
  void setAxesBaseline (QStringList & lst);

  void setMajorTicksType (const QValueList < int >&lst);
  void setMajorTicksType (const QStringList & lst);

  void setMinorTicksType (const QValueList < int >&lst);
  void setMinorTicksType (const QStringList & lst);

  int minorTickLength ();
  int majorTickLength ();
  void setAxisTicksLength (int axis, int majTicksType, int minTicksType,
			   int minLength, int majLength);
  void setTicksLength (int minLength, int majLength);
  void changeTicksLength (int minLength, int majLength);

  void setLabelsNumericFormat (const QStringList & l);
  void setLabelsNumericFormat (int axis, const QStringList & l);
  void setLabelsNumericFormat (int axis, int format, int prec,
			       const QString & formula);
  void setLabelsDateTimeFormat (int axis, int type,
				const QString & formatInfo);
  void setLabelsDayFormat (int axis, int format);
  void setLabelsMonthFormat (int axis, int format);

  const QStringList & axesLabelsFormatInfo ()
  {
    return axesFormatInfo;
  };
  void setLabelsTextFormat (int axis, int type, const QString & labelsColName,
			    Table * table);

  const QStringList & getAxesFormulas ()
  {
    return axesFormulas;
  };
  void setAxesFormulas (const QStringList & l)
  {
    axesFormulas = l;
  };
  void setAxisFormula (int pos, const QString & f)
  {
    axesFormulas[pos] = f;
  };

  // canvas frame 
  void drawCanvasFrame (bool frameOn, int width);
  void drawCanvasFrame (const QStringList & frame);
  void drawCanvasFrame (bool frameOn, int width, const QColor & color);
  QColor canvasFrameColor ();
  int canvasFrameWidth ();
  bool framed ();

  // plot title
  void setTitle (const QString & t);
  void setTitleFont (const QFont & fnt);
  void setTitleColor (const QColor & c);
  void setTitleAlignment (int align);

  bool titleSelected ();
  void selectTitle ();

  void removeTitle ();
  void initTitle (bool on, const QFont & fnt);

  // tools for modifing insertCurve data 
  bool selectPoint (const QPoint & pos);
  void highlightPoint (bool showIt);
  void selectCurve (const QPoint & pos);
  int selectedCurveID ()
  {
    return selectedCurve;
  };
  QString selectedCurveTitle ();

  void showCursor (bool showIt);
  void shiftPointCursor (int d);
  void shiftCurveCursor (int d);
  void moveBy (int dx, int dy);
  void move (const QPoint & pos);
  void removePoint ();
  void movePoints (bool enabled);
  bool movePointsActivated ();
  void removePoints (bool enabled);
  bool removePointActivated ();

  void copyCanvas (bool on);
  void enableCursor (bool on)
  {
    cursorEnabled = on;
  };
  bool enabledCursor ()
  {
    return cursorEnabled;
  };
  void showPlotPicker (bool on);
  bool pickerActivated ();

  void disableTools ();

  //translating curves
  void translateCurve (int direction);
  void translateCurveTo (const QPoint & p);
  bool translationInProgress ()
  {
    return translateOn;
  };
  void startCurveTranslation ();

  //! Returns TRUE if the data range selectors are enables, FALSE otherwise.
  bool selectorsEnabled ()
  {
    return rangeSelectorsEnabled;
  };

  //! Enables the data range selectors depending on the value of ON (ON = false will call disableRangeSelectors()).
  bool enableRangeSelectors (bool on);

  //! Disables the data range selectors tool.
  void disableRangeSelectors ();

  void moveRangeSelector (int d);
  void shiftRangeSelector (bool shift);

  //! Select the next/previous curve 
  void shiftCurveSelector (int d);

  //border and margin
  void changeMargin (int d);
  void changeSpacing (int s);
  int spacing(){ return d_spacing; };
  void setAxisTitleInverted(int axis, bool inv);
  void drawBorder (int width, const QColor & color);
  void setBorder (int width, const QColor & color);
  void setBackgroundColor (const QColor & color);
  void setCanvasBackground (const QColor & color);

  //functions in analysis.cpp file
  bool stepCheck( int );
  bool smoothSavGol (ApplicationWindow *app, int index, int order, int nl, int nr, int colIndex);
  bool smoothAverage (ApplicationWindow *app, int index, int points, int colIndex);
  bool PeakConv (ApplicationWindow *app, int index, int points, int colIndex,
		 int peakType);
  bool interpolate( int cindex, int spline, 
		    QMemArray <double> &x, QMemArray <double> &y );
  bool convoluteFFT (ApplicationWindow *app, int method,
		     const QString & sigTitle, const QString & resTitle,
		     bool normRes, double thres, double cutoff,
		     int colorIndex = 1);
  bool deconvoluteGS (ApplicationWindow *app,
		      const QString & sigTitle, const QString & resTitle,
		      bool normRes, int smooth, double damping, int iteration,
		      int constraint, int colorIndex = 1);
  
  QString integrateCurve (QString &curveTitle, double low, double up);
  QString curveFWHM (QString &, double, double, int, int, int);
  bool diffCurve (ApplicationWindow *app, const QString & curveTitle );
  bool fft (ApplicationWindow *app, int index, bool forward, double sampling,
	    bool resample, bool normalizeAmp, bool order, FFTWindow win);
  bool filterFFT (ApplicationWindow *app, int index, int filter_type, 
		  double lf, double hf,
		  bool resample, bool DCOffset, int colIndex);

  // functions called by fitDialog.cpp
  void deleteFitCurves ();

  //! Set start and end to selected X range of curve index or, if there's no selection, to the curve's total range.
  void range (int index, double *start, double *end);
  double minPositiveValue(int axis);
  double maxPositiveValue(int axis);
  //! Returns the number of points within the range.
  int numPoints (int index, double start, double end);
  //! Sets x and y to the curve points between start and end. Memory will be allocated with new double[].
  //! Returns the number of points within range == size of x and y arrays.
  int curveData (int index, double start, double end,
		 QMemArray <double> &x, QMemArray <double> &y);
    //! Same as curveData, but sorts the points by their x value.
  int sortedCurveData (int index, QValueList <double> &ranges,
		       QMemArray <double> &x, QMemArray <double> &y);
  int sortedCurveData (int index, double start, double end,
		       QMemArray <double> &x, QMemArray <double> &y);

  //! Set the number of fit curves
  void setFitID (int id)
  {
    fitID = id;
  };

  void addResultCurve (int n, QMemArray <double> &x, QMemArray <double> &y,
		       int colorIndex,
		       const QString & tableName, const QString & legend,
		       int prec, QwtPlotCurve *);

  //histograms
  void initHistogram (int index, const QMemArray < double >&Y, int it);
  void updateHistogram (Table * w, const QString & curveName, int curve);
  void updateHistogram (Table * w, const QString & curveName, int curve,
			bool automatic, double binSize, double begin,
			double end);
  void setBarsGap (int curve, int gapPercent, int offset);
  QString showHistogramStats (Table * w, const QString & curveName,
			      int curve);

  //image analyse tools
  bool lineProfile ();
  void calculateProfile (int average, bool ok);
  void calculateLineProfile (const QPoint & start, const QPoint & end);
  int averageImagePixel (const QImage & image, int px, int py, int average,
			 bool moreHorizontal);
  void showIntensityTable ();

  //user defined functions
  void modifyFunctionCurve (int curve, int type, const QStringList & formulas,
			    const QString & var, QValueList < double >&ranges,
			    int points);
  void addFunctionCurve (int type, const QStringList & formulas,
			 const QString & var, QValueList < double >&ranges,
			 int points, const QString & title = QString::null);
  //when reading from file
  void insertFunctionCurve (const QString & formula, int points,
			    int fileVersion);
  //! Returns a unique function name
  QString generateFunctionName ();

  void createWorksheet (const QString & name);
  void activateGraph ();
  void moveGraph (const QPoint & pos);
  void releaseGraph ();
  void drawFocusRect ();

  //vector curves
  void plotVectorCurve (Table * w, const QStringList & colList, int style);
  void setVectorsLook (int curve, const QColor & c, int width,
		       int arrowLength, int arrowAngle, bool filled,
		       int position);
  void updateVectorsLayout (Table * w, int curve, int colorIndex, int width,
			    int arrowLength, int arrowAngle, bool filled,
			    int position, const QString & xEndColName,
			    const QString & yEndColName);
  void updateVectorsData (Table * w, int curve);

  //box plots
  void openBoxDiagram (Table * w, const QStringList & l);
  void plotBoxDiagram (Table * w, const QStringList & names);

  void setCurveSymbol (int index, const QwtSymbol & s);
  void setCurvePen (int index, const QPen & p);
  void setCurveBrush (int index, const QBrush & b);
  void setCurveStyle (int index, int s);

  //resizing      
  bool ignoresResizeEvents ()
  {
    return ignoreResize;
  };
  void setIgnoreResizeEvents (bool ok)
  {
    //if(ok) fprintf(stderr,"setIgnoreResizeEvents : true\n");
    //else fprintf(stderr,"setIgnoreResizeEvents : false\n");

    ignoreResize = ok;
  };
  void resizeEvent (QResizeEvent * e);
  void scaleFonts (double factor);
  void scaleSymbols(double factor);
  void scaleSpaces(double factor);
  void scaleTickLength(double factor);
  void scaleFrame(double factor);
  void scaleGrid(double factor);

  void saveComponentSizes();
  void restoreComponentSizes();

  void modified ();
  void emitModified ();

  void updateSecondaryAxis (int axis);
  void enableAutoscaling (bool yes)
  {
    autoscale = yes;
  };

  bool autoscaleFonts ()
  {
    return autoScaleFonts;
  };
  double scaleFactor();
  void setAutoscaleFonts (bool yes);
  void setAutoScaleFrameWidth (bool on);
  void setAutoScaleGridWidth (bool on);
  void setAutoScaleLineWidth (bool on);

  static int obsoleteSymbolStyle (int type);
  static QString penStyleName (Qt::PenStyle style);
  static Qt::PenStyle getPenStyle (const QString & s);
  static Qt::PenStyle getPenStyle (int style);
  static Qt::BrushStyle getBrushStyle (int style);
  static void showPlotErrorMessage (QWidget * parent,
				    const QStringList & emptyColumns);

  void showTitleContextMenu ();
  void copyTitle ();
  void cutTitle ();

  void removeAxisTitle ();
  void cutAxisTitle ();
  void copyAxisTitle ();
  void showAxisTitleMenu (int axis);
  void showAxisContextMenu (int axis);
  void hideSelectedAxis ();
  void showGrids ();
  void showAxisDialog ();
  void showScaleDialog ();

  //! Initialize a multi peak fitting operations and creates a new MultiPeakFit object
  void multiPeakFit (ApplicationWindow * app, int profile, int peaks);
  void selectPeak (const QPoint & pos);
  bool selectPeaksOn ();

  //! Add a spectrogram to the graph
  void plotSpectrogram (Matrix * m, int type);
  //! Restores a spectrogram. Used when opening a project file.
  void restoreSpectrogram (ApplicationWindow * app, const QStringList & lst);

  QString generateUniqueName( const QString &);

  void swapCurves(int, int);
  void updateLegend();
  bool getDisplayRange(int, int*, int*);

  void resetColors(bool);
  void resetSymbols();
  //void graphColors();
  void setFonts(const QFont &, const QFont &, const QFont &, const QFont &);

  void setTextLayer(bool yes)
  {
    d_textlayer = yes;
  }
  bool isTextLayer()
  {
    return d_textlayer;
  }

signals:
  void highlightGraph (Graph *);
  void releaseGraph (Graph *);
  void moveGraph (Graph *, const QPoint & pos);
  void selectedGraph (Graph *);
  void closedGraph ();
  void drawTextOff ();
  void drawLineEnded (bool);
  void cursorInfo (const QString &,const QString &,int);
  void showInfo (const QString &);
  void showDisplayBar();
  void showPlotDialog (int);
  void showPieDialog ();
  void createTable (const QString &, int, int, const QString &);
  void createHiddenTable (const QString &, int, int, int, const QString &);
  void updateTable (const QString &, int, const QString &);
  void updateTableColumn (const QString &, QMemArray <double> &);
  void clearCell (const QString &, int);

  void viewImageDialog ();
  void viewTextDialog ();
  void viewLineDialog ();
  void viewTitleDialog ();
  void modifiedGraph ();
  void modifiedGraph (Graph *);
  void hiddenPlot (QWidget *);

  void modifiedFunction ();
  void modifiedPlotAssociation ();

  void showContextMenu ();
  void showCurveContextMenu (int);
  void showMarkerPopupMenu ();

  void showAxisDialog (int);
  void axisDblClicked (int);
  void xAxisTitleDblClicked ();
  void yAxisTitleDblClicked ();
  void rightAxisTitleDblClicked ();
  void topAxisTitleDblClicked ();

  void createTablePlot (const QString &, int, int, const QString &);
  void createIntensityTable (const QPixmap &);
  void createHistogramTable (const QString &, int, int, const QString &);
  void dataRangeChanged ();
  void showFitResults (const QString &);


private:
  double gm_x, gm_y, gm_width, gm_height;
  bool autoScaleFonts;
  bool autoScaleLineWidth;
  bool autoScaleFrameWidth;
  bool autoScaleGridWidth;
  int selectedAxis;
  QStringList axesFormulas;
  //! Stores columns used for axes with text labels or  time/date format info
  QStringList axesFormatInfo;
  QValueList < int >axisType;

  //! Structure used to define the grid
  gridOptions grid;

  MarkerType selectedMarkerType;
  QwtPlotMarker::LineStyle mrklStyle;
  QStringList associations;

  //! Stores the step the user specified for the four scale.
  //If step = 0.0, the step will be calculated automatically by the Qwt scale engine.
  QMemArray < double >d_user_step;
  //! Arrows/lines on plot keys
  QMemArray < int >d_lines;
  //! Images on plot keys
  QMemArray < int >d_images;
  //! Stores the identifiers (keys) of the text objects on the plot
  QMemArray < int >d_texts;

  QPtrList <CurveAttr> c_attr;

  //! array to hold component sizes
  QValueList <int> sizeStore;

  double d_frame_width, d_axes_width;
  double d_minor_ticklength, d_major_ticklength;
  double d_grid_major, d_grid_minor;

  // font sizes for default window size
  double axisFontSize[4];
  double axisTitleFontSize[4];
  double titleFontSize;

  double pieSize; // size of pie chart for default window size

  QPen mrkLinePen;
  QFont auxMrkFont, defaultMarkerFont;
  QColor auxMrkColor, auxMrkBkgColor;
  QPoint auxMrkStart, auxMrkEnd;
  Qt::PenStyle auxMrkStyle;
  QString auxMrkFileName, auxMrkText;

  int d_spacing; // space between axis title and axis labels
  double d_base_margin;
  double d_base_spacing;

  bool d_first_resize;

  int n_curves, selectedCurve, selectedPoint, startPoint, endPoint,
    selectedCursor;
  //int pieRay;
  //int selectedCol; 
  //int xCol;
  int widthLine, fitID, defaultMarkerFrame;
  QColor defaultTextMarkerColor, defaultTextMarkerBackground;
  int auxMrkAngle, auxMrkBkg, auxMrkWidth, averagePixels;
  int auxArrowHeadLength, auxArrowHeadAngle;
  int translationDirection;
  int selectedMarker, legendMarkerID, startID, endID;
  int mrkX, mrkY;		//x=0 et y=0 line markers keys
  bool startArrowOn, endArrowOn, drawTextOn, drawLineOn, drawArrowOn,
    subtractLineOn;

  //the following bool values tell which data tool is activated by the user
  bool removePointsEnabled, movePointsEnabled, translateOn;
  bool pickerEnabled, cursorEnabled, rangeSelectorsEnabled;
  bool piePlot;			//tells if the plot is a pie plot
  bool lineProfileOn;		// tells if pixel line profile is asked
  bool auxFilledArrowHead, ignoreResize;
  bool drawAxesBackbone, autoscale;

  int selected_peaks;
  MultiPeakFit *fitter;

  QColor defaultArrowColor;
  int defaultArrowLineWidth, defaultArrowHeadLength, defaultArrowHeadAngle;
  bool defaultArrowHeadFill;
  Qt::PenStyle defaultArrowLineStyle;

  bool d_pickerMarker;
  bool d_textlayer;
};

#endif // GRAPH_H
