/***************************************************************************
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor,                    *
 *   Boston, MA  02110-1301  USA                                           *
 *                                                                         *
 ***************************************************************************/
#include "graph3D.h"
#include "bars.h"
#include "cones.h"
#include "parser.h"
#include "nrutil.h"

#include <qapplication.h>
#include <qmessagebox.h>
#include <qfiledialog.h>
#include <qprinter.h>
#include <qimage.h>
#include <qclipboard.h>
#include <qdragobject.h>
#include <qregexp.h>

#include <qwt3d_surfaceplot.h>
#include <qwt3d_function.h>
#include <qwt3d_io.h>
#include <qwt3d_io_gl2ps.h>
#include <qwt3d_io_reader.h>
#include <qwt3d_coordsys.h>

#include <gsl/gsl_vector.h>
#include <math.h>
#include <fstream>

#define SWAP(type,a,b) do{type tmp=a; a=b; b=tmp;} while (0)
#define SQR(x) ((x)*(x))

UserFunction::UserFunction (const QString & s, SurfacePlot & pw):
Function (pw)
{
  formula = s;
  logscale = false;
  logMinimum = 0;
}

double
UserFunction::operator () (double x, double y)
{
  if (formula.isEmpty ())
    return 0.0;

  myParser parser;
  double result = 0.0;
  try
  {
    parser.DefineVar ("x", &x);
    parser.DefineVar ("y", &y);

    parser.SetExpr ((const std::string) formula.ascii ());
    result = parser.Eval ();
    if (logscale)
      {
	if(result>0)
	  result = log10(result);
	else
	  result = logMinimum;
      }
  }
  catch (mu::ParserError & e)
  {
    QMessageBox::critical (0, "QtiPlot - Input function error", e.GetMsg ().c_str());
  }
  return result;
}

UserFunction::~UserFunction ()
{
}

Graph3D::Graph3D (const QString & label, QWidget * parent, const char *name, WFlags f):myWidget (label, parent, name,
	  f)
{
  initPlot ();
}

void
Graph3D::initPlot ()
{
  worksheet = 0;
  matrix_ = 0;

  setBirthDate (QDateTime::currentDateTime ().toString (Qt::LocalDate));

  color_map = QString::null;
  animation_redraw_wait = 50;
  d_timer = new QTimer (this);
  connect (d_timer, SIGNAL (timeout ()), this, SLOT (rotate ()));

  //ignoreFonts = false;
  ignoreFonts = true;

  sp = new SurfacePlot (this, "3D SurfacePlot");
  sp->resize (500, 450);
  currentSize = QSize( 500, 450 );
  sp->installEventFilter (this);
  sp->setRotation (30, 0, 15);
  sp->setScale (1, 1, 1);
  //sp->setShift (0.15, 0, 0);
  //sp->setShift (0, 0, 0);
  sp->setZoom (0.9);
  sp->setOrtho (false);

  smoothMesh = true;
  sp->setSmoothMesh (smoothMesh);

  title = QString::null;
  sp->setTitle (title);

  titleCol = QColor (black);
  sp->setTitleColor (Qt2GL (titleCol));

  titleFnt = QFont ("Times New Roman", 14);
  titleFnt.setBold (TRUE);
  sp->setTitleFont (titleFnt.family (), titleFnt.pointSize (),
		    titleFnt.weight (), titleFnt.italic ());
  titleFontSize = titleFnt.pointSizeFloat();

  for(int i=0;i<3;i++)
    axisLabelFontSize[i] = 14;
  numberFontSize = 14;

  axesCol = QColor (black);
  labelsCol = QColor (black);
  numCol = QColor (black);
  meshCol = QColor (black);
  gridCol = QColor (black);
  bgCol = QColor (255, 255, 255);
  fromColor = QColor (red);
  toColor = QColor (blue);

  col_ = 0;

  legendOn = false;
  legendMajorTicks = 5;
  sp->showColorLegend (legendOn);
  sp->legend ()->setAutoScale (TRUE);
  sp->legend ()->setMajors (legendMajorTicks);

  labelsDist = 0;

  scaleType = QMemArray < int >(3);
  for (int j = 0; j < 3; j++)
    scaleType[j] = 0;

  pointStyle = None;
  func = 0;
  alpha = 1.0;
  barsRad = 0.007;
  pointSize = 5;
  smooth = false;
  crossHairRad = 0.02, crossHairLineWidth = 1;
  crossHairSmooth = true, crossHairBoxed = false;
  conesQuality = 32;
  conesRad = 0.01;

  //style_ = FILLEDMESH;
  style_ = NOPLOT;
  initCoord ();

  connect (sp, SIGNAL (rotationChanged (double, double, double)), this,
	   SLOT (rotationChanged (double, double, double)));
  connect (sp, SIGNAL (zoomChanged (double)), this,
	   SLOT (zoomChanged (double)));
  connect (sp, SIGNAL (scaleChanged (double, double, double)), this,
	   SLOT (scaleChanged (double, double, double)));
  connect (sp, SIGNAL (shiftChanged (double, double, double)), this,
	   SLOT (shiftChanged (double, double, double)));
}

void
Graph3D::setLabelString( int axis, const QString & s )
{
  int A1,A2,A3,A4;
  switch(axis)
    {
    case 0:
      A1 = X1; A2 = X2; A3 = X3; A4 = X4;
      break;
    case 1:
      A1 = Y1; A2 = Y2; A3 = Y3; A4 = Y4;
      break;
    case 2:
      A1 = Z1; A2 = Z2; A3 = Z3; A4 = Z4;
      break;
    }

  sp->coordinates ()->axes[A1].setLabelString (s);
  sp->coordinates ()->axes[A2].setLabelString (s);
  sp->coordinates ()->axes[A3].setLabelString (s);
  sp->coordinates ()->axes[A4].setLabelString (s);
}

void
Graph3D::setMajors( int axis, int val )
{
  int A1,A2,A3,A4;
  switch(axis)
    {
    case 0:
      A1 = X1; A2 = X2; A3 = X3; A4 = X4;
      break;
    case 1:
      A1 = Y1; A2 = Y2; A3 = Y3; A4 = Y4;
      break;
    case 2:
      A1 = Z1; A2 = Z2; A3 = Z3; A4 = Z4;
      break;
    }
  sp->coordinates ()->axes[A1].setMajors(val);
  sp->coordinates ()->axes[A2].setMajors(val);
  sp->coordinates ()->axes[A3].setMajors(val);
  sp->coordinates ()->axes[A4].setMajors(val);
}

void
Graph3D::setMinors( int axis, int val )
{
  int A1,A2,A3,A4;
  switch(axis)
    {
    case 0:
      A1 = X1; A2 = X2; A3 = X3; A4 = X4;
      break;
    case 1:
      A1 = Y1; A2 = Y2; A3 = Y3; A4 = Y4;
      break;
    case 2:
      A1 = Z1; A2 = Z2; A3 = Z3; A4 = Z4;
      break;
    }
  sp->coordinates ()->axes[A1].setMinors(val);
  sp->coordinates ()->axes[A2].setMinors(val);
  sp->coordinates ()->axes[A3].setMinors(val);
  sp->coordinates ()->axes[A4].setMinors(val);
}

void
Graph3D::initCoord ()
{
  sp->makeCurrent ();
  for(int i=0;i<3;i++)
    {
      setMajors(i,5);
      setMinors(i,5);
    }

  QString s = tr ("X axis");
  setLabelString( 0, s );
  labels << s;

  s = tr ("Y axis");
  setLabelString( 1, s );
  labels << s;

  s = tr ("Z axis");
  setLabelString( 2, s );
  labels << s;

  sp->setCoordinateStyle (BOX);
  sp->coordinates ()->setAutoScale (false);
}

void
Graph3D::setCoordinateSystem( double xl, double xr,
			      double yl, double yr,
			      double zl, double zr)
{
  sp->createCoordinateSystem (Triple (xl, yl, zl), Triple (xr, yr, zr));
  if (scaleType[2]==0)
    {
      sp->coordinates ()->axes[Z1].setScale (LINEARSCALE);
      sp->coordinates ()->axes[Z2].setScale (LINEARSCALE);
      sp->coordinates ()->axes[Z3].setScale (LINEARSCALE);
      sp->coordinates ()->axes[Z4].setScale (LINEARSCALE);
      sp->legend()->setScale(LINEARSCALE);
      //qDebug("its linear scale");
    }
  else
    {
      sp->coordinates ()->axes[Z1].setScale (LOG10SCALE);
      sp->coordinates ()->axes[Z2].setScale (LOG10SCALE);
      sp->coordinates ()->axes[Z3].setScale (LOG10SCALE);
      sp->coordinates ()->axes[Z4].setScale (LOG10SCALE);
      sp->legend()->setScale(LOG10SCALE);
      //qDebug("its log10 scale");
    }
  sp->legend ()->setLimits (zl, zr);
}

void
Graph3D::addFunction (const QString & s, double xl, double xr, double yl,
		      double yr, double zl, double zr, int scaletype)
{
  scaleType[2] = scaletype;
  if(scaletype==1)
    {
      if(zr<=0) zr = 10;
      if(zl<=0 || zl>zr) zl = zr*0.001;
      zl = log10(zl);
      zr = log10(zr);
    }

  sp->makeCurrent ();
  sp->resize (this->size ());
  currentSize = this->size();

  func = new UserFunction (s, *sp);

  func->setMesh (41, 31);
  func->setDomain (xl, xr, yl, yr);
  func->setMinZ (zl);
  func->setMaxZ (zr);
  func->setLogscale( scaletype==1, zl );
  func->create ();

  sp->legend ()->setLimits (zl, zr);

  if (sp->plotStyle () == NOPLOT)
    {
      sp->setPlotStyle (FILLED);
      style_ = FILLED;
      pointStyle = None;
      legendOn = true;
    }
  //sp->createCoordinateSystem (Triple (xl, yl, zl), Triple (xr, yr, zr));
  setCoordinateSystem( xl, xr, yl, yr, zl, zr );
  findBestLayout ();
}

void
Graph3D::insertFunction (const QString & s, double xl, double xr, double yl,
			 double yr, double zl, double zr)
{
  addFunction (s, xl, xr, yl, yr, zl, zr, scaleType[2]);
  update ();
}

void
Graph3D::addData (Table * table, int xcol, int ycol)
{
  worksheet = table;
  int r = table->tableRows ();
  int i, xmesh = 0, ymesh = 2;
  for (i = 0; i < r; i++)
    {
      if (!table->text (i, xcol).isEmpty ()
	  && !table->text (i, ycol).isEmpty ())
	xmesh++;
    }

  if (xmesh == 0)
    xmesh++;

  //qDebug("xmesh = %d",xmesh);

  double **data = matrix (0, xmesh - 1, 0, ymesh - 1);
  gsl_vector *x = gsl_vector_alloc (xmesh);
  gsl_vector *y = gsl_vector_alloc (xmesh);

  for (int j = 0; j < ymesh; j++)
    {
      int k = 0;
      for (i = 0; i < r; i++)
	{
	  if (!table->text (i, xcol).isEmpty ()
	      && !table->text (i, ycol).isEmpty ())
	    {
	      double xv = table->text (i, xcol).toDouble ();
	      double yv = table->text (i, ycol).toDouble ();

	      //qDebug("k = %d",k);
	      gsl_vector_set (x, k, xv);
	      gsl_vector_set (y, k, yv);

	      data[k][j] = yv;
	      k++;
	    }
	}
    }

  double maxy = gsl_vector_max (y);
  double maxz = 0.6 * maxy;
  sp->makeCurrent ();
  sp->legend ()->setLimits (gsl_vector_min (y), maxy);
  sp->loadFromData (data, xmesh, ymesh, gsl_vector_min (x),
		    gsl_vector_max (x), 0, maxz);

  findBestLayout ();

  gsl_vector_free (x);
  gsl_vector_free (y);
  free_matrix (data, 0, xmesh - 1, 0, ymesh - 1);
}

void
Graph3D::addData (Table * table, const QString & colName)
{
  int ycol = table->colIndex (colName);
  int xcol = table->colX (ycol);
  plotAssociation = table->colName (xcol) + "(X)," + colName + "(Y)";

  addData (table, xcol, ycol);
}

void
Graph3D::addData (Table * table, const QString & xColName,
		  const QString & yColName)
{
  int xcol = table->colIndex (xColName);
  plotAssociation = table->colName (xcol) + "(X)," + yColName + "(Y)";

  addData (table, xcol, table->colIndex (yColName));
}

void
Graph3D::changeMatrix (Matrix * m)
{
  if (matrix_ == m)
    return;

  addMatrixData (m);
}

void
Graph3D::addMatrixData (Matrix * m)
{
  plotAssociation = "matrix<" + m->getName() + ">";

  double x1 = m->xStart();
  double x2 = m->xEnd();
  double xl = QMIN(x1,x2);
  double xr = QMAX(x1,x2);
  double y1 = m->yStart();
  double y2 = m->yEnd();
  double yl = QMIN(y1,y2);
  double yr = QMAX(y1,y2);
  double zl,zr;
  m->range(&zl,&zr);
  addMatrixData(m,xl,xr,yl,yr,zl,zr,0);
}

void
Graph3D::addMatrixData (Matrix * m, double xl, double xr,
			double yl, double yr,
			double zl, double zr,
			int scaletype)
{
  matrix_ = m;
  plotAssociation = "matrix<" + m->getName() + ">";
  scaleType[2] = scaletype;
  updateScalesFromMatrix (xl, xr, yl, yr, zl, zr);
  findBestLayout();
}

void
Graph3D::addData (Table * table, const QString & xColName,
		  const QString & yColName, double xl, double xr, double yl,
		  double yr, double zl, double zr, int scaletype)
{
  scaleType[2] = scaletype;
  if(scaletype==1)
    {
      if(zr<=0) zr = 10;
      if(zl<=0 || zl>zr) zl = zr*0.001;
      zl = log10(zl);
      zr = log10(zr);
    }

  worksheet = table;
  int r = table->tableRows ();
  int xcol = table->colIndex (xColName);
  int ycol = table->colIndex (yColName);

  QString s = table->colName (xcol) + "(X),";
  s += yColName + "(Y)";
  plotAssociation = s;

  int i, j, xmesh = 0, ymesh = 2;
  double xv, yv;

  for (i = 0; i < r; i++)
    {
      if (!table->text (i, xcol).isEmpty ()
	  && !table->text (i, ycol).isEmpty ())
	{
	  xv = table->text (i, xcol).toDouble ();
	  if (xv >= xl && xv <= xr)
	    xmesh++;
	}
    }

  if (xmesh == 0)
    xmesh++;

  double **data = matrix (0, xmesh - 1, 0, ymesh - 1);
  for (j = 0; j < ymesh; j++)
    {
      int k = 0;
      for (i = 0; i < r; i++)
	{
	  if (!table->text (i, xcol).isEmpty ()
	      && !table->text (i, ycol).isEmpty ())
	    {
	      xv = table->text (i, xcol).toDouble ();
	      if (xv >= xl && xv <= xr)
		{
		  yv = table->text (i, ycol).toDouble ();
		  if(scaletype==1)
		    {
		      if(yv>0) yv = log10(yv);
		      else yv = zl;
		    }
		  if (yv > zr)
		    data[k][j] = zr;
		  else if (yv < zl)
		    data[k][j] = zl;
		  else
		    data[k][j] = yv;
		  k++;
		}
	    }
	}
    }
  sp->makeCurrent ();
  sp->loadFromData (data, xmesh, ymesh, xl, xr, yl, yr);

  setCoordinateSystem( xl, xr, yl, yr, zl, zr );
  sp->legend ()->setMajors (legendMajorTicks);

  free_matrix (data, 0, xmesh - 1, 0, ymesh - 1);
}

void
Graph3D::insertNewData (Table * table, const QString & colName)
{
  int zCol = table->colIndex (colName);
  int yCol = table->colY (zCol);
  int xCol = table->colX (zCol);

  addData (table, xCol, yCol, zCol, Trajectory);
  update ();
}

void
Graph3D::changeDataColumn (Table * table, const QString & colName)
{
  int zCol = table->colIndex (colName);
  int yCol = table->colY (zCol);
  int xCol = table->colX (zCol);

  plotAssociation =
    table->colName (xCol) + "(X)," + table->colName (yCol) + "(Y),";
  plotAssociation += colName + "(Z)";

  updateDataXYZ (table, xCol, yCol, zCol);
  findBestLayout ();
}

void
Graph3D::addData (Table * table, int xCol, int yCol, int zCol, int type)
{
  worksheet = table;
  int r = table->tableRows ();

  QString s = table->colName (xCol) + "(X),";
  s += table->colName (yCol) + "(Y),";
  s += table->colName (zCol) + "(Z)";
  plotAssociation = s;

  int i, j, columns = 0;
  for (i = 0; i < r; i++)
    {
      if (!table->text (i, xCol).isEmpty ()
	  && !table->text (i, yCol).isEmpty ()
	  && !table->text (i, zCol).isEmpty ())
	columns++;
    }

  if (columns == 0)
    columns++;

  Qwt3D::Triple ** data = allocateData (columns, columns);
  for (j = 0; j < columns; j++)
    {
      int k = 0;
      for (i = 0; i < r; i++)
	{
	  if (!table->text (i, xCol).isEmpty ()
	      && !table->text (i, yCol).isEmpty ()
	      && !table->text (i, zCol).isEmpty ())
	    {
	      double xv = table->text (i, xCol).toDouble ();
	      double yv = table->text (i, yCol).toDouble ();
	      double zv = table->text (i, zCol).toDouble ();

	      data[k][j] = Triple (xv, yv, zv);
	      k++;
	    }
	}
    }

  sp->makeCurrent ();
  sp->loadFromData (data, columns, columns, false, false);

  double start, end;
  sp->coordinates ()->axes[Z1].limits (start, end);
  sp->legend ()->setLimits (start, end);
  sp->legend ()->setMajors (legendMajorTicks);

  if (type == Scatter)
    {
      Dot d (pointSize, smooth);
      sp->setPlotStyle (d);
      pointStyle = Dots;
      style_ = Qwt3D::USER;
    }
  else if (type == Trajectory)
    {
      legendOn = false;
      sp->showColorLegend (legendOn);
    }
  else
    {
      sp->setPlotStyle (Bar (barsRad));
      pointStyle = VerticalBars;
      style_ = Qwt3D::USER;
    }

  findBestLayout ();
  deleteData (data, columns);
}

void
Graph3D::addData (Table * table, int xCol, int yCol, int zCol,
		  double xl, double xr, double yl, double yr, double zl,
		  double zr, int scaletype)
{
  scaleType[2]=scaletype;
  if(scaletype==1)//logscale
    {
      if(zr<=0) zr = 10;
      if(zl<=0 || zl>zr) zl = zr * 0.001;
      zl = log10(zl);
      zr = log10(zr);
    }

  worksheet = table;
  int r = table->tableRows ();

  QString s = table->colName (xCol) + "(X),";
  s += table->colName (yCol) + "(Y),";
  s += table->colName (zCol) + "(Z)";
  plotAssociation = s;

  int i, j, columns = 0;
  double xv, yv;
  for (i = 0; i < r; i++)
    {
      if (!table->text (i, xCol).isEmpty ()
	  && !table->text (i, yCol).isEmpty ()
	  && !table->text (i, zCol).isEmpty ())
	{
	  xv = table->text (i, xCol).toDouble ();
	  yv = table->text (i, yCol).toDouble ();
	  if (xv >= xl && xv <= xr && yv >= yl && yv <= yr)
	    columns++;
	}
    }

  if (columns == 0)
    columns++;

  Qwt3D::Triple ** data = allocateData (columns, columns);
  for (j = 0; j < columns; j++)
    {
      int k = 0;
      for (i = 0; i < r; i++)
	{
	  if (!table->text (i, xCol).isEmpty ()
	      && !table->text (i, yCol).isEmpty ()
	      && !table->text (i, zCol).isEmpty ())
	    {
	      xv = table->text (i, xCol).toDouble ();
	      yv = table->text (i, yCol).toDouble ();
	      if (xv >= xl && xv <= xr && yv >= yl && yv <= yr)
		{
		  double zv = table->text (i, zCol).toDouble ();
		  if(scaletype==1)
		    {
		      if(zv>0) zv = log10(zv);
		      else zv = zl;
		    }
		  if (zv > zr)
		    data[k][j] = Triple (xv, yv, zr);
		  else if (zv < zl)
		    data[k][j] = Triple (xv, yv, zl);
		  else
		    data[k][j] = Triple (xv, yv, zv);
		  k++;
		}
	    }
	}
    }

  sp->makeCurrent ();
  sp->loadFromData (data, columns, columns, false, false);
  //sp->createCoordinateSystem (Triple (xl, yl, zl), Triple (xr, yr, zr));
  //sp->legend ()->setLimits (zl, zr);
  setCoordinateSystem( xl, xr, yl, yr, zl, zr );
  sp->legend ()->setMajors (legendMajorTicks);

  deleteData (data, columns);
}

void
Graph3D::updateData (Table * table)
{
  if (func)			// function plot
    return;

  QStringList lst = QStringList::split(',',plotAssociation,false);
  int pos = lst[0].findRev('(');
  QString xColName = lst[0].left(pos);
  pos = lst[1].findRev('(');
  QString yColName = lst[1].left(pos);

  int xCol = table->colIndex (xColName);
  int yCol = table->colIndex (yColName);

  if (lst.count() > 2)
    {
      pos = lst[2].findRev('(');
      QString zColName = lst[2].left(pos);
      int zCol = table->colIndex (zColName);
      updateDataXYZ (table, xCol, yCol, zCol);
    }
  else
    updateDataXY (table, xCol, yCol);

  findBestLayout ();
  update ();
}

void
Graph3D::updateDataXY (Table * table, int xCol, int yCol)
{
  int r = table->tableRows ();
  int i, j, xmesh = 0, ymesh = 2;
  double ymin;
  bool first = true;

  for (i = 0; i < r; i++)
    {
      if (!table->text (i, xCol).isEmpty ()
	  && !table->text (i, yCol).isEmpty ())
	{
	  xmesh++;
	  double y = table->text(i,yCol).toDouble();
	  if(y>0)
	    {
	      if(first){ ymin = y; first = false; }
	      else if( y<ymin ) ymin = y;
	    }
	}
    }

  if (xmesh < 2)
    {
      sp->setPlotStyle (NOPLOT);
      update ();
      return;
    }

  double **data = matrix (0, xmesh - 1, 0, ymesh - 1);
  gsl_vector *x = gsl_vector_alloc (xmesh);
  gsl_vector *y = gsl_vector_alloc (xmesh);

  for (j = 0; j < ymesh; j++)
    {
      int k = 0;
      for (i = 0; i < r; i++)
	{
	  if (!table->text (i, xCol).isEmpty ()
	      && !table->text (i, yCol).isEmpty ())
	    {
	      double xv = table->text (i, xCol).toDouble ();
	      double yv = table->text (i, yCol).toDouble ();
	      if(scaleType[2]==1)//logscale
		{
		  if(yv>0) yv = log10(yv);
		  else yv = log10(ymin);
		}

	      gsl_vector_set (x, k, xv);
	      gsl_vector_set (y, k, yv);

	      data[k][j] = yv;
	      k++;
	    }
	}
    }

  double minx = gsl_vector_min (x);
  double maxx = gsl_vector_max (x);
  double minz = gsl_vector_min (y);
  double maxz = gsl_vector_max (y);
  double miny, maxy;

  sp->makeCurrent ();
  resetNonEmptyStyle ();
  sp->coordinates ()->axes[Y1].limits (miny, maxy);	//actual Y scale limits
  sp->loadFromData (data, xmesh, ymesh, minx, maxx, miny, maxy);
  if(scaleType[2]==1)
    {
      sp->coordinates ()->axes[Z1].setScale (LOG10SCALE);
      sp->coordinates ()->axes[Z2].setScale (LOG10SCALE);
      sp->coordinates ()->axes[Z3].setScale (LOG10SCALE);
      sp->coordinates ()->axes[Z4].setScale (LOG10SCALE);
      sp->legend()->setScale(LOG10SCALE);
    }
  sp->legend ()->setLimits (minz, maxz);
  sp->legend ()->setMajors (legendMajorTicks);

  gsl_vector_free (x);
  gsl_vector_free (y);
  free_matrix (data, 0, xmesh - 1, 0, ymesh - 1);
}

void
Graph3D::updateDataXYZ (Table * table, int xCol, int yCol, int zCol)
{
  int r = table->tableRows ();
  int i, j, columns = 0;
  double zmin;
  bool first = true;

  for (i = 0; i < r; i++)
    {
      if (!table->text (i, xCol).isEmpty ()
	  && !table->text (i, yCol).isEmpty ()
	  && !table->text (i, zCol).isEmpty ())
	{
	  columns++;
	  double z = table->text(i, zCol).toDouble();
	  if(z>0)
	    {
	      if(first) { zmin = z; first = false; }
	      else if(z<zmin) zmin = z;
	    }
	}
    }

  if (columns < 2)
    {
      sp->setPlotStyle (NOPLOT);
      update ();
      return;
    }

  Qwt3D::Triple ** data = allocateData (columns, columns);
  gsl_vector *z = gsl_vector_alloc (columns);

  for (j = 0; j < columns; j++)
    {
      int k = 0;
      for (i = 0; i < r; i++)
	{
	  if (!table->text (i, xCol).isEmpty ()
	      && !table->text (i, yCol).isEmpty ()
	      && !table->text (i, zCol).isEmpty ())
	    {
	      double xv = table->text (i, xCol).toDouble ();
	      double yv = table->text (i, yCol).toDouble ();
	      double zv = table->text (i, zCol).toDouble ();
	      if(scaleType[2]==1)
		{
		  if(zv>0)
		    zv = log10(zv);
		  else
		    zv = log10(zmin);
		}
	      gsl_vector_set (z, k, zv);
	      data[k][j] = Triple (xv, yv, zv);
	      k++;
	    }
	}
    }

  double minz = gsl_vector_min (z);
  double maxz = gsl_vector_max (z);
  gsl_vector_free (z);

  sp->makeCurrent ();
  resetNonEmptyStyle ();

  sp->loadFromData (data, columns, columns, false, false);
  if(scaleType[2]==1)
    {
      sp->coordinates ()->axes[Z1].setScale (LOG10SCALE);
      sp->coordinates ()->axes[Z2].setScale (LOG10SCALE);
      sp->coordinates ()->axes[Z3].setScale (LOG10SCALE);
      sp->coordinates ()->axes[Z4].setScale (LOG10SCALE);
      sp->legend()->setScale(LOG10SCALE);
    }
  sp->legend ()->setLimits (minz, maxz);
  sp->legend ()->setMajors (legendMajorTicks);

  deleteData (data, columns);
}

void
Graph3D::updateMatrixData()
{
  if(!matrix_) return;
  double x1 = matrix_->xStart();
  double x2 = matrix_->xEnd();
  double xl = QMIN(x1,x2);
  double xr = QMAX(x1,x2);
  double y1 = matrix_->yStart();
  double y2 = matrix_->yEnd();
  double yl = QMIN(y1,y2);
  double yr = QMAX(y1,y2);
  double zl,zr;
  matrix_->range(&zl,&zr);
  updateScalesFromMatrix(xl, xr, yl, yr, zl, zr);
  findBestLayout();
}

void
Graph3D::resetNonEmptyStyle ()
{
  if (sp->plotStyle () != Qwt3D::NOPLOT)
    return;			// the plot was not previousely emptied

  if (style_ == Qwt3D::USER)
    {				// reseting the right user plot style
      switch (pointStyle)
	{
	case None:
	  break;

	case Dots:
	  sp->setPlotStyle (Dot (pointSize, smooth));
	  break;

	case VerticalBars:
	  sp->setPlotStyle (Bar (barsRad));
	  break;

	case HairCross:
	  sp->
	    setPlotStyle (CrossHair
			  (crossHairRad, crossHairLineWidth, crossHairSmooth,
			   crossHairBoxed));
	  break;

	case Cones:
	  sp->setPlotStyle (Cone3D (conesRad, conesQuality));
	  break;
	}
    }
  else
    sp->setPlotStyle (style_);
}

UserFunction *
Graph3D::userFunction ()
{
  if (func)
    return func;
  else
    return 0;
}

void
Graph3D::update ()
{
  sp->makeCurrent ();

  resetAxesLabels ();

  sp->updateData ();
  sp->updateGL ();
}

void
Graph3D::adjustLabels (int val)
{
  if (labelsDist != val)
    {
      labelsDist = val;
      sp->coordinates ()->adjustLabels (val);
      sp->makeCurrent ();
      sp->updateGL ();
    }
  emit modified ();
}

QFont
Graph3D::numbersFont ()
{
  return sp->coordinates ()->axes[X1].numberFont ();
}

void
Graph3D::setNumbersFont (const QFont & font)
{
  sp->coordinates ()->setNumberFont (font);
  sp->legend()->setNumberFont(font);
  int w = currentSize.width ();
  int h = currentSize.height ();
  double scale = sqrt((double)(500*450) / (double)(w*h));
  numberFontSize = font.pointSizeFloat() * scale;
  sp->makeCurrent ();
  sp->updateGL ();
}

void
Graph3D::setNumbersFont (const QStringList & lst)
{
  QFont fnt =
    QFont (lst[1], lst[2].toInt (), lst[3].toInt (), lst[4].toInt ());
  setNumbersFont( fnt );
}

void
Graph3D::setAxisLabelFont(int axis, const QFont & fnt )
{
  int A1,A2,A3,A4;
  switch(axis)
    {
    case 0:
      A1 = X1; A2 = X2; A3 = X3; A4 = X4;
      break;
    case 1:
      A1 = Y1; A2 = Y2; A3 = Y3; A4 = Y4;
      break;
    case 2:
      A1 = Z1; A2 = Z2; A3 = Z3; A4 = Z4;
      break;
    }

  sp->coordinates ()->axes[A1].setLabelFont (fnt);
  sp->coordinates ()->axes[A2].setLabelFont (fnt);
  sp->coordinates ()->axes[A3].setLabelFont (fnt);
  sp->coordinates ()->axes[A4].setLabelFont (fnt);

  int w = currentSize.width ();
  int h = currentSize.height ();
  double scale = sqrt( (double)(500*450) / (double)(w*h) );
  axisLabelFontSize[axis] = fnt.pointSizeFloat() * scale;
}

void
Graph3D::setAxisLabelFont(int axis, const QStringList & lst)
{
  QFont fnt =
    QFont (lst[1], lst[2].toInt (), lst[3].toInt (), lst[4].toInt ());
  setAxisLabelFont( axis, fnt );
}

QStringList
Graph3D::axisTickLengths ()
{
  QStringList lst;
  double majorl, minorl;

  sp->coordinates ()->axes[X1].ticLength (majorl, minorl);
  lst << QString::number (majorl);
  lst << QString::number (minorl);

  sp->coordinates ()->axes[Y1].ticLength (majorl, minorl);
  lst << QString::number (majorl);
  lst << QString::number (minorl);

  sp->coordinates ()->axes[Z1].ticLength (majorl, minorl);
  lst << QString::number (majorl);
  lst << QString::number (minorl);

  return lst;
}

void
Graph3D::updateTickLength (int axis, double majorLength, double minorLength)
{
  double majorl, minorl;
  int A1,A2,A3,A4;
  switch(axis)
    {
    case 0:
      A1 = X1; A2 = X2; A3 = X3; A4 = X4;
      break;
    case 1:
      A1 = Y1; A2 = Y2; A3 = Y3; A4 = Y4;
      break;
    case 2:
      A1 = Z1; A2 = Z2; A3 = Z3; A4 = Z4;
      break;
    }
  sp->coordinates ()->axes[A1].ticLength (majorl, minorl);
  if (majorl != majorLength || minorl != minorLength)
    {
      sp->coordinates ()->axes[A1].setTicLength (majorLength,
						 minorLength);
      sp->coordinates ()->axes[A2].setTicLength (majorLength,
						 minorLength);
      sp->coordinates ()->axes[A3].setTicLength (majorLength,
						 minorLength);
      sp->coordinates ()->axes[A4].setTicLength (majorLength,
						 minorLength);
    }
  sp->updateGL ();
}

void
Graph3D::setTickLengths (const QStringList & lst)
{
  double majorl, minorl;
  QStringList tick_length = lst;
  if (int (lst.count ()) > 6)
    tick_length.remove (tick_length.first ());

  majorl = tick_length[0].toDouble ();
  minorl = tick_length[1].toDouble ();
  updateTickLength(0, majorl, minorl);

  majorl = tick_length[2].toDouble ();
  minorl = tick_length[3].toDouble ();
  updateTickLength(1, majorl, minorl);

  majorl = tick_length[4].toDouble ();
  minorl = tick_length[5].toDouble ();
  updateTickLength(2, majorl, minorl);

}

void
Graph3D::rotationChanged (double, double, double)
{
  emit modified ();
}

void
Graph3D::scaleChanged (double, double, double)
{
  emit modified ();
}

void
Graph3D::shiftChanged (double, double, double)
{
  emit modified ();
}

void
Graph3D::zoomChanged (double)
{
  emit modified ();
}

void
Graph3D::resetAxesLabels ()
{
  setLabelString( 0, labels[0] );
  setLabelString( 1, labels[1] );
  setLabelString( 2, labels[2] );
}

void
Graph3D::setAxesLabels (const QStringList & l)
{
  labels = l;
  resetAxesLabels();
}

void
Graph3D::updateLabel (int axis, const QString & label, const QFont & f)
{
  if (labels[axis] != label)
    {
      setLabelString( axis, label );
      labels[axis] = label;
    }
  int A1;
  switch (axis)
    {
    case 0 : A1 = X1; break;
    case 1 : A1 = Y1; break;
    case 2 : A1 = Z1; break;
    }
  if (sp->coordinates ()->axes[A1].labelFont () != f)
    setAxisLabelFont(axis,f);

  sp->makeCurrent ();
  sp->updateGL ();
  emit modified ();
}

QFont
Graph3D::axisLabelFont (int axis)
{
  switch(axis)
    {
    case 0:
      return sp->coordinates ()->axes[X1].labelFont ();
    case 1:
      return sp->coordinates ()->axes[Y1].labelFont ();
    case 2:
      return sp->coordinates ()->axes[Z1].labelFont ();
    }
}

double
Graph3D::xStart ()
{
  double start, stop;
  sp->coordinates ()->axes[X1].limits (start, stop);
  return start;
}

double
Graph3D::xStop ()
{
  double start, stop;
  sp->coordinates ()->axes[X1].limits (start, stop);
  return stop;
}

double
Graph3D::yStart ()
{
  double start, stop;
  sp->coordinates ()->axes[Y1].limits (start, stop);
  return start;
}

double
Graph3D::yStop ()
{
  double start, stop;
  sp->coordinates ()->axes[Y1].limits (start, stop);
  return stop;
}

double
Graph3D::zStart ()
{
  double start, stop;
  sp->coordinates ()->axes[Z1].limits (start, stop);
  return start;
}

double
Graph3D::zStop ()
{
  double start, stop;
  sp->coordinates ()->axes[Z1].limits (start, stop);
  return stop;
}

QStringList
Graph3D::scaleLimits ()
{
  QStringList limits;
  double start, stop;
  int majors, minors;

  sp->coordinates ()->axes[X1].limits (start, stop);
  majors = sp->coordinates ()->axes[X1].majors ();
  minors = sp->coordinates ()->axes[X1].minors ();

  limits << QString::number (start);
  limits << QString::number (stop);
  limits << QString::number (majors);
  limits << QString::number (minors);
  limits << QString::number (scaleType[0]);

  sp->coordinates ()->axes[Y1].limits (start, stop);
  majors = sp->coordinates ()->axes[Y1].majors ();
  minors = sp->coordinates ()->axes[Y1].minors ();

  limits << QString::number (start);
  limits << QString::number (stop);
  limits << QString::number (majors);
  limits << QString::number (minors);
  limits << QString::number (scaleType[1]);

  sp->coordinates ()->axes[Z1].limits (start, stop);
  majors = sp->coordinates ()->axes[Z1].majors ();
  minors = sp->coordinates ()->axes[Z1].minors ();

  if(scaleType[2]==1)//logscale
    {
      start = pow(10,start);
      stop = pow(10,stop);
    }
  limits << QString::number (start);
  limits << QString::number (stop);
  limits << QString::number (majors);
  limits << QString::number (minors);
  limits << QString::number (scaleType[2]);

  return limits;
}

QStringList
Graph3D::scaleTicks ()
{
  QStringList limits;
  int majors, minors;

  majors = sp->coordinates ()->axes[X1].majors ();
  minors = sp->coordinates ()->axes[X1].minors ();
  limits << QString::number (majors);
  limits << QString::number (minors);

  majors = sp->coordinates ()->axes[Y1].majors ();
  minors = sp->coordinates ()->axes[Y1].minors ();
  limits << QString::number (majors);
  limits << QString::number (minors);

  majors = sp->coordinates ()->axes[Z1].majors ();
  minors = sp->coordinates ()->axes[Z1].minors ();
  limits << QString::number (majors);
  limits << QString::number (minors);

  return limits;
}

void
Graph3D::updateScale (int axis, const QStringList & options)
{
  QString st = QString::number (scaleType[axis]);
  double start, stop;
  double xl, xr, yl, yr;
  int majors, minors, newMaj, newMin;

  sp->makeCurrent ();
  switch (axis)
    {
    case 0:
      majors = sp->coordinates ()->axes[X1].majors ();
      minors = sp->coordinates ()->axes[X1].minors ();
      sp->coordinates ()->axes[X1].limits (xl, xr);

      if (xl != options[0].toDouble () || xr != options[1].toDouble ())
	{
	  xl = options[0].toDouble ();
	  xr = options[1].toDouble ();
	  sp->coordinates ()->axes[Y1].limits (yl, yr);
	  sp->coordinates ()->axes[Z1].limits (start, stop);

	  updateScales (xl, xr, yl, yr, start, stop);
	}

      newMaj = options[2].toInt ();
      if (majors != newMaj)
	setMajors( 0, newMaj );
      newMin = options[3].toInt ();
      if (minors != newMin)
	setMinors( 0, newMin );
      break;

    case 1:
      majors = sp->coordinates ()->axes[Y1].majors ();
      minors = sp->coordinates ()->axes[Y1].minors ();
      sp->coordinates ()->axes[Y1].limits (yl, yr);

      if (yl != options[0].toDouble () || yr != options[1].toDouble ())
	{
	  yl = options[0].toDouble ();
	  yr = options[1].toDouble ();
	  sp->coordinates ()->axes[X1].limits (xl, xr);
	  sp->coordinates ()->axes[Z1].limits (start, stop);
	  updateScales (xl, xr, yl, yr, start, stop);
	}

      newMaj = options[2].toInt ();
      if (majors != newMaj)
	setMajors( 1, newMaj );
      newMin = options[3].toInt ();
      if (minors != newMin)
	setMinors( 1, newMin );

      break;

    case 2:
      majors = sp->coordinates ()->axes[Z1].majors ();
      minors = sp->coordinates ()->axes[Z1].minors ();
      sp->coordinates ()->axes[Z1].limits (start, stop);

      if (st != options[4] || start != options[0].toDouble ()
                           || stop != options[1].toDouble ())
	{
	  scaleType[axis] = options[4].toInt();//this must be done before updateScales().
	  start = options[0].toDouble ();
	  stop = options[1].toDouble ();
	  sp->coordinates ()->axes[X1].limits (xl, xr);
	  sp->coordinates ()->axes[Y1].limits (yl, yr);
	  updateScales (xl, xr, yl, yr, start, stop);
	}

      newMaj = options[2].toInt ();
      if (majors != newMaj)
	setMajors( 2, newMaj );
      newMin = options[3].toInt ();
      if (minors != newMin)
	setMinors( 2, newMin );

      break;
    }
  findBestLayout();
  update ();
  emit modified ();
}

void
Graph3D::updateScales (double xl, double xr, double yl, double yr, double zl,
		       double zr)
{
  QApplication::setOverrideCursor (waitCursor);

  if (func)
    updateFunctionGraph (xl, xr, yl, yr, zl, zr);
  else if (matrix_)
    updateScalesFromMatrix (xl, xr, yl, yr, zl, zr);
  else
    {
      QStringList lst = QStringList::split(',',plotAssociation,false);

      int pos = lst[0].findRev('(');
      QString xColName = lst[0].left(pos);
      int xCol = worksheet->colIndex (xColName);

      pos = lst[1].findRev('(');
      QString yColName = lst[1].left(pos);
      int yCol = worksheet->colIndex (yColName);

      if (lst.count() > 2)
	{
	  pos = lst[2].findRev('(');
	  QString zColName = lst[2].left(pos);
	  int zCol = worksheet->colIndex (zColName);

	  updateScales (xl, xr, yl, yr, zl, zr, xCol, yCol, zCol);
	}
      else
	updateScales (xl, xr, yl, yr, zl, zr, xCol, yCol);
    }

  QApplication::restoreOverrideCursor ();
}

void
Graph3D::updateFunctionGraph(double xl, double xr, double yl,
			     double yr, double zl, double zr)
{
  if(scaleType[2]==1)//logscale
    {
      if( zr <= 0 ) zr = 10;
      if( zl <= 0 || zl > zr ) zl = zr * 0.001;
      zl = log10(zl);
      zr = log10(zr);
      func->setLogscale(true, zl);
    }
  else
    func->setLogscale(false, 0);

  func->setMinZ (zl);
  func->setMaxZ (zr);
  func->create ();
  //sp->createCoordinateSystem (Triple (xl, yl, zl),
  //			      Triple (xr, yr, zr));
  //sp->legend ()->setLimits (zl, zr);
  setCoordinateSystem( xl, xr, yl, yr, zl, zr );
}

void
Graph3D::updateScalesFromMatrix (double xl, double xr, double yl,
                                 double yr, double zl, double zr)
{
  if ( xr<=xl || yr<=yl || zr<=zl ) return;

  if(scaleType[2]==1)//logscale
    {
      if( zr <= 0 ) zr = 10;
      if( zl <= 0 ) zl = zr * 0.001;
      zl = log10(zl);
      zr = log10(zr);
    }

  double x1 = matrix_->xStart();
  double x2 = matrix_->xEnd();
  double xmin = QMIN(x1,x2);
  double xmax = QMAX(x1,x2);
  double y1 = matrix_->yStart();
  double y2 = matrix_->yEnd();
  double ymin = QMIN(y1,y2);
  double ymax = QMAX(y1,y2);
  int rows = matrix_->numRows();
  int cols = matrix_->numCols();

  if (xl<xmin) xl = xmin;
  if (xr>xmax) xr = xmax;
  if (yl<ymin) yl = ymin;
  if (yr>ymax) yr = ymax;

  double dx = ( xmax - xmin ) / (rows - 1);
  double dy = ( ymax - ymin ) / (cols - 1);

  int nr = qRound( (xr-xl)/dx ) + 1;
  int nc = qRound( (yr-yl)/dy ) + 1;

  if( nr < 2 || nc < 2 ) return;

  double **data_matrix = matrix (0, nr - 1, 0, nc - 1);
  if(xl==x1 && xr==x2 && yl==y1 && yr==y2)
    {
      for(int i=0;i<nr;i++)
	for(int j=0;j<nc;j++)
	  data_matrix[i][j] = matrix_->text(i,j).toDouble();
    }
  else
    matrix_->interpolate( xl, xr, nr, yl, yr, nc, 1, data_matrix );

  for (int j = 0; j < nc; j++)
    for (int i = 0; i < nr; i++)
      {
	double val = data_matrix[i][j];
	if(scaleType[2]==1)//logscale
	  {
	    if(val>0)
	      val = log10(val);
	    else
	      val = zl;
	  }
	
	if (val > zr)
	  data_matrix[i][j] = zr;
	else if (val < zl)
	  data_matrix[i][j] = zl;
	else
	  data_matrix[i][j] = val;
      }

  sp->loadFromData (data_matrix, nr, nc, xl, xr, yl, yr);
  free_matrix (data_matrix, 0, nr - 1, 0, nc - 1);

  setCoordinateSystem( xl, xr, yl, yr, zl, zr );
  sp->legend ()->setMajors (legendMajorTicks);
  //update ();
}

void
Graph3D::updateScales (double xl, double xr, double yl, double yr, double zl,
		       double zr, int xcol, int ycol)
{
  if(scaleType[2]==1)//logscale
    {
      if( zr <= 0 ) zr = 10;
      if( zl <= 0 || zl > zr ) zl = zr * 0.001;
      zl = log10(zl);
      zr = log10(zr);
    }

  int r = worksheet->tableRows ();
  int i, j, xmesh = 0, ymesh = 2;
  double xv, yv;

  for (i = 0; i < r; i++)
    {
      if (!worksheet->text (i, xcol).isEmpty ()
	  && !worksheet->text (i, ycol).isEmpty ())
	{
	  xv = worksheet->text (i, xcol).toDouble ();
	  if (xv >= xl && xv <= xr)
	    xmesh++;
	}
    }

  if (xmesh == 0)
    xmesh++;

  double **data = matrix (0, xmesh - 1, 0, ymesh - 1);

  for (j = 0; j < ymesh; j++)
    {
      int k = 0;
      for (i = 0; i < r; i++)
	{
	  if (!worksheet->text (i, xcol).isEmpty ()
	      && !worksheet->text (i, ycol).isEmpty ())
	    {
	      xv = worksheet->text (i, xcol).toDouble ();
	      if (xv >= xl && xv <= xr)
		{
		  yv = worksheet->text (i, ycol).toDouble ();
		  if(scaleType[2]==1)//logscale
		    {
		      if(yv>0)
			yv = log10(yv);
		      else
			yv = zl;
		    }
		  if (yv > zr)
		    data[k][j] = zr;
		  else if (yv < zl)
		    data[k][j] = zl;
		  else
		    data[k][j] = yv;
		  k++;
		}
	    }
	}
    }

  sp->loadFromData (data, xmesh, ymesh, xl, xr, yl, yr);

  setCoordinateSystem( xl, xr, yl, yr, zl, zr );
  free_matrix (data, 0, xmesh - 1, 0, ymesh - 1);
}

void
Graph3D::updateScales (double xl, double xr, double yl, double yr, double zl,
		       double zr, int xCol, int yCol, int zCol)
{
  if(scaleType[2]==1)//logscale
    {
      if( zr <= 0 ) zr = 10;
      if( zl <= 0 || zl > zr ) zl = zr * 0.001;
      zl = log10(zl);
      zr = log10(zr);
    }

  int r = worksheet->tableRows ();
  int i, j, columns = 0;
  double xv, yv, zv;
  for (i = 0; i < r; i++)
    {
      if (!worksheet->text (i, xCol).isEmpty ()
	  && !worksheet->text (i, yCol).isEmpty ()
	  && !worksheet->text (i, zCol).isEmpty ())
	{
	  xv = worksheet->text (i, xCol).toDouble ();
	  yv = worksheet->text (i, yCol).toDouble ();
	  if (xv >= xl && xv <= xr && yv >= yl && yv <= yr)
	    columns++;
	}
    }

  if (columns == 0)
    columns++;

  Qwt3D::Triple ** data = allocateData (columns, columns);
  for (j = 0; j < columns; j++)
    {
      int k = 0;
      for (i = 0; i < r; i++)
	{
	  if (!worksheet->text (i, xCol).isEmpty ()
	      && !worksheet->text (i, yCol).isEmpty ()
	      && !worksheet->text (i, zCol).isEmpty ())
	    {
	      xv = worksheet->text (i, xCol).toDouble ();
	      yv = worksheet->text (i, yCol).toDouble ();
	      if (xv >= xl && xv <= xr && yv >= yl && yv <= yr)
		{
		  zv = worksheet->text (i, zCol).toDouble ();
		  if(scaleType[2]==1)//logscale
		    {
		      if(zv>0)
			zv = log10(zv);
		      else
			zv = zl;
		    }
		  if (zv > zr)
		    data[k][j] = Triple (xv, yv, zr);
		  else if (zv < zl)
		    data[k][j] = Triple (xv, yv, zl);
		  else
		    data[k][j] = Triple (xv, yv, zv);
		  k++;
		}
	    }
	}
    }
  sp->loadFromData (data, columns, columns, false, false);
  //sp->createCoordinateSystem (Triple (xl, yl, zl), Triple (xr, yr, zr));
  setCoordinateSystem( xl, xr, yl, yr, zl, zr );
  deleteData (data, columns);
}

void
Graph3D::setTicks (const QStringList & options)
{
  if (int (options.count ()) == 6)
    {
      setMajors( 0, options[0].toInt () );
      setMinors( 0, options[1].toInt () );
      setMajors( 1, options[2].toInt () );
      setMinors( 1, options[3].toInt () );
      setMajors( 2, options[4].toInt () );
      setMinors( 2, options[5].toInt () );
    }
  else
    {
      setMajors( 0, options[1].toInt () );
      setMinors( 0, options[2].toInt () );
      setMajors( 1, options[3].toInt () );
      setMinors( 1, options[4].toInt () );
      setMajors( 2, options[5].toInt () );
      setMinors( 2, options[6].toInt () );
    }
}

void
Graph3D::setColors (const QColor & meshColor, const QColor & axesColor,
		    const QColor & numColor, const QColor & labelColor,
		    const QColor & bgColor, const QColor & gridColor)
{
  if (meshCol != meshColor)
    {
      sp->setMeshColor (Qt2GL (meshColor));
      meshCol = meshColor;
    }

  if (axesCol != axesColor)
    {
      sp->coordinates ()->setAxesColor (Qt2GL (axesColor));
      axesCol = axesColor;
    }

  if (numCol != numColor)
    {
      sp->coordinates ()->setNumberColor (Qt2GL (numColor));
      numCol = numColor;
    }

  if (labelsCol != labelColor)
    {
      sp->coordinates ()->setLabelColor (Qt2GL (labelColor));
      labelsCol = labelColor;
    }

  if (bgCol != bgColor)
    {
      sp->setBackgroundColor (Qt2GL (bgColor));
      bgCol = bgColor;
    }

  if (gridCol != gridColor)
    {
      sp->coordinates ()->setGridLinesColor (Qt2GL (gridColor));
      gridCol = gridColor;
    }
}

void
Graph3D::setColors (const QStringList & colors)
{
  meshCol = QColor (colors[1]);
  sp->setMeshColor (Qt2GL (meshCol));

  axesCol = QColor (colors[2]);
  sp->coordinates ()->setAxesColor (Qt2GL (axesCol));

  numCol = QColor (colors[3]);
  sp->coordinates ()->setNumberColor (Qt2GL (numCol));

  labelsCol = QColor (colors[4]);
  sp->coordinates ()->setLabelColor (Qt2GL (labelsCol));

  bgCol = QColor (colors[5]);
  sp->setBackgroundColor (Qt2GL (bgCol));

  gridCol = QColor (colors[6]);
  sp->coordinates ()->setGridLinesColor (Qt2GL (gridCol));

  if ((int) colors.count () > 7)
    {
      QColor min = QColor (colors[7]);
      QColor max = QColor (colors[8]);
      alpha = colors[9].toDouble ();

      if ((int) colors.count () == 11)
	setDataColorMap (colors[10]);
      else
	setDataColors (min, max);
    }
}

void
Graph3D::updateColors (const QColor & meshColor, const QColor & axesColor,
		       const QColor & numColor, const QColor & labelColor,
		       const QColor & bgColor, const QColor & gridColor)
{
  if (meshCol != meshColor)
    {
      sp->setMeshColor (Qt2GL (meshColor));
      meshCol = meshColor;
    }

  if (axesCol != axesColor)
    {
      sp->coordinates ()->setAxesColor (Qt2GL (axesColor));
      axesCol = axesColor;
    }

  if (numCol != numColor)
    {
      sp->coordinates ()->setNumberColor (Qt2GL (numColor));
      numCol = numColor;
    }

  if (labelsCol != labelColor)
    {
      sp->coordinates ()->setLabelColor (Qt2GL (labelColor));
      labelsCol = labelColor;
    }

  if (bgCol != bgColor)
    {
      sp->setBackgroundColor (Qt2GL (bgColor));
      bgCol = bgColor;
    }

  if (gridCol != gridColor)
    {
      sp->coordinates ()->setGridLinesColor (Qt2GL (gridColor));
      gridCol = gridColor;
    }

  sp->updateData ();
  sp->updateGL ();
  emit modified ();
}

void
Graph3D::scaleFonts (double factor)
{
  //fprintf(stderr,"factor = %f\n",factor );

  QFont font = sp->coordinates ()->axes[X1].numberFont ();

  font.setPointSizeFloat (numberFontSize * factor);
  sp->coordinates ()->setNumberFont (font);
  sp->legend()->setNumberFont(font);

  //fprintf(stderr,"setting Title Font 2\n");
  titleFnt.setPointSizeFloat (factor * titleFontSize);
  sp->setTitleFont (titleFnt.family (), titleFnt.pointSize (),
		    titleFnt.weight (), titleFnt.italic ());

  //fprintf(stderr,"setting Xaxis Font\n");
  font = axisLabelFont (0);
  font.setPointSizeFloat (factor * axisLabelFontSize[0]);
  setAxisLabelFont( 0, font );

  //fprintf(stderr,"setting Yaxis Font\n");
  font = axisLabelFont (1);
  font.setPointSizeFloat (factor * axisLabelFontSize[1]);
  setAxisLabelFont( 1, font );

  //fprintf(stderr,"setting Zaxis Font\n");
  font = axisLabelFont (2);
  font.setPointSizeFloat (factor * axisLabelFontSize[2]);
  setAxisLabelFont( 2, font );
}

void
Graph3D::resizeEvent (QResizeEvent * e)
{
  QSize size = e->size ();
  QSize oldSize = e->oldSize ();

  //fprintf(stderr,"Graph3D::resizeEvent\n");
  //fprintf(stderr,"old size = %d %d\n",oldSize.width(),oldSize.height());
  //fprintf(stderr,"new size = %d %d\n",size.width(),size.height());

  sp->makeCurrent ();
  sp->resize (size);
  currentSize = size;

  //if (!ignoreFonts && this->isVisible ())
  if (!ignoreFonts)
    {
      if (oldSize.height()<=40) return;

      double ratio = sqrt( (double)size.height() * (double)size.width() /
			  (500*450) );
      scaleFonts (ratio);
    }
  else{
    //fprintf(stderr,"ignoreFonts\n");
  }

  sp->updateGL ();
  emit resizedWindow (this);
  emit modified ();
}

void
Graph3D::contextMenuEvent (QContextMenuEvent * e)
{
  emit showContextMenu ();
  e->accept ();
}

void
Graph3D::setFramed ()
{
  if (sp->coordinates ()->style () == FRAME)
    return;

  sp->makeCurrent ();
  sp->setCoordinateStyle (FRAME);
}

void
Graph3D::setBoxed ()
{
  if (sp->coordinates ()->style () == BOX)
    return;

  sp->makeCurrent ();
  sp->setCoordinateStyle (BOX);
}

void
Graph3D::setNoAxes ()
{
  if (sp->coordinates ()->style () == NOCOORD)
    return;

  sp->makeCurrent ();
  sp->setCoordinateStyle (NOCOORD);
}

void
Graph3D::setNoGrid ()
{
  if (sp->plotStyle () == FILLED)
    return;

  sp->makeCurrent ();
  sp->setPlotStyle (FILLED);
  legendOn = true;
  sp->showColorLegend (legendOn);
  sp->updateData ();
  sp->updateGL ();

  style_ = FILLED;
  pointStyle = None;
}

void
Graph3D::setFilledMesh ()
{
  if (sp->plotStyle () == FILLEDMESH)
    return;

  sp->makeCurrent ();
  sp->setPlotStyle (FILLEDMESH);
  legendOn = true;
  sp->showColorLegend (legendOn);
  sp->updateData ();
  sp->updateGL ();

  style_ = FILLEDMESH;
  pointStyle = None;
}

void
Graph3D::setHiddenLineGrid ()
{
  if (sp->plotStyle () == HIDDENLINE)
    return;

  sp->makeCurrent ();
  sp->setPlotStyle (HIDDENLINE);
  legendOn = false;
  sp->showColorLegend (legendOn);
  sp->updateData ();
  sp->updateGL ();

  style_ = HIDDENLINE;
  pointStyle = None;
}

void
Graph3D::setLineGrid ()
{
  if (sp->plotStyle () == WIREFRAME)
    return;

  sp->makeCurrent ();
  sp->setPlotStyle (WIREFRAME);
  legendOn = false;
  sp->showColorLegend (legendOn);
  sp->updateData ();
  sp->updateGL ();

  pointStyle = None;
  style_ = WIREFRAME;
}

void
Graph3D::setPointsMesh ()
{
  if (!sp || pointStyle == Dots)
    return;

  pointStyle = Dots;
  style_ = Qwt3D::USER;

  sp->makeCurrent ();
  sp->setPlotStyle (Dot (pointSize, smooth));
  legendOn = true;
  sp->showColorLegend (legendOn);
  sp->updateData ();
  sp->updateGL ();
}

void
Graph3D::setConesMesh ()
{
  if (!sp || pointStyle == Cones)
    return;

  QApplication::setOverrideCursor (waitCursor);

  pointStyle = Cones;
  style_ = Qwt3D::USER;

  sp->makeCurrent ();
  sp->setPlotStyle (Cone3D (conesRad, conesQuality));
  legendOn = true;
  sp->showColorLegend (legendOn);
  sp->updateData ();
  sp->updateGL ();

  QApplication::restoreOverrideCursor ();
}

void
Graph3D::setCrossMesh ()
{
  if (!sp || pointStyle == HairCross)
    return;

  pointStyle = HairCross;
  style_ = Qwt3D::USER;

  sp->makeCurrent ();
  sp->
    setPlotStyle (CrossHair
		  (crossHairRad, crossHairLineWidth, crossHairSmooth,
		   crossHairBoxed));
  legendOn = true;
  sp->showColorLegend (legendOn);
  sp->updateData ();
  sp->updateGL ();
}

void
Graph3D::clearData ()
{
  if (matrix_)
    matrix_ = 0;
  else if (worksheet)
    worksheet = 0;
  else if (func)
    {
      delete func;
      func = 0;
    }
  plotAssociation = QString::null;

  sp->makeCurrent ();
  sp->loadFromData (0, 0, 0, false, false);
  sp->updateData ();
  sp->updateGL ();
}

void
Graph3D::setBarsPlot ()
{
  if (pointStyle == VerticalBars)
    return;

  QApplication::setOverrideCursor (waitCursor);

  pointStyle = VerticalBars;
  style_ = Qwt3D::USER;

  sp->makeCurrent ();
  sp->setPlotStyle (Bar (barsRad));
  sp->updateData ();
  sp->updateGL ();
  QApplication::restoreOverrideCursor ();
}

void
Graph3D::setFloorData ()
{
  if (sp->floorStyle () == FLOORDATA)
    return;

  sp->makeCurrent ();
  sp->setFloorStyle (FLOORDATA);
  sp->updateData ();
  sp->updateGL ();
}

void
Graph3D::setFloorIsolines ()
{
  if (sp->floorStyle () == FLOORISO)
    return;

  sp->makeCurrent ();
  sp->setFloorStyle (FLOORISO);
  sp->updateData ();
  sp->updateGL ();
}

void
Graph3D::setEmptyFloor ()
{
  if (sp->floorStyle () == NOFLOOR)
    return;

  sp->makeCurrent ();
  sp->setFloorStyle (NOFLOOR);
  sp->updateData ();
  sp->updateGL ();
}

void
Graph3D::setMeshLineWidth (int lw)
{
  if ((int) sp->meshLineWidth () == lw)
    return;

  sp->makeCurrent ();
  sp->setMeshLineWidth ((double) lw);
  sp->updateData ();
  sp->updateGL ();
}

int
Graph3D::grids ()
{
  return sp->coordinates ()->grids ();
}

void
Graph3D::setGrid (Qwt3D::SIDE s, bool b)
{
  if (!sp)
    return;

  int sum = sp->coordinates ()->grids ();

  if (b)
    sum |= s;
  else
    sum &= ~s;

  sp->coordinates ()->setGridLines (sum != Qwt3D::NOSIDEGRID, FALSE, sum);
  sp->updateGL ();
  emit modified ();
}

void
Graph3D::setGrid (int grids)
{
  if (!sp)
    return;

  sp->coordinates ()->setGridLines (TRUE, FALSE, grids);
}

void
Graph3D::setLeftGrid (bool b)
{
  setGrid (Qwt3D::LEFT, b);
}

void
Graph3D::setRightGrid (bool b)
{
  setGrid (Qwt3D::RIGHT, b);
}

void
Graph3D::setCeilGrid (bool b)
{
  setGrid (Qwt3D::CEIL, b);
}

void
Graph3D::setFloorGrid (bool b)
{
  setGrid (Qwt3D::FLOOR, b);
}

void
Graph3D::setFrontGrid (bool b)
{
  setGrid (Qwt3D::FRONT, b);
}

void
Graph3D::setBackGrid (bool b)
{
  setGrid (Qwt3D::BACK, b);
}

void
Graph3D::print ()
{
  QPrinter printer;
  printer.setOrientation (QPrinter::Landscape);
  printer.setColorMode (QPrinter::Color);
  printer.setFullPage (FALSE);

  if (printer.setup ())
    {
      if (IO::save (sp, "qtiplot.png", "PNG"))
	{
	  QPixmap p;
	  p.load ("qtiplot.png", "PNG", QPixmap::Color);

	  QPainter paint (&printer);
	  paint.drawPixmap (QPoint (0, 0), p);
	  paint.end ();

	  QFile f ("qtiplot.png");
	  f.remove ();
	}
      else
	QMessageBox::about (0, tr ("QtiPlot - IO Error"),
			    tr ("Could not print: <h4>"+getName()+"</h4>.") );
    }
}

void
Graph3D::copyImage ()
{
  bool ok = IO::save (sp, "qtiplot.png", "PNG");
  if (ok)
    {
      QPixmap p;
      p.load ("qtiplot.png", "PNG", QPixmap::Auto);
      QImage image = p.convertToImage ();
      QApplication::clipboard ()->setData (new QImageDrag (image, sp, 0));
      QFile f ("qtiplot.png");
      f.remove ();
    }
  else
    QMessageBox::about (0, tr ("QtiPlot - IO Error"),
			tr ("Could not copy: <h4>" + getName() + "</h4>."));
}

void
Graph3D::saveImage ()
{
  sp->makeCurrent ();
  int i;
  QStringList list = IO::outputFormatList ();
  list.sort ();
//QString filter="*.eps;;",selectedFilter,aux;

  QString filter = "", selectedFilter, aux;

  for (i = 0; i < (int) list.count (); i++)
    {
      aux = "*." + (list[i]).lower () + ";;";
      filter += aux;
    }

  QString fname =
    QFileDialog::getSaveFileName (QString::null, filter, 0, "file dialog",
				  tr ("Choose a filename to save under"),
				  &selectedFilter, TRUE);
  if (!fname.isEmpty ())
    {
      QFileInfo fi (fname);
      QString baseName = fi.fileName ();
      QString filetype = selectedFilter.remove ('*');

      if (baseName.contains ('.') == 0)
	fname.append (filetype);

      if (QFile::exists (fname) &&
	  QMessageBox::question (0,
				 tr ("QtiPlot - Overwrite File?"),
				 tr
				 ("A file called: <p><b>%1</b><p>already exists. "
				  "Do you want to overwrite it?").arg (fname),
				 tr ("&Yes"), tr ("&No"), QString::null, 0,
				 1))
	return;
      else
	{
	  QFile f (fname);
	  if (!f.open (IO_WriteOnly))
	    {
	      QMessageBox::critical (0, tr ("QtiPlot - Export Error"),
				     tr ("Could not write to file: <br><h4>" +
					 fname +
					 "</h4><p>Please verify that you have the right to write to this location!").
				     arg (fname));
	      return;
	    }
	  IO::save (sp, fname, filetype.upper ().remove ('.'));
	}
    }
}

void
Graph3D::saveImageToFile (const QString & fname, const QString & format)
{
  IO::save (sp, fname, format);
}

bool
Graph3D::eventFilter (QObject * object, QEvent * e)
{
  if (e->type () == QEvent::MouseButtonDblClick
      && object == (QObject *) this->sp)
    {
      emit showOptionsDialog ();
      return TRUE;
    }
  else if (e->type () == QEvent::ContextMenu && object == titleBar)
    {
      emit showTitleBarMenu ();
      ((QContextMenuEvent *) e)->accept ();
      return true;
    }
  return QObject::eventFilter (object, e);
}

void
Graph3D::setPointOptions (double size, bool s)
{
  if (pointSize == size && smooth == s)
    return;

  pointSize = size;
  smooth = s;
}

double
Graph3D::barsRadius ()
{
  if (sp->plotStyle () == Qwt3D::USER && sp->plotStyle () != Qwt3D::POINTS)
    return barsRad;
  else
    return 0.0;
}

void
Graph3D::setBarsRadius (double rad)
{
  if (barsRad == rad)
    return;

  barsRad = rad;
}

void
Graph3D::updateBars (double rad)
{
  if (barsRad == rad)
    return;

  barsRad = rad;
  sp->setPlotStyle (Bar (barsRad));
  update ();
}

void
Graph3D::updatePoints (double size, bool sm)
{
  if (pointStyle == Dots && pointSize == size && smooth == sm)
    return;

  pointSize = size;
  smooth = sm;
  pointStyle = Dots;

  sp->setPlotStyle ( Dot(pointSize, smooth) );

  update ();
  emit modified ();
  emit custom3DActions (this);
}

void
Graph3D::updateCones (double rad, int quality)
{
  if (pointStyle == Cones && conesRad == rad && conesQuality == quality)
    return;

  conesRad = rad;
  conesQuality = quality;
  pointStyle = Cones;
  sp->setPlotStyle (Cone3D (conesRad, conesQuality));
  update ();
  emit modified ();
  emit custom3DActions (this);
}

void
Graph3D::setConesOptions (double rad, int quality)
{
  conesRad = rad;
  conesQuality = quality;
}

void
Graph3D::updateCross (double rad, double linewidth, bool smooth, bool boxed)
{
  if (pointStyle == HairCross && crossHairRad == rad &&
      crossHairSmooth == smooth && crossHairBoxed == boxed &&
      crossHairLineWidth == linewidth)
    return;

  crossHairRad = rad;
  crossHairLineWidth = linewidth;
  crossHairSmooth = smooth;
  crossHairBoxed = boxed;
  pointStyle = HairCross;

  sp->setPlotStyle (CrossHair (rad, linewidth, smooth, boxed));
  update ();
  emit modified ();
  emit custom3DActions (this);
}

void
Graph3D::setCrossOptions (double rad, double linewidth, bool smooth,
			  bool boxed)
{
  crossHairRad = rad;
  crossHairLineWidth = linewidth;
  crossHairSmooth = smooth;
  crossHairBoxed = boxed;
}

void
Graph3D::setStyle (Qwt3D::COORDSTYLE coord, Qwt3D::FLOORSTYLE floor,
		   Qwt3D::PLOTSTYLE plot, Graph3D::PointStyle point)
{
  sp->setCoordinateStyle (coord);
  sp->setFloorStyle (floor);

  if (point == None)
    sp->setPlotStyle (plot);
  else if (point == VerticalBars)
    sp->setPlotStyle (Bar (barsRad));
  else if (point == Dots)
    sp->setPlotStyle (Dot (pointSize, smooth));
  else if (point == HairCross)
    sp->
      setPlotStyle (CrossHair
		    (crossHairRad, crossHairLineWidth, crossHairSmooth,
		     crossHairBoxed));
  else if (point == Cones)
    sp->setPlotStyle (Cone3D (conesRad, conesQuality));

  pointStyle = point;
  style_ = sp->plotStyle ();
}

void
Graph3D::customPlotStyle (int style)
{
  sp->makeCurrent ();
  if (sp->plotStyle () == style)
    return;

  switch (style)
    {
    case WIREFRAME:
      {
	sp->setPlotStyle (WIREFRAME);
	style_ = WIREFRAME;
	pointStyle = None;

	legendOn = false;
	sp->showColorLegend (legendOn);
	break;
      }

    case FILLED:
      {
	sp->setPlotStyle (FILLED);
	style_ = FILLED;
	pointStyle = None;
	legendOn = true;
	sp->showColorLegend (legendOn);
	break;
      }

    case FILLEDMESH:
      {
	sp->setPlotStyle (FILLEDMESH);
	style_ = FILLEDMESH;
	pointStyle = None;
	legendOn = true;
	sp->showColorLegend (legendOn);
	break;
      }

    case HIDDENLINE:
      {
	sp->setPlotStyle (HIDDENLINE);
	style_ = HIDDENLINE;
	pointStyle = None;
	legendOn = false;
	sp->showColorLegend (legendOn);
	break;
      }

    case Qwt3D::POINTS:
      {
	pointSize = 3;
	smooth = true;
	pointStyle = Dots;
	style_ = Qwt3D::USER;
	legendOn = true;
	sp->showColorLegend (legendOn);

	Dot d (pointSize, smooth);
	sp->setPlotStyle (d);
	break;
      }

    case Qwt3D::USER:
      {
	pointStyle = VerticalBars;
	style_ = Qwt3D::USER;
	sp->setPlotStyle (Bar (barsRad));
	break;
      }
    }

  sp->updateData ();
  sp->updateGL ();
}

void
Graph3D::setStyle (const QStringList & st)
{
  if (st[1] == "nocoord")
    sp->setCoordinateStyle (NOCOORD);
  else if (st[1] == "frame")
    sp->setCoordinateStyle (FRAME);
  else if (st[1] == "box")
    sp->setCoordinateStyle (BOX);

  if (st[2] == "nofloor")
    sp->setFloorStyle (NOFLOOR);
  else if (st[2] == "flooriso")
    sp->setFloorStyle (FLOORISO);
  else if (st[2] == "floordata")
    sp->setFloorStyle (FLOORDATA);

  if (st[3] == "filledmesh")
    {
      sp->setPlotStyle (FILLEDMESH);
      legendOn = true;
      sp->showColorLegend (legendOn);
    }
  else if (st[3] == "filled")
    {
      sp->setPlotStyle (FILLED);
      legendOn = true;
      sp->showColorLegend (legendOn);
    }
  else if (st[3] == "points")
    {
      pointSize = st[4].toDouble ();
      legendOn = true;
      smooth = false;
      if (st[5] == "1")
	smooth = true;

      sp->setPlotStyle (Dot (pointSize, smooth));
      pointStyle = Dots;
    }
  else if (st[3] == "wireframe")
    {
      sp->setPlotStyle (WIREFRAME);
      legendOn = false;
      sp->showColorLegend (legendOn);
    }
  else if (st[3] == "hiddenline")
    {
      sp->setPlotStyle (HIDDENLINE);
      legendOn = false;
      sp->showColorLegend (legendOn);
    }
  else if (st[3] == "bars")
    {
      barsRad = (st[4]).toDouble ();
      sp->setPlotStyle (Bar (barsRad));
      pointStyle = VerticalBars;
	  legendOn = true;

    }
  else if (st[3] == "cones")
    {
      conesRad = (st[4]).toDouble ();
      conesQuality = (st[5]).toInt ();
	  legendOn = true;

      sp->setPlotStyle (Cone3D (conesRad, conesQuality));
      pointStyle = Cones;
    }
  else if (st[3] == "cross")
    {
      crossHairRad = (st[4]).toDouble ();
      crossHairLineWidth = (st[5]).toDouble ();
	  legendOn = true;

      crossHairSmooth = false;
      if (st[6] == "1")
	crossHairSmooth = true;

      crossHairBoxed = false;
      if (st[7] == "1")
	crossHairBoxed = true;

      sp->
	setPlotStyle (CrossHair
		      (crossHairRad, crossHairLineWidth, crossHairSmooth,
		       crossHairBoxed));
      pointStyle = HairCross;
    }

  style_ = sp->plotStyle ();
}

void
Graph3D::setRotation (double xVal, double yVal, double zVal)
{
  sp->setRotation (xVal, yVal, zVal);
}

void
Graph3D::setScale (double xVal, double yVal, double zVal)
{
  sp->setScale (xVal, yVal, zVal);
}

void
Graph3D::setShift (double xVal, double yVal, double zVal)
{
  sp->setShift (xVal, yVal, zVal);
}

void
Graph3D::setZoom (double val)
{
  sp->setZoom (val);
}

void
Graph3D::updateZoom (double val)
{
  if (sp->zoom () == val)
    return;

  QApplication::setOverrideCursor (waitCursor);

  sp->makeCurrent ();
  sp->setZoom (val);
  sp->updateData ();
  sp->updateGL ();

  QApplication::restoreOverrideCursor ();
}

void
Graph3D::updateScaling (double xVal, double yVal, double zVal)
{
  if (sp->xScale () == xVal && sp->yScale () == yVal && sp->zScale () == zVal)
    return;

  QApplication::setOverrideCursor (waitCursor);

  sp->setScale (xVal, yVal, zVal);
  sp->updateData ();
  sp->updateGL ();

  QApplication::restoreOverrideCursor ();
}

Qwt3D::PLOTSTYLE Graph3D::plotStyle ()
{
  return sp->plotStyle ();
}

Qwt3D::FLOORSTYLE Graph3D::floorStyle ()
{
  return sp->floorStyle ();
}

Qwt3D::COORDSTYLE Graph3D::coordStyle ()
{
  return sp->coordinates ()->style ();
}

QString
Graph3D::formula ()
{
  if (func)
    return func->function ();
  else
    return plotAssociation;
}

QString
Graph3D::saveToString (const QString & geometry)
{
  QString s = "<SurfacePlot>\n";
  s += getName() + "\t";
  s += birthDate () + "\n";
  s += geometry;
  s += "SurfaceFunction\t";

  sp->makeCurrent ();
  if (func)
    s += func->function () + "\t";
  else
    {
      s += plotAssociation;
      s += "\t";
    }

  double start, stop;
  sp->coordinates ()->axes[X1].limits (start, stop);
  s += QString::number (start) + "\t";
  s += QString::number (stop) + "\t";
  sp->coordinates ()->axes[Y1].limits (start, stop);
  s += QString::number (start) + "\t";
  s += QString::number (stop) + "\t";
  sp->coordinates ()->axes[Z1].limits (start, stop);
  if(scaleType[2]==1)//logscale
    {
      start = pow(10,start);
      stop = pow(10,stop);
    }
  s += QString::number (start) + "\t";
  s += QString::number (stop) + "\t";

  s += QString::number (scaleType[0]) + "\t";
  s += QString::number (scaleType[1]) + "\t";
  s += QString::number (scaleType[2]) + "\n";

  QString st;
  if (sp->coordinates ()->style () == Qwt3D::NOCOORD)
    st = "nocoord";
  else if (sp->coordinates ()->style () == Qwt3D::BOX)
    st = "box";
  else
    st = "frame";
  s += "Style\t" + st + "\t";

  switch (sp->floorStyle ())
    {
    case NOFLOOR:
      st = "nofloor";
      break;

    case FLOORISO:
      st = "flooriso";
      break;

    case FLOORDATA:
      st = "floordata";
      break;
    }
  s += st + "\t";

  switch (sp->plotStyle ())
    {
    case USER:
      if (pointStyle == VerticalBars)
	st = "bars\t" + QString::number (barsRad);
      else if (pointStyle == Dots)
	{
	  st = "points\t" + QString::number (pointSize);
	  st += "\t" + QString::number (smooth);
	}
      else if (pointStyle == Cones)
	{
	  st = "cones\t" + QString::number (conesRad);
	  st += "\t" + QString::number (conesQuality);
	}
      else if (pointStyle == HairCross)
	{
	  st = "cross\t" + QString::number (crossHairRad);
	  st += "\t" + QString::number (crossHairLineWidth);
	  st += "\t" + QString::number (crossHairSmooth);
	  st += "\t" + QString::number (crossHairBoxed);
	}
      break;

    case WIREFRAME:
      st = "wireframe";
      break;

    case HIDDENLINE:
      st = "hiddenline";
      break;

    case FILLED:
      st = "filled";
      break;

    case FILLEDMESH:
      st = "filledmesh";
      break;

    default:
      ;
    }
  s += st + "\n";

  s += "grids\t";
  s += QString::number (sp->coordinates ()->grids ()) + "\n";

  s += "title\t";
  s += title + "\t";
  s += titleCol.name () + "\t";
  s += titleFnt.family () + "\t";
  s += QString::number (titleFnt.pointSize ()) + "\t";
  s += QString::number (titleFnt.weight ()) + "\t";
  s += QString::number (titleFnt.italic ()) + "\n";

  s += "colors\t";
  s += meshCol.name () + "\t";
  s += axesCol.name () + "\t";
  s += numCol.name () + "\t";
  s += labelsCol.name () + "\t";
  s += bgCol.name () + "\t";
  s += gridCol.name () + "\t";
  s += fromColor.name () + "\t";
  s += toColor.name () + "\t";
  s += QString::number (alpha) + "\t";
  s += color_map + "\n";

  s += "axesLabels\t";
  s += labels.join ("\t") + "\n";

  s += "tics\t";
  QStringList tl = scaleTicks ();
  s += tl.join ("\t") + "\n";

  s += "tickLengths\t";
  tl = axisTickLengths ();
  s += tl.join ("\t") + "\n";

  s += "options\t";
  s += QString::number (legendOn) + "\t";
  s += QString::number (sp->resolution ()) + "\t";
  s += QString::number (labelsDist) + "\n";

  s += "numbersFont\t";
  QFont fnt = sp->coordinates ()->axes[X1].numberFont ();
  s += fnt.family () + "\t";
  s += QString::number (fnt.pointSize ()) + "\t";
  s += QString::number (fnt.weight ()) + "\t";
  s += QString::number (fnt.italic ()) + "\n";

  s += "xAxisLabelFont\t";
  fnt = sp->coordinates ()->axes[X1].labelFont ();
  s += fnt.family () + "\t";
  s += QString::number (fnt.pointSize ()) + "\t";
  s += QString::number (fnt.weight ()) + "\t";
  s += QString::number (fnt.italic ()) + "\n";

  s += "yAxisLabelFont\t";
  fnt = sp->coordinates ()->axes[Y1].labelFont ();
  s += fnt.family () + "\t";
  s += QString::number (fnt.pointSize ()) + "\t";
  s += QString::number (fnt.weight ()) + "\t";
  s += QString::number (fnt.italic ()) + "\n";

  s += "zAxisLabelFont\t";
  fnt = sp->coordinates ()->axes[Z1].labelFont ();
  s += fnt.family () + "\t";
  s += QString::number (fnt.pointSize ()) + "\t";
  s += QString::number (fnt.weight ()) + "\t";
  s += QString::number (fnt.italic ()) + "\n";

  s += "rotation\t";
  s += QString::number (sp->xRotation ()) + "\t";
  s += QString::number (sp->yRotation ()) + "\t";
  s += QString::number (sp->zRotation ()) + "\n";

  s += "zoom\t";
  s += QString::number (sp->zoom ()) + "\n";

  s += "scaling\t";
  s += QString::number (sp->xScale ()) + "\t";
  s += QString::number (sp->yScale ()) + "\t";
  s += QString::number (sp->zScale ()) + "\n";

  s += "shift\t";
  s += QString::number (sp->xShift ()) + "\t";
  s += QString::number (sp->yShift ()) + "\t";
  s += QString::number (sp->zShift ()) + "\n";

  s += "LineWidth\t";
  s += QString::number (sp->meshLineWidth ()) + "\n";

  s +=
    "WindowLabel\t" + windowLabel () + "\t" +
    QString::number (captionPolicy ()) + "\n";

  s += "Orthogonal\t";
  s += QString::number (sp->ortho ()) + "\n";

  s += "</SurfacePlot>\n";
  return s;
}

void
Graph3D::showColorLegend (bool show)
{
  if (legendOn == show)
    return;

  //  if((sp->plotStyle () == FILLED)||(sp->plotStyle () == FILLEDMESH))
  //  {
  sp->makeCurrent ();
  sp->showColorLegend (show);
  
  legendOn = show;
  sp->updateGL ();
  emit modified ();
  //   }
}

void
Graph3D::setResolution (int r)
{
  if (sp->resolution () == r)
    return;

  sp->makeCurrent ();
  sp->setResolution (r);
  sp->updateData ();
  sp->updateGL ();
  emit modified ();
}

void
Graph3D::setTitleFont (const QFont & font)
{
  if (titleFnt != font)
    {
      //fprintf(stderr,"setting Title Font 6\n");
      titleFnt = font;
      sp->setTitleFont (font.family (), font.pointSize (), font.weight (),
			font.italic ());
      int w = currentSize.width ();
      int h = currentSize.height ();
      double scale = sqrt((double)(500*450) / (double)(w*h));
      titleFontSize = font.pointSizeFloat() * scale;
    }
}

void
Graph3D::setTitle (const QString & s, const QColor & color,
		   const QFont & font)
{
  if (title != s)
    {
      title = s;
      sp->setTitle (title);
    }

  titleCol = color;
  sp->setTitleColor (Qt2GL (color));

  setTitleFont( font );
}

void
Graph3D::setTitle (const QStringList & lst)
{
  QFont font =
    QFont (lst[3], lst[4].toInt (), lst[5].toInt (), lst[6].toInt ());
  setTitle( lst[1], lst[2], font );
}

void
Graph3D::updateTitle (const QString & s, const QColor & color,
		      const QFont & font)
{
  setTitle( s, color, font );
  sp->makeCurrent ();
  sp->updateGL ();
  emit modified ();
}

void
Graph3D::setOptions (const QStringList & lst)
{
  legendOn = false;
  if (lst[1].toInt () == 1)
    legendOn = true;
  sp->showColorLegend (legendOn);
  sp->setResolution (lst[2].toInt ());
  adjustLabels (lst[3].toInt ());
}

void
Graph3D::setOptions (bool legend, int r, int dist)
{
  sp->showColorLegend (legend);
  legendOn = legend;
  sp->setResolution (r);
  adjustLabels (dist);
}

Qwt3D::Triple ** Graph3D::allocateData (int columns, int rows)
{
  Qwt3D::Triple ** data = new Qwt3D::Triple *[columns];

  for (int i = 0; i < columns; ++i)
    {
      data[i] = new Qwt3D::Triple[rows];
    }
  return data;
}

void
Graph3D::deleteData (Qwt3D::Triple ** data, int columns)
{
  for (int i = 0; i < columns; i++)
    {
      delete[]data[i];
    }
  delete[]data;
}

QColor
Graph3D::minDataColor ()
{
  return fromColor;
}

QColor
Graph3D::maxDataColor ()
{
  return toColor;
}

void
Graph3D::setDataColors (const QColor & cMin, const QColor & cMax)
{
  //fprintf(stderr,"setDataColors\n");

  fromColor = cMin;
  toColor = cMax;

  Qwt3D::ColorVector cv;

  int size = 255;
  double dsize = size;

  double r1 = cMax.red () / dsize;
  double r2 = cMin.red () / dsize;

  double stepR = (r1 - r2) / dsize;

  double g1 = cMax.green () / dsize;
  double g2 = cMin.green () / dsize;

  double stepG = (g1 - g2) / dsize;

  double b1 = cMax.blue () / dsize;
  double b2 = cMin.blue () / dsize;

  double stepB = (b1 - b2) / dsize;

  RGBA rgb;

  for (int i = 0; i < size; i++)
    {
      rgb.r = r1 - i * stepR;
      rgb.g = g1 - i * stepG;
      rgb.b = b1 - i * stepB;
      rgb.a = alpha;

      cv.push_back (rgb);
    }

  col_ = new StandardColor (sp);
  col_->setColorVector (cv);
  sp->setDataColor (col_);

  color_map = QString::null;

  if (legendOn)
    {
      sp->showColorLegend (false);
      sp->showColorLegend (legendOn);
    }
}

void
Graph3D::changeTransparency (double t)
{
  if (alpha == t)
    return;

  alpha = t;

  Qwt3D::StandardColor * color = (StandardColor *) sp->dataColor ();
  color->setAlpha (t);
  sp->showColorLegend (legendOn);
  sp->updateData ();
  sp->updateGL ();
  emit modified ();
}

void
Graph3D::setTransparency (double t)
{
  if (alpha == t)
    return;

  alpha = t;

  Qwt3D::StandardColor * color = (StandardColor *) sp->dataColor ();
  color->setAlpha (t);
}

void
Graph3D::showWorksheet ()
{
  if (worksheet)
    worksheet->showMaximized ();
  else if (matrix_)
    matrix_->showMaximized ();
}

void
Graph3D::setSmoothMesh (bool smooth)
{
  if (smoothMesh == smooth)
    return;

  smoothMesh = smooth;
  sp->setSmoothMesh (smoothMesh);
  sp->coordinates ()->setLineSmooth (smoothMesh);
  sp->updateData ();
  sp->updateGL ();
}

QString
Graph3D::saveAsTemplate (const QString & geometryInfo)
{
  QString s = saveToString (geometryInfo);
  QStringList lst = QStringList::split ('\n', s, false);
  QStringList l = QStringList::split ('\t', lst[3], true);
  l[1] = QString::null;
  lst[3] = l.join ("\t");
  return lst.join ("\n");
}

/*!
  Turns 3D animation on or off
*/
void
Graph3D::animate (bool on)
{
  if (on)
    d_timer->start (animation_redraw_wait);	// Wait this many msecs before redraw
  else
    d_timer->stop ();
}

void
Graph3D::rotate ()
{
  if (!sp)
    return;

  sp->setRotation (int (sp->xRotation () + 1) % 360,
		   int (sp->yRotation () + 1) % 360,
		   int (sp->zRotation () + 1) % 360);
}

void
Graph3D::setDataColorMap (const QString & fileName)
{
  //fprintf(stderr,"setDataColorMap\n");

  ColorVector cv;
  if (!openColorMap (cv, fileName))
    return;

  color_map = fileName;

  col_ = new StandardColor (sp);
  col_->setColorVector (cv);

  sp->setDataColor (col_);
  sp->updateData ();
  sp->showColorLegend (legendOn);
  sp->updateGL ();
}

bool
Graph3D::openColorMap (ColorVector & cv, QString fname)
{
  if (fname.isEmpty ())
    return false;

  using std::ifstream;

  ifstream file (QWT3DLOCAL8BIT (fname));
  if (!file)
    return false;

  RGBA rgb;

  while (file)
    {
      file >> rgb.r >> rgb.g >> rgb.b;
      file.ignore (10000, '\n');
      if (!file.good ())
	break;
      else
	{
	  rgb.a = 1;
	  rgb.r /= 255;
	  rgb.g /= 255;
	  rgb.b /= 255;
	  cv.push_back (rgb);
	}
    }
  return true;
}

void
Graph3D::findBestLayout ()
{
  double start, end;
  sp->coordinates ()->axes[X1].limits (start, end);
  double xScale = fabs (end - start);

  sp->coordinates ()->axes[Y1].limits (start, end);
  double yScale = fabs (end - start);

  sp->coordinates ()->axes[Z1].limits (start, end);
  double zScale = fabs (end - start);

  sp->setScale (1/xScale, 1/yScale, 1/zScale);

  //double d = (sp->hull ().maxVertex - sp->hull ().minVertex).length ();
  //sp->setZoom (d / sqrt (3.0));
  double d = sqrt( SQR(xScale)+SQR(yScale)+SQR(zScale) );  
  sp->setZoom ( d / sqrt(3.0) );

/*
double majl = 0.1/yScale;
updateTickLength(0, majl, 0.6*majl);
majl = 0.1/xScale;
updateTickLength(1, majl, 0.6*majl);
updateTickLength(2, majl, 0.6*majl);
*/
}

void 
Graph3D::updateTableName( const QString &oldName, const QString &newName )
{
  if (func) return;
  if(plotAssociation.contains(oldName))
  {
    QStringList lst = QStringList::split (',', plotAssociation, false);
    lst.gres(QRegExp("^"+oldName+"_"),newName+"_");
    plotAssociation = lst.join(",");
    //qDebug("plotAssociation changed to %s",plotAssociation.ascii());
  }
}

void 
Graph3D::updateColName( const QString &oldName, const QString &newName )
{
  if(func) return;
  if(plotAssociation.contains(oldName))
  {
    QStringList lst = QStringList::split(',',plotAssociation,false);
    lst.gres(QRegExp("^"+oldName+"\\("),newName+"(");
    plotAssociation = lst.join(",");
    //qDebug("plotAssociation changed to %s",plotAssociation.ascii());
  }
}

void 
Graph3D::updateMatrixName( const QString &oldName, const QString &newName )
{
  if(matrix_ && plotAssociation.contains ("<"+oldName+">"))
    plotAssociation.replace (oldName, newName);
}

Graph3D::~Graph3D ()
{
  if (func)
    delete func;

  delete sp;
}

