/***************************************************************************
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor,                    *
 *   Boston, MA  02110-1301  USA                                           *
 *                                                                         *
 ***************************************************************************/
#include "interpolationDialog.h"
#include "application.h"
#include "graph.h"
#include "parser.h"
#include "colorBox.h"

#include <qvariant.h>
#include <qpushbutton.h>
#include <qlabel.h>
#include <qcombobox.h>
#include <qlayout.h>
#include <qbuttongroup.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qmessagebox.h>

interpolationDialog::interpolationDialog (QWidget * parent, const char *name,
					  bool modal, WFlags fl):
QDialog (parent, name, modal, fl)
{
  if (!name)
    setName ("interpolationDialog");
  setCaption (tr ("QtiPlot - Interpolation Options"));

  QButtonGroup *GroupBox1 =
    new QButtonGroup (2, QGroupBox::Horizontal, tr (""), this, "GroupBox1");

  new QLabel (tr ("Make curve from"), GroupBox1, "TextLabel1", 0);
  boxName = new QComboBox (GroupBox1, "boxShow");

  new QLabel (tr ("Spline"), GroupBox1, "TextLabel2", 0);
  boxMethod = new QComboBox (GroupBox1, "boxMethod");

  new QLabel (tr ("From Xmin"), GroupBox1, "TextLabel4", 0);
  boxStart = new QLineEdit (GroupBox1, "boxStart");
  boxStart->setText (tr ("0"));

  new QLabel (tr ("To Xmax"), GroupBox1, "TextLabel5", 0);
  boxEnd = new QLineEdit (GroupBox1, "boxEnd");
  boxEnd->setText (tr ("10"));

  new QLabel (tr ("X Step"), GroupBox1, "TextLabel3", 0);
  boxStep = new QLineEdit (GroupBox1, "boxStep");
  boxStep->setText(tr("1"));

  new QLabel (tr ("Color"), GroupBox1, "TextLabel52", 0);
  boxColor = new ColorBox (FALSE, GroupBox1);
  boxColor->setColor (QColor (red));

  QButtonGroup *GroupBox2 =
    new QButtonGroup (1, QGroupBox::Horizontal, tr (""), this, "GroupBox2");
  GroupBox2->setFlat (TRUE);
  GroupBox2->setLineWidth (0);

  buttonApply = new QPushButton (GroupBox2, "buttonApply");
  buttonApply->setAutoDefault (TRUE);
  buttonApply->setDefault (TRUE);

  buttonCancel = new QPushButton (GroupBox2, "buttonCancel");
  buttonCancel->setAutoDefault (TRUE);

  QHBoxLayout *hlayout = new QHBoxLayout (this, 5, 5, "hlayout");
  hlayout->addWidget (GroupBox1);
  hlayout->addWidget (GroupBox2);

  languageChange ();

  // signals and slots connections
  connect (boxName, SIGNAL (activated (int)), this,
	   SLOT (activateCurve (int)));
  connect (buttonApply, SIGNAL (clicked ()), this, SLOT (apply ()));
  connect (buttonCancel, SIGNAL (clicked ()), this, SLOT (reject ()));
}

interpolationDialog::~interpolationDialog ()
{
}


void
interpolationDialog::languageChange ()
{
  buttonApply->setText (tr ("&Make"));
  buttonCancel->setText (tr ("&Close"));

  boxMethod->insertItem (tr ("Linear"));
  boxMethod->insertItem (tr ("Cubic Spline"));
  boxMethod->insertItem (tr ("Non-rounded Akima"));
}

void
interpolationDialog::apply ()
{
  QString curve = boxName->currentText ();
  QStringList clist = d_graph->analyzableCurvesList ();
  if (clist.contains (curve) <= 0)
    {
      QMessageBox::critical (this, tr ("QtiPlot - Warning"),
			     tr
			     ("The curve <b> %1 </b> doesn't exist anymore! Operation aborted!").
			     arg (curve));
      boxName->clear ();
      boxName->insertStringList (clist);
      return;
    }

  double from, to, step;
  try
  {
    myParser parser;
    parser.SetExpr (boxStart->text ().ascii ());
    from = parser.Eval ();
  }
  catch (mu::ParserError & e)
  {
    QMessageBox::critical (this, tr ("QtiPlot - Start limit error"),
			   e.GetMsg ().c_str());
    boxStart->setFocus ();
    return;
  }

  try
  {
    myParser parser;
    parser.SetExpr (boxEnd->text ().ascii ());
    to = parser.Eval ();
  }
  catch (mu::ParserError & e)
  {
    QMessageBox::critical (this, tr ("QtiPlot - End limit error"),
			   e.GetMsg ().c_str());
    boxEnd->setFocus ();
    return;
  }

  try
  {
    myParser parser;
    parser.SetExpr (boxStep->text ().ascii ());
    step = parser.Eval ();
  }
  catch (mu::ParserError & e)
  {
    QMessageBox::critical (this, tr ("QtiPlot - X Step error"),
			   e.GetMsg ().c_str());
    boxStep->setFocus ();
    return;
  }

  if (from >= to)
    {
      QMessageBox::critical (this, tr ("QtiPlot - Input error"),
			     tr
			     ("Please enter x limits that satisfy: from < to!"));
      boxEnd->setFocus ();
      return;
    }

  if (to-from<step)
    {
      QMessageBox::critical (this, tr ("QtiPlot - Input error"),
			     tr
			     ("Please enter X Step smaller than the range!"));
      boxStep->setFocus ();
      return;
    }

  ApplicationWindow *app = (ApplicationWindow *)this->parent();

  int cindex = d_graph->curveIndex( boxName->currentText() );
  int spline = boxMethod->currentItem();

  int points = qRound( fabs( (to-from)/step ) ) + 1;
  QMemArray <double> x(points), y(points);

  for (int i = 0; i < points; i++)
    x[i] = from + i * step;
  if ( d_graph->interpolate( cindex, spline, x, y ) )
    {
      QString label, wlabel;
      switch (spline)
        {
        case 0:
          label = "LinearInt";
          wlabel = "Linear";
          break;
        case 1:
          label = "CubicInt";
          wlabel ="Cubic Spline";
          break;
        case 2:
          label = "AkimaInt";
          wlabel = "Akima Spline";
          break;
        }
      wlabel += " interpolation of " + boxName->currentText();
      QString title = app->generateUniqueName( label );
      d_graph->addResultCurve (points, x, y, boxColor->currentItem(),
			       title, wlabel, app->tbl_output_precision,
			       d_graph->curve(cindex) );
    }

  close();
}

void
interpolationDialog::setGraph (Graph * g)
{
  d_graph = g;
  boxName->insertStringList (g->analyzableCurvesList (), -1);

  if (g->selectorsEnabled ())
    {
      int index = g->selectedCurveID();
      boxName->setCurrentItem (index);
      activateCurve (index);
    }
  else
    activateCurve (0);

  connect (d_graph, SIGNAL (closedGraph ()), this, SLOT (close ()));
  connect (d_graph, SIGNAL (dataRangeChanged ()), this,
	   SLOT (changeDataRange ()));
};

void
interpolationDialog::activateCurve (int index)
{
  QString ctitle = boxName->text(index);
  int cindex = d_graph->curveIndex(ctitle);
  QwtPlotCurve *c = d_graph->curve (cindex);
  if (!c || c->rtti () != QwtPlotItem::Rtti_PlotCurve)
    return;

  double start,end;
  d_graph->range(cindex, &start, &end);

  boxStart->setText (QString::number( QMIN(start,end) ) );
  boxEnd->setText (QString::number( QMAX(start,end) ) );
  boxStep->setText(QString::number( fabs(end-start)/500,'g',1));
};

void
interpolationDialog::changeDataRange ()
{
  double start = d_graph->selectedXStartValue ();
  double end = d_graph->selectedXEndValue ();
  boxStart->setText (QString::number (QMIN (start, end), 'g', 15));
  boxEnd->setText (QString::number (QMAX (start, end), 'g', 15));
}
