/***************************************************************************
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor,                    *
 *   Boston, MA  02110-1301  USA                                           *
 *                                                                         *
 ***************************************************************************/
#include "matrix.h"
#include "nrutil.h"

#include <qdatetime.h>
#include <qlayout.h>
#include <qaccel.h>
#include <qregexp.h>
#include <qmessagebox.h>
#include <qapplication.h>
#include <qclipboard.h>
#include <qdragobject.h>
#include <qprinter.h>
#include <qpainter.h>
#include <qpaintdevicemetrics.h>
#include <qsimplerichtext.h>

#include <math.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_sort.h>
#include <gsl/gsl_sort_vector.h>
#include <gsl/gsl_interp.h>
#include <gsl/gsl_fft_halfcomplex.h>
#include <gsl/gsl_fft_real.h>
#include <gsl/gsl_fft_complex.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_math.h>

#define SWAP(a,b) do{int tmp=a; a=b; b=tmp;}while(0)

Matrix::Matrix (ScriptingEnv * env, int r, int c, const QString & label, QWidget * parent, const char *name, WFlags f):myWidget (label, parent, name, f),scripted(env)
{
  init (r, c);
}

void
Matrix::init (int rows, int cols)
{
  formula_str = "";
  selectedCol = 0;
  LeftButton = FALSE;
  txt_format = 'f';
  num_precision = 6;
  x_start = 0.0;
  x_end = 10.0;
  y_start = 0.0;
  y_end = 10.0;
  dMatrix = 0;
  specifications = "";
  d_autoscale = false;

  QDateTime dt = QDateTime::currentDateTime ();
  setBirthDate (dt.toString (Qt::LocalDate));

  d_table = new QTable (rows, cols, this, "d_table");
  d_table->setFocusPolicy (QWidget::StrongFocus);
  d_table->setFocus ();
  d_table->setSelectionMode (QTable::Single);
  d_table->setRowMovingEnabled (false);
  d_table->setColumnMovingEnabled(false);

/*
//!TODO: enable column moving, right now it doesn't work because of the event filter
//installed on hHeader
d_table->setColumnMovingEnabled(true);
connect(hHeader, SIGNAL(indexChange (int, int, int)), this, SLOT(notifyChanges()));
*/

  QColor background = QColor (255, 255, 128);
  d_table->setPaletteBackgroundColor (background);
  d_table->setBackgroundColor (background);

  QVBoxLayout *hlayout = new QVBoxLayout (this, 0, 0);
  hlayout->addWidget (d_table);

  QHeader *hHeader = (QHeader *) d_table->horizontalHeader ();
  hHeader->installEventFilter (this);
  hHeader->setMouseTracking (TRUE);
  hHeader->setMovingEnabled(false);

  QHeader *vHeader = (QHeader *) d_table->verticalHeader ();
  vHeader->setResizeEnabled (false);
  vHeader->installEventFilter (this);
  vHeader->setMovingEnabled(false);

  int w, h;
  if (cols > 3)
    w = 3 * hHeader->sectionSize (0);
  else
    w = cols * hHeader->sectionSize (0);

  if (rows > 11)
    h = 11 * vHeader->sectionSize (0);
  else
    h = (rows + 1) * vHeader->sectionSize (0);
  setGeometry (50, 50, w + 55, h);

  QAccel *accel = new QAccel (this);
  accel->connectItem (accel->insertItem (Key_Tab),
		      this, SLOT (moveCurrentCell ()));
  accel->connectItem (accel->insertItem (CTRL + Key_A),
		      this, SLOT (selectAll ()));

  connect (d_table, SIGNAL (valueChanged (int, int)), this,
	   SLOT (cellEdited (int, int)));
  connect (vHeader, SIGNAL (indexChange (int, int, int)), this,
	   SLOT (notifyChanges ()));
}

void
Matrix::selectAll ()
{
  d_table->
    addSelection (QTableSelection
		  (0, 0, d_table->numRows (), d_table->numCols ()));
}

void
Matrix::moveCurrentCell ()
{
  int cols = d_table->numCols ();
  int row = d_table->currentRow ();
  int col = d_table->currentColumn ();
  d_table->clearSelection (TRUE);

  if (col + 1 < cols)
    {
      d_table->setCurrentCell (row, col + 1);
      d_table->selectCells (row, col + 1, row, col + 1);
    }
  else
    {
      d_table->setCurrentCell (row + 1, 0);
      d_table->selectCells (row + 1, 0, row + 1, 0);
    }
}

void
Matrix::cellEdited (int row, int col)
{
  QString text = d_table->text (row, col).replace (",", ".");
  bool ok = false;
  double res = text.toDouble (&ok);
  if (!text.isEmpty () && ok)
    setValue( row, col, res );
  else
    {
      Script *script =
	scriptEnv->newScript (d_table->text (row, col), this,
			      QString ("<%1_%2_%3>").arg (getName()).arg (row +
									1).
			      arg (col + 1));
      connect (script, SIGNAL (error (const QString &, const QString &, int)),
	       scriptEnv,
	       SIGNAL (error (const QString &, const QString &, int)));

      script->setInt (row + 1, "row");
      script->setInt (row + 1, "i");
      script->setInt (col + 1, "col");
      script->setInt (col + 1, "j");
      QVariant ret = script->eval ();
      if (ret.type () == QVariant::Int || ret.type () == QVariant::UInt
	  || ret.type () == QVariant::LongLong
	  || ret.type () == QVariant::ULongLong)
	d_table->setText (row, col, ret.toString ());
      else if (ret.canCast (QVariant::Double))
	setValue( row, col, ret.toDouble () );
      else
	d_table->setText (row, col, "");
    }

  emit modifiedWindow (this);
}

QString
Matrix::text (int row, int col)
{
  if (dMatrix)
    return QString::number (dMatrix[row][col], txt_format, num_precision);
  else
    return d_table->text (row, col);
}

void
Matrix::setText (int row, int col, const QString & text)
{
  d_table->setText(row, col, text);
}

void
Matrix::setValue(int row, int col, double val)
{
  d_table->setText(row, col, QString::number(val, txt_format, num_precision));
}

bool
Matrix::isEmptyRow (int row)
{
  bool empty = TRUE;
  int cols = d_table->numCols ();
  for (int i = 0; i < cols; i++)
    {
      QString text = d_table->text (row, i);
      if (!text.isEmpty ())
	{
	  empty = FALSE;
	  break;
	}
    }
  return empty;
}

void
Matrix::setCoordinates (double xs, double xe, double ys, double ye)
{
  if (x_start == xs && x_end == xe && y_start == ys && y_end == ye)
    return;

  x_start = xs;
  x_end = xe;
  y_start = ys;
  y_end = ye;

  d_autoscale = false;

  emit modifiedWindow (this);
}

QString
Matrix::saveToString (const QString & info)
{
  QString s = "<matrix>\n";
  s += getName() + "\t";
  s += QString::number (d_table->numRows ()) + "\t";
  s += QString::number (d_table->numCols ()) + "\t";
  s += birthDate () + "\n";
  s += info;
  s += "ColWidth\t" + QString::number (d_table->columnWidth (0)) + "\n";
  s += "<formula>\n" + formula_str + "\n</formula>\n";
  s +=
    "TextFormat\t" + QString (txt_format) + "\t" +
    QString::number (num_precision) + "\n";
  s +=
    "WindowLabel\t" + windowLabel () + "\t" +
    QString::number (captionPolicy ()) + "\n";
  s +=
    "Coordinates\t" + QString::number (x_start, 'g',
				       15) + "\t" + QString::number (x_end,
								     'g',
								     15) +
    "\t";
  s +=
    QString::number (y_start, 'g', 15) + "\t" + QString::number (y_end, 'g',
								 15) + "\n";
  s += saveText ();
  s += "</matrix>\n";
  return s;
}

QString
Matrix::saveAsTemplate (const QString & info)
{
  QString s = "<matrix>\t";
  s += QString::number (d_table->numRows ()) + "\t";
  s += QString::number (d_table->numCols ()) + "\n";
  s += info;
  s += "ColWidth\t" + QString::number (d_table->columnWidth (0)) + "\n";
  s += "<formula>\n" + formula_str + "\n</formula>\n";
  s +=
    "TextFormat\t" + QString (txt_format) + "\t" +
    QString::number (num_precision) + "\n";
  s +=
    "Coordinates\t" + QString::number (x_start, 'g',
				       15) + "\t" + QString::number (x_end,
								     'g',
								     15) +
    "\t";
  s +=
    QString::number (y_start, 'g', 15) + "\t" + QString::number (y_end, 'g',
								 15) + "\n";
  return s;
}

void
Matrix::setSpecifications( const QString & s )
{
  if ( specifications == s ) return;
  specifications = s;
}

QString & Matrix::getSpecifications()
{
  return specifications;
}

void
Matrix::restore(QString & spec)
{
  int r = d_table->numRows ();
  int c = d_table->numCols ();

  QTextStream t (spec, IO_ReadOnly);
  t.readLine (); // "<matrix>"
  QString s = t.readLine (); // name, rows, cols, birthdate
  QStringList list = QStringList::split ('\t', s, TRUE);

  QString newCaption = list[0];
  QString oldCaption = getName();
  if (oldCaption != newCaption) this->setName (newCaption);

  int rows = list[1].toInt ();
  if (rows != r) d_table->setNumRows (rows);
  int cols = list[2].toInt ();
  if (cols != c) d_table->setNumCols (cols);

  int i,j;
  for(i=0;i<rows;i++)
    for(j=0;j<cols;j++)
      setText( i, j, "" );

  t.readLine(); // geometry info

  s = t.readLine(); // ColWidth
  QStringList fields = QStringList::split ('\t', s, true);
  setColumnsWidth (fields[1].toInt ());

  s = t.readLine(); // <formula>
  s = t.readLine();
  QString formula="";
  while( s != "</formula>" ) 
    {
      formula += s + "\n";
      s = t.readLine();
    }
  formula.truncate (formula.length () - 1);
  setFormula (formula);

  s = t.readLine(); // TextFormat
  fields = QStringList::split ('\t', s, true);
  if (fields[1] == "f")
    setTextFormat ('f', fields[2].toInt ());
  else
    setTextFormat ('e', fields[2].toInt ());
  
  s = t.readLine(); // WindowLabel
  fields = QStringList::split ('\t', s, true);
  setWindowLabel (fields[1]);
  setCaptionPolicy ((myWidget::CaptionPolicy) fields[2].toInt ());
  //app->setListViewLabel (getName(), fields[1]);

  s = t.readLine(); // Coordinates
  fields = QStringList::split ('\t', s, true);
  setCoordinates (fields[1].toDouble (), fields[2].toDouble (),
		  fields[3].toDouble (), fields[4].toDouble ());

  s = t.readLine(); // <data>
  s = t.readLine();
  while( s != "</data>" )
    {
      fields = QStringList::split ('\t', s, true);
      int row = fields[0].toInt ();
      for (int col = 0; col < cols; col++)
        setText (row, col, fields[col + 1]);
      s = t.readLine();
    }
 
  specifications = spec;
  emit modifiedWindow(this);
}


void
Matrix::restore (const QStringList & lst)
{
  QStringList l;
  QStringList::const_iterator i = lst.begin ();

  l = QStringList::split ('\t', *i++, true);
  setColumnsWidth (l[1].toInt ());

  l = QStringList::split ('\t', *i++, true);
  if (l[0] == "Formula")
    formula_str = l[1];
  else if (l[0] == "<formula>")
    {
      for (formula_str = ""; i != lst.end () && *i != "</formula>"; i++)
	formula_str += *i + "\n";
      formula_str.truncate (formula_str.length () - 1);
      i++;
    }

  l = QStringList::split ('\t', *i++, true);
  if (l[1] == "f")
    setTextFormat ('f', l[2].toInt ());
  else
    setTextFormat ('e', l[2].toInt ());

  l = QStringList::split ('\t', *i++, true);
  x_start = l[1].toDouble ();
  x_end = l[2].toDouble ();
  y_start = l[3].toDouble ();
  y_end = l[4].toDouble ();
}

QString
Matrix::saveText ()
{
  QString text = "<data>\n";
  int cols = d_table->numCols ();
  int rows = d_table->numRows ();

  for (int i = 0; i < rows; i++)
    {
      if (!isEmptyRow (i))
	{
	  text += QString::number (i) + "\t";
	  for (int j = 0; j < cols - 1; j++)
	    text += d_table->text (i, j) + "\t";

	  text += d_table->text (i, cols - 1) + "\n";
	}
    }
  return text + "</data>\n";
}

void
Matrix::setFormula (const QString & s)
{
  if (formula_str == s)
    return;

  formula_str = s;
}

void
Matrix::setNumericFormat (const QChar & f, int prec)
{
  if (txt_format == f && num_precision == prec)
    return;

  QApplication::setOverrideCursor (waitCursor);

  txt_format = f;
  num_precision = prec;

  int rows = d_table->numRows ();
  int cols = d_table->numCols ();
  for (int i = 0; i < rows; i++)
    {
      for (int j = 0; j < cols; j++)
	{
	  QString t = d_table->text (i, j);
	  if (!t.isEmpty ())
	    {
	      double val = dMatrix[i][j];
	      t = t.setNum (val, txt_format, num_precision);
	      d_table->setText (i, j, t);
	    }
	}
    }

  emit modifiedWindow (this);
  QApplication::restoreOverrideCursor ();
}

void
Matrix::setTextFormat (const QChar & format, int precision)
{
  if (txt_format == format && num_precision == precision)
    return;

  txt_format = format;
  num_precision = precision;
}

int
Matrix::columnsWidth ()
{
  return d_table->columnWidth (0);
}

void
Matrix::setColumnsWidth (int width)
{
  if (width == columnsWidth ())
    return;

  for (int i = 0; i < d_table->numCols (); i++)
    d_table->setColumnWidth (i, width);

  emit modifiedWindow (this);
}

void
Matrix::setMatrixDimensions (int rows, int cols)
{
  int r = d_table->numRows ();
  int c = d_table->numCols ();

  if (r == rows && c == cols)
    return;

  if (rows < r || cols < c)
    {
      QString text = "Deleting rows/columns from the matrix!";
      text += "<p>Do you really want to continue?";
      switch (QMessageBox::
	      information (0, "QtiPlot", tr (text), tr ("Yes"), tr ("Cancel"),
			   0, 1))
	{
	case 0:
	  QApplication::setOverrideCursor (waitCursor);
	  if (cols != c)
	    d_table->setNumCols (cols);
	  if (rows != r)
	    d_table->setNumRows (rows);
	  QApplication::restoreOverrideCursor ();
	  emit modifiedWindow (this);
	  break;

	case 1:
	  return;
	  break;
	}
    }
  else
    {
      QApplication::setOverrideCursor (waitCursor);
      if (cols != c)
	d_table->setNumCols (cols);
      if (rows != r)
	d_table->setNumRows (rows);
      QApplication::restoreOverrideCursor ();
      emit modifiedWindow (this);
    }
}

int
Matrix::numRows ()
{
  return d_table->numRows ();
}

int
Matrix::numCols ()
{
  return d_table->numCols ();
}

double
Matrix::determinant ()
{
  int rows = d_table->numRows ();
  int cols = d_table->numCols ();

  if (rows != cols)
    {
      QMessageBox::critical (0, tr ("QtiPlot - Error"),
			     tr
			     ("Calculation failed, the matrix is not square!"));
      return GSL_POSINF;
    }

  QApplication::setOverrideCursor (waitCursor);

  gsl_matrix *A = gsl_matrix_alloc (rows, cols);
  int i, j;
  for (i = 0; i < rows; i++)
    {
      for (j = 0; j < cols; j++)
	{
	  QString s = d_table->text (i, j);
	  gsl_matrix_set (A, i, j, s.toDouble ());
	}
    }

  int s;
  gsl_permutation *p = gsl_permutation_alloc (rows);
  gsl_linalg_LU_decomp (A, p, &s);

  double det = gsl_linalg_LU_det (A, s);

  gsl_matrix_free (A);
  gsl_permutation_free (p);

  QApplication::restoreOverrideCursor ();
  return det;
}

void
Matrix::invert ()
{
  int rows = d_table->numRows ();
  int cols = d_table->numCols ();

  if (rows != cols)
    {
      QMessageBox::critical (0, tr ("QtiPlot - Error"),
			     tr
			     ("Inversion failed, the matrix is not square!"));
      return;
    }

  QApplication::setOverrideCursor (waitCursor);

  gsl_matrix *A = gsl_matrix_alloc (rows, cols);
  int i, j;
  for (i = 0; i < rows; i++)
    {
      for (j = 0; j < cols; j++)
	{
	  QString s = d_table->text (i, j);
	  gsl_matrix_set (A, i, j, s.toDouble ());
	}
    }

  int s;
  gsl_permutation *p = gsl_permutation_alloc (cols);
  gsl_linalg_LU_decomp (A, p, &s);

  gsl_matrix *inverse = gsl_matrix_alloc (rows, cols);
  gsl_linalg_LU_invert (A, p, inverse);

  gsl_matrix_free (A);
  gsl_permutation_free (p);

  for (i = 0; i < rows; i++)
    for (j = 0; j < cols; j++)
      setValue( i, j, gsl_matrix_get(inverse, i, j) );
  
  gsl_matrix_free (inverse);
  QApplication::restoreOverrideCursor ();

  emit modifiedWindow (this);
}

void
Matrix::transpose ()
{
  int i, j;
  int rows = d_table->numRows ();
  int cols = d_table->numCols ();

  QTable *t = new QTable (rows, cols);
  for (i = 0; i < rows; i++)
    {
      for (j = 0; j < cols; j++)
	t->setText (i, j, d_table->text (i, j));
    }

  d_table->setNumCols (rows);
  d_table->setNumRows (cols);

  for (i = 0; i < cols; i++)
    {
      for (j = 0; j < rows; j++)
	d_table->setText (i, j, t->text (j, i));
    }

  delete t;

  double dummy = x_start;
  x_start = y_start;
  y_start = dummy;
  dummy = x_end;
  x_end = y_end;
  y_end = dummy;

  d_autoscale = true;

  emit modifiedWindow (this);
}

void
Matrix::saveCellsToMemory ()
{
  //  qDebug("saveCellsToMemory");
  int rows = d_table->numRows ();
  int cols = d_table->numCols ();
  dMatrix = matrix (0, rows - 1, 0, cols - 1);
  for (int i = 0; i < rows; i++)
    {
      for (int j = 0; j < cols; j++)
	{
	  QString s = d_table->text (i, j);
	  dMatrix[i][j] = s.toDouble ();
	}
    }
}

void
Matrix::forgetSavedCells ()
{
  //  qDebug("forgetSavedCells");
  int rows = d_table->numRows ();
  int cols = d_table->numCols ();
  free_matrix (dMatrix, 0, rows - 1, 0, cols - 1);
  dMatrix = 0;
}

bool
Matrix::calculate (int startRow, int endRow, int startCol, int endCol)
{
  QApplication::setOverrideCursor (waitCursor);

  Script *script =
    scriptEnv->newScript (formula_str, this, QString ("<%1>").arg (getName()));
  connect (script, SIGNAL (error (const QString &, const QString &, int)),
	   scriptEnv, SIGNAL (error (const QString &, const QString &, int)));
  connect (script, SIGNAL (print (const QString &)),
	   parent ()->parent ()->parent (),
	   SLOT (scriptPrint (const QString &)));
  if (!script->compile ())
    {
      QApplication::restoreOverrideCursor ();
      return false;
    }

  int rows = d_table->numRows ();
  int cols = d_table->numCols ();
  if (endCol >= cols)
    d_table->setNumCols (endCol + 1);
  if (endRow >= rows)
    d_table->setNumRows (endRow + 1);

  QVariant ret;
  saveCellsToMemory ();
  double dx = fabs (x_end - x_start) / (double) (d_table->numRows () - 1);
  double dy = fabs (y_end - y_start) / (double) (d_table->numCols () - 1);
  for (int row = startRow; row <= endRow; row++)
    for (int col = startCol; col <= endCol; col++)
      {
	script->setInt (row + 1, "i");
	script->setInt (row + 1, "row");
	script->setDouble (x_start + row * dx, "x");
	script->setInt (col + 1, "j");
	script->setInt (col + 1, "col");
	script->setDouble (y_start + col * dy, "y");
	ret = script->eval ();
	if (ret.type () == QVariant::Int || ret.type () == QVariant::UInt
	    || ret.type () == QVariant::LongLong
	    || ret.type () == QVariant::ULongLong)
	  d_table->setText (row, col, ret.toString ());
	else if (ret.canCast (QVariant::Double))
	  setValue( row, col, ret.toDouble() );
	else
	  {
	    d_table->setText (row, col, "");
	    QApplication::restoreOverrideCursor ();
	    return false;
	  }
      }
  forgetSavedCells ();

  emit modifiedWindow (this);
  QApplication::restoreOverrideCursor ();
  return true;
}

void
Matrix::clearSelection ()
{
  bool colSelection = false;
  for (int i = 0; i < d_table->numCols (); i++)
    {
      if (d_table->isColumnSelected (i, TRUE))
	{
	  colSelection = true;
	  for (int j = 0; j < d_table->numRows (); j++)
	    d_table->setText (j, i, "");
	}
    }

  if (colSelection)
    {
      emit modifiedWindow (this);
      return;
    }
  else
    {
      QTableSelection sel = d_table->selection (d_table->currentSelection ());
      for (int i = sel.topRow (); i <= sel.bottomRow (); i++)
	{
	  for (int j = sel.leftCol (); j <= sel.rightCol (); j++)
	    d_table->setText (i, j, "");
	}
      emit modifiedWindow (this);
    }
}

void
Matrix::copySelection ()
{
  QString text;
  int i, j;
  int rows = d_table->numRows ();
  int cols = d_table->numCols ();

  QMemArray < int >selection (1);
  int c = 0;
  for (i = 0; i < cols; i++)
    {
      if (d_table->isColumnSelected (i, TRUE))
	{
	  c++;
	  selection.resize (c);
	  selection[c - 1] = i;
	}
    }
  if (c > 0)
    {
      for (i = 0; i < rows; i++)
	{
	  for (j = 0; j < c - 1; j++)
	    text += d_table->text (i, selection[j]) + "\t";
	  text += d_table->text (i, selection[c - 1]) + "\n";
	}
    }
  else
    {
      QTableSelection sel = d_table->selection (d_table->currentSelection ());
      int top = sel.topRow ();
      int bottom = sel.bottomRow ();
      int left = sel.leftCol ();
      int right = sel.rightCol ();
      for (i = top; i <= bottom; i++)
	{
	  for (j = left; j < right; j++)
	    text += d_table->text (i, j) + "\t";
	  text += d_table->text (i, right) + "\n";
	}
    }

// Copy text into the clipboard
  //QApplication::clipboard ()->setData (new QTextDrag (text, d_table, 0));
  QApplication::clipboard ()->setText( text );
}

void
Matrix::cutSelection ()
{
  copySelection ();
  clearSelection ();
}

bool
Matrix::rowsSelected ()
{
  QTableSelection sel = d_table->selection (d_table->currentSelection ());
  for (int i = sel.topRow (); i <= sel.bottomRow (); i++)
    {
      if (!d_table->isRowSelected (i, true))
	return false;
    }
  return true;
}

void
Matrix::deleteSelectedRows ()
{
  QTableSelection sel = d_table->selection (d_table->currentSelection ());
  int top = sel.topRow ();
  int bottom = sel.bottomRow ();

  QMemArray < int >rows (bottom - top + 1);
  for (int i = top; i <= bottom; i++)
    rows[i - top] = i;

  d_table->removeRows (rows);
  emit modifiedWindow (this);
}

void
Matrix::insertColumn ()
{
  int cr = d_table->currentColumn ();
  if (d_table->isColumnSelected (cr, true))
    {
      d_table->insertColumns (cr, 1);
      emit modifiedWindow (this);
    }
}

bool
Matrix::columnsSelected ()
{
  for (int i = 0; i < d_table->numCols (); i++)
    {
      if (d_table->isColumnSelected (i, true))
	return true;
    }
  return false;
}

void
Matrix::deleteSelectedColumns ()
{
  QMemArray < int >cols;
  int n = 0;
  for (int i = 0; i < d_table->numCols (); i++)
    {
      if (d_table->isColumnSelected (i, true))
	{
	  n++;
	  cols.resize (n);
	  cols[n - 1] = i;
	}
    }
  d_table->removeColumns (cols);
  emit modifiedWindow (this);
}

void
Matrix::insertRow ()
{
  int cr = d_table->currentRow ();
  if (d_table->isRowSelected (cr, true))
    {
      d_table->insertRows (cr, 1);
      emit modifiedWindow (this);
    }
}

void
Matrix::appendRow ()
{
  d_table->insertRows (d_table->numRows(), 1);
  emit modifiedWindow (this);
}

// Paste text from the clipboard
void
Matrix::paste(bool transposed)
{
  QApplication::setOverrideCursor (waitCursor);

  /*
  QString text;
  if (!QTextDrag::decode (QApplication::clipboard ()->data (), text)
      || text.isNull ())
    {
      QApplication::restoreOverrideCursor ();
      return;
    }
  */

  QString text = QApplication::clipboard ()->text();

  QTextStream ts (&text, IO_ReadOnly);
  QString s = ts.readLine ();
  QStringList cellTexts = QStringList::split ('\t', s, TRUE);
  int cols = int (cellTexts.count ());
  int rows = 1;
  while (!ts.atEnd ())
    {
      rows++;
      s = ts.readLine ();
    }
  ts.reset ();

  int i, j, top, bottom, right, left, firstCol;
  QTableSelection sel = d_table->selection (d_table->currentSelection ());
  if (!sel.isEmpty ())
    {				// not columns but only cells are selected
      top = sel.topRow ();
      bottom = sel.bottomRow ();
      left = sel.leftCol ();
      right = sel.rightCol ();
    }
  else
    {
      top = 0;
      bottom = d_table->numRows () - 1;
      left = 0;
      right = d_table->numCols () - 1;

      firstCol = -1;
      for (i = 0; i < d_table->numCols (); i++)
	{
	  if (d_table->isColumnSelected (i, TRUE))
	    {
	      firstCol = i;
	      break;
	    }
	}
      if (firstCol >= 0)
	{			// columns are selected
	  left = firstCol;
	  int selectedColsNumber = 0;
	  for (i = 0; i < d_table->numCols (); i++)
	    {
	      if (d_table->isColumnSelected (i, TRUE))
		selectedColsNumber++;
	    }
	  right = firstCol + selectedColsNumber - 1;
	}
    }

  QTextStream ts2 (&text, IO_ReadOnly);
  int r = bottom - top + 1;
  int c = right - left + 1;

  if(transposed) SWAP(rows,cols);

  QApplication::restoreOverrideCursor ();
  if (rows > r || cols > c)
    {
      switch (QMessageBox::information (0, "QtiPlot",
					tr ("The text in the clipboard is larger than your current selection!\
	        \nDo you want to insert cells?"), tr ("Yes"), tr ("No"), tr ("Cancel"), 0,
					0))
	{
	case 0:
	  if (cols > c)
	    d_table->insertColumns (right+1, cols - c);
	  if (rows > r)
	    d_table->insertRows (bottom+1, rows - r);
	  break;
	case 1:
	  if (rows>r) rows = r;
	  if (cols>c) cols = c;
	  break;
	case 2:
	  return;
	  break;
	}
    }

  if(transposed) SWAP(rows,cols);

  QApplication::setOverrideCursor (waitCursor);
  bool numeric;
  double value;
  for (i = 0; i < rows; i++)
    {
      s = ts2.readLine ();
      cellTexts = QStringList::split ('\t', s, TRUE);
      for (j = 0; j < cols; j++)
	{
	  int dr,dc;
	  if(!transposed)
	    { dr = top + i; dc = left + j; }
	  else
	    { dr = top + j; dc = left + i; }
	  value = cellTexts[j].stripWhiteSpace().toDouble(&numeric);
	  if (numeric)
	    setValue( dr, dc, value );
	  else
	    d_table->setText (dr, dc, cellTexts[j]);
	}
    }

  emit modifiedWindow (this);
  QApplication::restoreOverrideCursor ();
}

void
Matrix::pasteSelection()
{
  paste(false);
}

// Paste text from the clipboard as transposed matrix
void
Matrix::pasteTransposed()
{
  paste(true);
}

void
Matrix::contextMenuEvent (QContextMenuEvent * e)
{
  emit showContextMenu ();
  e->accept ();
}

void
Matrix::customEvent (QCustomEvent * e)
{
  if (e->type () == SCRIPTING_CHANGE_EVENT)
    scriptingChangeEvent ((ScriptingChangeEvent *) e);
}

bool
Matrix::eventFilter (QObject * object, QEvent * e)
{
  QHeader *hheader = d_table->horizontalHeader ();
  QHeader *vheader = d_table->verticalHeader ();

  if (e->type () == QEvent::MouseButtonPress && object == (QObject *) hheader)
    {
      const QMouseEvent *me = (const QMouseEvent *) e;
      int offset = hheader->offset ();
      int col = hheader->sectionAt (me->pos ().x () + offset );
      if (me->button () == QMouseEvent::LeftButton &&
          me->state () == Qt::ControlButton)
        {
	  selectedCol = col;
          d_table->selectColumn (selectedCol);
          d_table->setCurrentCell (0, selectedCol);
          return true;
        }
      else if (me->button () == QMouseEvent::RightButton
	      && !d_table->isColumnSelected (col, true))
        {
          d_table->clearSelection();
	  selectedCol = col;
          d_table->selectColumn(selectedCol);
          d_table->setCurrentCell(0,selectedCol);
        }
    }
  else if (e->type () == QEvent::MouseButtonPress
	   && object == (QObject *) vheader)
    {
      const QMouseEvent *me = (const QMouseEvent *) e;
      int offset = vheader->offset ();
      int row = vheader->sectionAt (me->pos ().y () + offset);
      if (me->button () == QMouseEvent::RightButton
	  && !d_table->isRowSelected(row,0) )
        {
          d_table->clearSelection ();
          d_table->selectRow (row);
          d_table->setCurrentCell (row, 0);
        }
    }
  else if (e->type () == QEvent::MouseButtonDblClick
	   && object == (QObject *) hheader)
    {
      const QMouseEvent *me = (const QMouseEvent *) e;
      selectedCol = hheader->sectionAt (me->pos ().x () + hheader->offset ());
      return true;
    }
  else if (e->type () == QEvent::ContextMenu && object == titleBar)
    {
      emit showTitleBarMenu ();
      ((QContextMenuEvent *) e)->accept ();
      return true;
    }
  
  return QObject::eventFilter (object, e);
}

void
Matrix::print ()
{
  QPrinter printer;
  printer.setColorMode (QPrinter::GrayScale);
  if (printer.setup ())
    {
      printer.setFullPage (TRUE);
      QPainter p;
      if (!p.begin (&printer))
	return;			// paint on printer
      QPaintDeviceMetrics metrics (p.device ());
      int dpiy = metrics.logicalDpiY ();
      const int margin = (int) ((1 / 2.54) * dpiy);	// 1 cm margins

      QHeader *hHeader = d_table->horizontalHeader ();
      QHeader *vHeader = d_table->verticalHeader ();

      int rows = d_table->numRows ();
      int cols = d_table->numCols ();
      int height = margin;
      int i, vertHeaderWidth = vHeader->width ();
      int right = margin + vertHeaderWidth;

      // print header
      p.setFont (QFont ());
      QRect br =
	p.boundingRect (br, Qt::AlignCenter, hHeader->label (0), -1, 0);
      p.drawLine (right, height, right, height + br.height ());
      QRect tr (br);

      for (i = 0; i < cols; i++)
	{
	  int w = d_table->columnWidth (i);
	  tr.setTopLeft (QPoint (right, height));
	  tr.setWidth (w);
	  tr.setHeight (br.height ());
	  p.drawText (tr, Qt::AlignCenter, hHeader->label (i), -1);
	  right += w;
	  p.drawLine (right, height, right, height + tr.height ());

	  if (right >= metrics.width () - 2 * margin)
	    break;
	}
      p.drawLine (margin + vertHeaderWidth, height, right - 1, height);	//first horizontal line
      height += tr.height ();
      p.drawLine (margin, height, right - 1, height);

      // print d_table values
      for (i = 0; i < rows; i++)
	{
	  right = margin;
	  QString text = vHeader->label (i) + "\t";
	  tr = p.boundingRect (tr, Qt::AlignCenter, text, -1, 0);
	  p.drawLine (right, height, right, height + tr.height ());

	  br.setTopLeft (QPoint (right, height));
	  br.setWidth (vertHeaderWidth);
	  br.setHeight (tr.height ());
	  p.drawText (br, Qt::AlignCenter, text, -1);
	  right += vertHeaderWidth;
	  p.drawLine (right, height, right, height + tr.height ());

	  for (int j = 0; j < cols; j++)
	    {
	      int w = d_table->columnWidth (j);
	      text = d_table->text (i, j) + "\t";
	      tr = p.boundingRect (tr, Qt::AlignCenter, text, -1, 0);
	      br.setTopLeft (QPoint (right, height));
	      br.setWidth (w);
	      br.setHeight (tr.height ());
	      p.drawText (br, Qt::AlignCenter, text, -1);
	      right += w;
	      p.drawLine (right, height, right, height + tr.height ());

	      if (right >= metrics.width () - 2 * margin)
		break;
	    }
	  height += br.height ();
	  p.drawLine (margin, height, right - 1, height);

	  if (height >= metrics.height () - margin)
	    {
	      printer.newPage ();
	      height = margin;
	      p.drawLine (margin, height, right, height);
	    }
	}
    }
}

void
Matrix::range (double *min, double *max)
{
  double d_min = d_table->text (0, 0).toDouble ();
  double d_max = d_min;
  for (int i = 0; i < d_table->numRows (); i++)
    {
      for (int j = 0; j < d_table->numCols (); j++)
	{
	  double aux = d_table->text (i, j).replace (",", ".").toDouble ();
	  if (aux <= d_min)
	    d_min = aux;

	  if (aux >= d_max)
	    d_max = aux;
	}
    }

  *min = d_min;
  *max = d_max;
}

QString
Matrix::areaStatistics(int prec)
{
  QTableSelection sel = d_table->selection (0);
  if(sel.isEmpty())
    {
      QMessageBox::critical (this, tr ("QtiPlot - Area Statistics error"),
			     tr("There is no selection!"));
      return QString::null; 
    }

  int top = sel.topRow();
  int bottom = sel.bottomRow();
  int left = sel.leftCol();
  int right = sel.rightCol();

  int cells = ( right - left + 1 ) * ( bottom - top +1 );

  QMemArray <double> dat(cells);
  gsl_vector *y = gsl_vector_alloc(cells);
  int aux=0;
  for (int i = top; i <= bottom; i++)
    for (int j = left; j <= right; j++)
      {
	QString dummy = d_table->text(i,j);
	double val = dummy.toDouble();
	gsl_vector_set( y, aux, val );
	dat[aux] = val;
	aux++;
      }

  double mean = gsl_stats_mean( dat.data(), 1, cells );
  double min,max;
  gsl_vector_minmax( y, &min, &max );
  double sd = gsl_stats_sd( dat.data(), 1, cells );
  double var = gsl_stats_variance( dat.data(), 1, cells );
  double sum = mean * cells;

  gsl_vector_free( y );

  QString text = tr("Name")+"\t"+tr("Values")+"\n";
  text += tr("Cells") + "\t" + QString::number(cells) + "\n";
  text += tr("Sum") + "\t" + QString::number(sum,'g',prec) + "\n";
  text += tr("Mean") + "\t" + QString::number(mean,'g',prec) + "\n";
  text += tr("Max") + "\t" + QString::number(max,'g',prec) + "\n";
  text += tr("Min") + "\t" + QString::number(min,'g',prec) + "\n";
  text += tr("Width") + "\t" + QString::number(max-min,'g',prec) + "\n";
  text += tr("StandardDev") + "\t" + QString::number(sd,'g',prec) + "\n";
  text += tr("Variance") + "\t" + QString::number(var,'g',prec) + "\n";

  return text;
}

// 2D interpolation
bool
Matrix::interpolate(double x1, double x2, int xn,
		    double y1, double y2, int yn,
		    int spline, double **mat)
{
  if(!dMatrix) saveCellsToMemory();

  const gsl_interp_type *method;
  switch (spline)
    {
    case 0:
      method = gsl_interp_linear;
      break;
    case 1:
      method = gsl_interp_cspline;
      break;
    case 2:
      method = gsl_interp_akima;
      break;
    }

  int n_rows = d_table->numRows();
  int n_cols = d_table->numCols();

  double dx = (x_end - x_start)/(n_rows-1);
  double dy = (y_end - y_start)/(n_cols-1);

  bool xinvert=(dx<0);
  bool yinvert=(dy<0);

  QMemArray <double> xtemp(n_rows), ytemp(n_cols);

  int r,c;

  for(r=0; r<n_rows; r++)
    {
      xtemp[r] = x_start + dx * r; 
      // x values must be monotonically increasing
      if(xinvert)
	xtemp[r] = -xtemp[r];
    }
    
  for(c=0; c<n_cols; c++)
    {
      ytemp[c] = y_start + dy * c; 
      // x values must be monotonically increasing
      if(yinvert)
	ytemp[c] = -ytemp[c];
    }

  gsl_interp_accel **acc =
    (gsl_interp_accel **)malloc( sizeof(gsl_interp_accel *) * n_rows );
  gsl_interp **sp = (gsl_interp **)malloc( sizeof(gsl_interp *) * n_rows );
  for(int r=0; r<n_rows; r++)
    {
      acc[r] = gsl_interp_accel_alloc ();
      sp[r] = gsl_interp_alloc (method, n_cols);
      gsl_interp_init (sp[r], ytemp.data(), dMatrix[r], n_cols);
    }

  QMemArray <double> ztemp(n_rows);
  dx = (x2-x1)/(xn-1);
  dy = (y2-y1)/(yn-1);

  for(int c=0; c<yn; c++)
    {
      double y = y1 + dy * c;
      if(yinvert) y = -y;
      for(int r=0; r<n_rows; r++)
	ztemp[r] = gsl_interp_eval(sp[r], ytemp.data(), dMatrix[r], y, acc[r]);

      gsl_interp_accel *acc2 = gsl_interp_accel_alloc();
      gsl_interp *sp2 = gsl_interp_alloc( method, n_rows );
      gsl_interp_init (sp2, xtemp.data(), ztemp.data(), n_rows);
      for(int r=0; r<xn; r++)
	{
	  double x = x1 + dx * r;
	  if(xinvert) x = -x;
	  mat[r][c] = gsl_interp_eval(sp2, xtemp.data(), ztemp.data(), x, acc2);
	}
      gsl_interp_free(sp2);
      gsl_interp_accel_free (acc2);
    }

  for(int r=0; r<n_rows; r++)
    {
      gsl_interp_free (sp[r]);
      gsl_interp_accel_free (acc[r]);
    }

  free(sp);
  free(acc);

  forgetSavedCells();

  return true;
}

void
Matrix::forwardFFT( double **real, double **imag )
{
  if(!dMatrix) saveCellsToMemory();

  int rows = d_table->numRows();
  int cols = d_table->numCols();
  
  gsl_fft_real_wavetable *rwave = gsl_fft_real_wavetable_alloc (cols);
  gsl_fft_real_workspace *rwork = gsl_fft_real_workspace_alloc (cols);

  QMemArray <double> result(cols*2);
  double **dat = matrix( 0, cols, 0, rows*2 );

  int i;
  for (int r=0; r<rows; r++)
    {
      gsl_fft_real_transform( dMatrix[r], 1, cols, rwave, rwork );
      gsl_fft_halfcomplex_unpack( dMatrix[r], result.data(), 1, cols );
      int d = (cols+1)/2;
      for(i=0;i<cols;i++)
	{
	  int i2 = (i+d) % cols;
	  dat[i][r*2] = result[i2*2];
	  dat[i][r*2+1] = result[i2*2+1];
	}
    }

  gsl_fft_real_wavetable_free (rwave);
  gsl_fft_real_workspace_free (rwork);

  gsl_fft_complex_wavetable *cwave = gsl_fft_complex_wavetable_alloc (rows);
  gsl_fft_complex_workspace *cwork = gsl_fft_complex_workspace_alloc (rows);

  for (int c=0; c<cols; c++)
    {
      gsl_fft_complex_forward (dat[c], 1, rows, cwave, cwork);
      int d = (rows+1) / 2; 
      for (i = 0; i < rows; i++)
	{
	  int i2 = (i+d) % rows;
	  real[i][c] = dat[c][i2*2];
	  imag[i][c] = dat[c][i2*2+1];
	}
    }

  gsl_fft_complex_wavetable_free (cwave);
  gsl_fft_complex_workspace_free (cwork);

  free_matrix( dat, 0, cols, 0, rows*2 );

  forgetSavedCells();
}

void
Matrix::inverseFFT( int rows, int cols, double **real, double **imag )
{
  double **dat = matrix( 0, cols, 0, rows*2 );
  QMemArray <double> buff(cols*2);

  gsl_fft_complex_wavetable *cwave = gsl_fft_complex_wavetable_alloc (rows);
  gsl_fft_complex_workspace *cwork = gsl_fft_complex_workspace_alloc (rows);

  for (int c = 0; c<cols; c++)
    {
      int d = (rows+1) / 2;
      for(int i=0; i<rows; i++ )
	{
	  int i2 = (i+d) % rows;
	  dat[c][i2*2] = real[i][c];
	  dat[c][i2*2+1] = imag[i][c];
	}
      gsl_fft_complex_inverse (dat[c], 1, rows, cwave, cwork);
    }
  
  gsl_fft_complex_wavetable_free (cwave);
  gsl_fft_complex_workspace_free (cwork);

  cwave = gsl_fft_complex_wavetable_alloc (cols);
  cwork = gsl_fft_complex_workspace_alloc (cols);
 
  for (int r=0; r<rows; r++)
    {
      int d = (cols+1)/2;
      for(int i=0;i<cols;i++)
        {
          int i2 = (i+d) % cols;
          buff[i2*2] = dat[i][r*2];
	  buff[i2*2+1] = dat[i][r*2+1];
        }
      gsl_fft_complex_inverse (buff.data(), 1, cols, cwave, cwork);
      // The result should be real matrix if the initial matrix(real,imag)
      // are tranformed from real matrix
      for(int c=0;c<cols;c++)
	setValue( r, c, buff[c*2] );
    }

  gsl_fft_complex_wavetable_free (cwave);
  gsl_fft_complex_workspace_free (cwork);
  
  free_matrix( dat, 0, cols, 0, rows*2 );
}

bool
Matrix::autoScale()
{
  bool dummy = d_autoscale;
  d_autoscale = false;
  return dummy;
}
