/***************************************************************************
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor,                    *
 *   Boston, MA  02110-1301  USA                                           *
 *                                                                         *
 ***************************************************************************/
#include "matrixFilterDialog.h"
#include "matrix.h"
#include "application.h"
#include "parser.h"
#include "nrutil.h"

#include <qvariant.h>
#include <qpushbutton.h>
#include <qhbox.h>
#include <qlabel.h>
#include <qcombobox.h>
#include <qcheckbox.h>
#include <qlayout.h>
#include <qbuttongroup.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qmessagebox.h>

matrixFilterDialog::matrixFilterDialog (QWidget * parent, const char *name,
					bool modal, WFlags fl):
  QDialog (parent, name, modal, fl)
{
  if (!name)
    setName ("matrixFilterDialog");

  QButtonGroup *GroupBox1 =
    new QButtonGroup (2, QGroupBox::Horizontal, tr (""), this, "GroupBox1");

  new QLabel (tr ("Filter matrix: "), GroupBox1, "TextLabel1", 0);
  boxName = new QComboBox (GroupBox1, "boxName");

  new QLabel (tr ("Method: "), GroupBox1, "TextLabel2", 0);
  boxMethod = new QComboBox (GroupBox1, "boxMethod");
  boxMethod->insertItem("Low Pass");
  boxMethod->insertItem("High Pass");
  boxMethod->insertItem("Gaussian");
  boxMethod->insertItem("Lorentzian");

  new QLabel (tr ("Threshold for X (Hz)"), GroupBox1, "TextLabel4", 0);
  boxXFreq = new QLineEdit (GroupBox1, "boxXFreq");
  boxXFreq->setText (tr ("1"));

  new QLabel (tr ("Threshold for Y (Hz)"), GroupBox1, "TextLabel5", 0);
  boxYFreq = new QLineEdit (GroupBox1, "boxYFreq");
  boxYFreq->setText (tr ("1"));

  QHBox *hbox1 = new QHBox (this, "hbox1");
  hbox1->setMargin (5);
  hbox1->setSpacing (5);

  buttonFilter = new QPushButton (hbox1, "buttonFilter");
  buttonFilter->setDefault (TRUE);

  buttonCancel = new QPushButton (hbox1, "buttonCancel");

  QVBoxLayout *hlayout = new QVBoxLayout (this, 5, 5, "hlayout");
  hlayout->addWidget (GroupBox1);
  hlayout->addWidget (hbox1);

  languageChange ();

  // signals and slots connections
  connect (buttonFilter, SIGNAL (clicked ()), this, SLOT (filter ()));
  connect (buttonCancel, SIGNAL (clicked ()), this, SLOT (reject ()));
}

matrixFilterDialog::~matrixFilterDialog ()
{
}


void
matrixFilterDialog::languageChange ()
{
  setCaption (tr ("QtiPlot - Filter options"));
  buttonFilter->setText (tr ("&Filter"));
  buttonCancel->setText (tr ("&Close"));
}

void
matrixFilterDialog::setMatrixNames(QStringList names, int index)
{
  boxName->clear();
  boxName->insertStringList( names );
  boxName->setCurrentItem( index );
}

void
matrixFilterDialog::filter ()
{
  double xfreq=0,yfreq=0;
  try
  {
    myParser parser;
    parser.SetExpr (boxXFreq->text ().ascii ());
    xfreq = parser.Eval ();
  }
  catch (mu::ParserError & e)
  {
    QMessageBox::critical (this, tr ("QtiPlot - Frequency input error"),
			   e.GetMsg ().c_str());
    boxXFreq->setFocus ();
    return;
  }

  if (xfreq < 0)
    {
      QMessageBox::critical (this, tr ("QtiPlot - Frequency input error"),
			     tr ("Please enter positive frequency values!"));
      boxXFreq->setFocus ();
      return;
    }

  try
    {
      myParser parser;
      parser.SetExpr (boxYFreq->text ().ascii ());
      yfreq = parser.Eval ();
    }
  catch (mu::ParserError & e)
    {
      QMessageBox::critical (this,
			     tr ("QtiPlot - Frequency input error"),
			     e.GetMsg ().c_str());
      boxYFreq->setFocus ();
      return;
    }

  if (yfreq < 0)
    {
      QMessageBox::critical (this,
			     tr("QtiPlot - Frequency input error"),
			     tr("Please enter positive frequency values!"));
      boxYFreq->setFocus ();
      return;
    }

  int filter_type = boxMethod->currentItem();

  ApplicationWindow *app = (ApplicationWindow *)this->parent();

  Matrix *m = app->matrix(boxName->currentText());
  if (!m) return;

  int rows = m->numRows();
  int cols = m->numCols();

  double **real = matrix( 0, rows, 0, cols );
  double **imag = matrix( 0, rows, 0, cols );
 
  m->forwardFFT( real, imag );

  double dfx = 1.0/fabs(m->xEnd()-m->xStart());
  double dfy = 1.0/fabs(m->yEnd()-m->yStart());

  int kx = rows / 2;
  int ky = cols / 2;
  for(int i=0;i<rows;i++)
    {
      double fx = (i - kx) * dfx;
      for(int j=0;j<cols;j++)
	{
	  double fy = (j - ky) * dfy;
	  double factor=0;
	  switch (filter_type) 
	    {
	    case LowPass:
	      factor = (fabs(fx) > xfreq || fabs(fy) > yfreq) ? 0 : 1;
	      break;
	    case HighPass:
	      factor = (fabs(fx) > xfreq || fabs(fy) > yfreq) ? 1 : 0;
	      break;
	    case Gauss:
	      factor = exp( - fx*fx/(xfreq*xfreq) - fy*fy/(yfreq*yfreq) );
	      break;
	    case Lorentz:
	      factor = exp( - 2 * fabs(fx) / xfreq - 2 * fabs(fy) / yfreq );
	      break;
	    }
	  real[i][j] *= factor;
	  imag[i][j] *= factor;
	}
    }
  
  QString caption = app->generateUniqueName(tr("FilteredMatrix"));
  Matrix *w = app->newMatrix (caption, rows, cols);
  QString label = tr("(")+QString::number(xfreq)+"Hz,"+QString::number(yfreq)+"Hz) ";
  switch (filter_type)
    {
    case LowPass:
      label += "LowPass";
      break;
    case HighPass:
      label += "HighPass";
      break;
    case Gauss:
      label += "Gauss";
      break;
    case Lorentz:
      label += "Lorentz";
      break;
    }
  label += " Filter of " + boxName->currentText();
  w->setWindowLabel( label );
  w->inverseFFT( rows, cols, real, imag );
  w->setCoordinates (m->xStart (), m->xEnd (), m->yStart (), m->yEnd ());
  w->show();

  free_matrix( imag, 0, rows, 0, cols );
  free_matrix( real, 0, rows, 0, cols );

  close();
}
