/***************************************************************************
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor,                    *
 *   Boston, MA  02110-1301  USA                                           *
 *                                                                         *
 ***************************************************************************/

#include "momentDialog.h"
#include "application.h"
#include "graph.h"
#include "worksheet.h"
#include "FunctionCurve.h"
#include "parser.h"

#include <qwt_plot_curve.h>

#include <qvariant.h>
#include <qpushbutton.h>
#include <qcheckbox.h>
#include <qlabel.h>
#include <qcombobox.h>
#include <qlayout.h>
#include <qbuttongroup.h>
#include <qlineedit.h>
#include <qhbuttongroup.h>
#include <qmessagebox.h>

momentDialog::momentDialog (QWidget * parent, const char *name, bool modal, 
			    WFlags fl):QDialog (parent, name, modal,fl)
{
  d_graph = 0;

  if (!name)
    setName ("momentDialog");
  setCaption (tr ("QtiPlot - Moment Analysis of Curves"));
  setSizeGripEnabled (true);
  setSizePolicy (QSizePolicy (QSizePolicy::Expanding, QSizePolicy::Fixed));

  ApplicationWindow *app = (ApplicationWindow *)parent;

  GroupBox1 =
    new QButtonGroup (2, QGroupBox::Horizontal, tr (""), this, "GroupBox1");

  new QLabel (tr ("Calculate Moments of"), GroupBox1, "TextLabel1", 0);
  boxName = new QComboBox (GroupBox1, "boxName");

  new QLabel (tr ("Lower limit"), GroupBox1, "TextLabel5", 0);
  boxStart = new QLineEdit (GroupBox1, "boxStart");

  new QLabel (tr ("Upper limit"), GroupBox1, "TextLabel6", 0);
  boxEnd = new QLineEdit (GroupBox1, "boxEnd");

  GroupBox2 = new QHButtonGroup (this, "GroupBox2");
  GroupBox2->setFlat (TRUE);
  GroupBox2->setLineWidth (0);

  buttonOk = new QPushButton (GroupBox2, "buttonOk");
  buttonOk->setAutoDefault (TRUE);
  buttonOk->setDefault (TRUE);

  buttonCancel = new QPushButton (GroupBox2, "buttonCancel");
  buttonCancel->setAutoDefault (TRUE);

  buttonHelp = new QPushButton (GroupBox2, "buttonHelp");
  buttonHelp->setAutoDefault (TRUE);

  QVBoxLayout *
    vlayout = new QVBoxLayout (this, 5, 5, "vlayout");
  vlayout->addWidget (GroupBox1);
  vlayout->addWidget (GroupBox2);

  languageChange ();

  // signals and slots connections
  connect (buttonOk, SIGNAL (clicked ()), this, SLOT (accept ()));
  connect (buttonCancel, SIGNAL (clicked ()), this, SLOT (reject ()));
  connect (buttonHelp, SIGNAL (clicked ()), this, SLOT (help ()));
  connect (boxName, SIGNAL (activated (int)), this,
           SLOT (activateCurve (int)));

}

momentDialog::~momentDialog ()
{
}

void
momentDialog::languageChange ()
{
  buttonOk->setText (tr ("&OK"));
  buttonCancel->setText (tr ("&Cancel"));
  buttonHelp->setText (tr ("&Help"));
}

void
momentDialog::setGraph (Graph *g, const QString &name)
{
  d_graph = g;
  graphname = name;

  QStringList names = d_graph->analyzableCurvesList();
  names.prepend("All Curves");
  boxName->insertStringList (names, -1);

  if (d_graph->selectorsEnabled ())
    {
      int index = 0;
      QString selected = d_graph->selectedCurveTitle();
      if((index = names.findIndex( selected )) >=0 )
        boxName->setCurrentItem( index );
      activateCurve(index);
    }
  else
    {
      activateCurve(1);
    }

  connect (d_graph, SIGNAL (closedGraph ()), this, SLOT (close ()));
  connect (d_graph, SIGNAL (dataRangeChanged ()), this,
           SLOT (changeDataRange ()));
}

void
momentDialog::activateCurve (int index)
{
  if(index == 0) return;

  QString ctitle = boxName->text(index);
  int cindex = d_graph->curveIndex(ctitle);
  QwtPlotCurve *c = d_graph->curve(cindex);
  if (!c || c->rtti () != QwtPlotItem::Rtti_PlotCurve)
    return;

  double start,end;
  d_graph->range(cindex, &start, &end);
  boxStart->setText (QString::number (QMIN (start, end), 'g', 15));
  boxEnd->setText (QString::number (QMAX (start, end), 'g', 15));
}

void
momentDialog::changeDataRange ()
{
  double start = d_graph->selectedXStartValue ();
  double end = d_graph->selectedXEndValue ();
  boxStart->setText (QString::number (QMIN (start, end), 'g', 15));
  boxEnd->setText (QString::number (QMAX (start, end), 'g', 15));
}

void
momentDialog::accept ()
{
  ApplicationWindow *app = (ApplicationWindow *)this->parent();

  QStringList curves;
  if (boxName->currentItem() == 0)
    curves = d_graph->analyzableCurvesList();
  else
    curves.append(boxName->currentText());

  double start,end;

  try
    {
      myParser parser;
      parser.SetExpr( boxStart->text().ascii() );
      start = parser.Eval();
    }
  catch(mu::ParserError &e)
    {
      QMessageBox::critical (this, tr ("QtiPlot - Lower limit error"),
                             e.GetMsg ().c_str());
      boxStart->setFocus ();
      return;
    }

  try
    {
      myParser parser;
      parser.SetExpr( boxEnd->text().ascii() );
      end = parser.Eval();
    }
  catch(mu::ParserError &e)
    {
      QMessageBox::critical (this, tr ("QtiPlot - Higher limit error"),
                             e.GetMsg ().c_str());
      boxEnd->setFocus ();
      return;
    }

  if (start >= end)
    {
      QMessageBox::warning (0, tr ("QtiPlot - Parameter Error"),
                            tr("Please input valid values") );
      return;
    }
  
  QString text = 
    tr("Curve") + "\t"
    + tr("From") + "\t"
    + tr("To") + "\t"
    + tr("0th") + "\t"
    + tr("1st") + "\t"
    + tr("2nd") + "\t"
    + tr("mean") + "\t"
    + tr("stddev") + "\n";

  QMemArray <double> x,y;

  for (int cv = 0; cv < (int)curves.count(); cv++)
    {
      int cindex = d_graph->curveIndex( curves[cv] );
      QwtPlotCurve *c = d_graph->curve( cindex );
      
      int dsize = d_graph->sortedCurveData( cindex, start, end, x, y );
      if (dsize<2) break;
      
      double sum0=0,sum1=0,sum2=0;
      for(int i=0;i<dsize-1;i++)
	{
	  double x1 = x[i];
	  double y1 = y[i];
	  double x2 = x[i+1];
	  double y2 = y[i+1];
	  double dy = y2 - y1;
	  double c0 = y1 * x2 - y2 * x1;
	  double c1 = x2 + x1;
	  double c2 = c1 * x2 + x1 * x1;
	  double c3 = c2 * x2 + x1 * x1 * x1;
	  sum0 += c0 + dy * c1 / 2;
	  sum1 += c0 * c1 / 2 + dy * c2 / 3;
	  sum2 += c0 * c2 / 3 + dy * c3 / 4;
	}
      
      double mean = sum1 / sum0;
      double sd = sqrt( sum2 / sum0 - mean * mean );
      int prec = app->tbl_output_precision;
      
      text += 
	curves[cv] + "\t"
	+ QString::number(start,'g',prec) + "\t"
	+ QString::number(end,'g',prec) + "\t"
	+ QString::number(sum0,'g',prec) + "\t"
	+ QString::number(sum1,'g',prec) + "\t"
	+ QString::number(sum2,'g',prec) + "\t"
	+ QString::number(mean,'g',prec) + "\t"
	+ QString::number(sd,'g',prec) + "\n";
    }

  QString tname = app->generateUniqueName("Moment");
  QString legend = tr("Moment Analysis of ") + graphname;
  int prec = app->tbl_output_precision;
  QString caption = tname + "\t" + legend;
  Table *w = app->newTable( caption, curves.count(), 8, text );
  QStringList ct;
  ct << "1";
  for (int i=0;i<6; i++)
    ct << tr("0;0/")+QString::number(app->tbl_output_precision);
  w->setColumnTypes( ct );
  
  close ();
}

void
momentDialog::help ()
{
  QMessageBox::about (0, tr ("QtiPlot - Help for Moment Analysis"),
                      tr(
"This function calculates zeroth, first, and second moment of the spectrum.\n\
For convenience, center of mass (mean value) and standard deviation are also\n\
derived from the calculated results."));
}
