/***************************************************************************
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor,                    *
 *   Boston, MA  02110-1301  USA                                           *
 *                                                                         *
 ***************************************************************************/
#include "muParserScripting.h"
#include "worksheet.h"
#include "matrix.h"
#include "parser.h"

#include <qstringlist.h>
#include <qasciidict.h>

//using namespace mu; // already declared in parser.h

const char *muParserScripting::langName = "muParser";

//functions additional to myParser
const myParser::mathFunction muParserScripting::math_functions[] = {
  {"binomial", 2, NULL, binomial, NULL, NULL,
   "binomial(p,int n):\n Random integer from the binomial distribution, the number of successes in n independent trials with probability p."},
  {"erand", 1, erand, NULL, NULL, NULL,
   "erand(x):\n Random value from the exponential distribution with mean x."},
  {"grand",2,NULL,grand,NULL, NULL,
   "grand(a,b):\n Random value from the Gamma distribution with parameters a and b where a=1 gives the exponential distribution."},
  {"nrand", 1, nrand, NULL, NULL, NULL, 
   "nrand(x):\n Random number from the normal distribution with mean zero and standard deviation x."},
  {"poisson", 1, poisson, NULL, NULL, NULL,
   "poisson(x):\n Random integer from the Poisson distribution with mean x."},
  {"rand", 1, myrand, NULL, NULL, NULL,
   "rand(x):\n Random value smaller than x."},
  {0, 0, NULL, NULL, NULL, NULL, 0}
};

muParserScript::muParserScript (ScriptingEnv * env, const QString & code,
				QObject * context, const QString & name):
Script (env, code, context, name)
{
  variables.setAutoDelete (true);
  rvariables.setAutoDelete (true);

  for (const myParser::mathFunction * i =
	 muParserScripting::math_functions; i->name; i++)
    if (i->numargs == 1 && i->fun1 != NULL)
      parser.DefineFun (i->name, i->fun1);
    else if (i->numargs == 2 && i->fun2 != NULL)
      parser.DefineFun (i->name, i->fun2);
    else if (i->numargs == 3 && i->fun3 != NULL)
      parser.DefineFun (i->name, i->fun3);
    else if (i->numargs == 4 && i->fun4 != NULL)
      parser.DefineFun (i->name, i->fun4);

  if (Context->isA ("Table"))
    parser.DefineFun ("col", mu_col, false);
  else if (Context->isA ("Matrix"))
    parser.DefineFun ("cell", mu_cell);

  rparser = parser;
  parser.SetVarFactory (mu_addVariable);
  rparser.SetVarFactory (mu_addVariableR);

}

double
muParserScript::col (const char *arg)
{
  if (!Context->isA ("Table"))
    throw Parser::exception_type ("col() works only on tables!");

  //qDebug("in muParserScript::col");
  //qDebug("arg = %s",arg);

  QString arg2 = QString::fromUtf8(arg);

  //qDebug("arg2 = %s",arg2.ascii());
  QStringList items = QStringList::split(',',arg2,false);
  
  /*
  QStringList items;
  QString item = "";
  for (int i = 0; arg[i] != QChar::null; i++)
    if (arg[i] == '"')
      {
	item += "\"";
	for (i++; arg[i] != QChar::null && arg[i] != '"'; i++)
	  if (arg[i] == '\\')
	    { // This allows the use of '"' in column name
	      item += "\\";
	      item += arg[++i];
	    }
	  else
	    item += arg[i];
	item += "\"";
      }
    else if (arg[i] == ',')
      {
	items << item;
	item = "";
      }
    else
      item += arg[i];
  items << item;
  */

  Table *table = (Table *) Context;
  int col, row;
  Parser local_parser (rparser);
  if (items[0].startsWith ("\"") && items[0].endsWith ("\""))
    {
      col = table->colNames().findIndex(items[0].
					mid (1, items[0].length () - 2));
      if (col < 0){
	//qDebug("no column");
	throw Parser::
	  exception_type (tr ("There's no column named %1 in table %2!").
			  arg (items[0]).arg(Context->name()).ascii());
      }
    }
  else
    {
      // for backwards compatibility
      col = table->colNames ().findIndex (items[0]);
      if (col < 0)
	{
	  local_parser.SetExpr ((const char *)items[0].utf8());
	  col = qRound (local_parser.Eval ()) - 1;
	}
    }
  if (items.count () == 2)
    {
      local_parser.SetExpr ((const char *)items[1].utf8());
      row = qRound (local_parser.Eval ()) - 1;
    }
  else if (variables["i"])
    row = (int) *(variables["i"]) - 1;
  else
    return 0;

  rvariables.clear ();
  if (row < 0 || row >= table->tableRows ())
    {
      //qDebug("no row");
      throw Parser::exception_type(tr("There's no row %1 in table %2!").
				   arg(row+1).arg(Context->name()).ascii());
    }
  if (col < 0 || col >= table->tableCols ())
    {
      //qDebug("no column 2");
      throw Parser::exception_type (tr("There's no column %1 in table %2!").
				    arg(col+1).arg(Context->name()).ascii());
    }
  if (table->text (row, col).isEmpty ())
    throw new EmptySourceError ();
  else
    return (table->text (row, col)).toDouble ();
}

double
muParserScript::cell (int row, int col)
{
  if (!Context->isA ("Matrix"))
    throw Parser::exception_type ("cell() works only on matrices!");
  Matrix *matrix = (Matrix *) Context;
  if (row < 1 || row > matrix->numRows ())
    throw Parser::exception_type (tr ("There's no row %1 in matrix %2!").
				  arg(row).arg(Context->name()).ascii());
  if (col < 1 || col > matrix->numCols())
    throw Parser::exception_type (tr ("There's no column %1 in matrix %2!").
				  arg(col).arg(Context->name()).ascii());
  if (matrix->text (row - 1, col - 1).isEmpty ())
    throw new EmptySourceError ();
  else
  return (matrix->text (row - 1, col - 1)).toDouble ();
}

double *
muParserScript::addVariable (const char *name)
{
  double *valptr = new double;
  if (!valptr)
    throw Parser::exception_type ("Out of memory");
  *valptr = 0;
  variables.insert (name, valptr);
  rparser.DefineVar (name, valptr);
  return valptr;
}

double *
muParserScript::addVariableR (const char *name)
{
  double *valptr = new double;
  if (!valptr)
    throw Parser::exception_type ("Out of memory");
  *valptr = 0;
  rvariables.insert (name, valptr);
  return valptr;
}

bool
muParserScript::setDouble (double val, const char *name)
{
  double *valptr = variables[name];
  if (!valptr)
    {
      valptr = new double;
      if (!valptr)
	{
	  emit_error (tr ("Out of memory"), 0);
	  return false;
	}
      try
      {
	parser.DefineVar (name, valptr);
	rparser.DefineVar (name, valptr);
	variables.insert (name, valptr);
      }
      catch (ParserError & e)
      {
	delete valptr;
	emit_error (QString::fromLocal8Bit(e.GetMsg ().c_str()), 0);
	return false;
      }
    }
  *valptr = val;
  return true;
}

bool
muParserScript::setInt (int val, const char *name)
{
  return setDouble ((double) val, name);
}

bool
muParserScript::setQObject (QObject *, const char *)
{
  return false;
}

QString
muParserScript::compileColArg (const QString & in)
{
  //qDebug("in muParserScript::compileColArg");
  //qDebug("in = %s",in.ascii());

  QString out = "\"";
  for (int i = 0; in[i] != QChar::null; i++)
    {
      if (in[i] == 'c' && in.mid (i, 4) == "col(")
	{
	  out += "col(";
	  QString arg = "";
	  int paren = 1;
	  for (i += 4; in[i] != QChar::null && paren > 0; i++)
	    {
	      /* 
	      if (in[i] == '"')
		{
		  arg += "\"";
		  for (i++; in[i] != QChar::null && in[i] != '"'; i++)
		    if (in[i] == '\\')
		      { // This allows you to use '"' in column name
		        arg += "\\";
		        arg += in[++i];
		      }
		    else
		      arg += in[i];
		  arg += "\"";
		}
	      else
	      */
	      if (in[i] == '(')
		{
		  paren++;
		  arg += "(";
		}
	      else if (in[i] == ')')
		{
		  paren--;
		  if (paren > 0)
		    arg += ")";
		}
	      else
		arg += in[i];
	    }
	  out += compileColArg(arg).replace ('"', "\\\"");
	  out += ")";
	  i--;
	}
      else if (in[i] == '\\')
	out += "\\\\";
      else if (in[i] == '"')
	out += "\\\"";
      else
	out += in[i];
    }

  out += "\"";

  //qDebug("out = %s",out.ascii());

  return out;
}

bool
muParserScript::compile (bool for_eval)
{
  //qDebug("in muParserScript::compile");
  //qDebug("Code = %s",Code.ascii());

  muCode.clear ();
  QString muCodeLine = "";
  for (int i = 0; Code[i] != QChar::null; i++)
    {
      if (Code[i] == 'c' && Code.mid (i, 4) == "col(")
	{
	  muCodeLine += "col(";
	  QString arg = "";
	  int paren = 1;
	  for (i += 4; Code[i] != QChar::null && paren > 0; i++)
	    {
	      /*
	      if (Code[i] == '"')
		{
		  arg += "\"";
		  for (i++; Code[i] != '"' && Code[i] != QChar::null; i++)
		    if (Code[i] == '\\')
		      { // This allows you to use '"' in column name
			arg += "\\";
			arg += Code[++i];
		      }
		    else
		      arg += Code[i];
		  arg += "\"";
		}
		else*/
	      if (Code[i] == '(')
		{
		  paren++;
		  arg += "(";
		}
	      else if (Code[i] == ')')
		{
		  paren--;
		  if (paren > 0)
		    arg += ")";
		}
	      else if (Code[i] != ' ') // trim white space
		arg += Code[i];
	    }
	  muCodeLine += compileColArg (arg);
	  muCodeLine += ")";
	  i--;
	}
      else if (Code[i] == '#')
	for (i++; Code[i] != '\n' && Code[i] != QChar::null; i++);
      else if (Code[i] == '\n')
	{
	  //muCodeLine = muCodeLine.stripWhiteSpace ();
	  if (!muCodeLine.isEmpty ())
	    muCode += muCodeLine;
	  muCodeLine = "";
	}
      else if (Code[i] != ' ') // trim white space
	muCodeLine += Code[i];
    }
  //muCodeLine = muCodeLine.stripWhiteSpace ();
  if (!muCodeLine.isEmpty ())
    {
      //qDebug("adding %s",muCodeLine.ascii());
      muCode += muCodeLine;
    }
  compiled = Script::isCompiled;
  return true;
}

muParserScript *
  muParserScript::current = NULL;

QVariant
muParserScript::eval ()
{
  if (compiled != Script::isCompiled && !compile ())
    return QVariant ();
  double val;
  try
  {
    current = this;
    for (QStringList::iterator i = muCode.begin (); i != muCode.end (); i++)
      {
	//qDebug("evaluating %s",(*i).ascii());
	parser.SetExpr ((const char *)(*i).utf8());
	val = parser.Eval ();
      }
  }
  catch (EmptySourceError *)
  {
    return QVariant ("");
  }
  catch (ParserError & e)
  {
	  emit_error (QString::fromLocal8Bit(e.GetMsg ().c_str()), 0);
    return QVariant ();
  }
  return QVariant (val);
}

bool
muParserScript::exec ()
{
  if (compiled != Script::isCompiled && !compile ())
    return false;
  try
  {
    current = this;
    for (QStringList::iterator i = muCode.begin (); i != muCode.end (); i++)
      {
	//qDebug("executing %s",(*i).ascii());
	parser.SetExpr ((const char *)(*i).utf8());
	parser.Eval ();
      }
  }
  catch (EmptySourceError *)
  {
    return true;
  }
  catch (ParserError & e)
  {
	  emit_error (QString::fromLocal8Bit(e.GetMsg ().c_str()), 0);
    return false;
  }
  return true;
}


const QStringList
muParserScripting::mathFunctions () const
{
  QStringList lst;

  for(const myParser::mathFunction *i = myParser::math_functions; i->name; i++)
    lst << i->name;

  for (const myParser::mathFunction * i = math_functions; i->name; i++)
    lst << i->name;
  
  lst.sort();

  return lst;
}

const QString
muParserScripting::mathFunctionDoc (const QString & name) const
{
  for(const myParser::mathFunction *i = myParser::math_functions; i->name; i++)
    if (name == i->name)
      return i->description;

  for (const myParser::mathFunction *i = math_functions; i->name; i++)
    if (name == i->name)
      return i->description;

  return QString::null;
}
