/***************************************************************************
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor,                    *
 *   Boston, MA  02110-1301  USA                                           *
 *                                                                         *
 ***************************************************************************/
#include "smoothCurveDialog.h"
#include "application.h"
#include "graph.h"
#include "parser.h"
#include "colorBox.h"

#include <qvariant.h>
#include <qpushbutton.h>
#include <qcheckbox.h>
#include <qlabel.h>
#include <qcombobox.h>
#include <qlayout.h>
#include <qbuttongroup.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qmessagebox.h>

smoothCurveDialog::smoothCurveDialog (int method, QWidget * parent,
				      const char *name, bool modal,
				      WFlags fl):
QDialog (parent, name, modal, fl)
{
  smooth_type = method;

  if (!name)
    setName ("smoothCurveDialog");
  setCaption (tr ("QtiPlot - Smoothing Options"));

  QButtonGroup *GroupBox1 =
    new QButtonGroup (2, QGroupBox::Horizontal, tr (""), this, "GroupBox1");

  new QLabel (tr ("Curve"), GroupBox1, "TextLabel1", 0);
  boxName = new QComboBox (GroupBox1, "boxShow");

  if (method == SavitzkyGolay)
    {
      new QLabel (tr ("Polynomial Order"), GroupBox1, "TextLabel2", 0);
      boxOrder = new QSpinBox (0, 9, 1, GroupBox1, "boxOrder");
      boxOrder->setValue (2);

      new QLabel (tr ("Points to the Left"), GroupBox1, "TextLabel3", 0);
      boxPointsLeft = new QSpinBox (1, 100, 1, GroupBox1, "boxPointsLeft");
      boxPointsLeft->setValue (2);

      new QLabel (tr ("Points to the Right"), GroupBox1, "TextLabel3", 0);
      boxPointsRight = new QSpinBox (1, 100, 1, GroupBox1, "boxPointsRight");
      boxPointsRight->setValue (2);
    }
  else
    {
      new QLabel (tr ("Points"), GroupBox1, "TextLabel3", 0);
      boxPointsLeft =
	new QSpinBox (1, 200, 1, GroupBox1, "boxPointsLeft");
      boxPointsLeft->setValue (5);
    }

  new QLabel (tr ("Color"), GroupBox1, "TextLabel52", 0);
  boxColor = new ColorBox (FALSE, GroupBox1);
  boxColor->setColor (QColor (red));

  QButtonGroup *GroupBox2 =
    new QButtonGroup (1, QGroupBox::Horizontal, tr (""), this, "GroupBox2");
  GroupBox2->setFlat (TRUE);
  GroupBox2->setLineWidth (0);

  btnSmooth = new QPushButton (GroupBox2, "btnSmooth");
  btnSmooth->setAutoDefault (TRUE);
  btnSmooth->setDefault (TRUE);

  btnDone = new QPushButton(GroupBox2, "btnDone" );
  btnDone->setAutoDefault (TRUE);

  btnCancel = new QPushButton (GroupBox2, "buttonCancel");
  btnCancel->setAutoDefault (TRUE);

  QHBoxLayout *hlayout = new QHBoxLayout (this, 5, 5, "hlayout");
  hlayout->addWidget (GroupBox1);
  hlayout->addWidget (GroupBox2);

  languageChange ();

  smoothed_index = -1;

  // signals and slots connections
  connect (btnSmooth, SIGNAL (clicked ()), this, SLOT (smooth ()));
  connect (btnDone, SIGNAL( clicked()), this, SLOT (accept()));
  connect (btnCancel, SIGNAL (clicked ()), this, SLOT (cancel ()));
  connect (boxName, SIGNAL (activated (int)), this,
	   SLOT (activateCurve (int)));
}

smoothCurveDialog::~smoothCurveDialog ()
{
}

void
smoothCurveDialog::languageChange ()
{
  btnSmooth->setText (tr ("&Smooth"));
  btnDone->setText (tr("&Done"));
  btnCancel->setText (tr ("&Cancel"));
}

void
smoothCurveDialog::accept()
{
  close();
}

void
smoothCurveDialog::cancel()
{
  ApplicationWindow *app = (ApplicationWindow *) this->parent();

  if(smoothed_index!=-1)
    {
      QwtPlotCurve *sm_curve = d_graph->curve(smoothed_index);
      QString title = sm_curve->title().text();
      QString tableName = title.section( '_', 0, 0 );
      Table *w = app->table(tableName);
      w->askOnCloseEvent (false);
      w->close ();
    }
  reject();
}


void
smoothCurveDialog::smooth ()
{
  int index = d_graph->curveIndex(boxName->currentText());

  if (index < 0)
    return;

  ApplicationWindow *app = (ApplicationWindow *) this->parent();

  if(smoothed_index!=-1)
    {
      QwtPlotCurve *sm_curve = d_graph->curve(smoothed_index);
      QString title = sm_curve->title().text();
      QString tableName = title.section( '_', 0, 0 );
      Table *w = app->table(tableName);
      w->askOnCloseEvent (false);
      w->close ();
    }

  bool result;
  if (smooth_type == SavitzkyGolay)
    {
      int nl = boxPointsLeft->value();
      int nr = boxPointsRight->value();
      int order = boxOrder->value();
      if( nl+nr < order)
	{
	  QMessageBox::critical(0, tr ("QtiPlot - Error"),
				tr("The sum of left and right points must be equal or greater than the order value."));
	  return; 
	}
      result = d_graph->smoothSavGol(app, index, order, nl, nr, 
				     boxColor->currentItem());
    }
  else if (smooth_type == Average)
    result = d_graph->smoothAverage (app, index, boxPointsLeft->value (),
				     boxColor->currentItem ());
  else if (smooth_type == GaussConv)
    result = d_graph->PeakConv(app, index, boxPointsLeft->value (),
			       boxColor->currentItem (), 0);
  else if (smooth_type == LorentzConv)
    result = d_graph->PeakConv(app, index, boxPointsLeft->value (),
			       boxColor->currentItem (), 1);
  if (result)
    smoothed_index = d_graph->curves()-1; // newest curve
  else
    smoothed_index = -1;

}

void
smoothCurveDialog::setGraph (Graph * g)
{
  d_graph = g;
  boxName->insertStringList (g->analyzableCurvesList (), -1);
  activateCurve (0);
}

void
smoothCurveDialog::activateCurve (int index)
{
  if (smooth_type == Average)
    {
      QString ctitle = boxName->text(index);
      int cindex = d_graph->curveIndex(ctitle);
      QwtPlotCurve *c = d_graph->curve (cindex);
      if (!c || c->rtti () != QwtPlotItem::Rtti_PlotCurve)
	return;

      boxPointsLeft->setMaxValue (c->dataSize () / 2);
    }
}
