/***************************************************************************
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor,                    *
 *   Boston, MA  02110-1301  USA                                           *
 *                                                                         *
 ***************************************************************************/
#include "worksheet.h"
#include "sortDialog.h"
#include "application.h"
#include "colFormulaDialog.h"
#include "gslutils.h"

#include <qpopupmenu.h>
#include <qfiledialog.h>
#include <qmessagebox.h>
#include <qregexp.h>
#include <qtextstream.h>
#include <qclipboard.h>
#include <qapplication.h>
#include <qcursor.h>
#include <qdatetime.h>
#include <qprinter.h>
#include <qpainter.h>
#include <qpaintdevicemetrics.h>
#include <qdragobject.h>
#include <qlayout.h>
#include <qaccel.h>
#include <qprogressdialog.h>
#include <qeventloop.h>
#include <qlineedit.h>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_sort.h>
#include <gsl/gsl_sort_vector.h>
#include <gsl/gsl_interp.h>
#include <gsl/gsl_fft_halfcomplex.h>
#include <gsl/gsl_fft_real.h>
#include <gsl/gsl_fft_complex.h>
#include <gsl/gsl_statistics.h>

#define SWAP(a,b) do{int tmp=a; a=b; b=tmp;}while(0)

Table::Table (ScriptingEnv * env, const QString & fname, const QString & sep,
	      int ignoredLines, bool renameCols, bool stripSpaces,
	      bool simplifySpaces, const QString & label, QWidget * parent,
	      const char *name, WFlags f):
myWidget (label, parent, name, f),
scripted (env)
{
  //fprintf(stderr,"New Table!!\n");
  importASCII (fname, sep, ignoredLines, renameCols, stripSpaces,
	       simplifySpaces, 0); // import to new table
}

Table::Table (ScriptingEnv * env, int r, int c, const QString & label,
	      QWidget * parent, const char *name, WFlags f):
myWidget (label, parent, name, f),
scripted (env)
{
  init (r, c);
}

void
Table::init (int rows, int cols)
{
  selectedCol = 0;
  savedCol = -1;

  QDateTime dt = QDateTime::currentDateTime ();
  setBirthDate (dt.toString (Qt::LocalDate));

  worksheet = new QTable (rows, cols, this, "table");
  worksheet->setFocusPolicy (QWidget::StrongFocus);
  worksheet->setFocus ();
  worksheet->setSelectionMode (QTable::Single);
  worksheet->setRowMovingEnabled (false);
  worksheet->setColumnMovingEnabled(false);
// TODO: would be useful, but then we have to interpret
// signal QHeader::indexChange ( int section, int fromIndex, int toIndex )
// and update some variables
// worksheet->setColumnMovingEnabled(true);

  connect (worksheet->verticalHeader (), SIGNAL (indexChange (int, int, int)),
	   this, SLOT (notifyChanges ()));

  QVBoxLayout *hlayout = new QVBoxLayout (this, 0, 0, "hlayout1");
  hlayout->addWidget (worksheet);

  for (int i = 0; i < cols; i++)
    {
      commandes << "";
      colTypes << Numeric;
      col_format << "0/6";
      comments << "";
      col_label << QString::number (i + 1);
      col_plot_type << Y;
    }

  QHeader *head = (QHeader *) worksheet->horizontalHeader ();
  head->setMouseTracking (true);
  head->setResizeEnabled (true);
  head->installEventFilter (this);
  head->setMovingEnabled(false);

  connect (head, SIGNAL (sizeChange (int, int, int)), this,
	   SLOT (colWidthModified (int, int, int)));

  col_plot_type[0] = X;
  setHeaderColType ();

  int w = 4 * (worksheet->horizontalHeader ())->sectionSize (0);
  int h;
  if (rows > 11)
    h = 11 * (worksheet->verticalHeader ())->sectionSize (0);
  else
    h = (rows + 1) * (worksheet->verticalHeader ())->sectionSize (0);
  setGeometry (50, 50, w + 45, h);

  worksheet->verticalHeader ()->setResizeEnabled (false);
  worksheet->verticalHeader ()->installEventFilter (this);
  worksheet->verticalHeader ()->setMovingEnabled(false);

  QAccel *accel = new QAccel (this);
  accel->connectItem (accel->insertItem (Key_Tab), this,
		      SLOT (moveCurrentCell ()));
  accel->connectItem (accel->insertItem (CTRL + Key_A), this,
		      SLOT (selectAllTable ()));

  connect (worksheet, SIGNAL (valueChanged (int, int)), this,
	   SLOT (cellEdited (int, int)));
}

void
Table::colWidthModified (int, int, int)
{
  emit modifiedWindow (this);
}

void
Table::setBackgroundColor (const QColor & col)
{
  worksheet->setPaletteBackgroundColor (col);
}

void
Table::setTextColor (const QColor & col)
{
  worksheet->setPaletteForegroundColor (col);
}

void
Table::setTextFont (const QFont & fnt)
{
  worksheet->setFont (fnt);
  QFontMetrics fm (fnt);
  int lm = fm.width (QString::number (10 * worksheet->numRows ()));
  worksheet->setLeftMargin (lm);
}

void
Table::setHeaderColor (const QColor & col)
{
  worksheet->horizontalHeader ()->setPaletteForegroundColor (col);
}

void
Table::setHeaderFont (const QFont & fnt)
{
  worksheet->horizontalHeader ()->setFont (fnt);
}

void
Table::print ()
{
  QPrinter printer;
  printer.setColorMode (QPrinter::GrayScale);
  if (printer.setup ())
    {
      printer.setFullPage (TRUE);
      QPainter p;
      if (!p.begin (&printer))
	return;			// paint on printer
      QPaintDeviceMetrics metrics (p.device ());
      int dpiy = metrics.logicalDpiY ();
      const int margin = (int) ((1 / 2.54) * dpiy);	// 2 cm margins

      QHeader *hHeader = worksheet->horizontalHeader ();
      QHeader *vHeader = worksheet->verticalHeader ();

      int rows = worksheet->numRows ();
      int cols = worksheet->numCols ();
      int height = margin;
      int i, vertHeaderWidth = vHeader->width ();
      int right = margin + vertHeaderWidth;

      // print header
      p.setFont (hHeader->font ());
      QRect br =
	p.boundingRect (br, Qt::AlignCenter, hHeader->label (0), -1, 0);
      p.drawLine (right, height, right, height + br.height ());
      QRect tr (br);

      for (i = 0; i < cols; i++)
	{
	  int w = worksheet->columnWidth (i);
	  tr.setTopLeft (QPoint (right, height));
	  tr.setWidth (w);
	  tr.setHeight (br.height ());
	  p.drawText (tr, Qt::AlignCenter, hHeader->label (i), -1);
	  right += w;
	  p.drawLine (right, height, right, height + tr.height ());

	  if (right >= metrics.width () - 2 * margin)
	    break;
	}
      p.drawLine (margin + vertHeaderWidth, height, right - 1, height);	//first horizontal line
      height += tr.height ();
      p.drawLine (margin, height, right - 1, height);

      p.setFont (worksheet->font ());
      // print table values
      for (i = 0; i < rows; i++)
	{
	  right = margin;
	  QString text = vHeader->label (i) + "\t";
	  tr = p.boundingRect (tr, Qt::AlignCenter, text, -1, 0);
	  p.drawLine (right, height, right, height + tr.height ());

	  br.setTopLeft (QPoint (right, height));
	  br.setWidth (vertHeaderWidth);
	  br.setHeight (tr.height ());
	  p.drawText (br, Qt::AlignCenter, text, -1);
	  right += vertHeaderWidth;
	  p.drawLine (right, height, right, height + tr.height ());

	  for (int j = 0; j < cols; j++)
	    {
	      int w = worksheet->columnWidth (j);
	      text = worksheet->text (i, j) + "\t";
	      tr = p.boundingRect (tr, Qt::AlignCenter, text, -1, 0);
	      br.setTopLeft (QPoint (right, height));
	      br.setWidth (w);
	      br.setHeight (tr.height ());
	      p.drawText (br, Qt::AlignCenter, text, -1);
	      right += w;
	      p.drawLine (right, height, right, height + tr.height ());

	      if (right >= metrics.width () - 2 * margin)
		break;
	    }
	  height += br.height ();
	  p.drawLine (margin, height, right - 1, height);

	  if (height >= metrics.height () - margin)
	    {
	      printer.newPage ();
	      height = margin;
	      p.drawLine (margin, height, right, height);
	    }
	}
    }
}

void
Table::cellEdited (int row, int col)
{
  if (columnType (col) == Table::Text)
    {
      //qDebug("coltype is Text");
      emit modifiedData (this, colName (col));
      emit modifiedWindow (this);
      return;
    }

  char f;
  int precision;
  columnNumericFormat (col, f, precision);
  QString text = worksheet->text(row, col).replace (",", ".");

  bool ok = false;
  double res = text.toDouble (&ok);
  if (text.isEmpty ())
    worksheet->setText (row, col, "");
  else if(ok)
    worksheet->setText (row, col, QString::number (res, f, precision));
  else
    {
      Script *script =
	scriptEnv->newScript (worksheet->text (row, col), this,
			      QString ("<%1_%2_%3>").arg (getName()).arg (row + 1).arg (col + 1));
      connect (script, SIGNAL (error (const QString &, const QString &, int)),
	       scriptEnv,
	       SIGNAL (error (const QString &, const QString &, int)));

      script->setInt (row + 1, "i");
      script->setInt (col + 1, "j");
      QVariant ret = script->eval ();
      if (ret.type () == QVariant::Int || ret.type () == QVariant::UInt
	  || ret.type () == QVariant::LongLong
	  || ret.type () == QVariant::ULongLong)
	worksheet->setText (row, col, ret.toString ());
      else if (ret.canCast (QVariant::Double))
	worksheet->setText (row, col,
			    QString::number (ret.toDouble (), f, precision));
      else
	worksheet->setText (row, col, "");
    }

  emit modifiedData (this, colName (col));
  emit modifiedWindow (this);
}

int
Table::colX (int col)
{
  int i, xcol = -1;
  for (i = col - 1; i >= 0; i--)
    {
      if (col_plot_type[i] == X)
	return i;
    }
  for (i = col + 1; i < (int) worksheet->numCols (); i++)
    {
      if (col_plot_type[i] == X)
	return i;
    }
  return xcol;
}

int
Table::colY (int col)
{
  int i, yCol = -1;
  for (i = col - 1; i >= 0; i--)
    {
      if (col_plot_type[i] == Y)
	return i;
    }
  for (i = col + 1; i < (int) worksheet->numCols (); i++)
    {
      if (col_plot_type[i] == Y)
	return i;
    }
  return yCol;
}

void
Table::setPlotDesignation (PlotDesignation pd)
{
  QStringList list = selectedColumns ();
  for (int i = 0; i < (int) list.count (); i++)
    col_plot_type[colIndex (list[i])] = pd;

  setHeaderColType ();
  emit modifiedWindow (this);
}

void
Table::columnNumericFormat (int col, int &f, int &precision)
{
  QStringList format = QStringList::split ('/', col_format[col], false);
  f = format[0].toInt ();
  precision = format[1].toInt ();
}

void
Table::columnNumericFormat (int col, char &f, int &precision)
{
  //fprintf(stderr,"columnNumericFormat of col %d\n",col);
  QStringList format = QStringList::split ('/', col_format[col], false);
  switch (format[0].toInt ())
    {
    case 0:
      f = 'g';
      break;

    case 1:
      f = 'f';
      break;

    case 2:
      f = 'e';
      break;
    }
  precision = format[1].toInt ();
}

int
Table::columnWidth (int col)
{
  return worksheet->columnWidth (col);
}

QStringList
Table::columnWidths ()
{
  QStringList widths;
  for (int i = 0; i < worksheet->numCols (); i++)
    widths << QString::number (worksheet->columnWidth (i));

  return widths;
}

void
Table::setColWidths (const QStringList & widths)
{
  for (int i = 0; i < (int) widths.count (); i++)
    worksheet->setColumnWidth (i, widths[i].toInt ());
}

void
Table::setColumnTypes (const QStringList & ctl)
{
  for (int i = 0; i < (int) ctl.count (); i++)
    {
      QStringList l = QStringList::split (';', ctl[i], true);
      colTypes[i] = l[0].toInt ();

      if ((int) l.count () > 0 && !l[1].isEmpty ())
	col_format[i] = l[1];
      else
	col_format[i] = "0/6";
    }
}

QString
Table::saveColumnWidths ()
{
  QString s = "ColWidth\t";
  int i;
  for (i = 0; i < worksheet->numCols (); i++)
    s += QString::number (worksheet->columnWidth (i)) + "\t";
  s.truncate( s.length() - 1 );

  return s + "\n";
}

QString
Table::saveColumnTypes ()
{
  QString s = "ColType\t";
  int i;
  for (i = 0; i < worksheet->numCols(); i++)
    s += QString::number (colTypes[i]) + ";" + col_format[i] + "\t";
  s.truncate( s.length() - 1 );

  return s + "\n";
}

void
Table::setCommandes (const QStringList & com)
{
  commandes.clear ();
  for (int i = 0; i < (int) com.size () && i < tableCols (); i++)
    commandes << com[i].stripWhiteSpace ();
}

void
Table::setCommand (int col, const QString & com)
{
  if (col < (int) commandes.size ())
    commandes[col] = com.stripWhiteSpace ();
}

void
Table::setCommandes (const QString & com)
{
  QStringList lst = QStringList::split ('\t', com, true);
  lst.pop_front ();
  setCommandes (lst);
}

bool
Table::calculate (int col, int startRow, int endRow, bool notify)
{
  if(commandes[col]=="") return false;

  QApplication::setOverrideCursor (waitCursor);

  Script *colscript =
    scriptEnv->newScript (commandes[col], this,
			  QString ("<%1>").arg (colName (col)));
  connect (colscript, SIGNAL (error (const QString &, const QString &, int)),
	   scriptEnv, SIGNAL (error (const QString &, const QString &, int)));
  connect (colscript, SIGNAL (print (const QString &)),
	   parent ()->parent ()->parent (),
	   SLOT (scriptPrint (const QString &)));

  if (!colscript->compile ())
    {
      QApplication::restoreOverrideCursor ();
      return false;
    }
  if (endRow+1 > tableRows ())
    resizeRows (endRow + 1);

  colscript->setInt (col + 1, "j");
  colscript->setInt (startRow + 1, "sr");
  colscript->setInt (endRow + 1, "er");
  QVariant ret;
  saveColToMemory (col);
  for (int i = startRow; i <= endRow; i++)
    {
      colscript->setInt (i + 1, "i");
      ret = colscript->eval ();
      if (ret.type () == QVariant::Double)
	{
	  int prec;
	  char f;
	  columnNumericFormat (col, f, prec);
	  worksheet->setText (i, col,
			      QString::number (ret.toDouble (), f, prec));
	}
      else if (ret.canCast (QVariant::String))
	worksheet->setText (i, col, ret.toString ());
      else
	{
	  QApplication::restoreOverrideCursor ();
	  return false;
	}
    }
  forgetSavedCol ();

  emit modifiedData (this, colName (col));
  if (notify) emit modifiedWindow (this);
  QApplication::restoreOverrideCursor ();
  return true;
}

QTableSelection
Table::getSelection ()
{
  QTableSelection sel;
  if (worksheet->numSelections () == 0)
    {
      sel.init (worksheet->currentRow (), worksheet->currentColumn ());
      sel.expandTo (worksheet->currentRow (), worksheet->currentColumn ());
    }
  else if (worksheet->currentSelection () > 0)
    sel = worksheet->selection (worksheet->currentSelection ());
  else
    sel = worksheet->selection (0);
  return sel;
}

bool
Table::calculate ()
{
  QTableSelection sel = getSelection ();
  bool success = true;
  for (int col = sel.leftCol (); col <= sel.rightCol (); col++)
    if (!calculate (col, sel.topRow (), sel.bottomRow ()))
      success = false;
  emit modifiedWindow(this);
  return success;
}

QString
Table::saveCommandes ()
{
  QString s = "<com>\n";
  for (int col = 0; col < tableCols (); col++)
    if (!commandes[col].isEmpty ())
      {
	s += "<col nr=\"" + QString::number (col) + "\">\n";
	s += commandes[col];
	s += "\n</col>\n";
      }
  s += "</com>\n";
  return s;
}

QString
Table::saveComments ()
{
  QString s = "Comments\t";
  for (int i = 0; i < worksheet->numCols (); i++)
    s += comments[i] + "\t";
  s.truncate( s.length() - 1 );
  return s + "\n";
}

QString
Table::saveToString (const QString & geometry)
{
  QString s = "<table>\n";
  s += getName() + "\t";
  s += QString::number (worksheet->numRows ()) + "\t";
  s += QString::number (worksheet->numCols ()) + "\t";
  s += birthDate () + "\n";
  s += geometry;
  s += saveHeader ();
  s += saveColumnWidths ();
  s += saveCommandes ();
  s += saveColumnTypes ();
  s += saveComments ();
  s +=
    "WindowLabel\t" + windowLabel () + "\t" +
    QString::number (captionPolicy ()) + "\n";
  s += saveText ();
  return s += "</table>\n";
}

QString
Table::saveHeader ()
{
  QString s = "header\t";
  for (int j = 0; j < worksheet->numCols (); j++)
    {
      if (col_plot_type[j] == X)
	s += colLabel (j) + "[X]\t";
      else if (col_plot_type[j] == Y)
	s += colLabel (j) + "[Y]\t";
      else if (col_plot_type[j] == Z)
	s += colLabel (j) + "[Z]\t";
      else if (col_plot_type[j] == xErr)
	s += colLabel (j) + "[xEr]\t";
      else if (col_plot_type[j] == yErr)
	s += colLabel (j) + "[yEr]\t";
      else
	s += colLabel (j) + "\t";
    }
  s.truncate( s.length() - 1 );
  return s += "\n";
}

int
Table::firstXCol ()
{
  int xcol = -1;
  for (int j = 0; j < worksheet->numCols (); j++)
    {
      if (col_plot_type[j] == X)
	return j;
    }
  return xcol;
}

void
Table::setColComment (const QString & s)
{
  if (comments[selectedCol] == s)
    return;

  comments[selectedCol] = s;
}

void
Table::changeColWidth (int width, bool allCols)
{
  int cols = worksheet->numCols ();
  if (allCols)
    {
      for (int i = 0; i < cols; i++)
	worksheet->setColumnWidth (i, width);

      emit modifiedWindow (this);
    }
  else
    {
      if (worksheet->columnWidth (selectedCol) == width)
	return;

      worksheet->setColumnWidth (selectedCol, width);
      emit modifiedWindow (this);
    }
}

bool
Table::changeColName (const QString & text, bool enumerateRightCols)
{
  QString caption = getName(); //Table name

  if(enumerateRightCols)
    {
      int cols = worksheet->numCols() - selectedCol;
      QStringList labels = colNames().grep(QRegExp("^"+text));
      for(int i = 0; i<(int)labels.count(); i++)
	{
	  int num = labels[i].remove(text).toInt();
	  if (num > 0 && num <= cols)
	    {
	      QMessageBox::critical 
		(0, tr ("QtiPlot - Error"),
		 tr ("There is already a column called : <b>" +
		     text + QString::number(num) +
		     "</b> in table <b>" + caption +
		     "</b>!<p>Please choose another name!"));
	      return false;
	    }
	}
      if(caption.startsWith(text))
	{
	  int num = caption.remove(text).toInt();
	  if(num > 0 && num <= cols)
	    {
	      QMessageBox::critical
		(0, tr ("QtiPlot - Error"),
		 tr("The column name must be different from the table name!"
		     "<p>Please choose another name!"));
	      return false;
	    }
	}
      int n = 1;
      caption = getName(); //caption may be modified
      for (int i = selectedCol; i < (int) worksheet->numCols (); i++)
	{
	  QString oldLabel = col_label[i];
	  QString newLabel = text + QString::number (n);
	  commandes.gres ("col(\"" + oldLabel + "\")",
			  "col(\"" + newLabel + "\")", true);
	  col_label[i] = newLabel;
	  emit changedColHeader(caption + "_" + oldLabel, 
				caption + "_" + newLabel);
	  n++;
	}
    }
  else
    {  
      QString oldName = colName (selectedCol);
      QString newName = caption + "_" + text;
      if(newName==oldName) return true;
      QStringList labels = colNames ();
      if (labels.contains (text) > 0)
	{
	  QMessageBox::critical (0, tr ("QtiPlot - Error"),
				 tr ("There is already a column called : <b>" +
				     text + "</b> in table <b>" + caption +
				     "</b>!<p>Please choose another name!"));
	  return false;
	}
      if (caption == text)
	{
	  QMessageBox::critical
	    (0, tr ("QtiPlot - Error"),
	     tr("The column name must be different from the table name : <b>"
		+ caption + "</b></b>!<p>Please choose another name!"));
	  return false;
	}
      commandes.gres ("col(\"" + col_label[selectedCol] + "\")",
		      "col(\"" + text + "\")", true);
      col_label[selectedCol] = text;
      emit changedColHeader (oldName, newName);
    }

  setHeaderColType ();
  emit modifiedWindow (this);
  return true;
}

void
Table::setColName (int col, const QString & text)
{
  if (text.isEmpty () || col < 0 || col >= worksheet->numCols ())
    return;

  col_label[col] = text;
}

QStringList
Table::selectedColumns ()
{
  QStringList names;
  for (int i = 0; i < worksheet->numCols (); i++)
    {
      if (worksheet->isColumnSelected (i, true))
	names << getName() + "_" + col_label[i];
    }
  return names;
}

QStringList
Table::YColumns ()
{
  QStringList names;
  for (int i = 0; i < worksheet->numCols (); i++)
    {
      if (col_plot_type[i] == Y)
	names << getName() + "_" + col_label[i];
    }
  return names;
}

QStringList
Table::selectedYColumns ()
{
  QStringList names;
  for (int i = 0; i < worksheet->numCols (); i++)
    {
      if (worksheet->isColumnSelected (i, true) && col_plot_type[i] == Y)
	names << getName() + "_" + col_label[i];
    }
  return names;
}

QStringList
Table::selectedErrColumns ()
{
  QStringList names;
  for (int i = 0; i < worksheet->numCols (); i++)
    {
      if (worksheet->isColumnSelected (i, true) &&
	  (col_plot_type[i] == yErr || col_plot_type[i] == xErr))
	names << getName() + "_" + col_label[i];
    }
  return names;
}

QStringList
Table::drawableColumnSelection ()
{
  QStringList names;
  for (int i = 0; i < worksheet->numCols (); i++)
    {
      if (worksheet->isColumnSelected (i, true) && col_plot_type[i] == Y)
	names << getName() + "_" + col_label[i];
    }

  for (int j = 0; j < worksheet->numCols (); j++)
    {
      if (worksheet->isColumnSelected (j, true)
	  && (col_plot_type[j] == yErr || col_plot_type[j] == xErr))
	names << getName() + "_" + col_label[j];
    }
  return names;
}

QStringList
Table::selectedYLabels ()
{
  QStringList names;
  for (int i = 0; i < worksheet->numCols (); i++)
    {
      if (worksheet->isColumnSelected (i, true) && col_plot_type[i] == Y)
	names << col_label[i];
    }
  return names;
}

QStringList
Table::columnsList ()
{
  QStringList names;
  for (int i = 0; i < worksheet->numCols (); i++)
    names << getName() + "_" + col_label[i];

  return names;
}

int
Table::firstSelectedColumn ()
{
  for (int i = 0; i < worksheet->numCols (); i++)
    {
      if (worksheet->isColumnSelected (i, TRUE))
	return i;
    }
  return -1;
}

int
Table::selectedRows ()
{
  int r = 0;
  for (int i = 0; i < worksheet->numRows (); i++)
    {
      if (worksheet->isRowSelected (i, TRUE))
	r++;
    }
  return r;
}

int
Table::selectedColsNumber ()
{
  int c = 0;
  for (int i = 0; i < worksheet->numCols (); i++)
    {
      if (worksheet->isColumnSelected (i, TRUE))
	c++;
    }
  return c;
}

QString
Table::colName (int col)
{				//returns the table name + horizontal header text
  if (col < 0 || col >= worksheet->numCols ())
    return QString::null;

  return getName() + "_" + col_label[col];
}

QMemArray < double >
Table::col (int ycol)
{
  int i;
  int rows = worksheet->numRows ();
  int cols = worksheet->numCols ();
  QMemArray < double >Y (rows);
  if (ycol <= cols)
    {
      for (i = 0; i < rows; i++)
	Y[i] = worksheet->text (i, ycol).toDouble ();
    }
  return Y;
}

void
Table::insertCols (int start, int count)
{
  start--;			//insert new columns before the start/selected column

  int max = 0;
  int cols = worksheet->numCols ();
  QRegExp rx("\\D");
  for (int i = 0; i < cols; i++)
    {
      if (col_label[i].contains(rx) == 0)
	{
	  int index = col_label[i].toInt ();
	  if (index > max)
	    max = index;
	}
    }
  max++;

  QStringList::Iterator it = commandes.at (++start);
  commandes.insert (it++, count, QString::null);
  it = col_format.at (start);
  col_format.insert (it++, count, "0/6");
  it = comments.at (start);
  comments.insert (it++, count, QString::null);
  it = col_label.at (start);
  col_label.insert (it++, count, QString::null);

  QValueListIterator < int >itt = colTypes.at (start);
  colTypes.insert (itt++, count, Numeric);
  itt = col_plot_type.at (start);
  col_plot_type.insert (itt++, count, Y);

  worksheet->insertColumns (start, count);
  for (int j = 0; j < count; j++)
    col_label[start + j] = QString::number (max + j);

  setHeaderColType ();
}

void
Table::insertCol ()
{
  insertCols (selectedCol, 1);
  emit modifiedWindow (this);
}

void
Table::insertRow ()
{
  int cr = worksheet->currentRow ();
  if (worksheet->isRowSelected (cr, true))
    {
      worksheet->insertRows (cr, 1);
      emit modifiedWindow (this);
    }
}

void
Table::addRow ()
{
  int rows = worksheet->numRows();
  int col = worksheet->currentColumn();
  worksheet->insertRows (rows, 1);
  worksheet->ensureCellVisible (rows, col);
  emit modifiedWindow (this);
}

void
Table::addCol (PlotDesignation pd)
{
  worksheet->clearSelection ();
  int index, max = 0, cols = worksheet->numCols ();
  for (int i = 0; i < cols; i++)
    {
      bool ok;
      index = col_label[i].toInt (&ok);
      if (ok && index > max)
	max = index;
    }
  worksheet->insertColumns (cols, 1);
  worksheet->ensureCellVisible (0, cols);

  comments << QString::null;
  commandes << "";
  colTypes << Numeric;
  col_format << "0/6";
  col_label << QString::number (max + 1);
  col_plot_type << pd;

  setHeaderColType ();
  emit modifiedWindow (this);
}

void
Table::addColumns (int c)
{
  int i, index, max = 0, cols = worksheet->numCols ();
  for (i = 0; i < cols; i++)
    {
      bool ok;
      index = col_label[i].toInt (&ok);
      if (ok && index > max)
	max = index;
    }
  max++;
  worksheet->insertColumns (cols, c);
  for (i = 0; i < c; i++)
    {
      comments << QString::null;
      commandes << "";
      colTypes << Numeric;
      col_format << "0/6";
      col_label << QString::number (max + i);
      col_plot_type << Y;
    }
}

void
Table::clearCol ()
{
  int rows = worksheet->numRows ();
  for (int i = 0; i < rows; i++)
    {
      worksheet->setText (i, selectedCol, "");
    }
  QString name = colName (selectedCol);
  emit modifiedData (this, name);
  emit modifiedWindow (this);
}

void
Table::clearCell (int row, int col)
{
  worksheet->setText (row, col, "");

  QString name = colName (col);
  emit modifiedData (this, name);
  emit modifiedWindow (this);
}

void
Table::deleteSelectedRows ()
{
  QTableSelection sel = worksheet->selection (0);
  int top = sel.topRow ();
  int bottom = sel.bottomRow ();
  int numberOfRows = bottom - top + 1;
  QMemArray < int >rowsToDelete (numberOfRows);
  for (int i = 0; i < numberOfRows; i++)
    rowsToDelete[i] = top + i;
  worksheet->removeRows (rowsToDelete);
  notifyChanges ();
}

void
Table::cutSelection ()
{
  copySelection ();
  clearSelection ();
}

void
Table::selectAllTable ()
{
  worksheet->
    addSelection (QTableSelection
		  (0, 0, worksheet->numRows (), worksheet->numCols ()));
}

void
Table::deselect ()
{
  worksheet->clearSelection ();
}

void
Table::clearSelection ()
{
  QStringList list = selectedColumns ();
  int i, n = int (list.count ());

  if (n > 0)
    {
      int rows = worksheet->numRows ();
      for (i = 0; i < n; i++)
	{
	  QString name = list[i];
	  int id = colIndex (name);
	  //selectedCol = id;
	  //clearCol ();
	  for (int j = 0; j < rows; j++)
	    worksheet->setText (j, id, "");
	  emit modifiedData (this, name);
	}
      emit modifiedWindow (this);	  
    }
  else
    {
      QTableSelection sel = worksheet->selection (0);
      int top = sel.topRow ();
      int bottom = sel.bottomRow ();
      int left = sel.leftCol ();
      int right = sel.rightCol ();

      if (sel.isEmpty ())
	{
	  int row = worksheet->currentRow ();
	  int col = worksheet->currentColumn ();
	  worksheet->setText (row, col, "");

	  QString name = colName (col);
	  emit modifiedData (this, name);
	  emit modifiedWindow (this);
	}
      else
	{
	  for (i = top; i <= bottom; i++)
	    {
	      for (int j = left; j <= right; j++)
		worksheet->setText (i, j, "");
	    }

	  for (i = left; i <= right; i++)
	    {
	      QString name = colName (i);
	      emit modifiedData (this, name);
	    }

	  emit modifiedWindow (this);
	}
    }
}

QString
Table::partialColStatistics(int prec)
{
  double min,max;
  int min_index, max_index;

  int rows = selectedRows();
  if(rows==0) return QString::null;

  QMemArray <double> dat(rows);
  QString text = "Col\tMean\tStandardDev\tVariance\tSum\tiMax\tMax\tiMin\tMin\tN\n";
  int cols = worksheet->numCols();
  for(int c = 0; c < cols; c++)
    { 
      int num = 0;
      for( int r = 0; r < worksheet->numRows(); r++)
	{
	  if(worksheet->isRowSelected(r) && !worksheet->text(r,c).isEmpty())
	    {
	      num++;
	      double value = worksheet->text(r,c).toDouble();
	      dat[num-1] = value;
	      if(num==1)
		{
		  min = value;
		  min_index = r;
		  max = value;
		  max_index = r;
		}
	      else if (value < min)
		{
		  min = value;
		  min_index = r;
		}
	      else if (value > max)
		{
		  max = value;
		  max_index = r;
		}
	    }	  
	}
      double mean =  gsl_stats_mean (dat.data(), 1, num);
      text += QString::number(c+1) + "\t";
      text += QString::number(mean,'g',prec) + "\t";
      text += QString::number(gsl_stats_sd(dat.data(),1,num),'g',prec) + "\t";
      text += QString::number(gsl_stats_variance(dat.data(),1,num),'g',prec) + "\t";
      text += QString::number(mean*num,'g',prec) + "\t";
      text += QString::number(max_index+1) + "\t";
      text += QString::number(max,'g',prec) + "\t";
      text += QString::number(min_index+1) + "\t";
      text += QString::number(min,'g',prec) + "\t";
      text += QString::number(num) + "\n";
    }

  return text;
}

QString
Table::partialRowStatistics(int prec)
{
  double min,max;
  int min_index, max_index;

  int cols = selectedColumns().count();
  if(cols==0) return QString::null;

  QMemArray <double> dat(cols);
  QString text = "Row\tMean\tStandardDev\tVariance\tSum\tiMax\tMax\tiMin\tMin\tN\n";
  int rows = worksheet->numRows();
  for(int r = 0; r < rows; r++)
    { 
      int num = 0;
      for( int c = 0; c < worksheet->numCols(); c++)
	{
	  if(worksheet->isColumnSelected(c,true)
	     && !worksheet->text(r,c).isEmpty())
	    {
	      num++;
	      double value = worksheet->text(r,c).toDouble();
	      dat[num-1] = value;
	      if(num==1)
		{
		  min = value;
		  min_index = c;
		  max = value;
		  max_index = c;
		}
	      else if (value < min)
		{
		  min = value;
		  min_index = c;
		}
	      else if (value > max)
		{
		  max = value;
		  max_index = c;
		}
	    }	  
	}
      double mean =  gsl_stats_mean (dat.data(), 1, num);
      text += QString::number(r+1) + "\t";
      text += QString::number(mean,'g',prec) + "\t";
      text += QString::number(gsl_stats_sd(dat.data(),1,num),'g',prec) + "\t";
      text += QString::number(gsl_stats_variance(dat.data(),1,num),'g',prec) + "\t";
      text += QString::number(mean*num,'g',prec) + "\t";
      text += QString::number(max_index+1) + "\t";
      text += QString::number(max,'g',prec) + "\t";
      text += QString::number(min_index+1) + "\t";
      text += QString::number(min,'g',prec) + "\t";
      text += QString::number(num) + "\n";
    }

  return text;
}

QString
Table::momentAnalysis(int prec)
{
  QStringList list = selectedYColumns ();
  int i, n = int (list.count ());

  if (n == 0)
    {
      QMessageBox::critical (this, tr ("QtiPlot - Moment Analysis error"),
			     tr
			     ("There is no selected Y Column!"));
      return QString::null; 
    }

  QString text = 
    tr("Column") + "\t" 
    + tr("n") + "\t" 
    + tr("0th") + "\t" 
    + tr("1st") + "\t"
    + tr("2nd") + "\t"
    + tr("mean") + "\t"
    + tr("stddev") + "\n";
  
  int rows = worksheet->numRows();
  int num;
  double sum0,sum1,sum2;
  for (i = 0; i < n; i++)
    {
      QString name = list[i];
      int ycol = colIndex (name);
      int xcol = colX(ycol);
      if(xcol>=0)
	{
	  num=0; sum0=0; sum1=0; sum2=0;
	  for (int j = 0; j < rows; j++)
	    if(worksheet->text(j,xcol)!="")
	      {	  
		double x = (worksheet->text (j, xcol)).toDouble();
		double y = (worksheet->text (j, ycol)).toDouble();
		num++;
		sum0 += y;
		sum1 += x * y;
		sum2 += x * x * y;
	      }
	  double mean = sum1 / sum0;
	  double sd = sqrt( sum2 / sum0 - mean * mean );
	  text += name + "\t" 
	    + QString::number(num) + "\t"
	    + QString::number(sum0,'g',prec) + "\t"
	    + QString::number(sum1,'g',prec) + "\t"
	    + QString::number(sum2,'g',prec) + "\t"
	    + QString::number(mean,'g',prec) + "\t"
	    + QString::number(sd,'g',prec) + "\n";
	}
    }
  return text;
}

void
Table::copySelection ()
{
  QString text="";
  int i, j;
  int rows = worksheet->numRows ();
  int cols = worksheet->numCols ();

  QMemArray < int >selection (1);
  int c = 0;
  for (i = 0; i < cols; i++)
    {
      if (worksheet->isColumnSelected (i, true))
	{
	  c++;
	  selection.resize (c);
	  selection[c - 1] = i;
	}
    }
  if (c > 0)
    {
      for (i = 0; i < rows; i++)
	{
	  for (j = 0; j < c - 1; j++)
	    text += worksheet->text (i, selection[j]) + "\t";
	  text += worksheet->text (i, selection[c - 1]) + "\n";
	}
    }
  else
    {
      QTableSelection sel = worksheet->selection (0);
      if(sel.isEmpty())
	{
	  int row = worksheet->currentRow();
	  int col = worksheet->currentColumn();
	  text += worksheet->text(row,col) + "\n";
	}
      else
	{
	  int right = sel.rightCol ();
	  for (i = sel.topRow (); i <= sel.bottomRow (); i++)
	    {
	      for (j = sel.leftCol (); j < right; j++)
		text += worksheet->text (i, j) + "\t";
	      text += worksheet->text (i, right) + "\n";
	    }
	}
    }

// Copy text into the clipboard
  //QApplication::clipboard ()->setData (new QTextDrag (text, worksheet, 0));
  QApplication::clipboard ()->setText(text);
}

// Paste text from the clipboard
void
Table::paste(bool transposed)
{
  /*
  QString text;
  if (!QTextDrag::decode (QApplication::clipboard ()->data (), text)
      || text.isNull ())
    return;
  */
  QString text=QApplication::clipboard ()->text();
  if(text.isNull()) return;

  QApplication::setOverrideCursor (waitCursor);

  QTextStream ts (&text, IO_ReadOnly);
  QString s = ts.readLine ();
  QStringList cellTexts = QStringList::split ('\t', s, TRUE);

  int cols = int (cellTexts.count ());
  int rows = 1;

  while (!ts.atEnd ())
    {
      rows++;
      s = ts.readLine ();
    }
  ts.reset ();

  int i, j, top, bottom, right, left, firstCol = firstSelectedColumn ();
  QTableSelection sel = worksheet->selection (0);
  if (!sel.isEmpty ())
    {				// not columns but only cells are selected
      top = sel.topRow ();
      bottom = sel.bottomRow ();
      left = sel.leftCol ();
      right = sel.rightCol ();
    }
  else
    {
      if (cols == 1 && rows == 1)
	{
	  top = bottom = worksheet->currentRow ();
	  left = right = worksheet->currentColumn ();
	}
      else
	{
	  top = 0;
	  bottom = worksheet->numRows () - 1;
	  left = 0;
	  right = worksheet->numCols () - 1;
	  if (firstCol >= 0)
	    {			// columns are selected
	      left = firstCol;
	      right = firstCol + selectedColsNumber () - 1;
	    }
	}
    }

  QTextStream ts2 (&text, IO_ReadOnly);
  int r = bottom - top + 1;
  int c = right - left + 1;

  QApplication::restoreOverrideCursor ();

  if ((rows > r || cols > c) && (rows <= c && cols <= r)) 
    transposed = true;

  if (transposed) SWAP( rows, cols );
  if (rows > r || cols > c)
    {
      switch (QMessageBox::information (0, "QtiPlot",
					tr ("The text in the clipboard is larger than your current selection!\
	        \nDo you want to insert cells?"), tr ("Yes"), tr ("No"), tr ("Cancel"), 0,
					0))
	{
	case 0: // yes
	  if (cols > c)
	    insertCols (right+1, cols - c);
	  if (rows > r)
	    worksheet->insertRows (bottom+1, rows - r);
	  break;
	case 1: // no
	  if(rows>r) rows = r;
	  if(cols>c) cols = c;
	  break;
	case 2: // cancel
	  return;
	  break;
	}
    }

  if(transposed) SWAP( rows, cols );

  //fprintf(stderr,"rows = %d, cols = %d\n",rows,cols);

  QApplication::setOverrideCursor (waitCursor);
  int prec;
  char f;
  bool numeric;
  double value;

  for (i = 0; i < rows; i++)
    {
      s = ts2.readLine ();
      //fprintf(stderr,"s = %s\n",s.ascii());
      cellTexts = QStringList::split ('\t', s, TRUE);
      for (j = 0; j < cols; j++)
	{
	  int dr,dc;
	  if(!transposed)
	    { dr = top+i; dc = left+j; }
	  else
	    { dr = top+j; dc = left+i; }
	  value = cellTexts[j].stripWhiteSpace().toDouble(&numeric);
	  if (numeric)
	    {
	      columnNumericFormat (dc, f, prec);
	      worksheet->setText (dr, dc, QString::number (value, f, prec));
	    }
	  else
	    worksheet->setText (dr, dc, cellTexts[j]);
	}
    }
  if(!transposed)
    {
      for (j = 0; j < cols; j++)
	emit modifiedData (this, colName (left+j));
    }
  else
    {
      for (j = 0; j < rows; j++)
	emit modifiedData (this, colName (left+j));
    }

  emit modifiedWindow (this);
  QApplication::restoreOverrideCursor ();
}

void
Table::pasteSelection ()
{
  paste(false);
}

void
Table::pasteTransposed ()
{
  paste(true);
}

void
Table::removeCol ()
{
  QStringList list = selectedColumns ();
  removeCol (list);
}

void
Table::removeCol (const QStringList & list)
{
  int n = list.count();
  if(n==0) return;

  QApplication::setOverrideCursor (waitCursor);

  QMemArray <int> columns(n);
  int nc = 0;
  for (int i = 0; i < n; i++)
    {
      //qDebug("list[%d] = %s",i,list[i].ascii());
      int id = colIndex (list[i]);
      if (id >= 0)
	columns[nc++] = id;
    }
  columns.resize(nc);
  columns.sort();
  worksheet->removeColumns(columns);
  for (int i = nc-1; i>=0; i--)
    {
      int id = columns[i];
      QStringList::Iterator it = commandes.at (id);
      if (it != commandes.end ())
	commandes.remove (it);

      it = col_label.at (id);
      if (it != col_label.end ())
	col_label.remove (it);

      it = col_format.at (id);
      if (it != col_format.end ())
	col_format.remove (it);

      it = comments.at (id);
      if (it != comments.end ())
	comments.remove (it);

      QValueListIterator < int >itt = colTypes.at (id);
      if (itt != colTypes.end ())
	colTypes.remove (itt);
      
      itt = col_plot_type.at (id);
      if (itt != col_plot_type.end ())
	col_plot_type.remove (itt);

    }
  setHeaderColType();
  emit removedCols (list);
  emit modifiedWindow (this);
  QApplication::restoreOverrideCursor ();
}

void
Table::correlate ()
{
  QStringList s = selectedColumns ();
  if ((int) s.count () != 2)
    {
      QMessageBox::warning (0, tr ("QtiPlot - Error"),
			    tr
			    ("Please select two columns for this operation!"));
      return;
    }

  int col1 = colIndex (s[0]);
  int col2 = colIndex (s[1]);
  int i, rows = worksheet->numRows ();
  int N = 16;
  while (N < rows)
    N *= 2;			// round N up to the nearest power of 2

  QMemArray <double> tmpdata(N), tmpdata2(N);
  tmpdata.fill(0);
  tmpdata2.fill(0);
  for (i = 0; i < rows; i++)
    {
      tmpdata[i] = worksheet->text (i, col1).toDouble ();
      tmpdata2[i] = worksheet->text (i, col2).toDouble ();
    }
  
  QApplication::setOverrideCursor (waitCursor);
// calculate the FFTs of the two functions
  if (gsl_fft_real_radix2_transform (tmpdata.data(), 1, N) == 0 &&
      gsl_fft_real_radix2_transform (tmpdata2.data(), 1, N) == 0)
    {
      // multiply the FFT by its complex conjugate
      for (i = 0; i < N / 2; i++)
	{
	  if (i == 0 || i == (N / 2) - 1)
	    tmpdata[i] *= tmpdata[i];
	  else
	    {
	      int ni = N - i;
	      double dReal =
		tmpdata[i] * tmpdata2[i] + tmpdata[ni] * tmpdata2[ni];
	      double dImag =
		tmpdata[i] * tmpdata2[ni] - tmpdata[ni] * tmpdata2[i];
	      tmpdata[i] = dReal;
	      tmpdata[ni] = dImag;
	    }
	}
    }
  else
    {
      QApplication::restoreOverrideCursor ();
      QMessageBox::warning (0, "QtiPlot",
			    tr ("Error in GSL forward FFT operation!"));
      return;
    }

  gsl_fft_halfcomplex_radix2_inverse (tmpdata.data(), 1, N);	//inverse FFT

  int cols = worksheet->numCols ();
  int cols2 = cols + 1;
  addCol ();
  addCol ();
  int n = rows / 2;
  for (i = 0; i < rows; i++)
    {
      double y = 0;
      if (i < n)
	y = tmpdata[N - n + i];
      else
	y = tmpdata[i - n];

      worksheet->setText (i, cols, QString::number (i - n));
      worksheet->setText (i, cols2, QString::number (y));
    }

  s = colNames ();
  QStringList l = s.grep ("Lag");
  QString id = QString::number ((int) l.size () + 1);
  QString label = "Corr" + id;

  col_label[cols] = "Lag" + id;
  col_label[cols2] = label;
  col_plot_type[cols] = X;
  setHeaderColType ();

  emit plotCol (this, QStringList (getName() + "_" + label), 0);
  emit modifiedWindow (this);
  QApplication::restoreOverrideCursor ();
}

void
Table::normalizeSelection ()
{
  QStringList s = selectedColumns ();
  for (int i = 0; i < (int) s.count (); i++)
    {
      selectedCol = colIndex (s[i]);
      normalizeCol ();
    }

  emit modifiedWindow (this);
}

void
Table::invertColumns ()
{
  QStringList s = selectedColumns ();
  for (int i = 0; i < (int) s.count (); i++)
    {
      int col = colIndex (s[i]);
      int rows = worksheet->numRows();
      for(int i = 0; i < rows; i++)
	{
	  QString text = this->text (i, col);
	  double aux = text.toDouble ();
	  if (!text.isEmpty ())
	    worksheet->setText (i, col, QString::number( - aux ) );
	}

      QString name = colName (col);
      emit modifiedData (this, name);
    }

  emit modifiedWindow (this);
}

void
Table::normalizeTable ()
{
  for (int i = 0; i < worksheet->numCols (); i++)
    {
      selectedCol = i;
      normalizeCol ();
    }
  emit modifiedWindow (this);
}

void
Table::normalizeCol (int col)
{
  if (col < 0)
    col = selectedCol;
  double max = worksheet->text (0, col).toDouble ();
  double aux = 0.0;
  int i;
  int rows = worksheet->numRows ();
  for (i = 0; i < rows; i++)
    {
      QString text = this->text (i, col);
      aux = text.toDouble ();
      if (!text.isEmpty () && fabs (aux) > fabs (max))
	max = aux;
    }

  if (max == 1.0)
    return;

  for (i = 0; i < rows; i++)
    {
      QString text = this->text (i, col);
      aux = text.toDouble ();
      if (!text.isEmpty ())
	worksheet->setText (i, col, QString::number (aux / max));
    }

  QString name = colName (col);
  emit modifiedData (this, name);
}

void
Table::sortColumnsDialog ()
{
  QStringList s = selectedColumns ();
  sortDialog *sortd =
    new sortDialog (this, "sortDialog", TRUE,
		    WStyle_Tool | WDestructiveClose);
  connect (sortd, SIGNAL (sort (int, int, const QString &)), this,
	   SLOT (sortColumns (int, int, const QString &)));
  sortd->insertColumnsList (s);
  sortd->showNormal ();
  sortd->setActiveWindow ();
}

void
Table::sortTableDialog ()
{
  sortDialog *sortd =
    new sortDialog (this, "sortDialog", TRUE,
		    WStyle_Tool | WDestructiveClose);
  connect (sortd, SIGNAL (sort (int, int, const QString &)), this,
	   SLOT (sortTable (int, int, const QString &)));
  sortd->insertColumnsList (colNames ());
  sortd->showNormal ();
  sortd->setActiveWindow ();
}

void
Table::sortTable (int type, int order, const QString & leadCol)
{
  sortColumns (colNames (), type, order, leadCol);
}

void
Table::sortColumns (int type, int order, const QString & leadCol)
{
  sortColumns (selectedColumns (), type, order, leadCol);
}

void
Table::sortColumns (const QStringList & s, int type, int order,
		    const QString & leadCol)
{
  int cols = s.count ();
  if (!type)			// separately
    {
      for (int i = 0; i < cols; i++)
	{
	  selectedCol = colIndex (s[i]);
	  sortCol(order!=0);
	}
    }
  else				// together
    {
      int leadcol = colIndex (leadCol);
      if (columnType (leadcol) == Table::Text)
	{
	  QMessageBox::critical (this, tr ("QtiPlot - Error"),
				 tr("The leading column has the type set to 'Text'! Operation aborted!"));
	  return;
	}

      int rows = worksheet->numRows ();
      int non_empty_cells = 0;
      QMemArray <int> valid_cell (rows);
      QMemArray <double> data_double (rows);

      for (int j = 0; j < rows; j++)
	{
	  if (!worksheet->text (j, leadcol).isEmpty ())
	    {
	      data_double[non_empty_cells] =
		this->text (j, leadcol).toDouble ();
	      valid_cell[non_empty_cells] = j;
	      non_empty_cells++;
	    }
	}
      if (!non_empty_cells)
	{
	  QMessageBox::critical (this, tr ("QtiPlot - Error"),
				 tr
				 ("The leading column is empty! Operation aborted!"));
	  return;
	}

      // valid_cell.resize (non_empty_cells);
      //QMemArray <QString> data_string(non_empty_cells);
      QMemArray <size_t> p(non_empty_cells);

      // Find the permutation index for the lead col
      gsl_sort_index (p.data(), data_double.data(), 1, non_empty_cells);

      for (int i = 0; i < cols; i++)
	{ // Since we have the permutation index, sort all the columns
	  int scol = colIndex (s[i]);
	  if (columnType (scol) == Table::Text)
	    {
	      QStringList data_string;
	      for (int j = 0; j < non_empty_cells; j++){
		//data_string[j] = text(valid_cell[j], scol);
		data_string.append(text(valid_cell[j], scol));
	      }
	      if (!order)
		for (int j = 0; j < non_empty_cells; j++)
		  worksheet->setText(valid_cell[j], scol, data_string[p[j]]);
	      else
		for (int j = 0; j < non_empty_cells; j++)
		  worksheet->setText(valid_cell[j], scol,
				     data_string[p[non_empty_cells-j-1]]);
	    }
	  else
	    {
	      for (int j = 0; j < non_empty_cells; j++)
		data_double[j] = text (valid_cell[j], scol).toDouble ();
	      int prec;
	      char f;
	      columnNumericFormat (scol, f, prec);
	      if (!order)
		for (int j = 0; j < non_empty_cells; j++)
		  worksheet->setText (valid_cell[j], scol,
				      QString::number (data_double[p[j]], f,
						       prec));
	      else
		for (int j = 0; j < non_empty_cells; j++)
		  worksheet->setText (valid_cell[j], scol,
				      QString::number (data_double
					      [p[non_empty_cells-j-1]], f,
					      prec));
	    }
	  emit modifiedData (this, colName (scol));
	}
    }
  emit modifiedWindow (this);
}

void
Table::sortCol(bool descending)
{
  if (columnType (selectedCol) == Table::Text)
    return;

  int rows = worksheet->numRows ();
  QMemArray < int >aux (rows);
  QMemArray < double >r (rows);
  int i, n = 0;
  for (i = 0; i < rows; i++)
    {
      QString text = this->text (i, selectedCol);
      if (!text.isEmpty ())
	{
	  n++;
	  aux[i] = i;
	  r[i] = text.toDouble ();
	}
      else
	{
	  aux[i] = -1;
	  r[i] = 0.0;
	}
    }

  if (!n)
    return;

  gsl_vector *v = gsl_vector_alloc (n);
  int index;
  n = 0;
  for (i = 0; i < rows; i++)
    {
      index = aux[i];
      if (index >= 0)
	{
	  gsl_vector_set (v, n, r[index]);
	  n++;
	}
    }

  gsl_sort_vector (v);

  int prec;
  char f;
  columnNumericFormat (selectedCol, f, prec);

  double value;
  double m = 0;
  for (i = 0; i < rows; i++)
    {
      index = aux[i];
      if (index >= 0)
	{
	  m++;
	  if(descending)
	    value = gsl_vector_get(v,n-m);
	  else
	    value = gsl_vector_get(v,m-1);
	  worksheet->setText (i, selectedCol,
			      QString::number (value, f, prec));
	}
    }
  gsl_vector_free (v);
  QString name = colName (selectedCol);
  emit modifiedData (this, name);
  emit modifiedWindow (this);
}

void
Table::sortColAsc()
{
  sortCol( false );
}

void
Table::sortColDesc ()
{
  sortCol( true );
}

int
Table::tableRows ()
{
  return worksheet->numRows ();
}

int
Table::tableCols ()
{
  return worksheet->numCols ();
}

bool
Table::isEmptyRow (int row)
{
  int cols = worksheet->numCols ();

  for (int i = 0; i < cols; i++)
    if(!worksheet->text(row, i).isEmpty())
      return false;

  return true;
}

bool
Table::isEmptyColumn (int col)
{
  int rows = worksheet->numRows ();

  for (int i = 0; i < rows; i++)
    if(!worksheet->text(i, col).isEmpty())
      return false;

  return true;
}

QString
Table::saveText ()
{
  QString text = "<data>\n";
  int cols = worksheet->numCols ();
  int rows = worksheet->numRows ();

  for (int i = 0; i < rows; i++)
    {
      if (!isEmptyRow (i))
	{
	  text += QString::number (i) + "\t";
	  for (int j = 0; j < cols - 1; j++)
	    text += worksheet->text (i, j) + "\t";

	  text += worksheet->text (i, cols - 1) + "\n";
	}
    }
  return text + "</data>\n";
}

int
Table::nonEmptyRows ()
{
  int r = 0;
  for (int i = 0; i < worksheet->numRows (); i++)
    {
      if (!isEmptyRow (i))
	r++;
    }
  return r;
}

QString
Table::text (int row, int col)
{
  if (col == savedCol)
    return savedCells[row].replace (",", ".");
  else
    return worksheet->text (row, col).replace (",", ".");
}

void
Table::setText (int row, int col, const QString & text)
{
  worksheet->setText (row, col, text);
}

void
Table::saveColToMemory (int col)
{
  int rows = worksheet->numRows ();
  savedCells.clear ();
  for (int row = 0; row < rows; row++)
    savedCells << worksheet->text (row, col);
  savedCol = col;
}

void
Table::forgetSavedCol ()
{
  savedCells.clear ();
  savedCol = -1;
}

void
Table::setTextFormat (bool applyToAll)
{
  if (applyToAll)
    {
      for (int i = selectedCol; i < worksheet->numCols (); i++)
	colTypes[i] = Text;
    }
  else
    colTypes[selectedCol] = Text;
}

void
Table::setNumericFormat (int f, int prec, bool applyToAll)
{
  int cols = worksheet->numCols ();
  if (applyToAll)
    {
      for (int i = selectedCol; i < cols; i++)
	setColNumericFormat (f, prec, i);
    }
  else
    setColNumericFormat (f, prec, selectedCol);

  emit modifiedWindow (this);
}

void
Table::setColNumericFormat (int f, int prec, int col)
{
  int old_f, old_prec;
  columnNumericFormat (col, old_f, old_prec);
  if (colTypes[col] == Numeric && old_f == f && old_prec == prec)
    return;

  colTypes[col] = Numeric;
  col_format[col] = QString::number (f) + "/" + QString::number (prec);

  int rows = worksheet->numRows ();
  for (int i = 0; i < rows; i++)
    {
      QString t = text (i, col);
      if (!t.isEmpty ())
	{
	  double val = t.toDouble ();
	  if (!f)
	    t.setNum (val, 'g', prec);
	  else if (f == 1)
	    t.setNum (val, 'f', prec);
	  else if (f == 2)
	    t.setNum (val, 'e', prec);

	  worksheet->setText (i, col, t);
	}
    }
}

void
Table::setColumnsFormat (const QStringList & lst)
{
  if (col_format == lst)
    return;

  col_format = lst;
  emit modifiedWindow (this);
}

void
Table::setDateTimeFormat (int f, const QString & format, bool applyToAll)
{
  QApplication::setOverrideCursor (waitCursor);

  int cols = worksheet->numCols ();
  if (applyToAll)
    {
      for (int j = selectedCol; j < cols; j++)
	setDateTimeFormat (f, format, j);
    }
  else
    setDateTimeFormat (f, format, selectedCol);

  emit modifiedWindow (this);
  QApplication::restoreOverrideCursor ();
}

void
Table::setDateTimeFormat (int f, const QString & format, int col)
{
  switch (f)
    {
    case 2:
      setDateFormat (format, col);
      break;

    case 3:
      setTimeFormat (format, col);
      break;

    case 4:
      setMonthFormat (format, col);
      break;

    case 5:
      setDayFormat (format, col);
      break;

    default:
      break;
    }
}

void
Table::setDateFormat (const QString & format, int col)
{
  if (col_format[col] == format)
    return;

  colTypes[col] = Date;
  col_format[col] = format;
  int rows = worksheet->numRows ();

  for (int i = 0; i < rows; i++)
    {
      QString s = worksheet->text (i, col);
      if (!s.isEmpty ())
	{
	  QDate d = QDate::fromString (s, Qt::ISODate);
	  if (d.isValid ())
	    worksheet->setText (i, col, d.toString (format));
	  else
	    worksheet->setText (i, col, "-");
	}
    }
}

void
Table::setTimeFormat (const QString & format, int col)
{
  if (col_format[col] == format)
    return;

  int rows = worksheet->numRows ();

  colTypes[col] = Time;
  col_format[col] = format;

  for (int i = 0; i < rows; i++)
    {
      QString s = worksheet->text (i, col);
      if (!s.isEmpty ())
	{
	  QTime t = QTime::fromString (s, Qt::TextDate);
	  if (t.isValid ())
	    worksheet->setText (i, col, t.toString (format));
	  else
	    worksheet->setText (i, col, "-");
	}
    }
}

void
Table::setMonthFormat (const QString & format, int col)
{
  colTypes[col] = Month;
  int rows = worksheet->numRows ();
  if (format == "shortMonthName")
    {
      for (int i = 0; i < rows; i++)
	{
	  QString s = worksheet->text (i, col);
	  if (!s.isEmpty ())
	    {
	      int month = s.toInt () % 12;
	      if (!month)
		month = 12;

	      worksheet->setText (i, col, QDate::shortMonthName (month));
	    }
	}
    }
  else if (format == "longMonthName")
    {
      for (int i = 0; i < rows; i++)
	{
	  QString t = worksheet->text (i, col);
	  if (!t.isEmpty ())
	    {
	      int month = t.toInt () % 12;
	      if (!month)
		month = 12;

	      worksheet->setText (i, col, QDate::longMonthName (month));
	    }
	}
    }
}

void
Table::setDayFormat (const QString & format, int col)
{
  colTypes[col] = Day;
  int rows = worksheet->numRows ();
  if (format == "shortDayName")
    {
      for (int i = 0; i < rows; i++)
	{
	  QString t = worksheet->text (i, col);
	  if (!t.isEmpty ())
	    {
	      int day = t.toInt () % 7;
	      if (!day)
		day = 7;

	      worksheet->setText (i, col, QDate::shortDayName (day));
	    }
	}
    }
  else if (format == "longDayName")
    {
      for (int i = 0; i < rows; i++)
	{
	  QString t = worksheet->text (i, col);
	  if (!t.isEmpty ())
	    {
	      int day = t.toInt () % 7;
	      if (!day)
		day = 7;

	      worksheet->setText (i, col, QDate::longDayName (day));
	    }
	}
    }
}

void
Table::setRandomValues ()
{
  QApplication::setOverrideCursor (waitCursor);

  int i;
  double d;
  int rows = worksheet->numRows ();
  QStringList list = selectedColumns ();

  for (int j = 0; j < (int) list.count (); j++)
    {
      QString name = list[j];
      selectedCol = colIndex (name);

      int prec;
      char f;
      columnNumericFormat (selectedCol, f, prec);

      for (i = 0; i < rows; i++)
	{
	  d = rand_uniform();
	  worksheet->setText (i, selectedCol, QString::number (d, f, prec));
	}
      emit modifiedData (this, name);
    }

  emit modifiedWindow (this);
  QApplication::restoreOverrideCursor ();
}

void
Table::loadHeader (const QStringList & header)
{
  int cols = header.size();
  if (cols!=worksheet->numCols())
    {
      fprintf(stderr,"column size not match in Table::loadHeader\n");
      cols = worksheet->numCols();
    }
  QStringList col_label_new;
  QValueList <int> col_plot_type_new;
  for (int i = 0; i < cols; i++)
    {
      //QString s = header[i].replace ("_", "-");
      QString s = header[i];
      s.replace ("_", "-");
      if (s.contains ("[X]"))
	{
	  col_label_new << s.remove ("[X]");
	  col_plot_type_new << X;
	}
      else if (s.contains ("[Y]"))
	{
	  col_label_new << s.remove ("[Y]");
	  col_plot_type_new << Y;
	}
      else if (s.contains ("[Z]"))
	{
	  col_label_new << s.remove ("[Z]");
	  col_plot_type_new << Z;
	}
      else if (s.contains ("[xEr]"))
	{
	  col_label_new << s.remove ("[xEr]");
	  col_plot_type_new << xErr;
	}
      else if (s.contains ("[yEr]"))
	{
	  col_label_new << s.remove ("[yEr]");
	  col_plot_type_new << yErr;
	}
      else
	{
	  col_label_new << s;
	  col_plot_type_new << None;
	}
      //fprintf(stderr,"setting column header[%d]\n",i);
      //fprintf(stderr,"label = %s\n",col_label_new[i].ascii());
      //fprintf(stderr,"type  = %d\n",col_plot_type_new[i]);
    }
  col_label = col_label_new;
  col_plot_type = col_plot_type_new;
  setHeaderColType ();
}

void
Table::setHeader (const QStringList & header)
{
  col_label = header;
  setHeaderColType ();
}

int
Table::colIndex (const QString & name)
{
  //name is a column name and not necessarily include "_"
  int pos = name.find ("_", FALSE);
  QString label = name.right (name.length () - pos - 1);
  return col_label.findIndex (label);
}

void
Table::setHeaderColType ()
{
  QHeader *head = worksheet->horizontalHeader ();
  int xcols = 0;
  for (int j = 0; j < (int) worksheet->numCols (); j++)
    {
      if (col_plot_type[j] == X)
	xcols++;
    }
  QString numstr = "";
  if (xcols>1) numstr = "1";
  xcols = 0;
  for (int i = 0; i < (int) worksheet->numCols (); i++)
    {
      if (col_plot_type[i] == X)
	{
	  xcols++;
	  if (xcols>1) numstr = QString::number(xcols);
	  head->setLabel (i,col_label[i] + "[X" + numstr + "]");
	}
      else if (col_plot_type[i] == Y)
	head->setLabel (i, col_label[i] + "[Y" + numstr + "]");
      else if (col_plot_type[i] == Z)
	head->setLabel (i, col_label[i] + "[Z" + numstr + "]");
      else if (col_plot_type[i] == xErr)
	head->setLabel (i, col_label[i] + "[xEr]");
      else if (col_plot_type[i] == yErr)
	head->setLabel (i, col_label[i] + "[yEr]");
      else
	head->setLabel (i, col_label[i]);
    }
}

void
Table::setAscValues ()
{
  QApplication::setOverrideCursor (waitCursor);

  int rows = worksheet->numRows ();
  QStringList list = selectedColumns ();

  for (int j = 0; j < (int) list.count (); j++)
    {
      QString name = list[j];
      selectedCol = colIndex (name);

      int prec;
      char f;
      columnNumericFormat (selectedCol, f, prec);

      for (int i = 0; i < rows; i++)
	worksheet->setText (i, selectedCol, QString::number (i + 1, f, prec));

      emit modifiedData (this, name);
    }

  emit modifiedWindow (this);
  QApplication::restoreOverrideCursor ();
}

void
Table::plotL ()
{
  if (!valid2DPlot ())
    return;

  emit plotCol (this, drawableColumnSelection (), Graph::Line);
}

void
Table::plotP ()
{
  if (!valid2DPlot ())
    return;

  emit plotCol (this, drawableColumnSelection (), Graph::Scatter);
}

void
Table::plotLP ()
{
  if (!valid2DPlot ())
    return;

  emit plotCol (this, drawableColumnSelection (), Graph::LineSymbols);
}

void
Table::plotVB ()
{
  if (!valid2DPlot ())
    return;

  emit plotCol (this, drawableColumnSelection (), Graph::VerticalBars);
}

void
Table::plotHB ()
{
  if (!valid2DPlot ())
    return;

  emit plotCol (this, drawableColumnSelection (), Graph::HorizontalBars);
}

void
Table::plotArea ()
{
  if (!valid2DPlot ())
    return;

  emit plotCol (this, drawableColumnSelection (), Graph::Area);
}

bool
Table::noXColumn ()
{
  bool notSet = true;
  for (int i = 0; i < worksheet->numCols (); i++)
    {
      if (col_plot_type[i] == X)
	return false;
    }
  return notSet;
}

bool
Table::noYColumn ()
{
  bool notSet = true;
  for (int i = 0; i < worksheet->numCols (); i++)
    {
      if (col_plot_type[i] == Y)
	return false;
    }
  return notSet;
}

void
Table::plotPie ()
{
  if (noXColumn ())
    {
      QMessageBox::critical (0, tr ("QtiPlot - Error"),
			     tr
			     ("Please set a default X column for this table, first!"));
      return;
    }

  QStringList s = selectedColumns ();
  if (int (s.count ()) > 0)
    emit plotCol (this, s, Graph::Pie);
  else
  QMessageBox::warning (0, tr ("QtiPlot - Error"),
			tr ("Please select a column to plot!"));
}

void
Table::plotVerticalDropLines ()
{
  if (!valid2DPlot ())
    return;

  emit plotCol (this, drawableColumnSelection (), Graph::VerticalDropLines);
}

void
Table::plotSpline ()
{
  if (!valid2DPlot ())
    return;

  emit plotCol (this, drawableColumnSelection (), Graph::Spline);
}

void
Table::plotVertSteps ()
{
  if (!valid2DPlot ())
    return;

  emit plotCol (this, drawableColumnSelection (), Graph::VerticalSteps);
}

void
Table::plotHorSteps ()
{
  if (!valid2DPlot ())
    return;

  emit plotCol (this, drawableColumnSelection (), Graph::HorizontalSteps);
}

void
Table::plotHistogram ()
{
  if (!valid2DPlot ())
    return;

  emit plotCol (this, drawableColumnSelection (), Graph::Histogram);
}

void
Table::plotBoxDiagram ()
{
  if (!valid2DPlot ())
    return;

  emit plotCol (this, drawableColumnSelection (), Graph::Box);
}

void
Table::plotVectXYXY ()
{
  if (!valid2DPlot ())
    return;

  QStringList s = selectedColumns ();
  if ((int) s.count () == 4)
    emit plotCol (this, s, Graph::VectXYXY);
  else
  QMessageBox::warning (0, tr ("QtiPlot - Error"),
			tr
			("Please select four columns for this operation!"));
}

void
Table::plotVectXYAM ()
{
  if (!valid2DPlot ())
    return;

  QStringList s = selectedColumns ();
  if ((int) s.count () == 4)
    emit plotCol (this, s, Graph::VectXYAM);
  else
  QMessageBox::warning (0, tr ("QtiPlot - Error"),
			tr
			("Please select four columns for this operation!"));
}

bool
Table::valid2DPlot ()
{
  if (!selectedYColumns ().count ())
    {
      QMessageBox::warning (0, tr ("QtiPlot - Error"),
			    tr ("Please select a Y column to plot!"));
      return false;
    }
  else if (worksheet->numCols () < 2)
    {
      QMessageBox::critical (0, tr ("QtiPlot - Error"),
			     tr
			     ("You need at least two columns for this operation!"));
      return false;
    }
  else if (noXColumn ())
    {
      QMessageBox::critical (0, tr ("QtiPlot - Error"),
			     tr
			     ("Please set a default X column for this table, first!"));
      return false;
    }
  return true;
}

void
Table::plot3DRibbon ()
{
  if (!valid2DPlot ())
    return;

  emit plot3DRibbon (this, colName (selectedCol));
}

void
Table::plot3DBars ()
{
  if (!valid3DPlot ())
    return;

  emit plotXYZ (this, colName (selectedCol), 2);
}

void
Table::plot3DScatter ()
{
  if (!valid3DPlot ())
    return;

  emit plotXYZ (this, colName (selectedCol), 0);
}

void
Table::plot3DTrajectory ()
{
  if (!valid3DPlot ())
    return;

  emit plotXYZ (this, colName (selectedCol), 1);
}

bool
Table::valid3DPlot ()
{
  if (worksheet->numCols () < 2)
    {
      QMessageBox::critical (0, tr ("QtiPlot - Error"),
			     tr
			     ("You need at least two columns for this operation!"));
      return false;
    }
  if (selectedCol < 0 || col_plot_type[selectedCol] != Z)
    {
      QMessageBox::critical (0, tr ("QtiPlot - Error"),
			     tr
			     ("Please select a Z column for this operation!"));
      return false;
    }
  if (noXColumn ())
    {
      QMessageBox::critical (0, tr ("QtiPlot - Error"),
			     tr ("You need to define a X column first!"));
      return false;
    }
  if (noYColumn ())
    {
      QMessageBox::critical (0, tr ("QtiPlot - Error"),
			     tr ("You need to define a Y column first!"));
      return false;
    }
  return true;
}

void
Table::importASCII (const QString & fname, const QString & sep,
		    int ignoredLines, bool renameCols, bool stripSpaces,
		    bool simplifySpaces, int importFileAs)
// importFileAs
// 0 : new Table
// 1 : new Columns
// 2 : new Rows
// 3 : overwrite Table
{
  QFile f (fname);
  QTextStream t (&f);		// use a text stream
  if (f.open (IO_ReadOnly))
    {
      QApplication::setOverrideCursor (waitCursor);

      int i, rows = 1, cols = 0;
      for (i = 0; i < ignoredLines; i++)
	t.readLine ();
      
      QString s = t.readLine ();	//read first line after the ignored ones
      //ignore the lines that start with '#'
      while(s.at(0)=='#')
	s=t.readLine();

      //count row number
      QString s2;
      while (!t.atEnd ())
	{
	  s2=t.readLine ();
	  if(s2.at(0)!='#') rows++;
	}
      
      if (simplifySpaces)
	s = s.simplifyWhiteSpace ();
      else if (stripSpaces)
	s = s.stripWhiteSpace ();

      QString separator;
      if (sep=="Auto")
	{
	  if (s.contains('\t'))
	    separator = "\t";
	  else if (s.contains(','))
	    separator = ",";
	  else if (s.contains(';'))
	    separator = ";";
	  else
	    separator = " ";
	}
      else
	separator = sep;

      // count column number
      QStringList line = QStringList::split (separator, s, true);
      cols = (int) line.count ();

      //verify if the strings in the line used to rename
      //the columns are not all numbers
      bool allNumbers = true;
      line.gres (",", ".", false);	//Qt uses decimal dot
      for (i = 0; i < cols; i++)
	{
	  (line[i].stripWhiteSpace()).toDouble (&allNumbers);
	  if (!allNumbers)
	    break;
	}

      // if the first row is all numbers, the renameCols flag is canceled.
      // this is useful because the "use first row as a column name" option
      // can be kept on
      if (renameCols && !allNumbers)
	rows--;

      QApplication::restoreOverrideCursor ();

      int c,r;
      if(importFileAs)
	{
	  c = worksheet->numCols (); // original column number
	  r = worksheet->numRows (); // original row number
	}
      QStringList oldHeader;
      switch (importFileAs)
	{
	case 0 : //new Table 
	  init( rows, cols ); 
	  break;
	case 1 : //new columns
	  addColumns (cols);
	  if (r < rows)
	    worksheet->setNumRows (rows);
	  break;
	case 2 : //new rows
	  if (c < cols)
	    addColumns (cols - c);
	  worksheet->setNumRows (r + rows);
	  break;
	case 3 : //overwrite
	  if (r != rows)
	    worksheet->setNumRows (rows);
	  oldHeader = col_label;
	  if (c != cols)
	    {
	      if (c < cols)
		addColumns (cols - c);
	      else
		{
		  //worksheet->setNumCols (cols);
		  QStringList removeList;
		  for (i = cols; i < c; i++)
		    removeList << QString(name())+"_"+oldHeader[i];
		  removeCol( removeList );
		}
	    }
	  break;
	}

      f.reset ();
      for (i = 0; i < ignoredLines; i++)
	t.readLine ();

      int startRow = 0, startCol = 0;
      if (importFileAs == 1)
        startCol = c;
      else if (importFileAs == 2)
        startRow = r;

      if (renameCols && !allNumbers)
	{ //use first row as column names
	  s = t.readLine ();
	  //ignore the line starts with '#'
	  while(s.at(0)=='#')
	    s=t.readLine();
	  if (simplifySpaces)
	    s = s.simplifyWhiteSpace ();
	  else if (stripSpaces)
	    s = s.stripWhiteSpace ();
	  line = QStringList::split (separator, s, false);
          int end = startCol + (int) line.count ();
 	  for (i = startCol; i < end; i++)
	    col_label[i] = QString::null;

	  for (i = startCol; i < end; i++)
	    {
	      s = line[i-startCol].stripWhiteSpace();
	      if (s.contains("[X]"))
		col_plot_type[i] = X;
	      else if (s.contains("[Y]"))
		col_plot_type[i] = Y;
	      else if (s.contains("[Z]"))
		col_plot_type[i] = Z;
	      else if (s.contains("[xEr]"))
		col_plot_type[i] = xErr;
	      else if (s.contains("[yEr]"))
		col_plot_type[i] = yErr;
	      s = s.remove(QRegExp("\\[.*\\]"));
	      comments[i] = s;
	      s = s.replace("-","_").remove(QRegExp("\\W")).replace("_","-");
	      int n = (int) col_label.contains (s);
	      if (n)
		{		//avoid identical col names
		  while ((int) col_label.contains (s + QString::number (n)))
		    n++;
		  s += QString::number (n);
		}
	      col_label[i] = s;
	    }
	}
      setHeaderColType ();

      for (i = startRow; i < worksheet->numRows(); i++)
	{
	  s = t.readLine ();
	  //ignore the line starts with '#'
	  while(s.at(0)=='#')
	    s=t.readLine();
	  if (simplifySpaces)
	    s = s.simplifyWhiteSpace ();
	  else if (stripSpaces)
	    s = s.stripWhiteSpace ();
	  line = QStringList::split (separator, s, true);
	  int lc = (int) line.count ();
	  if (lc > cols)
	    {
	      addColumns (lc - cols);
	      cols = lc;
	    }
	  for (int j = 0; j < lc; j++)
	    worksheet->setText (i, startCol+j, line[j].stripWhiteSpace());
	}
      f.close ();

      if (importFileAs>=2) // overwrite or new rows
	{
	  for (i = 0; i < QMIN(c,cols); i++)
	    {
	      emit modifiedData (this, colName (i));
	      if (colLabel (i) != oldHeader[i])
		emit changedColHeader (getName() + "_" + oldHeader[i],
				       getName() + "_" + colLabel (i));
	    }
	}
    }
  emit modifiedWindow(this);
}

bool
Table::exportToASCIIFile (const QString & fname, const QString & separator,
			  bool withLabels, bool exportSelection)
{
  QFile f (fname);
  if (!f.open (IO_WriteOnly))
    {
      QApplication::restoreOverrideCursor ();
      QMessageBox::critical (0, tr ("QtiPlot - ASCII Export Error"),
			     tr ("Could not write to file: <br><h4>" + fname +
				 "</h4><p>Please verify that you have the right to write to this location!").
			     arg (fname));
      return false;
    }

  QString text;
  int rows = worksheet->numRows ();
  int cols = worksheet->numCols ();
  int selectedCols = 0;
  int topRow = 0, bottomRow = 0;

  QMemArray <int> sCols;
  if (exportSelection)
    {
      for (int i = 0; i < cols; i++)
	if (worksheet->isColumnSelected (i))
	  {
	    selectedCols++;
	    sCols.resize(selectedCols);
	    sCols[selectedCols-1] = i;
	  }

      for (int i = 0; i < rows; i++)
	if (worksheet->isRowSelected (i))
	  {
	    topRow = i;
	    break;
	  }

      for (int i = rows - 1; i >= 0; i--)
	if (worksheet->isRowSelected (i))
	  {
	    bottomRow = i;
	    break;
	  }
    }

  if (withLabels)
    {
      QStringList header = colNames ();
      if (exportSelection)
	{
	  for (int i = 0; i < selectedCols; i++)
	    {
	      int scol = sCols[i];
	      text += header[scol];
	      if (col_plot_type[scol] == X)
		text += "[X]";
	      else if (col_plot_type[scol] == Y)
		text += "[Y]";
	      else if (col_plot_type[scol] == Z)
		text += "[Z]";
	      else if (col_plot_type[scol] == xErr)
		text += "[xEr]";
	      else if (col_plot_type[scol] == yErr)
		text += "[yEr]";
	      if(i<selectedCols-1)
		text += separator;
	    }
	  text += "\n";
	}
      else
	{
	  for (int i = 0; i < cols; i++)
	    {
	      text += header[i];
	      if (col_plot_type[i] == X)
		text += "[X]";
	      else if (col_plot_type[i] == Y)
		text += "[Y]";
	      else if (col_plot_type[i] == Z)
		text += "[Z]";
	      else if (col_plot_type[i] == xErr)
		text += "[xEr]";
	      else if (col_plot_type[i] == yErr)
		text += "[yEr]";
	      if(i<cols-1)
		text += separator;
	    }
	  text += "\n";
	}
    }				// finished writting labels

  if (exportSelection)
    {
      for (int i = topRow; i <= bottomRow; i++)
	{
	  for (int j = 0; j < selectedCols-1; j++)
	    text += worksheet->text (i, sCols[j]) + separator;
	  text += worksheet->text (i, sCols[selectedCols-1]) + "\n";
	}
    }
  else
    {
      for (int i = 0; i < rows; i++)
	{
	  for (int j = 0; j < cols - 1; j++)
	    text += worksheet->text (i, j) + separator;
	  text += worksheet->text (i, cols - 1) + "\n";
	}
    }
  QTextStream t (&f);
  t << text;
  f.close ();
  return true;
}

void
Table::contextMenuEvent (QContextMenuEvent * e)
{
  QRect r =
    worksheet->horizontalHeader ()->sectionRect (worksheet->numCols () - 1);
  if (e->pos ().x () > r.right () + worksheet->verticalHeader ()->width ())
    emit showContextMenu (false);
  else
    emit showContextMenu (true);
  e->accept ();
}

void
Table::moveCurrentCell ()
{
  int cols = worksheet->numCols ();
  int row = worksheet->currentRow ();
  int col = worksheet->currentColumn ();
  worksheet->clearSelection (TRUE);

  if (col + 1 < cols)
    {
      worksheet->setCurrentCell (row, col + 1);
      worksheet->selectCells (row, col + 1, row, col + 1);
    }
  else
    {
      worksheet->setCurrentCell (row + 1, 0);
      worksheet->selectCells (row + 1, 0, row + 1, 0);
    }
}

bool
Table::eventFilter (QObject * object, QEvent * e)
{
  QHeader *hheader = worksheet->horizontalHeader ();
  QHeader *vheader = worksheet->verticalHeader ();

  if (e->type () == QEvent::MouseButtonPress && object == (QObject *) hheader)
    {
      const QMouseEvent *me = (const QMouseEvent *) e;
      int offset = hheader->offset ();
      selectedCol = hheader->sectionAt (me->pos ().x () + offset);

      if (me->button () == QMouseEvent::LeftButton &&
	  me->state () == Qt::ControlButton)
	{
	  worksheet->selectColumn (selectedCol);
	  worksheet->setCurrentCell (0, selectedCol);
	  return true;
	}
      else if (me->button () == QMouseEvent::RightButton &&
	       selectedColsNumber () <= 1)
	{
	  worksheet->clearSelection();
	  worksheet->selectColumn(selectedCol);
	  worksheet->setCurrentCell(0,selectedCol);
	}
    }
  else if (e->type () == QEvent::MouseButtonPress
	   && object == (QObject *) vheader)
    {
      const QMouseEvent *me = (const QMouseEvent *) e;
      int offset = vheader->offset ();
      if (me->button () == QMouseEvent::RightButton && selectedRows () <= 1)
	{
	  worksheet->clearSelection ();
	  int row = vheader->sectionAt (me->pos ().y () + offset);
	  worksheet->selectRow (row);
	  worksheet->setCurrentCell (row, 0);
	}
    }
  else if (e->type () == QEvent::MouseButtonDblClick && object == hheader)
    {
      const QMouseEvent *me = (const QMouseEvent *) e;
      selectedCol = hheader->sectionAt (me->pos ().x () + hheader->offset ());

      QRect rect = hheader->sectionRect (selectedCol);
      rect.setLeft (rect.right () - 2);
      rect.setWidth (4);

      if (rect.contains (me->pos ()))
	{
	  worksheet->adjustColumn (selectedCol);
	  emit modifiedWindow (this);
	}
      else
	emit optionsDialog ();
      return true;
    }
  else if (e->type () == QEvent::ContextMenu && object == titleBar)
    {
      emit showTitleBarMenu ();
      ((QContextMenuEvent *) e)->accept ();
      return true;
    }
  return QObject::eventFilter (object, e);
}

void
Table::customEvent (QCustomEvent * e)
{
  if (e->type () == SCRIPTING_CHANGE_EVENT)
    scriptingChangeEvent ((ScriptingChangeEvent *) e);
}

void
Table::setSpecifications (const QString & s)
{
  if (specifications == s)
    return;

  specifications = s;
}

QString & Table::getSpecifications ()
{
  return specifications;
}

void
Table::restore (QString & spec)
{
  int i, j, row;
  int c = worksheet->numCols ();
  int r = worksheet->numRows ();

  if(c==0) addColumns(1);

  QTextStream t (spec, IO_ReadOnly);
  t.readLine ();		//table tag
  QString s = t.readLine ();
  QStringList list = QStringList::split ('\t', s, TRUE);

  QString oldCaption = getName();
  QString newCaption = list[0];
  if (oldCaption != newCaption)
    this->setName (newCaption);

  int rows = list[1].toInt ();
  if (rows != r)
    worksheet->setNumRows (rows);

  int cols = list[2].toInt ();
  if (cols != c)
    worksheet->setNumCols (cols);

//clear all cells
  for (i = 0; i < rows; i++)
    for (j = 0; j < cols; j++)
      worksheet->setText (i, j, "");

  t.readLine ();		//table geometry useless info when restoring
  s = t.readLine ();		//header line

  list = QStringList::split ('\t', s, TRUE);
  list.remove (list.first ());

  if (!col_label.isEmpty () && col_label != list)
    {
      QStringList col_label_old = col_label;
      loadHeader (list);
      /*
      list.gres ("[X]", "");
      list.gres ("[Y]", "");
      list.gres ("[Z]", "");
      list.gres ("[xEr]", "");
      list.gres ("[yEr]", "");
      */

      if(cols==c)
	{
	  for (j = 0; j < cols; j++)
	    {
	      if (col_label[j] != col_label_old[j])
		emit changedColHeader (newCaption + "_" + col_label_old[j],
				       newCaption + "_" + col_label[j]);
	    }
	}
      else if (cols < c)
	{
	  QStringList rmcols;
	  for (j = 0; j < c; j++)
	    {
	      if (!col_label.contains (col_label_old[j]))
		rmcols << oldCaption + "_" + col_label_old[j];
	    }
	  if (!rmcols.isEmpty())
	    emit removedCols( rmcols );
	}
    }

  s = t.readLine ();		//colWidth line
  list = QStringList::split ('\t', s, TRUE);
  list.remove (list.first ());
  if (columnWidths () != list)
    setColWidths (list);

  s = t.readLine ();
  list = QStringList::split ('\t', s, TRUE);
  if (list[0] == "com")		//commandes line
    {
      list.remove (list.first ());
      if (list != commandes)
	commandes = list;
    }
  else // if (list[0] == "<com>")
    {				// commandes block
      commandes.clear ();
      for (int i = 0; i < tableCols (); i++)
	commandes << "";
      for (s = t.readLine (); s != "</com>"; s = t.readLine ())
	{
	  int col = s.mid (9, s.length () - 11).toInt ();
	  QString formula;
	  for (s = t.readLine (); s != "</col>"; s = t.readLine ())
	    {
	      formula += s + "\n";
	    }
	  formula.truncate (formula.length () - 1);
	  setCommand (col, formula);
	}
    }

  s = t.readLine ();		//colType line ?

  list = QStringList::split ('\t', s, TRUE);
  if (s.contains ("ColType", TRUE))
    {
      QStringList col_form;
      QValueList <int> col_t;
      list.remove (list.first ());
      for (i = 0; i < int (list.count ()); i++)
	{
	  QStringList l = QStringList::split (';', list[i], true);
	  col_t << l[0].toInt ();

	  if ((int) l.count () == 0)
	    fprintf(stderr,"failed to set col_format[%d]\n",i);

	  col_form << l[1];
	}
      col_format = col_form;
      colTypes = col_t;
    }
  else //if fileVersion < 65 set table values
    {
      row = list[0].toInt ();
      for (j = 0; j < cols; j++)
	worksheet->setText (row, j, list[j + 1]);
    }

  s = t.readLine ();		//comments line ?
  list = QStringList::split ('\t', s, TRUE);
  if (s.contains ("Comments", TRUE))
    {
      list.remove (list.first ());
      comments = list;
    }

  s = t.readLine ();
  list = QStringList::split ('\t', s, TRUE);
  if (s.contains ("WindowLabel", TRUE))
    {
      setWindowLabel (list[1]);
      setCaptionPolicy ((myWidget::CaptionPolicy) list[2].toInt ());
    }

  s = t.readLine ();
  if (s == "<data>")
    s = t.readLine ();

  while (!t.atEnd () && s != "</data>")
    {
      list = QStringList::split ('\t', s, TRUE);

      row = list[0].toInt ();
      for (j = 0; j < cols; j++)
	worksheet->setText (row, j, list[j + 1]);

      s = t.readLine ();
    }

  emit modifiedTable(this);

  specifications = spec;
}

void
Table::setNumRows (int rows)
{
  worksheet->setNumRows (rows);
}

void
Table::setNumCols (int cols)
{
  worksheet->setNumCols (cols);
}

void
Table::resizeRows (int r)
{
  int rows = worksheet->numRows ();
  if (rows == r)
    return;

  if (rows > r)
    {
      QString text = tr ("Rows will be deleted from the table!");
      text += "<p>" + tr ("Do you really want to continue?");
      switch (QMessageBox::
	      information (0, tr ("QtiPlot"), text, tr ("Yes"), tr ("Cancel"),
			   0, 1))
	{
	case 0:
	  QApplication::setOverrideCursor (waitCursor);
	  worksheet->setNumRows (r);
	  emit modifiedTable(this);

	  QApplication::restoreOverrideCursor ();
	  break;

	case 1:
	  return;
	  break;
	}
    }
  else
    {
      QApplication::setOverrideCursor (waitCursor);
      worksheet->setNumRows (r);
      QApplication::restoreOverrideCursor ();
    }

  emit modifiedWindow (this);
}

void
Table::resizeCols (int c)
{
  int i, cols = worksheet->numCols ();
  if (cols == c)
    return;

  if (cols > c)
    {
      QString text = tr ("Columns will be deleted from the table!");
      text += "<p>" + tr ("Do you really want to continue?");
      QMemArray < int >columns (cols - c);
      QStringList rmcols;
      switch (QMessageBox::
	      information (0, tr ("QtiPlot"), text, tr ("Yes"), tr ("Cancel"),
			   0, 1))
	{
	case 0:
	  QApplication::setOverrideCursor (waitCursor);
	  for (i = c; i < cols; i++)
	    {
	      rmcols << colName(i);
	      columns[i - c] = i;

	      QStringList::Iterator it = commandes.end ();
	      commandes.remove (--it);

	      it = comments.end ();
	      comments.remove (--it);
	      it = col_format.end ();
	      col_format.remove (--it);
	      it = col_label.end ();
	      col_label.remove (--it);

	      QValueListIterator < int >itt = colTypes.end ();
	      colTypes.remove (--itt);
	      itt = col_plot_type.end ();
	      col_plot_type.remove (--itt);
	    }
	  if(!rmcols.isEmpty())
	    emit removedCols( rmcols );
	  
	  worksheet->removeColumns (columns);
	  QApplication::restoreOverrideCursor ();
	  break;

	case 1:
	  return;
	  break;
	}
    }
  else
    {
      QApplication::setOverrideCursor (waitCursor);
      addColumns (c - cols);
      setHeaderColType ();
      QApplication::restoreOverrideCursor ();
    }
  emit modifiedWindow (this);
}

void
Table::convolute (int sign)
{
  QStringList s = selectedColumns ();
  if ((int) s.count () != 2)
    {
      QMessageBox::warning (0, tr ("QtiPlot - Error"),
			    tr
			    ("Please select two columns for this operation:\n the first represents the signal and the second the response function!"));
      return;
    }

  int col1 = colIndex (s[0]);
  int col2 = colIndex (s[1]);
  int i, rows = worksheet->numRows ();
  int m = 0;
  for (i = 0; i < rows; i++)
    {
      if (!worksheet->text (i, col2).isEmpty ())
	m++;
    }

  if (m >= rows / 2)
    {
      QMessageBox::warning (0, tr ("QtiPlot - Error"),
			    tr
			    ("The response dataset '%1' must be less than half the size of the signal dataset '%2'!").
			    arg (s[1]).arg (s[0]));
      return;
    }
  else if (m % 2 == 0)
    {
      QMessageBox::warning (0, tr ("QtiPlot - Error"),
			    tr("The response dataset '%1' must contain an odd number of points!").arg (s[1]));
      return;
    }

  QApplication::setOverrideCursor (waitCursor);

  int n = 16;			// tmp number of points
  while (n < rows + m / 2)      // m/2 is added to avoid the polution
    n *= 2;

  QMemArray <double> sig(n), res(m);
  sig.fill(0);

  for (i = 0; i < rows; i++)
    sig[i] = worksheet->text (i, col1).toDouble ();
  for (i = rows; i<n; i++)
    sig[i] = 0;
  for (i = 0; i < m; i++)
    res[i] = worksheet->text (i, col2).toDouble ();

  convlv (sig, res, sign);

  int cols = worksheet->numCols ();
  int cols2 = cols + 1;
  addCol ();
  addCol ();
  for (i = 0; i < rows; i++)
    {
      worksheet->setText (i, cols, QString::number (i + 1));
      worksheet->setText (i, cols2, QString::number (sig[i]));
    }

  QString label = "Conv";
  if (sign == -1)
    label = "Deconv";

  s = colNames ();
  QStringList l = s.grep ("Index");
  QString id = QString::number ((int) l.size () + 1);
  label += id;

  col_label[cols] = "Index" + id;
  col_label[cols2] = label;
  col_plot_type[cols] = X;
  setHeaderColType ();

  emit plotCol (this, QStringList (getName() + "_" + label), 0);
  emit modifiedWindow (this);
  QApplication::restoreOverrideCursor ();
}

void
Table::convlv (QMemArray <double> &sig, QMemArray <double> &dres, int sign)
{
  int n = sig.size();
  int m = dres.size();
  QMemArray <double> res(n);
  res.fill(0);

  int i, m2 = m / 2;
  for (i = 0; i < m2; i++)
    { //store the response in wrap around order, see Numerical Recipes doc
      res[i] = dres[m2 + i];
      res[n - m2 + i] = dres[i];
    }

  if (m2 % 2 == 1)
    res[m2] = dres[m - 1];

// calculate ffts
  gsl_fft_real_radix2_transform (res.data(), 1, n);
  gsl_fft_real_radix2_transform (sig.data(), 1, n);

  double re, im, size;
  for (i = 0; i <= n / 2; i++)
    {				// multiply/divide both ffts
      if (i == 0 || i == n / 2)
	{
	  if (sign == 1)
	    sig[i] *= res[i];
	  else
	    sig[i] /= res[i];
	}
      else
	{
	  int ni = n - i;
	  if (sign == 1)
	    {
	      re = res[i] * sig[i] - res[ni] * sig[ni];
	      im = res[i] * sig[ni] + res[ni] * sig[i];
	    }
	  else
	    {
	      size = res[i] * res[i] + res[ni] * res[ni];
	      re = res[i] * sig[i] + res[ni] * sig[ni];
	      im = res[i] * sig[ni] - res[ni] * sig[i];
	      re /= size;
	      im /= size;
	    }

	  sig[i] = re;
	  sig[ni] = im;
	}
    }
  gsl_fft_halfcomplex_radix2_inverse (sig.data(), 1, n); // inverse fft
}

void
Table::fft (ApplicationWindow *app, double sampling,
	    const QString & realColName, const QString & imagColName,
	    bool forward, bool normalize, bool order, FFTWindow win)
{
  QApplication::setOverrideCursor (waitCursor);

  int rows = worksheet->numRows ();

  QMemArray <double> dat(rows*2), amp(rows), x(rows);

  gsl_fft_complex_wavetable *wavetable = gsl_fft_complex_wavetable_alloc (rows);
  gsl_fft_complex_workspace *workspace = gsl_fft_complex_workspace_alloc (rows);

  int realCol = colIndex (realColName);
  int imCol = -1;
  if (!imagColName.isEmpty ())
    imCol = colIndex (imagColName);

  if (wavetable && workspace)
    { // zero-pad data array
      for (int i = 0; i < rows; i++)
        {
	  double factor=1.0;
	  if(win==Hanning)
	    factor = 0.5 * (1 - cos(M_PI * 2 * i / rows));
	  else if(win==Hamming)
	    factor = 0.54 - 0.46 * cos( M_PI * 2 * i / rows);

          int i2 = i*2;
	  dat[i2] = worksheet->text (i, realCol).toDouble () * factor;
	  if (imCol >= 0)
	    dat[i2+1] = worksheet->text (i, imCol).toDouble () * factor;
          else
            dat[i2+1] = 0;
	}
    }
  else
    {
      QMessageBox::warning (0, "QtiPlot",
			    tr
			    ("Could not allocate memory, operation aborted!"));
      return;
    }

  double df = 1.0 / (double) (rows * sampling);	//frequency sampling
  double aMax = 0.0;		//max amplitude
  QString label, wlabel, text;
  if (forward)
    {
      label = "FFT";
      wlabel = tr("ForwardFFT of ") + getName();
      text = tr ("Frequency");
      gsl_fft_complex_forward (dat.data(), 1, rows, wavetable, workspace);
    }
  else
    {
      label = "IFFT";
      wlabel = tr("InverseFFT of ") + getName();
      text = tr ("Time");
      gsl_fft_complex_inverse (dat.data(), 1, rows, wavetable, workspace);
    }

  gsl_fft_complex_wavetable_free (wavetable);
  gsl_fft_complex_workspace_free (workspace);

  QMemArray <double> yr(rows), yi(rows);

  if (order)
    {
      int n2 = rows / 2;
      int d = (rows+1) / 2;
      for (int i = 0; i < rows; i++)
        {
	  int i2 = (i+d) % rows;
          x[i] = (i-n2) * df;
          yr[i] = dat[i2*2];
          yi[i] = dat[i2*2+1];
        }
    }
  else
    {
      for (int i = 0; i < rows; i++)
	{
	  x[i] = i * df;
	  yr[i] = dat[i*2];
          yi[i] = dat[i*2+1];
	}
    }
  
  for (int i = 0; i < rows; i++)
    {
      double a = sqrt (yr[i] * yr[i] + yi[i] * yi[i]);
      amp[i] = a;
      if (a > aMax)
	aMax = a;
    }
  text +=
    "\t" + tr ("Real") + "\t" + tr ("Imaginary") + "\t" + tr ("Amplitude") +
    "\t" + tr ("Angle") + "\n";
  for (int i = 0; i < rows; i++)
    {
      text += QString::number (x[i]) + "\t";
      text += QString::number (yr[i]) + "\t";
      text += QString::number (yi[i]) + "\t";
      if (normalize)
	text += QString::number (amp[i] / aMax) + "\t";
      else
	text += QString::number (amp[i]) + "\t";
      text += QString::number (atan2(yi[i],yr[i])) + "\n";
    }

  QString title = app->generateUniqueName( label );
  emit createTable (title + "\t" + wlabel, rows, 5, text);
  QApplication::restoreOverrideCursor ();
}

void
Table::copy (Table * m)
{
  for (int i = 0; i < worksheet->numRows (); i++)
    {
      for (int j = 0; j < worksheet->numCols (); j++)
	worksheet->setText (i, j, m->text (i, j));
    }

  setColWidths (m->columnWidths ());
  col_label = m->colNames ();
  col_plot_type = m->plotDesignations ();
  setHeaderColType ();

  commandes = m->getCommandes ();
  setColumnTypes (m->columnTypes ());
  col_format = m->getColumnsFormat ();
  comments = m->colComments ();
}

QString
Table::saveAsTemplate (const QString & geometryInfo)
{
  QString s = "<table>\t" + QString::number (worksheet->numRows ()) + "\t";
  s += QString::number (worksheet->numCols ()) + "\n";
  s += geometryInfo;
  s += saveHeader ();
  s += saveColumnWidths ();
  s += saveCommandes ();
  s += saveColumnTypes ();
  s += saveComments ();
  return s;
}

void
Table::restore (const QStringList & lst)
{
  QStringList l;
  QStringList::const_iterator i = lst.begin ();

  l = QStringList::split ('\t', *i++, true);
  l.remove (l.first ());
  loadHeader (l);

  setColWidths (QStringList::
		split ('\t', (*i).right ((*i).length () - 9), FALSE));
  i++;

  l = QStringList::split ('\t', *i++, true);
  if (l[0] == "com")
    {
      l.remove (l.first ());
      setCommandes (l);
    }
  else if (l[0] == "<com>")
    {
      commandes.clear ();
      for (int col = 0; col < tableCols (); col++)
	commandes << "";
      for (; i != lst.end () && *i != "</com>"; i++)
	{
	  int col = (*i).mid (9, (*i).length () - 11).toInt ();
	  QString formula;
	  for (i++; i != lst.end () && *i != "</col>"; i++)
	    formula += *i + "\n";
	  formula.truncate (formula.length () - 1);
	  commandes[col] = formula;
	}
      i++;
    }

  l = QStringList::split ('\t', *i++, true);
  l.remove (l.first ());
  setColumnTypes (l);

  l = QStringList::split ('\t', *i++, true);
  l.remove (l.first ());
  setColComments (l);
}

void
Table::notifyChanges ()
{
  emit modifiedTable(this);
  emit modifiedWindow (this);
}

void
Table::clear ()
{
  for (int i = 0; i < worksheet->numCols (); i++)
    {
      for (int j = 0; j < worksheet->numRows (); j++)
	worksheet->setText (j, i, QString::null);
    }
  emit modifiedTable(this);
  emit modifiedWindow (this);
}

void
Table::generateFormula()
{
  ColFormulaDialog *cfd 
    = new ColFormulaDialog(this, "ColFormulaDialog",TRUE,
			   WStyle_Tool | WDestructiveClose);
  
  connect (cfd, SIGNAL (makeFormula(int, const QString &, const QString &)),
	   this, SLOT (makeFormula (int, const QString &, const QString &)));
  
  QStringList s = columnsList ();
  cfd->insertColumnsList (s);
  cfd->showNormal ();
  cfd->setActiveWindow ();
}

void
Table::makeFormula(int method, const QString & mapFunc, const QString & destCol)
{
  int index = colIndex(destCol);
  QString map = mapFunc;
  map.replace("%c","col(\"\\1\")");
  QStringList s = selectedColumns();
  s.gres(QRegExp("^.*_(.*)$"), map );
  QString formula;
  switch(method)
    {
    case 0://Sum
      formula = tr("sum(") + s.join(",") + ")";
      break;
    case 1://Mean
      formula = tr("avg(") + s.join(",") + ")";
      break;
    case 2://Product
      formula = s.join("*");
      break;
    case 3://Geometric Mean;
      int n = s.count();
      if(n==2)
	formula = tr("sqrt(")+s.join("*")+")";
      else
	formula = tr("pow(")+s.join("*")+",1/"+QString::number(n)+")";
      break;
    }
  setCommand(index,formula);
  calculate(index,0,worksheet->numRows()-1, true);
}

QString
Table::xyStatistics(int xcol, int ycol, int prec)
{
  //qDebug("xyStatistics");
  QString text = colName(xcol).section('_',1,1)
    + "\tN\tMin\tMax\tMean\tStandardDev\n";
  int rows = worksheet->numRows();
  QMap <QString, QValueList<double> > yvalues;
  for(int r = 0; r < rows; r++)
    {
      QString xval = worksheet->text(r,xcol);
      QString yval = worksheet->text(r,ycol);
      if (!xval.isEmpty () && !yval.isEmpty ())
	yvalues[xval].append(yval.toDouble());
    }
  QValueList <QString> keys = yvalues.keys();
  for(int j=0;j<(int)keys.count(); j++)
    {
      QString xval = keys[j];
      int num = yvalues[xval].count();
      QMemArray <double> dat(num);
      double sum=0,min=0,max=0;
      for( int k = 0; k<num; k++)
	{
	  double value = yvalues[xval][k]; 
	  dat[k] = value;
	  sum += value;
	  if(k==0)
	    {
	      min = value;
	      max = value;
	    }
	  else if (value < min)
	    min = value;
	  else if (value > max)
	    max = value;
	}
      double mean = sum / num;
      double std = gsl_stats_sd(dat.data(),1,num);
      text += xval + "\t" 
	+ QString::number(num) + "\t"
	+ QString::number(min,'g',prec) + "\t"
	+ QString::number(max,'g',prec) + "\t"
	+ QString::number(mean,'g',prec) + "\t"
	+ QString::number(std,'g',prec) + "\n";
    }      
  return text;
}

QString
Table::areaStatistics(int prec)
{
  QTableSelection sel = worksheet->selection (0);
  if(sel.isEmpty())
    {
      QMessageBox::critical (this, tr ("QtiPlot - Area Statistics error"),
                             tr("There is no selection!"));
      return QString::null;
    }

  int top = sel.topRow();
  int bottom = sel.bottomRow();
  int left = sel.leftCol();
  int right = sel.rightCol();

  int cells = ( right - left + 1 ) * ( bottom - top +1 );

  QMemArray <double> dat(cells);
  gsl_vector *y = gsl_vector_alloc(cells);
  int aux=0;
  for (int i = top; i <= bottom; i++)
    for (int j = left; j <= right; j++)
      {
        QString dummy = worksheet->text(i,j);
        double val = dummy.toDouble();
        gsl_vector_set( y, aux, val );
        dat[aux] = val;
        aux++;
      }

  double mean = gsl_stats_mean( dat.data(), 1, cells );
  double min,max;
  gsl_vector_minmax( y, &min, &max );
  double sd = gsl_stats_sd( dat.data(), 1, cells );
  double var = gsl_stats_variance( dat.data(), 1, cells );
  double sum = mean * cells;
  gsl_vector_free( y );

  QString text = tr("Name")+"\t"+tr("Values")+"\n";
  text += tr("Cells") + "\t" + QString::number(cells) + "\n";
  text += tr("Sum") + "\t" + QString::number(sum,'g',prec) + "\n";
  text += tr("Mean") + "\t" + QString::number(mean,'g',prec) + "\n";
  text += tr("Max") + "\t" + QString::number(max,'g',prec) + "\n";
  text += tr("Min") + "\t" + QString::number(min,'g',prec) + "\n";
  text += tr("Width") + "\t" + QString::number(max-min,'g',prec) + "\n";
  text += tr("StandardDev") + "\t" + QString::number(sd,'g',prec) + "\n";
  text += tr("Variance") + "\t" + QString::number(var,'g',prec) + "\n";

  return text;
}

int
Table::findRowIndex(int col, double value)
{
  for(int r=0;r<worksheet->numRows();r++)
    {
      QString text = worksheet->text(r,col);
      if(!text.isEmpty())
	{ 
	  double v = text.toDouble();
	  if(v==0 && value==0)
	    return r;
	  if( fabs(v-value)/QMAX(fabs(v),fabs(value)) < DBL_EPSILON )
	    return r;
	}
    }
  return -1;
}


bool
Table::interpolate( int col, int spline, 
		    QMemArray <double> &x, QMemArray <double> &y ) 
// spline
// 0 : linear
// 1 : cubic
// 2 : akima
{
  int points = x.size();
  if (points<1) return false;
  if (y.size() != points)
    y.resize( points );

  int xcol = colX( col );
  if (xcol == -1)
    {
      QMessageBox::critical (this,
			     tr("QtiPlot - Error"),
			     tr("The column %1 does not have corresponding X Column. Operation aborted!").arg(colName(col)) );
      return false;
    }
  int rows = worksheet->numRows();
  QMemArray <double> xtemp(rows), ytemp(rows);
  int n1=0;
  for(int r=0;r<rows;r++)
    {
      QString xval = worksheet->text(r,xcol);
      QString yval = worksheet->text(r,col);
      if (!xval.isEmpty() && !yval.isEmpty() )
	{
	  xtemp[n1] = xval.toDouble();
	  ytemp[n1] = yval.toDouble();
	  n1++;
	}
    }
  QMemArray <double> x1(n1), y1(n1);
  QMemArray <size_t> p(n1);
  gsl_sort_index( p.data(), xtemp.data(), 1, n1 );
  double pr_x=0;
  for (int i = 0; i < n1; i++)
    {
      if(i>0 && xtemp[p[i]] == pr_x)
        {
	  QMessageBox::critical (this,
				 tr("QtiPlot - Error"),
				 tr("Several cells have the same x value causing divisions by zero, operation aborted!"));
	  return false;
        }
      x1[i] = xtemp[p[i]];
      y1[i] = ytemp[p[i]];
      pr_x = xtemp[p[i]];
    }

  if (n1 < spline + 3)
    {
      QMessageBox::critical (this, tr ("QtiPlot - Error"),
                             tr("You need at least %1 points to perform this operation! Operation aborted!").arg(QString::number(spline+3)));
      return false;
    }

  const gsl_interp_type *method;
  switch (spline)
    {
    case 0:
      method = gsl_interp_linear;
      break;
    case 1:
      method = gsl_interp_cspline;
      break;
    case 2:
      method = gsl_interp_akima;
      break;
    }

  gsl_interp_accel *acc = gsl_interp_accel_alloc ();
  gsl_interp *sp = gsl_interp_alloc (method, n1);
  gsl_interp_init (sp, x1.data(), y1.data(), n1);

  for (int i = 0; i < points; i++)
    y[i] = gsl_interp_eval (sp, x1.data(), y1.data(), x[i], acc);

  gsl_interp_free (sp);
  gsl_interp_accel_free (acc);

  return true;
}

Table::~Table ()
{

}

