/***************************************************************************
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor,                    *
 *   Boston, MA  02110-1301  USA                                           *
 *                                                                         *
 ***************************************************************************/
#include "graph.h"
#include "canvaspicker.h"
#include "ErrorBar.h"
#include "LegendMarker.h"
#include "LineMarker.h"
#include "cursors.h"
#include "scalePicker.h"
#include "pie.h"
#include "ImageMarker.h"
#include "BarCurve.h"
#include "BoxCurve.h"
#include "Histogram.h"
#include "VectorCurve.h"
#include "scales.h"
#include "plot.h"
#include "parser.h"
#include "fileDialogs.h"
#include "colorBox.h"
#include "patternBox.h"
#include "symbolBox.h"
#include "FunctionCurve.h"
#include "Fitter.h"
#include "nrutil.h"
#include "Spectrogram.h"
#include "analysisDialog.h"

#include <qapplication.h>
#include <qbitmap.h>
#include <qclipboard.h>
#include <qcursor.h>
#include <qdragobject.h>
#include <qfiledialog.h>
#include <qimage.h>
#include <qmessagebox.h>
#include <qpixmap.h>
#include <qpicture.h>
#include <qpixmapcache.h>
#include <qpopupmenu.h>
#include <qptrlist.h>
#include <qregexp.h>

#include <qwt_painter.h>
#include <qwt_plot_canvas.h>
#include <qwt_plot_layout.h>
#include <qwt_plot_zoomer.h>
#include <qwt_scale_widget.h>
#include <qwt_scale_engine.h>
#include <qwt_text.h>
#include <qwt_text_label.h>
#include <qwt_color_map.h>

#include <gsl/gsl_sort.h>

#include <math.h>
#include <stddef.h>

static const char *cut_xpm[] = {
  "18 18 3 1",
  ". c None",
  "# c #000000",
  "a c #00007f",
  "..................",
  "..................",
  "..................",
  ".......#...#......",
  ".......#...#......",
  ".......#...#......",
  ".......##.##......",
  "........#.#.......",
  "........###.......",
  ".........#........",
  "........a#a.......",
  "........a.aaa.....",
  "......aaa.a..a....",
  ".....a..a.a..a....",
  ".....a..a.a..a....",
  ".....a..a..aa.....",
  "......aa..........",
  ".................."
};

static const char *copy_xpm[] = {
  "15 13 4 1",
  "# c None",
  ". c #000000",
  "b c #00007f",
  "a c #ffffff",
  "......#########",
  ".aaaa..########",
  ".aaaa.a.#######",
  ".a..a.bbbbbb###",
  ".aaaaabaaaabb##",
  ".a....baaaabab#",
  ".aaaaaba..abbbb",
  ".a....baaaaaaab",
  ".aaaaaba.....ab",
  "......baaaaaaab",
  "######ba.....ab",
  "######baaaaaaab",
  "######bbbbbbbbb"
};

static const char *unzoom_xpm[] = {
  "18 17 2 1",
  ". c None",
  "# c #000000",
  "..................",
  "...#..............",
  "..###.............",
  ".#.#.#.......##...",
  "...#.....##..##...",
  "...#..##.##.......",
  "...#..##....##....",
  "...#........##....",
  "...#...##.........",
  "...#...##.##.##...",
  ".#.#.#....##.##...",
  "..###.............",
  "...#...#......#...",
  "...#..#........#..",
  "...##############.",
  "......#........#..",
  ".......#......#..."
};


#define SWAP(type,a,b) do{type tmp=a; a=b; b=tmp;} while (0)

Graph::Graph (QWidget * parent, const char *name, WFlags f):
QWidget (parent, name, f)
{
  //fprintf(stderr,"graph.cpp : Graph::Graph\n");

  if (!name)
    setName ("graph");

  fitter = 0;
  n_curves = 0;
  widthLine = 1;
  mrkX = -1;
  mrkY = -1;
  fitID = 0;
  //selectedCol = 0;
  selectedPoint = -1;
  selectedCurve = -1;
  selectedMarker = -1;
  selectedCursor = -1;
  startPoint = 0;
  endPoint = -1;
  startID = -1;
  endID = -1;
  //pieRay = 100;
  lineProfileOn = FALSE;
  drawTextOn = FALSE;
  drawLineOn = FALSE;
  drawArrowOn = FALSE;
  subtractLineOn = FALSE;
  cursorEnabled = FALSE;
  movePointsEnabled = FALSE;
  removePointsEnabled = FALSE;
  pickerEnabled = FALSE;
  rangeSelectorsEnabled = FALSE;
  piePlot = FALSE;
  ignoreResize = false;

  drawAxesBackbone = true;
  autoscale = true;
  autoScaleFonts = true;
  autoScaleLineWidth = true;
  autoScaleFrameWidth = true;
  autoScaleGridWidth = true;
  translateOn = false;

  d_spacing = 2;
  d_base_spacing = 2;
  d_base_margin = 5;
  
  d_first_resize=true;

  d_textlayer = false;

  defaultArrowLineWidth = 1;
  defaultArrowColor = QColor (black);
  defaultArrowLineStyle = QPen::SolidLine;
  defaultArrowHeadLength = 8;
  defaultArrowHeadAngle = 25;
  defaultArrowHeadFill = true;

  defaultMarkerFont = QFont ();
  defaultMarkerFrame = 1;
  defaultTextMarkerColor = QColor (black);
  defaultTextMarkerBackground = QColor (white);

  d_user_step = QMemArray < double >(QwtPlot::axisCnt);
  for (int i = 0; i < QwtPlot::axisCnt; i++)
    {
      axisType << Numeric;
      axesFormatInfo << QString::null;
      axesFormulas << QString::null;
      d_user_step[i] = 0.0;
    }

  d_plot = new Plot (this);
  
  //d_plot->plotLayout()->setAlignCanvasToScales(true);
  
  cp = new CanvasPicker (this);

  titlePicker = new TitlePicker (d_plot);
  scalePicker = new ScalePicker (d_plot);

  d_zoomer[0] = new QwtPlotZoomer (QwtPlot::xBottom, QwtPlot::yLeft,
				   QwtPicker::DragSelection | QwtPicker::
				   CornerToCorner, QwtPicker::AlwaysOff,
				   d_plot->canvas ());
  d_zoomer[0]->setRubberBandPen (Qt::black);
  d_zoomer[1] = new QwtPlotZoomer (QwtPlot::xTop, QwtPlot::yRight,
				   QwtPicker::DragSelection | QwtPicker::
				   CornerToCorner, QwtPicker::AlwaysOff,
				   d_plot->canvas ());
  zoom (FALSE);

  setGeometry (QRect (0, 0, 520, 470));
  setFocusPolicy (QWidget::StrongFocus);
  setFocusProxy (d_plot);
  setMouseTracking (TRUE);

  grid.majorOnX = 0;
  grid.majorOnY = 0;
  grid.minorOnX = 0;
  grid.minorOnY = 0;
  grid.majorCol = 3;
  grid.majorStyle = 0;
  grid.majorWidth = 1;
  grid.minorCol = 15;
  grid.minorStyle = 2;
  grid.minorWidth = 1;
  grid.xZeroOn = 0;
  grid.yZeroOn = 0;
  grid.xAxis = QwtPlot::xBottom;
  grid.yAxis = QwtPlot::yLeft;
  setGridOptions (grid);

  //LegendMarker *mrk = new LegendMarker (d_plot);
  //mrk->setOrigin (QPoint (10, 20));
  //legendMarkerID = d_plot->insertMarker (mrk);
  //d_texts = QwtArray < int >(1);
  //d_texts[0] = legendMarkerID;
  legendMarkerID = -1; // no legend for an empty graph
  d_texts = QwtArray < int >(0);

  c_attr.clear();
  c_attr.setAutoDelete( true );

  connect (d_plot, SIGNAL (selectPlot ()), this, SLOT (activateGraph ()));
  connect (d_plot, SIGNAL (selectPlot ()), this, SLOT (drawFocusRect ()));
  connect (d_plot, SIGNAL (moveGraph (const QPoint &)), this,
	   SLOT (moveGraph (const QPoint &)));
  connect (d_plot, SIGNAL (releasedGraph ()), this, SLOT (releaseGraph ()));

  connect (cp, SIGNAL (moveGraph (const QPoint &)), this,
	   SLOT (moveGraph (const QPoint &)));
  connect (cp, SIGNAL (releasedGraph ()), this, SLOT (releaseGraph ()));
  connect (cp, SIGNAL (highlightGraph ()), this, SLOT (drawFocusRect ()));
  connect (cp, SIGNAL (selectPlot ()), this, SLOT (activateGraph ()));
  connect (cp, SIGNAL (drawTextOff ()), this, SIGNAL (drawTextOff ()));
  connect (cp, SIGNAL (viewImageDialog ()), this,
	   SIGNAL (viewImageDialog ()));
  connect (cp, SIGNAL (viewTextDialog ()), this, SIGNAL (viewTextDialog ()));
  connect (cp, SIGNAL (viewLineDialog ()), this, SIGNAL (viewLineDialog ()));
  connect (cp, SIGNAL (showPlotDialog (int)), this,
	   SIGNAL (showPlotDialog (int)));
  connect (cp, SIGNAL (showPieDialog ()), this, SIGNAL (showPieDialog ()));
  connect (cp, SIGNAL (showMarkerPopupMenu ()), this,
	   SIGNAL (showMarkerPopupMenu ()));
  connect (cp, SIGNAL (modified ()), this, SLOT (modified ()));
  connect (cp, SIGNAL (modified ()), this, SIGNAL (modifiedGraph ()));
  connect (cp, SIGNAL (calculateProfile (const QPoint &, const QPoint &)),
	   this,
	   SLOT (calculateLineProfile (const QPoint &, const QPoint &)));
  connect (cp, SIGNAL (showInfo(const QString &)), 
           this, SIGNAL(showInfo(const QString &)));
  connect (cp, SIGNAL (showDisplayBar()), this, SIGNAL(showDisplayBar()));

  connect (titlePicker, SIGNAL (highlightGraph ()), this,
	   SLOT (drawFocusRect ()));
  connect (titlePicker, SIGNAL (showTitleMenu ()), this,
	   SLOT (showTitleContextMenu ()));
  connect (titlePicker, SIGNAL (doubleClicked ()), this,
	   SIGNAL (viewTitleDialog ()));
  connect (titlePicker, SIGNAL (removeTitle ()), this, SLOT (removeTitle ()));
  connect (titlePicker, SIGNAL (clicked ()), this, SLOT (selectTitle ()));
  connect (titlePicker, SIGNAL (moveGraph (const QPoint &)), this,
	   SLOT (moveGraph (const QPoint &)));
  connect (titlePicker, SIGNAL (releasedGraph ()), this,
	   SLOT (releaseGraph ()));

  connect (scalePicker, SIGNAL (highlightGraph ()), this,
	   SLOT (drawFocusRect ()));
  connect (scalePicker, SIGNAL (clicked ()), this, SLOT (activateGraph ()));
  connect (scalePicker, SIGNAL (clicked ()), this, SLOT (deselectMarker ()));
  connect (scalePicker, SIGNAL (axisDblClicked (int)), this,
	   SIGNAL (axisDblClicked (int)));
  connect (scalePicker, SIGNAL (axisTitleRightClicked (int)), this,
	   SLOT (showAxisTitleMenu (int)));
  connect (scalePicker, SIGNAL (axisRightClicked (int)), this,
	   SLOT (showAxisContextMenu (int)));
  connect (scalePicker, SIGNAL (xAxisTitleDblClicked ()), this,
	   SIGNAL (xAxisTitleDblClicked ()));
  connect (scalePicker, SIGNAL (yAxisTitleDblClicked ()), this,
	   SIGNAL (yAxisTitleDblClicked ()));
  connect (scalePicker, SIGNAL (rightAxisTitleDblClicked ()), this,
	   SIGNAL (rightAxisTitleDblClicked ()));
  connect (scalePicker, SIGNAL (topAxisTitleDblClicked ()), this,
	   SIGNAL (topAxisTitleDblClicked ()));
  connect (scalePicker, SIGNAL (moveGraph (const QPoint &)), this,
	   SLOT (moveGraph (const QPoint &)));
  connect (scalePicker, SIGNAL (releasedGraph ()), this,
	   SLOT (releaseGraph ()));

  connect (d_zoomer[0], SIGNAL (zoomed (const QwtDoubleRect &)), this,
	   SLOT (zoomed (const QwtDoubleRect &)));
}

double
Graph::scaleFactor()
{
  int w = frameGeometry().width();
  int h = frameGeometry().height();
  return sqrt( (double)(w*h) / (double)(500*450) );
}

void
Graph::drawFocusRect ()
{
  if (translateOn || pickerEnabled)
    return;

  emit highlightGraph (this);
}

void
Graph::customLegend ()
{
  LegendMarker *mrk = (LegendMarker *) d_plot->marker (legendMarkerID);
  if (!mrk)
    return;

  mrk->setBackground (defaultMarkerFrame);
  mrk->setFont (defaultMarkerFont, this->frameGeometry());
  mrk->setTextColor (defaultTextMarkerColor);
  mrk->setBackgroundColor (defaultTextMarkerBackground);

  if (!n_curves)
    {
      mrk->setValue (50, 990);
      mrk->updateOrigin ();
    }
}

void
Graph::emitModified ()
{
  emit modifiedGraph ();
}

void
Graph::modified ()
{
  //fprintf(stderr,"Graph::modified\n");
  emit modifiedGraph (this);
}

void
Graph::moveGraph (const QPoint & pos)
{
  emit moveGraph (this, pos);
}

void
Graph::releaseGraph ()
{
  emit releaseGraph (this);
}

void
Graph::activateGraph ()
{
  emit selectedGraph (this);
  setFocus ();
}

void
Graph::deselectMarker ()
{
  if(selectedMarker>=0){
    selectedMarker = -1;
    d_plot->replot ();
  }
}

int
Graph::selectedMarkerKey ()
{
  return selectedMarker;
}

QwtPlotMarker *
Graph::selectedMarkerPtr ()
{
  return d_plot->marker (selectedMarker);
}

// orphaned
void
Graph::setSelectedMarker (int mrk)
{
  selectedMarker = mrk;
}

void
Graph::highlightLineMarker (int markerID)
{
  LineMarker *mrkL = (LineMarker *) d_plot->marker (markerID);
  if (!mrkL)
    return;

  selectedMarker = markerID;

  QwtPlotCanvas *canvas = d_plot->canvas ();

// draw highlight line + end resizing rectangles 
  QPainter painter (canvas);
  painter.save ();

  painter.setRasterOp (Qt::NotXorROP);
  painter.setPen (QPen (QColor (red), mrkL->width (), Qt::SolidLine));
  painter.drawLine (mrkL->startPoint (), mrkL->endPoint ());

  painter.setPen (QPen (QColor (black), mrkL->width (), Qt::SolidLine));
  painter.setBrush (QBrush (QColor (black), QBrush::SolidPattern));

  QRect sr = QRect (QPoint (0, 0), QSize (7, 7));
  sr.moveCenter (mrkL->startPoint ());
  painter.drawRect (sr);

  sr.moveCenter (mrkL->endPoint ());
  painter.drawRect (sr);
  painter.restore ();
}

void
Graph::highlightTextMarker (int markerID)
{
  LegendMarker *mrk = (LegendMarker *) d_plot->marker (markerID);
  if (!mrk)
    return;

  selectedMarker = markerID;

  QwtPlotCanvas *canvas = d_plot->canvas ();
  QPainter painter (canvas);
  painter.setPen (QPen (red, 2, Qt::SolidLine));
  painter.setRasterOp (Qt::NotXorROP);
  painter.drawRect (mrk->rect ());
}

void
Graph::highlightImageMarker (int markerID)
{
  ImageMarker *mrkI = (ImageMarker *) d_plot->marker (markerID);
  if (!mrkI)
    return;

  selectedMarker = markerID;
  QwtPlotCanvas *canvas = d_plot->canvas ();
  QPainter painter (canvas);
  painter.setPen (QPen (red, 2, Qt::SolidLine));
  painter.setRasterOp (Qt::NotXorROP);
  painter.drawRect (mrkI->rect ());
}

void
Graph::selectNextMarker ()
{
  QValueList < int >mrkKeys = d_plot->markerKeys ();
  int n = mrkKeys.size ();
  if (n == 0)
    return;

  int min_key = mrkKeys[0], max_key = mrkKeys[0];
  for (int i = 0; i < n; i++)
    {
      if (mrkKeys[i] >= max_key)
	max_key = mrkKeys[i];
      if (mrkKeys[i] <= min_key)
	min_key = mrkKeys[i];
    }

  d_plot->replot ();

  if (selectedMarker >= 0)
    {
      int key = selectedMarker + 1;
      if (key > max_key)
	key = min_key;
      while (key <= max_key)
	{
	  if (d_texts.contains (key))
	    {
	      highlightTextMarker (key);
	      break;
	    }
	  else if (d_lines.contains (key))
	    {
	      highlightLineMarker (key);
	      break;
	    }
	  else if (d_images.contains (key))
	    {
	      highlightImageMarker (key);
	      break;
	    }
	  else
	    key++;
	}
    }
  else
    {
      int key = min_key;
      while (selectedMarker < 0)
	{
	  if (d_texts.contains (key))
	    highlightTextMarker (key);
	  else if (d_lines.contains (key))
	    highlightLineMarker (key);
	  else if (d_images.contains (key))
	    highlightImageMarker (key);
	  key++;
	}
    }
}

void
Graph::replot ()
{
  d_plot->replot ();
}

void
Graph::movedPicker (const QPoint & pos, bool mark)
{
  static double x,y;

  QString info;
  info.sprintf ("x=%g; y=%g",
		d_plot->invTransform (QwtPlot::xBottom, pos.x ()),
		d_plot->invTransform (QwtPlot::yLeft, pos.y ()));
  emit showInfo (info);

  if (mark)
    {
      QwtPlotMarker mrk;
      mrk.
	setSymbol (QwtSymbol
		   (QwtSymbol::Cross, QBrush (Qt::NoBrush), QPen (red, 1),
		    QSize (15, 15)));

      QPainter painter (d_plot->canvas ());
      painter.setClipping (TRUE);
      QRect cr = d_plot->canvas ()->contentsRect ();
      painter.setClipRect (cr);
      painter.setRasterOp (Qt::NotXorROP);
     
      if(d_pickerMarker)
        {
          mrk.setValue(x,y);
          mrk.draw (&painter, d_plot->canvasMap (QwtPlot::xBottom),
	  	    d_plot->canvasMap (QwtPlot::yLeft), cr);
        }

      if (translateOn)
	{
	  const QwtPlotCurve *c = curve (selectedCurve);
	  if (!c)
	    return;

	  if (translationDirection)
            {
	      x = c->x (selectedPoint);
	      y = d_plot->invTransform( QwtPlot::yLeft, pos.y() );
            }
	  else
            {
	      x = d_plot->invTransform( QwtPlot::xBottom, pos.x() );
              y = c->y (selectedPoint);
            }
	}
      else
        {
	  x = d_plot->invTransform( QwtPlot::xBottom, pos.x() );
          y = d_plot->invTransform( QwtPlot::yLeft, pos.y() );
        }

      mrk.setValue(x,y);
      mrk.draw (&painter, d_plot->canvasMap (QwtPlot::xBottom),
		d_plot->canvasMap (QwtPlot::yLeft), cr);
      d_pickerMarker = true;
    }
}

bool
Graph::pickerActivated ()
{
  return pickerEnabled;
}

void
Graph::initFonts (const QFont & scaleTitleFnt, const QFont & numbersFnt)
{
  double scale = scaleFactor();
  for (int i = 0; i < QwtPlot::axisCnt; i++)
    {
      axisFontSize[i] = numbersFnt.pointSizeFloat() / scale; 
      axisTitleFontSize[i] = scaleTitleFnt.pointSizeFloat() / scale;

      d_plot->setAxisFont (i, numbersFnt);
      QwtText t = d_plot->axisTitle (i);
      t.setFont (scaleTitleFnt);
      d_plot->setAxisTitle (i, t);
    }
}

void
Graph::setAxisFont (int axis, const QFont & fnt)
{
  axisFontSize[axis] = fnt.pointSizeFloat() / scaleFactor();

  d_plot->setAxisFont (axis, fnt);
  d_plot->replot ();
  emit modifiedGraph ();
}

QFont
Graph::axisFont (int axis)
{
  return d_plot->axisFont (axis);
}

void
Graph::enableAxes (const QStringList & list)
{
  int i;
  for (i = 0; i < QwtPlot::axisCnt; i++)
    {
      bool ok = list[i + 1].toInt ();
      d_plot->enableAxis (i, ok);
    }

  for (i = 0; i < QwtPlot::axisCnt; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	scale->setMargin (0);
    }
  scalePicker->refresh ();
}

void
Graph::enableAxes (QMemArray < bool > axesOn)
{
  int i;
  for (i = 0; i < QwtPlot::axisCnt; i++)
    d_plot->enableAxis (i, axesOn[i]);

  for (i = 0; i < QwtPlot::axisCnt; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	scale->setMargin (0);
    }
  scalePicker->refresh ();
}

QMemArray < bool > Graph::enabledAxes ()
{
  QMemArray < bool > axesOn (4);
  for (int i = 0; i < QwtPlot::axisCnt; i++)
    axesOn[i] = d_plot->axisEnabled (i);
  return axesOn;
}

QValueList < int >
Graph::axesBaseline ()
{
  QValueList < int >baselineDist;
  for (int i = 0; i < QwtPlot::axisCnt; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	baselineDist << scale->margin ();
      else
	baselineDist << 0;
    }
  return baselineDist;
}

void
Graph::setAxesBaseline (const QValueList < int >&lst)
{
  for (int i = 0; i < QwtPlot::axisCnt; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	scale->setMargin (lst[i]);
    }
}

void
Graph::setAxesBaseline (QStringList & lst)
{
  //lst.remove (lst.first ());
  for (int i = 0; i < QwtPlot::axisCnt; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	scale->setMargin ((lst[i+1]).toInt ());
    }
}

QValueList < int >
Graph::axesType ()
{
  return axisType;
}

void
Graph::setLabelsNumericFormat (int axis, int format, int prec,
			       const QString & formula)
{
  axisType[axis] = Numeric;
  axesFormulas[axis] = formula;

  ScaleDraw *sd_old = (ScaleDraw *) d_plot->axisScaleDraw (axis);
  const QwtScaleDiv div = sd_old->scaleDiv ();

  if (format == Plot::Superscripts)
    {
      QwtSupersciptsScaleDraw *sd =
	new QwtSupersciptsScaleDraw (formula.ascii ());
      sd->setLabelFormat ('s', prec);
      sd->setScaleDiv (div);
      d_plot->setAxisScaleDraw (axis, sd);
    }
  else
    {
      ScaleDraw *sd = new ScaleDraw (formula.ascii ());
      sd->setScaleDiv (div);

      if (format == Plot::Automatic)
	sd->setLabelFormat ('g', prec);
      else if (format == Plot::Scientific)
	sd->setLabelFormat ('e', prec);
      else if (format == Plot::Decimal)
	sd->setLabelFormat ('f', prec);

      d_plot->setAxisScaleDraw (axis, sd);
    }
}

void
Graph::setLabelsNumericFormat (int axis, const QStringList & l)
{
  QwtScaleDraw *sd = d_plot->axisScaleDraw (axis);
  if (!sd->hasComponent (QwtAbstractScaleDraw::Labels) ||
      axisType[axis] != Numeric)
    return;

  int format = l[2 * axis].toInt ();
  int prec = l[2 * axis + 1].toInt ();
  setLabelsNumericFormat (axis, format, prec, axesFormulas[axis]);
}

void
Graph::setLabelsNumericFormat (const QStringList & l)
{
  for (int axis = 0; axis < 4; axis++)
    setLabelsNumericFormat (axis, l);
}

void
Graph::setAxesType (const QValueList < int >tl)
{
  axisType = tl;
}

QString
Graph::saveAxesLabelsType ()
{
  QString s = "AxisType\t";
  for (int i = 0; i < 4; i++)
    {
      int type = axisType[i];
      s += QString::number (type);
      if (type == Time || type == Date || type == Txt ||
	  type == ColHeader || type == Day || type == Month)
	s += ";" + axesFormatInfo[i];
      s += "\t";
    };

  return s + "\n";
}

QString
Graph::saveTicksType ()
{
  QValueList < int >ticksTypeList = d_plot->getMajorTicksType ();
  QString s = "MajorTicks\t";
  int i;
  for (i = 0; i < 4; i++)
    s += QString::number (ticksTypeList[i]) + "\t";
  s += "\n";

  ticksTypeList = d_plot->getMinorTicksType ();
  s += "MinorTicks\t";
  for (i = 0; i < 4; i++)
    s += QString::number (ticksTypeList[i]) + "\t";

  return s + "\n";
}

QStringList
Graph::enabledTickLabels ()
{
  QStringList lst;
  for (int axis = 0; axis < QwtPlot::axisCnt; axis++)
    {
      const QwtScaleDraw *sd = d_plot->axisScaleDraw (axis);
      lst << QString::number (sd->
			      hasComponent (QwtAbstractScaleDraw::Labels));
    }
  return lst;
}

QString
Graph::saveEnabledTickLabels ()
{
  QString s = "EnabledTickLabels\t";
  for (int axis = 0; axis < QwtPlot::axisCnt; axis++)
    {
      const QwtScaleDraw *sd = d_plot->axisScaleDraw (axis);
      s +=
	QString::number (sd->hasComponent (QwtAbstractScaleDraw::Labels)) +
	"\t";
    }
  return s + "\n";
}

QString
Graph::saveLabelsFormat ()
{
  QString s = "LabelsFormat\t";
  for (int axis = 0; axis < QwtPlot::axisCnt; axis++)
    {
      s += QString::number (d_plot->axisLabelFormat (axis)) + "\t";
      s += QString::number (d_plot->axisLabelPrecision (axis)) + "\t";
    }
  return s + "\n";
}

QString
Graph::saveAxesBaseline ()
{
  QString s = "AxesBaseline\t";
  for (int i = 0; i < QwtPlot::axisCnt; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	s += QString::number (scale->margin ()) + "\t";
      else
	s += "0\t";
    }
  return s + "\n";
}

QString
Graph::saveLabelsRotation ()
{
  QString s = "LabelsRotation\t";
  s += QString::number (labelsRotation (QwtPlot::xBottom)) + "\t";
  s += QString::number (labelsRotation (QwtPlot::xTop)) + "\n";
  return s;
}

void
Graph::setEnabledTickLabels (const QStringList & labelsOn)
{
  for (int axis = 0; axis < QwtPlot::axisCnt; axis++)
    {
      QwtScaleWidget *sc = d_plot->axisWidget (axis);
      if (sc)
	{
	  QwtScaleDraw *sd = d_plot->axisScaleDraw (axis);
	  sd->enableComponent (QwtAbstractScaleDraw::Labels,
			       labelsOn[axis] == "1");
	}
    }
}

void
Graph::setMajorTicksType (const QValueList < int >&lst)
{
  if (d_plot->getMajorTicksType () == lst)
    return;

  for (int i = 0; i < (int) lst.count (); i++)
    {
      ScaleDraw *sd = (ScaleDraw *) d_plot->axisScaleDraw (i);
      if (lst[i] == ScaleDraw::None || lst[i] == ScaleDraw::In)
	sd->enableComponent (QwtAbstractScaleDraw::Ticks, false);
      else
	{
	  sd->enableComponent (QwtAbstractScaleDraw::Ticks);
	  sd->setTickLength (QwtScaleDiv::MinorTick,
			     d_plot->minorTickLength ());
	  sd->setTickLength (QwtScaleDiv::MediumTick,
			     d_plot->minorTickLength ());
	  sd->setTickLength (QwtScaleDiv::MajorTick,
			     d_plot->majorTickLength ());
	}
      sd->setMajorTicksStyle ((ScaleDraw::TicksStyle) lst[i]);
    }
}

void
Graph::setMajorTicksType (const QStringList & lst)
{
  for (int i = 0; i < (int) lst.count (); i++)
    d_plot->setMajorTicksType (i, lst[i].toInt ());
}

void
Graph::setMinorTicksType (const QValueList < int >&lst)
{
  if (d_plot->getMinorTicksType () == lst)
    return;

  for (int i = 0; i < (int) lst.count (); i++)
    d_plot->setMinorTicksType (i, lst[i]);
}

void
Graph::setMinorTicksType (const QStringList & lst)
{
  for (int i = 0; i < (int) lst.count (); i++)
    d_plot->setMinorTicksType (i, lst[i].toInt ());
}

int
Graph::minorTickLength ()
{
  return d_plot->minorTickLength ();
}

int
Graph::majorTickLength ()
{
  return d_plot->majorTickLength ();
}

void
Graph::setAxisTicksLength (int axis, int majTicksType, int minTicksType,
			   int minLength, int majLength)
{
  QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (axis);
  if (!scale)
    return;

  d_plot->setTickLength (minLength, majLength);

  ScaleDraw *sd = (ScaleDraw *) d_plot->axisScaleDraw (axis);
  sd->setMajorTicksStyle ((ScaleDraw::TicksStyle) majTicksType);
  sd->setMinorTicksStyle ((ScaleDraw::TicksStyle) minTicksType);

  if (majTicksType == ScaleDraw::None && minTicksType == ScaleDraw::None)
    sd->enableComponent (QwtAbstractScaleDraw::Ticks, false);
  else
    sd->enableComponent (QwtAbstractScaleDraw::Ticks);

  if (majTicksType == ScaleDraw::None || majTicksType == ScaleDraw::In)
    majLength = 1;
  if (minTicksType == ScaleDraw::None || minTicksType == ScaleDraw::In)
    minLength = 0;

  sd->setTickLength (QwtScaleDiv::MinorTick, minLength);
  sd->setTickLength (QwtScaleDiv::MediumTick, minLength);
  sd->setTickLength (QwtScaleDiv::MajorTick, majLength);
}

void
Graph::setTicksLength (int minLength, int majLength)
{
  QValueList < int >majTicksType = d_plot->getMajorTicksType ();
  QValueList < int >minTicksType = d_plot->getMinorTicksType ();

  for (int i = 0; i < 4; i++)
    setAxisTicksLength (i, majTicksType[i], minTicksType[i], minLength,
			majLength);

  double factor = scaleFactor();
  d_minor_ticklength = minLength / factor;
  d_major_ticklength = majLength / factor;
}

void
Graph::changeTicksLength (int minLength, int majLength)
{
  if (d_plot->minorTickLength () == minLength &&
      d_plot->majorTickLength () == majLength)
    return;

  setTicksLength (minLength, majLength);

  d_plot->hide ();
  for (int i = 0; i < 4; i++)
    {
      if (d_plot->axisEnabled (i))
	{
	  d_plot->enableAxis (i, false);
	  d_plot->enableAxis (i, true);
	}
    }
  d_plot->replot ();
  d_plot->show ();

  emit modifiedGraph ();
}

void
Graph::showAxis (int axis, int type, const QString & formatInfo,
		 Table * table, bool axisOn, int majTicksType,
		 int minTicksType, bool labelsOn, const QColor & c,
		 int format, int prec, int rotation, int baselineDist,
		 const QString & formula, const QColor & labelsColor)
{
  d_plot->enableAxis (axis, axisOn);
  if (!axisOn)
    return;

  QValueList < int >majTicksTypeList = d_plot->getMajorTicksType ();
  QValueList < int >minTicksTypeList = d_plot->getMinorTicksType ();

  QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (axis);
  ScaleDraw *sclDraw = (ScaleDraw *) d_plot->axisScaleDraw (axis);

  if (d_plot->axisEnabled (axis) == axisOn &&
      majTicksTypeList[axis] == majTicksType &&
      minTicksTypeList[axis] == minTicksType &&
      axesColors ()[axis] == c.name () &&
      axesNumColors ()[axis] == labelsColor.name () &&
      prec == d_plot->axisLabelPrecision (axis) &&
      format == d_plot->axisLabelFormat (axis) &&
      labelsRotation (axis) == rotation &&
      axisType[axis] == type &&
      axesFormatInfo[axis] == formatInfo &&
      axesFormulas[axis] == formula &&
      scale->margin () == baselineDist &&
      sclDraw->hasComponent (QwtAbstractScaleDraw::Labels) == labelsOn)
    return;

  scale->setMargin (baselineDist);
  QPalette pal = scale->palette ();
  if (pal.color (QPalette::Active, QColorGroup::Foreground) != c)
    pal.setColor (QColorGroup::Foreground, c);
  if (pal.color (QPalette::Active, QColorGroup::Text) != labelsColor)
    pal.setColor (QColorGroup::Text, labelsColor);
  scale->setPalette (pal);

  if (!labelsOn)
    sclDraw->enableComponent (QwtAbstractScaleDraw::Labels, false);
  else
    {
      if (type == Numeric)
	setLabelsNumericFormat (axis, format, prec, formula);
      else if (type == Day)
	setLabelsDayFormat (axis, format);
      else if (type == Month)
	setLabelsMonthFormat (axis, format);
      else if (type == Time || type == Date)
	setLabelsDateTimeFormat (axis, type, formatInfo);
      else
	setLabelsTextFormat (axis, type, formatInfo, table);

      setAxisLabelRotation (axis, rotation);
    }

  sclDraw = (ScaleDraw *) d_plot->axisScaleDraw (axis);
  sclDraw->enableComponent (QwtAbstractScaleDraw::Backbone, drawAxesBackbone);

  setAxisTicksLength (axis, majTicksType, minTicksType,
		      d_plot->minorTickLength (), d_plot->majorTickLength ());

  if (axisOn && (axis == QwtPlot::xTop || axis == QwtPlot::yRight))
    updateSecondaryAxis (axis);	//synchronize scale divisions

  scalePicker->refresh ();
  d_plot->updateLayout ();	//This is necessary in order to enable/disable tick labels
  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::setLabelsDayFormat (int axis, int format)
{
  axisType[axis] = Day;
  axesFormatInfo[axis] = QString::number (format);

  ScaleDraw *sd_old = (ScaleDraw *) d_plot->axisScaleDraw (axis);
  const QwtScaleDiv div = sd_old->scaleDiv ();

  WeekDayScaleDraw *sd =
    new WeekDayScaleDraw ((WeekDayScaleDraw::NameFormat) format);
  sd->setScaleDiv (div);
  d_plot->setAxisScaleDraw (axis, sd);
}

void
Graph::setLabelsMonthFormat (int axis, int format)
{
  axisType[axis] = Month;
  axesFormatInfo[axis] = QString::number (format);

  ScaleDraw *sd_old = (ScaleDraw *) d_plot->axisScaleDraw (axis);
  const QwtScaleDiv div = sd_old->scaleDiv ();

  MonthScaleDraw *sd =
    new MonthScaleDraw ((MonthScaleDraw::NameFormat) format);
  sd->setScaleDiv (div);
  d_plot->setAxisScaleDraw (axis, sd);
}

void
Graph::setLabelsTextFormat (int axis, int type, const QString & labelsColName,
			    Table * table)
{
  if (type == Numeric || type == Time || type == Date)
    return;

  QStringList list;
  if (type == Txt)
    {
      if (!table)
	return;

      axesFormatInfo[axis] = labelsColName;
      int r = table->tableRows ();
      int col = table->colIndex (labelsColName);

      for (int i = 0; i < r; i++)
	list << table->text (i, col);
    }
  else if (type == ColHeader)
    {
      if (!table)
	return;

      axesFormatInfo[axis] = table->getName();
      for (int i = 0; i < table->tableCols (); i++)
	{
	  if (table->colPlotDesignation (i) == Table::Y)
	    list << table->colLabel (i);
	}
    }
  d_plot->setAxisScaleDraw (axis, new QwtTextScaleDraw (list));
  axisType[axis] = type;
}

void
Graph::setLabelsDateTimeFormat (int axis, int type,
				const QString & formatInfo)
{
  if (type < Time)
    return;

  QStringList list = QStringList::split (';', formatInfo, true);
  if ((int) list.count () < 2 || list[0].isEmpty () || list[1].isEmpty ())
    return;

  if (type == Time)
    {
      QTime t = QTime::fromString (list[0], Qt::TextDate);
      TimeScaleDraw *sd = new TimeScaleDraw (t, list[1]);
      sd->enableComponent (QwtAbstractScaleDraw::Backbone, drawAxesBackbone);
      d_plot->setAxisScaleDraw (axis, sd);
    }
  else if (type == Date)
    {
      QDate d = QDate::fromString (list[0], Qt::ISODate);
      DateScaleDraw *sd = new DateScaleDraw (d, list[1]);
      sd->enableComponent (QwtAbstractScaleDraw::Backbone, drawAxesBackbone);
      d_plot->setAxisScaleDraw (axis, sd);
    }

  axisType[axis] = type;
  axesFormatInfo[axis] = formatInfo;
}

void
Graph::setAxisLabelRotation (int axis, int rotation)
{
  if (axis == QwtPlot::xBottom)
    {
      if (rotation > 0)
	d_plot->setAxisLabelAlignment (axis,
				       Qt::AlignRight | Qt::AlignVCenter);
      else if (rotation < 0)
	d_plot->setAxisLabelAlignment (axis,
				       Qt::AlignLeft | Qt::AlignVCenter);
      else if (rotation == 0)
	d_plot->setAxisLabelAlignment (axis,
				       Qt::AlignHCenter | Qt::AlignBottom);
    }
  else if (axis == QwtPlot::xTop)
    {
      if (rotation > 0)
	d_plot->setAxisLabelAlignment (axis,
				       Qt::AlignLeft | Qt::AlignVCenter);
      else if (rotation < 0)
	d_plot->setAxisLabelAlignment (axis,
				       Qt::AlignRight | Qt::AlignVCenter);
      else if (rotation == 0)
	d_plot->setAxisLabelAlignment (axis, Qt::AlignHCenter | Qt::AlignTop);
    }
  d_plot->setAxisLabelRotation (axis, (double) rotation);
}

int
Graph::labelsRotation (int axis)
{
  ScaleDraw *sclDraw = (ScaleDraw *) d_plot->axisScaleDraw (axis);
  return (int) sclDraw->labelRotation ();
}

void
Graph::setYAxisTitleFont (const QFont & fnt)
{
  setAxisTitleFont( QwtPlot::yLeft, fnt );
}

void
Graph::setXAxisTitleFont (const QFont & fnt)
{
  setAxisTitleFont( QwtPlot::xBottom, fnt );
}

void
Graph::setRightAxisTitleFont (const QFont & fnt)
{
  setAxisTitleFont( QwtPlot::yRight, fnt );
}

void
Graph::setTopAxisTitleFont (const QFont & fnt)
{
  setAxisTitleFont( QwtPlot::xTop, fnt );
}

void
Graph::setAxisTitleFont (int axis, const QFont & fnt)
{
  axisTitleFontSize[axis] = fnt.pointSizeFloat() / scaleFactor();

  QwtText t = d_plot->axisTitle (axis);
  t.setFont (fnt);
  d_plot->setAxisTitle (axis, t);
  d_plot->replot ();
  emit modifiedGraph ();
}

QFont
Graph::axisTitleFont (int axis)
{
  return d_plot->axisTitle (axis).font ();
}

QColor
Graph::axisTitleColor (int axis)
{
  QColor c;
  QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (axis);
  if (scale)
    c = scale->title ().color ();
  return c;
}

void
Graph::setAxesNumColors (const QStringList & colors)
{
  for (int i = 0; i < 4; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	{
	  QPalette pal = scale->palette ();
	  pal.setColor (QColorGroup::Text, QColor (colors[i]));
	  scale->setPalette (pal);
	}
    }
}

void
Graph::setAxesColors (const QStringList & colors)
{
  for (int i = 0; i < 4; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	{
	  QPalette pal = scale->palette ();
	  pal.setColor (QColorGroup::Foreground, QColor (colors[i]));
	  scale->setPalette (pal);
	}
    }
}

QString
Graph::saveAxesColors ()
{
  QString s = "AxesColors\t";
  QStringList colors, numColors;
  QPalette pal;
  int i;
  for (i = 0; i < 4; i++)
    {
      colors << (Qt::black).name ();
      numColors << (Qt::black).name ();
    }

  for (i = 0; i < 4; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	{
	  pal = scale->palette ();
	  colors[i] =
	    pal.color (QPalette::Active, QColorGroup::Foreground).name ();
	  numColors[i] =
	    pal.color (QPalette::Active, QColorGroup::Text).name ();
	}
    }
  s += colors.join ("\t") + "\n";
  s += "AxesNumberColors\t" + numColors.join ("\t") + "\n";
  return s;
}

QStringList
Graph::axesColors ()
{
  QStringList colors;
  QPalette pal;
  int i;
  for (i = 0; i < 4; i++)
    colors << (Qt::black).name ();

  for (i = 0; i < 4; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	{
	  pal = scale->palette ();
	  colors[i] =
	    pal.color (QPalette::Active, QColorGroup::Foreground).name ();
	}
    }
  return colors;
}

QColor
Graph::axisColor (int axis)
{
  QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (axis);
  if (scale)
    return scale->palette ().color (QPalette::Active,
				    QColorGroup::Foreground);
  else
    return QColor (black);
}

QColor
Graph::axisNumbersColor (int axis)
{
  QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (axis);
  if (scale)
    return scale->palette ().color (QPalette::Active, QColorGroup::Text);
  else
    return QColor (black);
}

QStringList
Graph::axesNumColors ()
{
  QStringList colors;
  QPalette pal;
  int i;
  for (i = 0; i < 4; i++)
    colors << (Qt::black).name ();

  for (i = 0; i < 4; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	{
	  pal = scale->palette ();
	  colors[i] = pal.color (QPalette::Active, QColorGroup::Text).name ();
	}
    }
  return colors;
}

void
Graph::setTitleColor (const QColor & c)
{
  QwtText t = d_plot->title ();
  t.setColor (c);
  d_plot->setTitle (t);
  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::setTitleAlignment (int align)
{
  QwtText t = d_plot->title ();
  t.setRenderFlags (align);
  d_plot->setTitle (t);
  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::setTitleFont (const QFont & fnt)
{
 // font size for standard geometry 
  titleFontSize = fnt.pointSizeFloat() / scaleFactor();

  QwtText t = d_plot->title ();
  t.setFont (fnt);
  d_plot->setTitle (t);
  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::setYAxisTitle (const QString & text)
{
  d_plot->setAxisTitle (0, text);
  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::setXAxisTitle (const QString & text)
{
  d_plot->setAxisTitle (2, text);
  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::setRightAxisTitle (const QString & text)
{
  d_plot->setAxisTitle (1, text);
  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::setTopAxisTitle (const QString & text)
{
  d_plot->setAxisTitle (3, text);
  d_plot->replot ();
  emit modifiedGraph ();
}

int
Graph::axisTitleAlignment (int axis)
{
  return d_plot->axisTitle (axis).renderFlags ();
}

void
Graph::setAxesTitlesAlignment (const QStringList & align)
{
  for (int i = 0; i < 4; i++)
    {
      QwtText t = d_plot->axisTitle (i);
      t.setRenderFlags (align[i + 1].toInt ());
      d_plot->setAxisTitle (i, t);
    }
}

void
Graph::setXAxisTitleAlignment (int align)
{
  QwtText t = d_plot->axisTitle (QwtPlot::xBottom);
  t.setRenderFlags (align);
  d_plot->setAxisTitle (QwtPlot::xBottom, t);

  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::setYAxisTitleAlignment (int align)
{
  QwtText t = d_plot->axisTitle (QwtPlot::yLeft);
  t.setRenderFlags (align);
  d_plot->setAxisTitle (QwtPlot::yLeft, t);

  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::setTopAxisTitleAlignment (int align)
{
  QwtText t = d_plot->axisTitle (QwtPlot::xTop);
  t.setRenderFlags (align);
  d_plot->setAxisTitle (QwtPlot::xTop, t);
  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::setRightAxisTitleAlignment (int align)
{
  QwtText t = d_plot->axisTitle (QwtPlot::yRight);
  t.setRenderFlags (align);
  d_plot->setAxisTitle (QwtPlot::yRight, t);

  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::setAxisTitle (int axis, const QString & text)
{
  int a;
  switch (axis)
    {
    case 0:
      a = 2;
      break;

    case 1:
      a = 0;
      break;

    case 2:
      a = 3;
      break;

    case 3:
      a = 1;
      break;
    }

  d_plot->setAxisTitle (a, text);
  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::setGridOptions (const gridOptions & o)
{
  double factor = scaleFactor();
  d_grid_major = o.majorWidth / factor;
  d_grid_minor = o.minorWidth / factor;

  if (grid.majorCol == o.majorCol && grid.majorOnX == o.majorOnX &&
      grid.majorOnY == o.majorOnY && grid.majorStyle == o.majorStyle &&
      grid.majorWidth == o.majorWidth && grid.minorCol == o.minorCol &&
      grid.minorOnX == o.minorOnX && grid.minorOnY == o.minorOnY &&
      grid.minorStyle == o.minorStyle && grid.minorWidth == o.minorWidth &&
      grid.xAxis == o.xAxis && grid.yAxis == o.yAxis &&
      grid.xZeroOn == o.xZeroOn && grid.yZeroOn == o.yZeroOn)
    return;

  grid = o;

  QColor minColor = ColorBox::color (grid.minorCol);
  QColor majColor = ColorBox::color (grid.majorCol);

  Qt::PenStyle majStyle = getPenStyle (grid.majorStyle);
  Qt::PenStyle minStyle = getPenStyle (grid.minorStyle);

  QPen majPen = QPen (majColor, grid.majorWidth, majStyle);
  d_plot->grid ()->setMajPen (majPen);

  QPen minPen = QPen (minColor, grid.minorWidth, minStyle);
  d_plot->grid ()->setMinPen (minPen);

  if (grid.majorOnX)
    d_plot->grid ()->enableX (TRUE);
  else if (grid.majorOnX == 0)
    d_plot->grid ()->enableX (FALSE);

  if (grid.minorOnX)
    d_plot->grid ()->enableXMin (TRUE);
  else if (grid.minorOnX == 0)
    d_plot->grid ()->enableXMin (FALSE);

  if (grid.majorOnY)
    d_plot->grid ()->enableY (TRUE);
  else if (grid.majorOnY == 0)
    d_plot->grid ()->enableY (FALSE);

  if (grid.minorOnY)
    d_plot->grid ()->enableYMin (TRUE);
  else
    d_plot->grid ()->enableYMin (FALSE);

  d_plot->grid ()->setAxis (grid.xAxis, grid.yAxis);

  if (mrkX < 0 && grid.xZeroOn)
    {
      QwtPlotMarker *m = new QwtPlotMarker ();
      mrkX = d_plot->insertMarker (m);
      m->setAxis (grid.xAxis, grid.yAxis);
      m->setLineStyle (QwtPlotMarker::VLine);
      m->setValue (0.0, 0.0);
      m->setLinePen (QPen (black, 2, SolidLine));
    }
  else if (mrkX >= 0 && !grid.xZeroOn)
    {
      d_plot->removeMarker (mrkX);
      mrkX = -1;
    }

  if (mrkY < 0 && grid.yZeroOn)
    {
      QwtPlotMarker *m = new QwtPlotMarker ();
      mrkY = d_plot->insertMarker (m);
      m->setAxis (grid.xAxis, grid.yAxis);
      m->setLineStyle (QwtPlotMarker::HLine);
      m->setValue (0.0, 0.0);
      m->setLinePen (QPen (black, 2, SolidLine));
    }
  else if (mrkY >= 0 && !grid.yZeroOn)
    {
      d_plot->removeMarker (mrkY);
      mrkY = -1;
    }

  emit modifiedGraph ();
}

gridOptions
Graph::getGridOptions ()
{
  return grid;
}

QStringList
Graph::scalesTitles ()
{
  QStringList scaleTitles;
  int axis;
  for (int i = 0; i < QwtPlot::axisCnt; i++)
    {
      switch (i)
	{
	case 0:
	  axis = 2;
	  scaleTitles << d_plot->axisTitle (axis).text ();
	  break;

	case 1:
	  axis = 0;
	  scaleTitles << d_plot->axisTitle (axis).text ();
	  break;

	case 2:
	  axis = 3;
	  scaleTitles << d_plot->axisTitle (axis).text ();
	  break;

	case 3:
	  axis = 1;
	  scaleTitles << d_plot->axisTitle (axis).text ();
	  break;
	}
    }
  return scaleTitles;
}

void
Graph::updateSecondaryAxis (int axis)
{
  for (int i = 0; i < n_curves; i++)
    {
      QwtPlotCurve *c = this->curve (i);
      if (!c)
	continue;

      if (c->rtti () == QwtPlotItem::Rtti_PlotSpectrogram)
	{
	  Spectrogram *sp = (Spectrogram *) c;
	  if (sp->colorScaleAxis () == axis)
	    return;
	}

      if ((axis == QwtPlot::yRight && c->yAxis () == QwtPlot::yRight) ||
	  (axis == QwtPlot::xTop && c->xAxis () == QwtPlot::xTop))
	return;
    }

  int a = QwtPlot::xBottom;
  if (axis == QwtPlot::yRight)
    a = QwtPlot::yLeft;

  if (!d_plot->axisEnabled (a))
    return;

  QwtScaleEngine *se = d_plot->axisScaleEngine (a);
  const QwtScaleDiv *sd = d_plot->axisScaleDiv (a);

  QwtScaleEngine *sc_engine = 0;
  if (se->transformation ()->type () == QwtScaleTransformation::Log10)
    sc_engine = new QwtLog10ScaleEngine ();
  else if (se->transformation ()->type () == QwtScaleTransformation::Linear)
    sc_engine = new QwtLinearScaleEngine ();

  if (se->testAttribute (QwtScaleEngine::Inverted))
    sc_engine->setAttribute (QwtScaleEngine::Inverted);

  d_plot->setAxisScaleEngine (axis, sc_engine);
  d_plot->setAxisScaleDiv (axis, *sd);

  d_user_step[axis] = d_user_step[a];

  QwtScaleWidget *scale = d_plot->axisWidget (a);
  int start = scale->startBorderDist ();
  int end = scale->endBorderDist ();
  scale = d_plot->axisWidget (axis);
  scale->setMinBorderDist (start, end);
}


void
Graph::setAutoScale ()
{
  //fprintf(stderr,"setAutoScale\n");
  QwtPlotCurve *c = this->curve(0);
  if(!c) return;
  if(c->rtti () == QwtPlotItem::Rtti_PlotSpectrogram)
    {
      Spectrogram *sp = (Spectrogram *) c;
      Matrix *m = sp->matrix();
      double min = QMIN( m->xStart(), m->xEnd() );
      double max = QMAX( m->xStart(), m->xEnd() );
      QwtScaleEngine *se = d_plot->axisScaleEngine (QwtPlot::xBottom);
      if( se->testAttribute (QwtScaleEngine::Inverted) )
	d_plot->setAxisScale (QwtPlot::xBottom, max, min);
      else
	d_plot->setAxisScale (QwtPlot::xBottom, min, max);

      min = QMIN( m->yStart(), m->yEnd() );
      max = QMAX( m->yStart(), m->yEnd() );
      se = d_plot->axisScaleEngine (QwtPlot::yLeft);
      if( se->testAttribute (QwtScaleEngine::Inverted) )
	d_plot->setAxisScale (QwtPlot::yLeft, max, min);
      else
	d_plot->setAxisScale (QwtPlot::yLeft, min, max);

      min = sp->data().range().minValue();
      max = sp->data().range().maxValue();
      d_plot->setAxisScale (QwtPlot::yRight, min, max );
      sp->setLevelRange( min, max );
      sp->setLevelsNumber( sp->levels() );
    }
  else
    {
      for (int a = 0; a < QwtPlot::axisCnt; a++)
	{
	  const QwtScaleEngine *sc_eng = d_plot->axisScaleEngine (a);
	  QwtScaleTransformation *tr = sc_eng->transformation ();
	  if((int) tr->type () == QwtScaleTransformation::Log10)
	    {
	      double start = minPositiveValue(a);
	      double end   = maxPositiveValue(a);
	      start = pow( 10, floor( log10( start ) ) );
	      end = pow( 10, floor( log10( end ) ) + 1 );
	      int major = d_plot->axisMaxMajor (a);
	      int minor = d_plot->axisMaxMinor (a);
	      setScale(a,start,end,0,major,minor,1,false);
	    }
	  else
	    d_plot->setAxisAutoScale (a);
	}
    }

  d_plot->replot(); // this is necessary before setZoomBase()
  d_zoomer[0]->setZoomBase ();
  d_zoomer[1]->setZoomBase ();

  updateScale ();

  emit modifiedGraph ();
}

void
Graph::setAutoscaleFonts(bool yes)
{
  if(yes && !autoScaleFonts)
  {
    QFont font;
    int i;

    double factor = scaleFactor();

    // reset the basefontsize
    for (i = 0; i < (int) d_texts.size (); i++)
    {
      LegendMarker *mrk = (LegendMarker *) d_plot->marker (d_texts[i]);
      mrk->setBaseFontSize(mrk->getFont().pointSizeFloat() / factor); 
    }
    for (i = 0; i < QwtPlot::axisCnt; i++)
    {
      axisFontSize[i] = axisFont(i).pointSizeFloat() / factor;
      axisTitleFontSize[i] = axisTitleFont(i).pointSizeFloat() / factor;
    }
    titleFontSize = d_plot->title().font().pointSizeFloat() / factor;
  }

  autoScaleFonts = yes;
}

void
Graph::setAutoScaleFrameWidth(bool on)
{
  if(autoScaleFrameWidth != on)
    {
      autoScaleFrameWidth = on;
      int width = canvasFrameWidth();
      if(on && width > 0)
	d_frame_width = (double) width / scaleFactor();
    }
}

void
Graph::setAutoScaleGridWidth(bool on)
{
  if(autoScaleGridWidth != on)
    {
      autoScaleGridWidth = on;
      if(on)
	{
	  double factor = scaleFactor();
	  d_grid_major = grid.majorWidth / factor;
	  d_grid_minor = grid.minorWidth / factor;
	}
    }
}

void
Graph::setAutoScaleLineWidth(bool on)
{
  if(autoScaleLineWidth != on)
    {
      autoScaleLineWidth = on;
      if(on)
	{
	  double factor = scaleFactor();
	  for (int i = 0; i < n_curves; i++)
	    {
	      QwtPlotCurve *c = curve (i);
	      int ct = curveType(i);
	      if(ct==Graph::Line || ct==Graph::LineSymbols)
		{
		  QPen pen = c->pen();
		  c_attr.at(i)->line_width = pen.width() / factor;
		}
	      if(ct==Graph::Scatter || ct==Graph::LineSymbols)
		{
		  QwtSymbol s = c->symbol();
		  c_attr.at(i)->symbol_edge_width = s.pen().width() / factor;
		}
	    }
	}
    }
}

void
Graph::setScale (int axis, double start, double end, double step,
		 int majorTicks, int minorTicks, int type, bool inverted)
{
  QwtScaleEngine *sc_engine = 0;

  if (type)
    sc_engine = new QwtLog10ScaleEngine ();
  else
    sc_engine = new QwtLinearScaleEngine ();

  double min = QMIN(start,end);
  double max = QMAX(start,end);
  if (min==max) return;

  QwtPlotCurve *c = this->curve(0);
  if (c && c->rtti () == QwtPlotItem::Rtti_PlotSpectrogram)
    {
      Spectrogram *sp = (Spectrogram *) c;
      if (sp->colorScaleAxis () == axis)
	{
	  sp->setLevelRange( min, max );
	  sp->setLevelsNumber( sp->levels() );
	}
    }

  d_plot->setAxisMaxMajor (axis, majorTicks);
  d_plot->setAxisMaxMinor (axis, minorTicks);

  QwtScaleDiv div =
    sc_engine->divideScale (min, max, majorTicks, minorTicks, step);

  if (inverted)
    {
      sc_engine->setAttribute (QwtScaleEngine::Inverted);
      div.invert ();
    }

  d_plot->setAxisScaleEngine (axis, sc_engine);
  d_plot->setAxisScaleDiv (axis, div);

  d_plot->replot(); // this is necessary before setZoomBase()
  d_zoomer[0]->setZoomBase ();
  d_zoomer[1]->setZoomBase ();

  d_user_step[axis] = step;
}

void
Graph::copyCanvas (bool on)
{
  int w = d_plot->canvas ()->width ();
  int h = d_plot->canvas ()->height ();
  if (on)
    {
      QPixmap pix (w, h, -1);
      //pix = pix.grabWidget (d_plot->canvas (), 0, 0, -1, -1);
      pix = QPixmap::grabWidget (d_plot->canvas (), 0, 0, -1, -1);
      if(pix.isNull()){
	fprintf(stderr,"failed to grabWidget\n");
      }
      if(!QPixmapCache::insert ("field", pix)){
	fprintf(stderr,"failed to insert pixmap cache\n");
      }
    }
}

void
Graph::shiftCurveSelector (int d)
{
  //const QValueList < int >keys = d_plot->curveKeys ();

  int index = 0;
  if (selectedCurve >= 0)
    {
      for (int i = 0; i < n_curves; i++)
	{
	  if (selectedCurve == i)
	    {
	      index = i + d;
	      break;
	    }
	}
    }

  selectedCurve = (index + n_curves ) % n_curves;

  QwtPlotCurve *c = curve (selectedCurve);
  if (!c || c->dataSize () <= 0)
    return;

  int from,to;
  if(getDisplayRange(index,&from,&to))
    {
      startPoint = from;
      endPoint = to;
    }
  else
    {
      startPoint = 0;
      endPoint = c->dataSize () - 1;
    }

  QwtPlotMarker *m = d_plot->marker (startID);
  m->setValue (c->x (startPoint), c->y (startPoint));
  m = d_plot->marker (endID);
  m->setValue (c->x (endPoint), c->y (endPoint));
  d_plot->replot ();

  QString info;
  if (selectedCursor == startID)
    {
      info = tr ("Left") + " <=> ";
      selectedPoint = startPoint;
    }
  else if (selectedCursor == endID)
    {
      info = tr ("Right") + " <=> ";
      selectedPoint = endPoint;
    }

  emit cursorInfo (info,
		   associations[selectedCurve],
		   selectedPoint);

  emit dataRangeChanged ();

  QPainter painter (d_plot->canvas ());
  painter.setClipping (TRUE);
  painter.setClipRect (d_plot->canvas ()->contentsRect ());
  painter.setRasterOp (Qt::NotROP);

  c->draw (&painter,
	   d_plot->canvasMap (c->xAxis ()),
	   d_plot->canvasMap (c->yAxis ()), selectedPoint,
	   selectedPoint);
}

void
Graph::shiftRangeSelector (bool shift)
{
  QwtPlotCurve *c = curve (selectedCurve);
  if (!c || c->dataSize () <= 0 )
    return;

  QString info;
  if (shift)
    {
      QwtPlotMarker *m = d_plot->marker (selectedCursor);
      QwtSymbol symbol = m->symbol ();
      if (selectedCursor == startID)
	{
	  symbol.setPen (QPen (black, 2, Qt::SolidLine));
	  m->setSymbol (symbol);
	  m->setLinePen (QPen (black, 1, Qt::DashLine));

	  selectedCursor = endID;
	  m = d_plot->marker (selectedCursor);
	  symbol = m->symbol ();
	  symbol.setPen (QPen (red, 2, Qt::SolidLine));
	  m->setSymbol (symbol);
	  m->setLinePen (QPen (red, 1, Qt::DashLine));

	  selectedPoint = endPoint;
	  info = tr ("Right") + " <=> ";
	}
      else if (selectedCursor == endID)
	{
	  symbol.setPen (QPen (black, 2, Qt::SolidLine));
	  m->setSymbol (symbol);
	  m->setLinePen (QPen (black, 1, Qt::DashLine));

	  selectedCursor = startID;
	  m = d_plot->marker (selectedCursor);
	  symbol = m->symbol ();
	  symbol.setPen (QPen (red, 2, Qt::SolidLine));
	  m->setSymbol (symbol);
	  m->setLinePen (QPen (red, 1, Qt::DashLine));

	  selectedPoint = startPoint;
	  info = tr ("Left") + " <=> ";
	}
      d_plot->replot ();

      emit cursorInfo (info,
		       associations[selectedCurve],
		       selectedPoint);

      QPainter painter (d_plot->canvas ());
      painter.setClipping (TRUE);
      painter.setClipRect (d_plot->canvas ()->contentsRect ());
      painter.setRasterOp (Qt::NotROP);

      c->draw (&painter, d_plot->canvasMap (c->xAxis ()),
	       d_plot->canvasMap (c->yAxis ()), selectedPoint,
	       selectedPoint);
    }
}

void
Graph::moveRangeSelector (int d)
{
  QwtPlotCurve *c = curve (selectedCurve);
  if (!c)
    return;

  int points = c->dataSize ();
  if (points == 0)
    {
      QMessageBox::critical (this, tr ("QtiPlot - Warning"),
			     tr ("All the curves on this plot are empty!"));
      disableRangeSelectors ();
      return;
    }

  int index = selectedPoint + d;
  index = (index + points) % points;

  QString info;
  if (index != selectedPoint)
    selectedPoint = index;

  QwtPlotMarker *m = d_plot->marker (selectedCursor);
  m->setValue (c->x (selectedPoint), c->y (selectedPoint));
  d_plot->replot ();

  if (selectedCursor == startID)
    {
      startPoint = selectedPoint;
      info = tr ("Left") + " <=> ";
    }
  else if (selectedCursor == endID)
    {
      endPoint = selectedPoint;
      info = tr ("Right") + " <=> ";
    }

  emit cursorInfo (info,
		   associations[selectedCurve],
		   selectedPoint);

  emit dataRangeChanged ();

  QPainter painter (d_plot->canvas ());
  painter.setClipping (TRUE);
  painter.setClipRect (d_plot->canvas ()->contentsRect ());
  painter.setRasterOp (Qt::NotROP);

  c->draw (&painter, d_plot->canvasMap (c->xAxis ()),
	   d_plot->canvasMap (c->yAxis ()), selectedPoint,
	   selectedPoint);
}

bool
Graph::enableRangeSelectors (bool on)
{
  rangeSelectorsEnabled = on;

  int curveID;
  QwtPlotCurve *c = 0;
  bool success = FALSE;
  int from,to;
  for (int i = 0; i < n_curves; i++)
    {
      curveID = i;
      c = curve (i);
      if (c && c->rtti () == QwtPlotItem::Rtti_PlotCurve
	  && getDisplayRange(i, &from, &to) )
	{
	  //fprintf(stderr,"index = %d\n",i);
	  //fprintf(stderr,"from = %d  to = %d\n",from,to);
	  success = TRUE;
	  break;
	}
    }

  if (!success)
    {
      QMessageBox::critical (this, tr ("QtiPlot - Warning"),
			     tr ("All the curves on this plot are empty!"));

      disableRangeSelectors ();
      return FALSE;
    }

  startPoint = from;

  if (on)
    {
      QCursor cursor = QCursor (QPixmap (vizor_xpm), -1, -1);
      d_plot->canvas ()->setCursor (cursor);

      endPoint = to;
      selectedCurve = curveID;
      selectedPoint = from;
      d_plot->canvas ()->setFocus ();

      int d = 32;
      c = curve (selectedCurve);
      QwtSymbol symbol = c->symbol ();
      if (symbol.style () != QwtSymbol::NoSymbol)
	{
	  QSize sz = symbol.size ();
	  d += QMAX (sz.width (), sz.height ());
	}

      QwtPlotMarker *m = new QwtPlotMarker ();
      m->
	setSymbol (QwtSymbol
		   (QwtSymbol::Cross, QBrush (Qt::NoBrush), QPen (red, 2),
		    QSize (d, d)));
      m->setLineStyle (QwtPlotMarker::VLine);
      m->setLinePen (QPen (red, 1, Qt::DashLine));
      m->setValue (c->x (startPoint), c->y (startPoint));
      startID = d_plot->insertMarker (m);

      selectedCursor = startID;

      m = new QwtPlotMarker ();
      m->setLineStyle (QwtPlotMarker::VLine);
      m->setLinePen (QPen (black, 1, Qt::DashLine));
      m->
	setSymbol (QwtSymbol
		   (QwtSymbol::Cross, QBrush (Qt::NoBrush), QPen (black, 2),
		    QSize (d, d)));
      m->setValue (c->x (endPoint), c->y (endPoint));
      endID = d_plot->insertMarker (m);

      d_plot->replot ();

      QPainter painter (d_plot->canvas ());

      painter.setClipping (TRUE);
      painter.setClipRect (d_plot->canvas ()->contentsRect ());
      painter.setRasterOp (Qt::NotROP);

      c->draw (&painter,
	       d_plot->canvasMap (c->xAxis ()),
	       d_plot->canvasMap (c->yAxis ()), 0, 0);

      emit dataRangeChanged ();
    }
  else
    disableRangeSelectors ();

  return TRUE;
}

void
Graph::disableRangeSelectors ()
{
  d_plot->removeMarker (startID);
  d_plot->removeMarker (endID);
  d_plot->replot ();
  startID = -1;
  endID = -1;
  startPoint = 0;
  endPoint = -1;
  selectedCurve = -1;
  selectedPoint = -1;
  selectedCursor = -1;
  rangeSelectorsEnabled = FALSE;
  d_plot->canvas ()->setCursor (Qt::arrowCursor);
}

void
Graph::showPlotPicker (bool on)
{
  pickerEnabled = on;
  QCursor cursor = QCursor (QPixmap (cursor_xpm), -1, -1);
  if (on)
    {
      d_plot->canvas ()->setCursor (cursor);
      d_pickerMarker = false;
    }
  else
    {
      d_plot->canvas ()->setCursor (Qt::arrowCursor);
      d_plot->replot();
    }
}

void
Graph::startCurveTranslation ()
{
  showPlotPicker (true);
  emit showInfo (tr
		 ("Curve selected! Move cursor and click to choose a point and double-click/press 'Enter' to finish!"));
}

void
Graph::insertPlottedList (const QStringList & names)
{
  for (int i = 0; i < n_curves; i++)
    {
      QwtPlotItem *c = d_plot->curve( curveKey(i) );
      //if (c)
      if (c && c->rtti () == QwtPlotItem::Rtti_PlotCurve)
	c->setTitle (names[i]);
    }
}

QStringList
Graph::curvesList ()
{
  QStringList cList;
  for (int i = 0; i < n_curves; i++)
    {
      QwtPlotItem *c = d_plot->curve( curveKey(i) );
      if (c && c->rtti () == QwtPlotItem::Rtti_PlotCurve)
	cList << c->title ().text ();
    }
  return cList;
}

QStringList
Graph::analyzableCurvesList()
{
  QStringList cList;
  for (int i = 0; i < n_curves; i++)
    {
      QwtPlotItem *c = d_plot->curve( curveKey(i) );
      int t = curveType(i);
      if (c && c->rtti () == QwtPlotItem::Rtti_PlotCurve
	  && t != Pie && t != Histogram && t != XErrorBars && t != YErrorBars 
	  && t != VectXYXY && t != VectXYAM && t != ColorMap
	  && t != GrayMap && t != ContourMap)
	cList << c->title ().text ();
    }
  return cList;
}

QStringList
Graph::xyCurvesList()
{
  QStringList cList;
  for (int i = 0; i < n_curves; i++)
    {
      QwtPlotItem *c = d_plot->curve( curveKey(i) );
      int t = curveType(i);
      if (c && (c->rtti () == QwtPlotItem::Rtti_PlotCurve
		|| c->rtti() == FunctionCurve::RTTI)
	  && t != Pie && t != Histogram && t != XErrorBars && t != YErrorBars 
	  && t != VectXYXY && t != VectXYAM && t != ColorMap
	  && t != GrayMap && t != ContourMap)
	cList << c->title ().text ();
    }
  return cList;
}

QStringList
Graph::plotItemsList ()
{
  QStringList cList;
  for (int i = 0; i < n_curves; i++)
    {
      QwtPlotItem *c = d_plot->curve ( curveKey(i) );
      if (c)
	cList << c->title ().text ();
    }
  return cList;
}

QStringList
Graph::plotAssociations ()
{
  return associations;
}

void
Graph::setPlotAssociations (const QStringList & newList)
{
  associations = newList;
}

void
Graph::copyImage ()
{
  QPixmap pic = graphPixmap ();
  QImage image = pic.convertToImage ();
  QApplication::clipboard ()->setData (new QImageDrag (image, d_plot, 0));
}

QPixmap
Graph::graphPixmap ()
{
  return QPixmap::grabWidget (this);
/*
#ifdef Q_OS_MAC			// Mac
  return QPixmap::grabWidget (this);
#else
  int lw = d_plot->lineWidth ();
  QPixmap pic (d_plot->width () + 2 * lw, d_plot->height () + 2 * lw, -1,
	       QPixmap::BestOptim);
  pic.fill (QColor (255, 255, 255));
  QPainter paint;
  paint.begin (&pic);

  QRect rect = QRect (lw, lw, d_plot->width (), d_plot->height ());
  QwtPlotPrintFilter filter;
  filter.
    setOptions (QwtPlotPrintFilter::PrintAll | QwtPlotPrintFilter::
		PrintTitle | QwtPlotPrintFilter::PrintCanvasBackground);
  d_plot->print (&paint, rect, filter);
  activateLayout();

  paint.end ();
  return pic;
#endif
  */
}

void
Graph::exportImage (const QString & fileName, const QString & fileType,
		    int quality, bool transparent)
{
  QPixmap pic = graphPixmap ();

  if (transparent)
    {				//save transparency
      QBitmap mask (pic.size ());
      mask.fill (Qt::color1);
      QPainter p;
      p.begin (&mask);
      p.setPen (Qt::color0);

      QColor background = QColor (QColor (255, 255, 255));
      QRgb backgroundPixel = background.rgb ();
      QImage image = pic.convertToImage ();
      for (int y = 0; y < image.height (); y++)
	{
	  for (int x = 0; x < image.width (); x++)
	    {
	      QRgb rgb = image.pixel (x, y);
	      if (rgb == backgroundPixel)	// we want the frame transparent
		{
		  p.drawPoint (x, y);
		}
	    }
	}
      p.end ();
      pic.setMask (mask);
    }

  pic.save (fileName, fileType, quality);
}

void
Graph::exportToEPS (const QString & fname)
{
  exportToEPS(fname,0,QPrinter::Portrait,QPrinter::A4,QPrinter::Color); 
}

void
Graph::exportToEPS (const QString & fname, int res, QPrinter::Orientation o,
		    QPrinter::PageSize size, QPrinter::ColorMode col)
{
  QPrinter printer;
  if(res) printer.setResolution (res);
  printer.setPageSize (size);
  printer.setColorMode (col);
  printer.setOrientation (o);
  printer.setFullPage (TRUE);
  printer.setOutputToFile (TRUE);
  printer.setOutputFileName (fname);

  QwtPlotPrintFilter filter;
  int option = QwtPlotPrintFilter::PrintAll;
  if( canvasFrameWidth() > 0 )
    option |= QwtPlotPrintFilter::PrintFrameWithScales;
  filter.setOptions( option );

  QPaintDeviceMetrics mpr (&printer);
  int width = frameGeometry ().width ();
  int height = frameGeometry ().height ();
  int x = (mpr.width() - width) / 2;
  int y = (mpr.height() - height) / 2;

  if(x<0 || y<0)
    {
      double aspect = double (width)/ double(height);
      height = mpr.height();
      width = qRound(aspect * height);
      x = (mpr.width() - width) / 2;
      y = 0;
      if (x < 0)
	{
	  width = mpr.width();
	  height = qRound( width / aspect );
	  x = 0;
	  y = (mpr.height() - height) / 2;
	}
    }

  QPainter paint (&printer);
  QRect rect =  QRect(x, y, width, height);
  paint.setClipRect( rect, QPainter::CoordPainter ); // bounding box
  d_plot->print (&paint, rect, filter);
  activateLayout();
}

void
Graph::print(const QString & orientation)
{
  QPrinter printer;
  printer.setColorMode (QPrinter::Color);
  printer.setFullPage (TRUE);

  //printing should preserve plot aspect ratio, if possible
  int w = frameGeometry().width();
  int h = frameGeometry().height();
  double aspect = (double) w / h;

  if(orientation=="Portrait")
    printer.setOrientation (QPrinter::Portrait);
  else if(orientation=="Landscape")
    printer.setOrientation (QPrinter::Landscape);
  else if (aspect < 1)
    printer.setOrientation (QPrinter::Portrait);
  else
    printer.setOrientation (QPrinter::Landscape);

  if (printer.setup ())
    {
      QPaintDeviceMetrics mpr (&printer);
      int dpiy = mpr.logicalDpiY ();
      int margin = (int) ((1 / 2.54) * dpiy);	// 1 cm margins

      int height =  mpr.height () - 2 * margin;
      int width = qRound (aspect * height);
      int x = qRound( (mpr.width () - width) * 0.5); // centering
      int y = margin;
      if (x < margin)
	{
	  width = mpr.width() - 2 * margin;
	  height = qRound( (double)width / aspect ); //centering
	  x = margin;
	  y = qRound( (mpr.height() - height) * 0.5 );
	}

      double scale = sqrt( (double) width*height/ (500*450) );

      saveComponentSizes();

      scaleFonts( scale );
      scaleSymbols( scale );
      scaleSpaces( scale );
      scaleTickLength( scale );
      scaleFrame( scale );
      scaleGrid( scale );

      QRect rect (x, y, width, height);
      QPainter paint (&printer);

      QwtPlotPrintFilter filter;
      int option = QwtPlotPrintFilter::PrintAll;
      if( canvasFrameWidth() > 0 )
	option |= QwtPlotPrintFilter::PrintFrameWithScales;
      filter.setOptions( option );

      d_plot->print (&paint, rect, filter);
      activateLayout();


      scale = sqrt((double)w*h/(500*450));
      scaleSpaces( scale );
      scaleTickLength( scale );
      //scaleFonts( scale );
      //scaleSymbols( scale );
      //scaleFrame( scale );
      //scaleGrid( scale );

      restoreComponentSizes();

      d_plot->replot();
    }
}

void
Graph::exportToSVG (const QString & fname)
{
// enable workaround for Qt3 misalignments
  QwtPainter::setSVGMode (true);

  QPicture picture;
  QPainter p (&picture);
  QRect rect =  d_plot->rect ();
  d_plot->print (&p, rect);
  activateLayout();
  p.end ();

  picture.save (fname, "svg");
}

void
Graph::movePoints (bool enabled)
{
  movePointsEnabled = enabled;
  if (enabled)
    d_plot->canvas ()->setCursor (Qt::pointingHandCursor);
  else
    d_plot->canvas ()->setCursor (Qt::arrowCursor);
}

bool
Graph::movePointsActivated ()
{
  return movePointsEnabled;
}

void
Graph::removePoints (bool enabled)
{
  removePointsEnabled = enabled;
  QCursor cursor = QCursor (QPixmap (vizor_xpm), -1, -1);

  if (enabled)
    d_plot->canvas ()->setCursor (cursor);
  else
    d_plot->canvas ()->setCursor (Qt::arrowCursor);
}

bool
Graph::removePointActivated ()
{
  return removePointsEnabled;
}

void
Graph::translateCurve (int direction)
{
  translateOn = true;
  translationDirection = direction;
  d_plot->canvas ()->setCursor (QCursor (QPixmap (vizor_xpm), -1, -1));
}

void
Graph::translateCurveTo (const QPoint & p)
{
  QwtPlotCurve *c = curve (selectedCurve);
  if (!c)
    return;

  QStringList ass =
    QStringList::split (',', associations[selectedCurve], false);

  int n = c->dataSize ();
  QMemArray <double> dat(n);
  if (translationDirection)
    {
      QwtScaleEngine *se = d_plot->axisScaleEngine( c->yAxis() );
      if(se->transformation()->type() == QwtScaleTransformation::Log10)
	{
	  double factor =
	    d_plot->invTransform (c->yAxis(), p.y ()) / c->y (selectedPoint);
	  for (int i = 0; i < n; i++)
	    dat[i] = c->y (i) * factor;
	}
      else
	{
	  double dy =
	    d_plot->invTransform (c->yAxis(), p.y ()) - c->y (selectedPoint);
	  for (int i = 0; i < n; i++)
	    dat[i] = c->y (i) + dy;
	}
      emit updateTableColumn (ass[1], dat);
    }
  else
    {
      QwtScaleEngine *se = d_plot->axisScaleEngine( c->xAxis() );
      if(se->transformation()->type() == QwtScaleTransformation::Log10)
	{
	  double factor =
	    d_plot->invTransform (c->xAxis(), p.x ()) / c->x(selectedPoint);
	  for (int i = 0; i < n; i++)
	    dat[i] = c->x (i) * factor;
	}
      else
	{
	  double dx =
	    d_plot->invTransform (c->xAxis(), p.x ()) - c->x(selectedPoint);
	  for (int i = 0; i < n; i++)
	    dat[i] = c->x (i) + dx;
	}
      emit updateTableColumn (ass[0], dat);
    }
  showPlotPicker (false);
  translateOn = false;

//no need for setCurveData() or replot(), since the main application will call
//updateCurves() which does this job!
}

QString
Graph::selectedCurveTitle ()
{
  const QwtPlotItem *c = curve (selectedCurve);
  if (c)
    return c->title ().text ();
  else
    return QString::null;
}

void
Graph::multiPeakFit (ApplicationWindow * app, int profile, int peaks)
{
  showPlotPicker (true);
  selected_peaks = 0;
  fitter =
    new MultiPeakFit (app, this, (MultiPeakFit::PeakProfile) profile, peaks);
  fitter->enablePeakCurves (app->generatePeakCurves);
  fitter->setPeakCurvesColor (app->peakCurvesColor);
  fitter->setGenerateUniformFitPoints (app->generateUniformFitPoints);
  fitter->setGenerateFunction (app->generateFunction);
  fitter->setFitPoints( app->fitPoints );
  QString name;
  if(app->generateFunction)
    name = this->generateUniqueName( QString(fitter->name()) + tr("Fit") );
  else
    name = app->generateUniqueName( QString(fitter->name()) + tr("Fit") );
  fitter->setCurveName(name);
  d_plot->canvas ()->grabMouse ();
}

bool
Graph::selectPeaksOn ()
{
  if (fitter)
    return true;
  else
    return false;
}

void
Graph::selectPeak (const QPoint &)
{
  const QwtPlotCurve *c = curve (selectedCurve);
  if (!c)
    return;

  //double width = 1.0;
  //double width = fabs(c->maxXValue() - c->minXValue())/10.0;

  const QwtScaleDiv *scDiv = d_plot->axisScaleDiv (c->xAxis());
  double width = fabs( scDiv->hBound() - scDiv->lBound() )/50.0;
  //fprintf(stderr,"width = %f\n",width);
  fitter->setInitialGuess (3 * selected_peaks, c->y(selectedPoint) * width);
  fitter->setInitialGuess (3 * selected_peaks + 1, c->x(selectedPoint));
  fitter->setInitialGuess (3 * selected_peaks + 2, width);

  QwtPlotMarker *m = new QwtPlotMarker ();
  m->setLineStyle (QwtPlotMarker::VLine);
  m->setLinePen (QPen (green, 2, Qt::DashLine));

  double xv = c->x(selectedPoint);
  if( c->xAxis() == QwtPlot::xTop )
    {
      int x = d_plot->transform( c->xAxis(), xv );
      xv = d_plot->invTransform( QwtPlot::xBottom, x );
    }
  m->setXValue (xv);
  d_plot->insertMarker (m);
  d_plot->replot ();

  selected_peaks++;
  int peaks = fitter->peaks ();
  if (selected_peaks == peaks)
    {
      showPlotPicker (false);
      d_plot->canvas ()->releaseMouse ();

      if (fitter->setDataFromCurve (c->title ().text ()))
	{
	  QApplication::setOverrideCursor (waitCursor);
	  fitter->fit ();
	  delete fitter;
	  QApplication::restoreOverrideCursor ();
	}

      //remove peak line markers
      QValueList < int >mrks = d_plot->markerKeys ();
      int n = (int) mrks.count ();
      for (int i = 0; i < peaks; i++)
	d_plot->removeMarker (mrks[n - i - 1]);

      d_plot->replot ();
      return;
    }
  emit
    showInfo (tr
		("Peak %1 selected! Click to select a point and double-click/press 'Enter' to set the position of the next peak!").
		arg (QString::number (selected_peaks)));
}

bool
Graph::selectPoint (const QPoint & pos)
{
  int dist, point;
  int key = d_plot->closestCurve (pos.x (), pos.y (), dist, point);
  int index = curveIndex(key);
  
  if (index >= 0 && dist < 10)	//10 pixels tolerance
    {
      const QwtPlotCurve *c = curve (index);
      if (!c){
	qDebug("Graph::selectPoint : curve does not exist");
	return false;
      }

      if (rangeSelectorsEnabled && index != selectedCurve)
	{
	  int from,to;
	  if(!getDisplayRange(index, &from, &to))
	    { // This shouldn't happen
	      from = 0;
	      to = c->dataSize() - 1;
	    }
	  if (selectedCursor == startID)
	    {
	      startPoint = point;
	      endPoint = to;
	      if (point == to)
		endPoint = 0;
	    }
	  else if (selectedCursor == endID)
	    {
	      startPoint = from;
	      endPoint = point;
	      if (point==from)
		startPoint = to;
	    }

	  QwtPlotMarker *m = d_plot->marker (startID);
	  if (m)
	    m->setValue (c->x (startPoint), c->y (startPoint));

	  m = d_plot->marker (endID);
	  if (m)
	    m->setValue (c->x (endPoint), c->y (endPoint));
	}

      selectedCurve = index;
      selectedPoint = point;
      highlightPoint (TRUE);

      emit cursorInfo (QString::null,
		       associations[selectedCurve],
		       selectedPoint);

      return true;
    }
  else if (!rangeSelectorsEnabled && !translateOn)
    {				// deselect
      selectedCurve = -1;
      selectedPoint = -1;
      highlightPoint (FALSE);
      return false;
    }
  return false;
}

void
Graph::selectCurve (const QPoint & pos)
{
  int dist, point;
  int key = d_plot->closestCurve (pos.x (), pos.y (), dist, point);
  int index = curveIndex(key);
  if (index >= 0 && dist < 10)	// 10 pixels tolerance
    {
      selectedCurve = index;
      selectedPoint = point;
      showCursor (TRUE);

      emit cursorInfo (QString::null,
		       associations[selectedCurve],
		       point);
    }
  else				// deselect
    {
      showCursor (FALSE);
      selectedCurve = -1;
      selectedPoint = -1;
    }
}

bool
Graph::markerSelected ()
{
  bool selected = FALSE;
  if (selectedMarker >= 0)
    selected = TRUE;
  return selected;
}

void
Graph::removeMarker ()
{
  //fprintf(stderr,"removeMarker\n");
  if (selectedMarker >= 0)
    {
      d_plot->removeMarker (selectedMarker);
      d_plot->replot ();
      emit modifiedGraph ();

      if (selectedMarker == legendMarkerID)
	legendMarkerID = -1;

      if (d_lines.contains (selectedMarker) > 0)
	{
	  int index = d_lines.find (selectedMarker, 0);
	  int linesOnPlot = (int) d_lines.size ();
	  for (int i = index; i < linesOnPlot-1; i++)
	    d_lines[i] = d_lines[i + 1];
	  d_lines.resize (--linesOnPlot);
	}
      else if (d_texts.contains (selectedMarker) > 0)
	{
	  int index = d_texts.find (selectedMarker, 0);
	  int texts = d_texts.size ();
	  for (int i = index; i < texts-1; i++)
	    d_texts[i] = d_texts[i + 1];
	  d_texts.resize (--texts);
	}
      else if (d_images.contains (selectedMarker) > 0)
	{
	  int index = d_images.find (selectedMarker, 0);
	  int imagesOnPlot = d_images.size ();
	  for (int i = index; i < imagesOnPlot-1; i++)
	    d_images[i] = d_images[i + 1];
	  d_images.resize (--imagesOnPlot);
	}
      selectedMarker = -1;
    }
}

void
Graph::cutMarker ()
{
  copyMarker ();
  removeMarker ();
}

bool
Graph::arrowMarkerSelected ()
{
  bool arrow = FALSE;
  if (d_lines.contains (selectedMarker) > 0)
    arrow = TRUE;
  return arrow;
}

bool
Graph::imageMarkerSelected ()
{
  bool image = FALSE;
  if (d_images.contains (selectedMarker) > 0)
    image = TRUE;
  return image;
}

void
Graph::copyMarker ()
{
  if (selectedMarker < 0)
    {
      selectedMarkerType = None;
      return;
    }

  if (d_lines.contains (selectedMarker))
    {
      LineMarker *mrkL = (LineMarker *) d_plot->marker (selectedMarker);
      auxMrkStart = mrkL->startPoint ();
      auxMrkEnd = mrkL->endPoint ();
      selectedMarkerType = Arrow;
    }
  else if (d_images.contains (selectedMarker))
    {
      ImageMarker *mrkI = (ImageMarker *) d_plot->marker (selectedMarker);
      auxMrkStart = mrkI->getOrigin ();
      QRect rect = mrkI->rect ();
      auxMrkEnd = rect.bottomRight ();
      auxMrkFileName = mrkI->getFileName ();
      selectedMarkerType = Image;
    }
  else
    selectedMarkerType = Text;
}

void
Graph::pasteMarker ()
{
  if (selectedMarkerType == Arrow)
    {
      LineMarker *mrkL = new LineMarker ();
      int linesOnPlot = (int) d_lines.size ();
      d_lines.resize (++linesOnPlot);
      d_lines[linesOnPlot - 1] = d_plot->insertMarker (mrkL);

      mrkL->setColor (auxMrkColor);
      mrkL->setWidth (auxMrkWidth);
      mrkL->setStyle (auxMrkStyle);
      mrkL->
	setStartPoint (QPoint (auxMrkStart.x () + 10, auxMrkStart.y () + 10));
      mrkL->setEndPoint (QPoint (auxMrkEnd.x () + 10, auxMrkEnd.y () + 10));
      mrkL->drawStartArrow (startArrowOn);
      mrkL->drawEndArrow (endArrowOn);
      mrkL->setHeadLength (auxArrowHeadLength);
      mrkL->setHeadAngle (auxArrowHeadAngle);
      mrkL->fillArrowHead (auxFilledArrowHead);

      d_plot->replot ();

      selectedMarker = -1;
    }
  else if (selectedMarkerType == Image)
    {
      QString fn = auxMrkFileName;
      QRect rect = QRect (auxMrkStart, auxMrkEnd);

      QFileInfo fi (fn);
      QString baseName = fi.fileName ();
      QString type = baseName.section('.',-1);

      QPixmap photo;
      if (type.upper () == "JPG")
	photo.load (fn, "JPEG", QPixmap::Auto);
      else
	photo.load (fn, type.upper (), QPixmap::Auto);

      ImageMarker *mrk = new ImageMarker (photo);
      int imagesOnPlot = d_images.size ();
      d_images.resize (++imagesOnPlot);
      d_images[imagesOnPlot - 1] = d_plot->insertMarker (mrk);
      mrk->setFileName (fn);

      QPoint o = d_plot->canvas ()->mapFromGlobal (QCursor::pos ());
      if (!d_plot->canvas ()->contentsRect ().contains (o))
	o = QPoint (auxMrkStart.x () + 20, auxMrkStart.y () + 20);

      mrk->setOrigin (o);
      mrk->setSize (rect.size ());
      d_plot->replot ();
      selectedMarker = -1;
    }
  else
    {
      LegendMarker *mrk = new LegendMarker (d_plot);
      int texts = d_texts.size ();
      d_texts.resize (++texts);
      d_texts[texts - 1] = d_plot->insertMarker (mrk);

      QPoint o = d_plot->canvas ()->mapFromGlobal (QCursor::pos ());
      if (!d_plot->canvas ()->contentsRect ().contains (o))
	o = QPoint (auxMrkStart.x () + 20, auxMrkStart.y () + 20);

      mrk->setOrigin (o);
      mrk->setAngle (auxMrkAngle);
      mrk->setBackground (auxMrkBkg);
      mrk->setFont (auxMrkFont, this->frameGeometry());
      mrk->setText (auxMrkText);
      mrk->setTextColor (auxMrkColor);
      mrk->setBackgroundColor (auxMrkBkgColor);

      d_plot->replot ();
      selectedMarker = -1;
    }
}

void
Graph::setCopiedMarkerEnds (const QPoint & start, const QPoint & end)
{
  auxMrkStart = start;
  auxMrkEnd = end;
}

void
Graph::setCopiedTextOptions (int bkg, const QString & text,
			     const QFont & font, const QColor & color,
			     const QColor & bkgColor)
{
  auxMrkBkg = bkg;
  auxMrkText = text;
  auxMrkFont = font;
  auxMrkColor = color;
  auxMrkBkgColor = bkgColor;
}

void
Graph::setCopiedArrowOptions (int width, Qt::PenStyle style,
			      const QColor & color, bool start, bool end,
			      int headLength, int headAngle, bool filledHead)
{
  auxMrkWidth = width;
  auxMrkStyle = style;
  auxMrkColor = color;
  startArrowOn = start;
  endArrowOn = end;
  auxArrowHeadLength = headLength;
  auxArrowHeadAngle = headAngle;
  auxFilledArrowHead = filledHead;
}

bool
Graph::titleSelected ()
{
  return d_plot->titleLabel ()->hasFocus ();
}

void
Graph::selectTitle ()
{
  if (!d_plot->hasFocus ())
    {
      emit selectedGraph (this);
      QwtTextLabel *title = d_plot->titleLabel ();
      title->setFocus ();
    }

  selectedMarker = -1;
}

void
Graph::setTitle (const QString & t)
{
  d_plot->setTitle (t);
  emit modifiedGraph ();
}

void
Graph::removeTitle ()
{
  if (d_plot->titleLabel ()->hasFocus ())
    {
      /*
      QRect rect = d_plot->contentsRect ();

      QwtPlotLayout *layout = (QwtPlotLayout *) d_plot->plotLayout ();
      int y = 0;
      if (d_plot->axisEnabled (QwtPlot::xTop))
	y = layout->scaleRect (QwtPlot::xTop).y ();
      else
	y = layout->canvasRect ().y ();

      rect.setY (y);
      */
      d_plot->setTitle (QString::null);
      //d_plot->resize (rect.size ());
      //setGeometry (rect);

      emit modifiedGraph ();
    }
}

void
Graph::initTitle (bool on, const QFont & fnt)
{
  if (on)
    {
      QwtText t = d_plot->title ();
      t.setFont (fnt);


      t.setText (tr ("Title"));
      d_plot->setTitle (t);
    }
  titleFontSize = fnt.pointSizeFloat() / scaleFactor();
}

void
Graph::removeLegend ()
{
  //fprintf(stderr,"removeLegend\n");
  if (legendMarkerID >= 0)
    {
      int index = d_texts.find (legendMarkerID, 0);
      int texts = d_texts.size ();
      for (int i = index; i < texts-1; i++)
	d_texts[i] = d_texts[i + 1];
      d_texts.resize (--texts);

      d_plot->removeMarker (legendMarkerID);
      legendMarkerID = -1;
    }
}

void
Graph::updateImageMarker (int x, int y, int w, int h)
{
  ImageMarker *mrk = (ImageMarker *) d_plot->marker (selectedMarker);
  mrk->setRect (x, y, w, h);
  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::updateTextMarker (const QString & text, int angle, int bkg,
			 const QFont & fnt, const QColor & textColor,
			 const QColor & backgroundColor)
{
  LegendMarker *mrkL = (LegendMarker *) d_plot->marker (selectedMarker);
  mrkL->setText (text);
  mrkL->setAngle (angle);
  mrkL->setTextColor (textColor);
  mrkL->setBackgroundColor (backgroundColor);
  mrkL->setFont (fnt, this->frameGeometry());
  mrkL->setBackground (bkg);

  if(d_textlayer)
    {
      bool wch=false, hch=false;
      QRect mr = mrkL->rect();
      QRect geom = this->geometry();
      if(mr.right()+10 > geom.width())
	{
	  geom.setWidth( mr.right()+10 );
	  wch = true;
	}
      if(mr.bottom()+10 > geom.height() )
	{
	  geom.setHeight( mr.bottom()+10 );
	  hch = true;
	}
      if( wch || hch )
	{
	  bool dummy = ignoreResize;
	  setIgnoreResizeEvents (true);
	  int xp = d_plot->transform( QwtPlot::xBottom, mrkL->xValue() );
	  int yp = d_plot->transform( QwtPlot::yLeft, mrkL->yValue() );
	  setGeometry( geom );
	  d_plot->resize ( geom.size() );
	  mrkL->setOrigin( QPoint( xp, yp ) );
	  int w = ((QWidget *)parent())->width();
	  int h = ((QWidget *)parent())->height();
	  setRelativeGeometry((double)geom.x()/w,(double)geom.y()/h,
			      (double)geom.width()/w,(double)geom.height()/h );
	  mrkL->setFont (fnt, this->frameGeometry());
	  setIgnoreResizeEvents (dummy);
	}
    }

  d_plot->replot ();
  highlightTextMarker(selectedMarker); // keep selected
  emit modifiedGraph ();
}

LegendMarker *
Graph::legend ()
{
  if (legendMarkerID >= 0)
    return (LegendMarker *) d_plot->marker (legendMarkerID);
  else
    return 0;
}

QString
Graph::legendText ()
{
  QString text = "";
  for (int i = 0; i < n_curves; i++)
    {
      const QwtPlotCurve *c = curve (i);
      int ct = curveType(i);
      if (c && c->rtti () != QwtPlotItem::Rtti_PlotSpectrogram
	  && ct != XErrorBars && ct != YErrorBars)
	{
	  text += "\\c{";
	  text += QString::number (i + 1);
	  text += "}";
	  text += c->title ().text ();
	  text += "\n";
	}
    }
  return text;
}

QString
Graph::pieLegendText ()
{
  QString text = "";
  //QValueList < int >keys = d_plot->curveKeys ();
  const QwtPlotCurve *c = curve( 0 );
  if (c)
    {
      for (int i = 0; i < int (c->dataSize ()); i++)
	{
	  text += "\\p{";
	  text += QString::number (i + 1);
	  text += "} ";
	  text += QString::number (i + 1);
	  text += "\n";
	}
    }
  return text;
}

void
Graph::updateLegend()
{
  LegendMarker *leg = legend();
  if(!leg) return;
  if (piePlot)
    leg->setText (pieLegendText ());
  else
    leg->setText (legendText ());

  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::moveBy (int dx, int dy)
{
  if (dx == 0 && dy == 0)
    return;

  const QwtPlotCurve *c = curve (selectedCurve);
  if (!c)
    return;

  const int x = d_plot->transform (c->xAxis (),
				   c->x (selectedPoint)) + dx;
  const int y = d_plot->transform (c->yAxis (),
				   c->y (selectedPoint)) + dy;

  move (QPoint (x, y));
}

void
Graph::removePoint ()
{
  QwtPlotCurve *c = curve (selectedCurve);
  if (!c)
    return;

  if (c->rtti () == FunctionCurve::RTTI)
    {
      QMessageBox::critical (0, tr ("QtiPlot - Remove point error"),
			     tr
			     ("This function is not available for function curves!"));
      return;
    }

  QString assoc = associations[selectedCurve];
  emit clearCell( assoc, selectedPoint );

  this->setFocus ();
  selectedPoint = -1;
  selectedCurve = -1;
}

// Move the selected point
void
Graph::move (const QPoint & pos)
{
  QwtPlotCurve *c = curve (selectedCurve);
  if (!c)
    return;

  if (c->rtti () == FunctionCurve::RTTI)
    {
      QMessageBox::critical (0, tr ("QtiPlot - Move point error"),
			     tr
			     ("This function is not available for function curves!"));
      return;
    }

  QwtArray < double >xData (c->dataSize ());
  QwtArray < double >yData (c->dataSize ());

  for (int i = 0; i < c->dataSize (); i++)
    {
      if (i == selectedPoint)
	{
	  xData[i] = d_plot->invTransform (c->xAxis (), pos.x ());
	  yData[i] = d_plot->invTransform (c->yAxis (), pos.y ());
	}
      else
	{
	  xData[i] = c->x (i);
	  yData[i] = c->y (i);
	}
    }
  c->setData (xData, yData);
  d_plot->replot ();

  QString text =
    QString::number (xData[selectedPoint]) + "\t" +
    QString::number (yData[selectedPoint]);

  emit updateTable (associations[selectedCurve], selectedPoint, text);

  QString info;
  info.sprintf ("x=%g; y=%g", xData[selectedPoint], yData[selectedPoint]);
  emit showInfo (info);
}

// Hightlight the selected point
void
Graph::highlightPoint (bool showIt)
{
  QwtPlotCurve *c = curve (selectedCurve);
  if (!c)
    return;

  QPainter painter (d_plot->canvas ());

  painter.setClipping (TRUE);
  painter.setClipRect (d_plot->canvas ()->contentsRect ());

  if (showIt)
    painter.setRasterOp (Qt::NotROP);

  c->draw (&painter,
	   d_plot->canvasMap (c->xAxis ()),
	   d_plot->canvasMap (c->yAxis ()), selectedPoint,
	   selectedPoint);
}

void
Graph::showCursor (bool showIt)
{
  QwtPlotCurve *c = curve (selectedCurve);
  if (!c || c->rtti () != QwtPlotItem::Rtti_PlotCurve)
    return;

  d_plot->replot ();

  if (showIt)
    {
      QPainter paint (d_plot->canvas ());
      paint.setRasterOp (Qt::NotXorROP);
      QwtPlotMarker mrkCross;
      mrkCross.setValue (c->x (selectedPoint), c->y (selectedPoint));
      mrkCross.setLinePen (QPen (red, 1));
      mrkCross.setLineStyle (QwtPlotMarker::Cross);
      mrkCross.setSymbol (QwtSymbol (QwtSymbol::Rect,
				     QBrush (Qt::NoBrush), QPen (black, 1),
				     QSize (25, 25)));
      mrkCross.draw (&paint, d_plot->canvasMap (c->xAxis()),
		     d_plot->canvasMap (c->yAxis()),
		     d_plot->canvas ()->contentsRect ());
    }
}

// Select the next/previous neighbour of the selected point
void
Graph::shiftPointCursor (int d)
{
  const QwtPlotCurve *c = curve (selectedCurve);
  if (!c || c->rtti () != QwtPlotItem::Rtti_PlotCurve)
    return;

  int index = selectedPoint + d;
  index = ( index + c->dataSize () ) % c->dataSize ();

  emit cursorInfo (QString::null,
		   associations[selectedCurve],
		   index);

  if (index != selectedPoint)
    {
      showCursor (FALSE);
      selectedPoint = index;
      showCursor (TRUE);
    }
}

void
Graph::shiftCurveCursor (int d)
{				// Select the next/previous curve 
  int index = 0;
  if (selectedCurve >= 0)
    {
      for (int i = 0; i < n_curves; i++)
	{
	  if (selectedCurve == i)
	    {
	      index = i + d;
	      break;
	    }
	}
    }
  index = (index + n_curves ) % n_curves;

  if (selectedCurve != index)
    {
      showCursor (FALSE);
      selectedCurve = index;
      showCursor (TRUE);
    }

  const QwtPlotCurve *c = curve (selectedCurve);
  if (!c || c->rtti () != QwtPlotItem::Rtti_PlotCurve)
    return;

  emit cursorInfo (QString::null,
		   associations[selectedCurve],
		   selectedPoint);
}

// called only from associationDialog.cpp
void
Graph::changePlotAssociation (Table * t, int curve, 
			      int ctype, const QString & text)
{
  if (associations[curve] == text)
    return;

  QApplication::setOverrideCursor (waitCursor);
  if (ctype == XErrorBars || ctype == YErrorBars)
    {
      QwtErrorPlotCurve *er = (QwtErrorPlotCurve *) this->curve (curve);
      if (!er)
	return;
      int type = QwtErrorPlotCurve::Vertical;
      if (ctype == XErrorBars)
	type = QwtErrorPlotCurve::Horizontal;
      if (er->direction () != type)
	{
	  er->setDirection (type);
	  d_plot->replot ();
	}

      QString s = text;
      if (associations[curve] != s)
	{
	  associations[curve] = text;
	  updateData (t, curve);
	  updateErrorBarsData (t, curve);
	}
      else
	associations[curve] = text;
    }
  else if (ctype == VectXYXY || ctype == VectXYAM)
    { //vectors curve
      //QStringList ls = QStringList::split (',', text, FALSE);
      QwtPlotCurve *c = (QwtPlotCurve *) d_plot->curve( curveKey(curve) );
      if (c)
	{
	  c->setTitle ( text.section(',',1,1) );	//update curve name
	}
      associations[curve] = text;
      updateVectorsData (t, curve);
      updateLegend();
    }
  else
    {				// change error bars associations depending on this curve
      QString old_as = associations[curve];
      for (int i = 0; i < n_curves; i++)
	{
	  QString as = associations[i];
	  int ct = curveType(i);
	  if (as.contains (old_as)
	      && (ct == XErrorBars || ct == YErrorBars))
	    {
	      associations[i] = text + "," + as.section(',',2,2);
	    }
	}

      //update curve name
      //QStringList ls = QStringList::split (',', text, FALSE);
      QwtPlotCurve *c = (QwtPlotCurve *) d_plot->curve( curveKey(curve) );
      if (c)
	{
	  c->setTitle ( text.section(',',1,1) );
	}
      associations[curve] = text;
      updateData (t, curve);
      updateLegend();
    }
  emit modifiedGraph ();
  emit modifiedPlotAssociation ();
  QApplication::restoreOverrideCursor ();
}

void
Graph::updateCurveData (Table * w, const QString & yColName, int curve)
{
  if (piePlot)
    updatePieCurveData (w, yColName, curve);
  else
    {
      int plotType = curveType(curve);
      if (plotType == XErrorBars || plotType == YErrorBars)
	updateErrorBarsData (w, curve);
      else if (plotType == Histogram)
	updateHistogram (w, yColName, curve);
      else if (plotType == VectXYXY || plotType == VectXYAM)
	updateVectorsData (w, curve);
      else if (plotType == Box)
	updateBoxData (w, yColName, curve);
      else
	updateData (w, curve);
    }
}

void
Graph::updateData (Table * w, int curve)
{
  int curveID = curveKey(curve);
  int ct = curveType(curve);
  QStringList cols = QStringList::split (',', associations[curve], false);
  QString xcName = cols[0];
  QString ycName = cols[1];

  int xcol = w->colIndex (xcName);
  int ycol = w->colIndex (ycName);
  if (xcol < 0 || ycol < 0)
    {
      removeCurve (curve, true);
      return;
    }

  QMemArray < double >X, Y;
  int i, it = 0;
  int rows = w->tableRows ();

  int xColType = w->columnType (xcol);
  int yColType = w->columnType (ycol);

  QStringList xLabels, yLabels;	// store text labels

  QTime time0;
  QDate date;
  if (xColType == Table::Time)
    {
      for (i = 0; i < rows; i++)
	{
	  QString xval = w->text (i, xcol);
	  if (!xval.isEmpty ())
	    {
	      time0 = QTime::fromString (xval, Qt::TextDate);
	      if (time0.isValid ())
		break;
	    }
	}
    }
  else if (xColType == Table::Date)
    {
      for (i = 0; i < rows; i++)
	{
	  QString xval = w->text (i, xcol);
	  if (!xval.isEmpty ())
	    {
	      date = QDate::fromString (xval, Qt::ISODate);
	      if (date.isValid ())
		break;
	    }
	}
    }

  for (i = 0; i < rows; i++)
    {
      QString xval = w->text (i, xcol);
      QString yval = w->text (i, ycol);
      if (!xval.isEmpty () && !yval.isEmpty ())
	{
	  it++;
	  X.resize (it);
	  Y.resize (it);
	  if (xColType == Table::Text)
	    {
	      xLabels << xval;
	      X[it - 1] = (double) it;
	    }
	  else if (xColType == Table::Time)
	    {
	      QTime time = QTime::fromString (xval, Qt::TextDate);
	      if (time.isValid ())
		X[it - 1] = time0.msecsTo (time);
	    }
	  else if (xColType == Table::Date)
	    {
	      QDate d = QDate::fromString (xval, Qt::ISODate);
	      if (d.isValid ())
		X[it - 1] = (double) date.daysTo (d);
	    }
	  else
	    X[it - 1] = xval.toDouble ();

	  if (yColType == Table::Text)
	    {
	      yLabels << yval;
	      Y[it - 1] = (double) it;
	    }
	  else
	    Y[it - 1] = yval.toDouble ();
	}
    }

  if (!it)
    {
      removeCurve (curve, true);
      return;
    }
  else
    {				//update curve data
      QwtPlotCurve *c = (QwtPlotCurve *) d_plot->curve (curveID);
      if (ct == HorizontalBars)
	c->setData (Y, X, it);
      else
	c->setData (X, Y, it);

      if (xColType == Table::Text)
	{
	  if (ct == HorizontalBars)
	    {
	      axisType[QwtPlot::yLeft] = Txt;
	      axesFormatInfo[QwtPlot::yLeft] = xcName;
	      d_plot->setAxisScaleDraw (QwtPlot::yLeft,
					new QwtTextScaleDraw (xLabels));
	    }
	  else
	    {
	      axisType[QwtPlot::xBottom] = Txt;
	      axesFormatInfo[QwtPlot::xBottom] = xcName;
	      d_plot->setAxisScaleDraw (QwtPlot::xBottom,
					new QwtTextScaleDraw (xLabels));
	    }
	}
      else if (xColType == Table::Time)
	{
	  if (ct == HorizontalBars)
	    {
	      /*
	      QStringList lst =
		QStringList::split (';', axesFormatInfo[QwtPlot::yLeft],
				    true);
	      QString fmtInfo = time0.toString (Qt::TextDate) + ";" + lst[1];
	      */
	      QString fmtInfo = time0.toString (Qt::TextDate) + ";"
		+ axesFormatInfo[QwtPlot::yLeft].section(';',1,1);
	      setLabelsDateTimeFormat (QwtPlot::yLeft, Time, fmtInfo);
	    }
	  else
	    {
	      /*
	      QStringList lst =
		QStringList::split (';', axesFormatInfo[QwtPlot::xBottom],
				    true);
	      QString fmtInfo = time0.toString (Qt::TextDate) + ";" + lst[1];
	      */
	      QString fmtInfo = time0.toString (Qt::TextDate) + ";"
		+ axesFormatInfo[QwtPlot::xBottom].section(';',1,1);
	      setLabelsDateTimeFormat (QwtPlot::xBottom, Time, fmtInfo);
	    }
	}
      else if (xColType == Table::Date)
	{
	  if (ct == HorizontalBars)
	    {
	      /*
	      QStringList lst =
		QStringList::split (';', axesFormatInfo[QwtPlot::yLeft],
				    true);
	      QString fmtInfo = date.toString (Qt::ISODate) + ";" + lst[1];
	      */
	      QString fmtInfo = date.toString (Qt::ISODate) + ";"
		+ axesFormatInfo[QwtPlot::yLeft].section(';',1,1);
	      setLabelsDateTimeFormat (QwtPlot::yLeft, Date, fmtInfo);
	    }
	  else
	    {
	      /*
	      QStringList lst =
		QStringList::split (';', axesFormatInfo[QwtPlot::xBottom],
				    true);
	      QString fmtInfo = date.toString (Qt::ISODate) + ";" + lst[1];
	      */
	      QString fmtInfo = date.toString (Qt::ISODate) + ";"
		+  axesFormatInfo[QwtPlot::xBottom].section(';',1,1);
	      setLabelsDateTimeFormat (QwtPlot::xBottom, Date, fmtInfo);
	    }
	}

      if (yColType == Table::Text)
	{
	  axisType[QwtPlot::yLeft] = Txt;
	  axesFormatInfo[QwtPlot::yLeft] = ycName;
	  d_plot->setAxisScaleDraw (QwtPlot::yLeft,
				    new QwtTextScaleDraw (yLabels));
	}
    }

// move error bars following the data change
  for (i = 0; i < n_curves; i++)
    {
      int ct = curveType(i);
      if (ct == XErrorBars || ct == YErrorBars)
	{
	  QStringList lst = QStringList::split (',', associations[i], false);
	  if (lst[0] == cols[0] && lst[1] == cols[1])
	    {
	      if (!it)
		removeCurve (i, true);
	      else
		{
		  QwtErrorPlotCurve *er =
		    (QwtErrorPlotCurve *) d_plot->curve( curveKey(i) );
		  if (er) {
		    if (ct == HorizontalBars)
		      er->setData (Y, X, it);
		    else
		      er->setData (X, Y, it);
		  }
		}
	    }
	}
    }
}

void
Graph::updateErrorBarsData (Table * w, int curve)
{
  //fprintf(stderr,"updateErrorBarsData\n");
  QwtErrorPlotCurve *er = (QwtErrorPlotCurve *) this->curve (curve);
  if (!er)
    return;

  QStringList cols = QStringList::split (',', associations[curve], FALSE);
  int xcol = w->colIndex (cols[0]);
  int ycol = w->colIndex (cols[1]);
  int errcol = w->colIndex (cols[2]);
  if ((xcol<0) || (ycol<0) || (errcol < 0))
    {
      removeCurve (curve, true);
      return;
    }
  int rows = w->tableRows ();

  QMemArray < double > err;
  int it=0;
  for (int j = 0; j < rows; j++)
    {
      QString xval = w->text (j, xcol);
      QString yval = w->text (j, ycol);
      QString errval = w->text (j, errcol);
      if (!xval.isEmpty () && !yval.isEmpty ())
	{
	  it++;
	  err.resize(it);
	  err[it-1]=errval.toDouble();
	}
    }

  QMemArray < double >err_old = er->errors ();
  if (err_old == err)
    return;

  er->setErrors (err);
  updatePlot ();
}

void
Graph::updatePieCurveData (Table * w, const QString & yColName, int curve)
{
  QMemArray < double >X;
  int it = 0;
  int ycol = w->colIndex (yColName);
  for (int i = 0; i < w->tableRows (); i++)
    {
      QString xval = w->text (i, ycol);
      if (!xval.isEmpty ())
	{
	  X.resize (++it);
	  X[it - 1] = xval.toDouble ();
	}
    }
  QwtPlotCurve *c = (QwtPlotCurve *) d_plot->curve( curveKey(curve) );
  if (c)
    {
      c->setData (X, X, it);
      d_plot->replot ();
    }
}

void
Graph::updateVectorsData (Table * w, int curve)
{
  VectorCurve *vect = (VectorCurve *) d_plot->curve( curveKey(curve) );
  if (!vect)
    return;

  QStringList cols = QStringList::split (',', associations[curve], false);
  int xcol = w->colIndex (cols[0]);
  int ycol = w->colIndex (cols[1]);
  int endXCol = w->colIndex (cols[2]);
  int endYCol = w->colIndex (cols[3]);

  QMemArray < double >X (1), Y (1), endX (1), endY (1);

  int i, it = 0;
  for (i = 0; i < w->tableRows (); i++)
    {
      QString xval = w->text (i, xcol);
      QString yval = w->text (i, ycol);
      QString xend = w->text (i, endXCol);
      QString yend = w->text (i, endYCol);
      if (!xval.isEmpty () && !yval.isEmpty () && !xend.isEmpty ()
	  && !yend.isEmpty ())
	{
	  it++;
	  X.resize (it);
	  Y.resize (it);
	  endX.resize (it);
	  endY.resize (it);
	  X[it - 1] = xval.toDouble ();
	  Y[it - 1] = yval.toDouble ();
	  endX[it - 1] = xend.toDouble ();
	  endY[it - 1] = yend.toDouble ();
	}
    }

  if (!it)
    removeCurve (curve, true);

  QwtPlotCurve *c = (QwtPlotCurve *) d_plot->curve( curveKey(curve) );
  if (c)
    c->setData (X, Y, it);

  vect->setVectorEnd (endX, endY);
  updatePlot ();
}

QString
Graph::saveEnabledAxes ()
{
  QString list = "EnabledAxes\t";
  for (int i = 0; i < QwtPlot::axisCnt; i++)
    list += QString::number (d_plot->axisEnabled (i)) + "\t";
  
  list += "\n";
  return list;
}

bool
Graph::framed ()
{
  bool frameOn = FALSE;

  QwtPlotCanvas *canvas = (QwtPlotCanvas *) d_plot->canvas ();
  if (canvas->lineWidth () > 0)
    frameOn = TRUE;

  return frameOn;
}

QColor
Graph::canvasFrameColor ()
{
  QwtPlotCanvas *canvas = (QwtPlotCanvas *) d_plot->canvas ();
  QPalette pal = canvas->palette ();
  return pal.color (QPalette::Active, QColorGroup::Foreground);
}

int
Graph::canvasFrameWidth ()
{
  QwtPlotCanvas *canvas = (QwtPlotCanvas *) d_plot->canvas ();
  return canvas->lineWidth ();
}

void
Graph::drawCanvasFrame (const QStringList & frame)
{
  QwtPlotCanvas *canvas = (QwtPlotCanvas *) d_plot->canvas ();
  int width = frame[1].toInt ();
  canvas->setLineWidth( width );

  d_frame_width = (double)width / scaleFactor();

  QPalette pal = canvas->palette ();
  pal.setColor (QColorGroup::Foreground, QColor (frame[2]));
  canvas->setPalette (pal);
}

void
Graph::drawCanvasFrame (bool frameOn, int width, const QColor & color)
{
  //qDebug("drawCanvasFrame");

  QwtPlotCanvas *canvas = (QwtPlotCanvas *) d_plot->canvas ();
  QPalette pal = canvas->palette ();

  d_frame_width = (double)width / scaleFactor();

  if (frameOn && canvas->lineWidth () == width &&
      pal.color (QPalette::Active, QColorGroup::Foreground) == color)
    return;

  if (frameOn)
    {
      canvas->setLineWidth (width);
      pal.setColor (QColorGroup::Foreground, color);
      canvas->setPalette (pal);
    }
  else
    {
      canvas->setLineWidth (0);
      pal.setColor (QColorGroup::Foreground, QColor (black));
      canvas->setPalette (pal);
    }
  emit modifiedGraph ();
}

void
Graph::drawCanvasFrame (bool frameOn, int width)
{
  if (frameOn)
    {
      QwtPlotCanvas *canvas = (QwtPlotCanvas *) d_plot->canvas ();
      canvas->setLineWidth (width);
    }

  d_frame_width = width / scaleFactor();
}

void
Graph::drawAxesBackbones (bool yes)
{
  if (drawAxesBackbone == yes)
    return;

  drawAxesBackbone = yes;

  for (int i = 0; i < QwtPlot::axisCnt; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	{
	  ScaleDraw *sclDraw = (ScaleDraw *) d_plot->axisScaleDraw (i);
	  sclDraw->enableComponent (QwtAbstractScaleDraw::Backbone, yes);
	  scale->repaint ();
	}
    }

  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::loadAxesOptions (const QString & s)
{
  if (s == "1")
    return;

  drawAxesBackbone = false;

  for (int i = 0; i < QwtPlot::axisCnt; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	{
	  ScaleDraw *sclDraw = (ScaleDraw *) d_plot->axisScaleDraw (i);
	  sclDraw->enableComponent (QwtAbstractScaleDraw::Backbone, false);
	  scale->repaint ();
	}
    }
}

void
Graph::setAxesLinewidth (int width)
{
  if (d_plot->axesLinewidth () == width)
    return;
  
  loadAxesLinewidth (width);

  for (int i = 0; i < QwtPlot::axisCnt; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	{
	  scale->setPenWidth (width);
	  scale->repaint ();
	}
    }

  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::loadAxesLinewidth (int width)
{
  d_plot->setAxesLinewidth (width);
  d_axes_width = width / scaleFactor();
}

QString
Graph::saveCanvas ()
{
  QString s = "";
  int w = d_plot->canvas ()->lineWidth ();
  if (w > 0)
    {
      s += "CanvasFrame\t" + QString::number (w) + "\t";
      s += canvasFrameColor ().name () + "\n";
    }
  s += "CanvasBackground\t" + d_plot->canvasBackground ().name () + "\n";
  return s;
}

QString
Graph::saveFonts ()
{
  int i;
  QString s;
  QStringList list, axesList;
  QFont f;
  list << "TitleFont";
  f = d_plot->title ().font ();
  list << f.family ();
  list << QString::number (f.pointSize ());
  list << QString::number (f.weight ());
  list << QString::number (f.italic ());
  list << QString::number (f.underline ());
  list << QString::number (f.strikeOut ());
  s = list.join ("\t") + "\n";

  for (i = 0; i < d_plot->axisCnt; i++)
    {
      f = d_plot->axisTitle (i).font ();
      list[0] = "ScaleFont" + QString::number (i);
      list[1] = f.family ();
      list[2] = QString::number (f.pointSize ());
      list[3] = QString::number (f.weight ());
      list[4] = QString::number (f.italic ());
      list[5] = QString::number (f.underline ());
      list[6] = QString::number (f.strikeOut ());
      s += list.join ("\t") + "\n";
    }

  for (i = 0; i < d_plot->axisCnt; i++)
    {
      f = d_plot->axisFont (i);
      list[0] = "AxisFont" + QString::number (i);
      list[1] = f.family ();
      list[2] = QString::number (f.pointSize ());
      list[3] = QString::number (f.weight ());
      list[4] = QString::number (f.italic ());
      list[5] = QString::number (f.underline ());
      list[6] = QString::number (f.strikeOut ());
      s += list.join ("\t") + "\n";
    }
  return s;
}

QString
Graph::saveAxesFormulas ()
{
  QString s;
  for (int i = 0; i < 4; i++)
    if (!axesFormulas[i].isEmpty ())
      {
	s += "<AxisFormula pos=\"" + QString::number (i) + "\">\n";
	s += axesFormulas[i];
	s += "\n</AxisFormula>\n";
      }
  return s;
}

QString
Graph::saveScale ()
{
  QString s;
  for (int i = 0; i < QwtPlot::axisCnt; i++)
    {
      s += "scale\t" + QString::number (i) + "\t";
      const QwtScaleDiv *scDiv = d_plot->axisScaleDiv (i);
      QwtValueList lst = scDiv->ticks (QwtScaleDiv::MajorTick);

      s +=
	QString::number (QMIN (scDiv->lBound (), scDiv->hBound ()), 'g',
			 15) + "\t";
      s +=
	QString::number (QMAX (scDiv->lBound (), scDiv->hBound ()), 'g',
			 15) + "\t";
      s += QString::number (d_user_step[i], 'g', 15) + "\t";
      s += QString::number (d_plot->axisMaxMajor (i)) + "\t";
      s += QString::number (d_plot->axisMaxMinor (i)) + "\t";

      const QwtScaleEngine *sc_eng = d_plot->axisScaleEngine (i);
      QwtScaleTransformation *tr = sc_eng->transformation ();
      s += QString::number ((int) tr->type ()) + "\t";
      s +=
	QString::number (sc_eng->testAttribute (QwtScaleEngine::Inverted)) +
	"\n";
    }
  return s;
}

void
Graph::setXAxisTitleColor (const QColor & c)
{
  QwtScaleWidget *scale =
    (QwtScaleWidget *) d_plot->axisWidget (QwtPlot::xBottom);
  if (scale)
    {
      QwtText t = scale->title ();
      t.setColor (c);
      scale->setTitle (t);
      emit modifiedGraph ();
    }
}

void
Graph::setYAxisTitleColor (const QColor & c)
{
  QwtScaleWidget *scale =
    (QwtScaleWidget *) d_plot->axisWidget (QwtPlot::yLeft);
  if (scale)
    {
      QwtText t = scale->title ();
      t.setColor (c);
      scale->setTitle (t);
      emit modifiedGraph ();
    }
}

void
Graph::setRightAxisTitleColor (const QColor & c)
{
  QwtScaleWidget *scale =
    (QwtScaleWidget *) d_plot->axisWidget (QwtPlot::yRight);
  if (scale)
    {
      QwtText t = scale->title ();
      t.setColor (c);
      scale->setTitle (t);
      emit modifiedGraph ();
    }
}

void
Graph::setTopAxisTitleColor (const QColor & c)
{
  QwtScaleWidget *scale =
    (QwtScaleWidget *) d_plot->axisWidget (QwtPlot::xTop);
  if (scale)
    {
      QwtText t = scale->title ();
      t.setColor (c);
      scale->setTitle (t);
      emit modifiedGraph ();
    }
}

void
Graph::setAxesTitleColor (const QStringList & l)
{
  for (int i = 0; i < int (l.count () - 1); i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	{
	  QwtText title = scale->title ();
	  title.setColor (QColor (l[i + 1]));
	  scale->setTitle (title);
	}
    }
}

QString
Graph::saveAxesTitleColors ()
{
  QString s = "AxesTitleColors\t";
  for (int i = 0; i < 4; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      QColor c;
      if (scale)
	c = scale->title ().color ();
      else
	c = QColor (Qt::black);

      s += c.name () + "\t";
    }
  return s + "\n";
}

QString
Graph::saveTitle ()
{
  QString s = "PlotTitle\t";
  s += d_plot->title ().text ().replace ("\n", "<br>") + "\t";
  s += d_plot->title ().color ().name () + "\t";
  s += QString::number (d_plot->title ().renderFlags ()) + "\n";
  return s;
}

QString
Graph::saveScaleTitles ()
{
  int a;
  QString s = "";
  for (int i = 0; i < 4; i++)
    {
      switch (i)
	{
	case 0:
	  a = 2;
	  break;

	case 1:
	  a = 0;
	  break;

	case 2:
	  a = 3;
	  break;

	case 3:
	  a = 1;
	  break;
	}
      s += (d_plot->axisTitle (a)).text ().replace ("\n", "<br>") + "\t";
    }
  return s + "\n";
}

QString
Graph::saveAxesTitleAlignement ()
{
  QString s = "AxesTitleAlignment\t";
  QStringList axes;
  int i;
  for (i = 0; i < 4; i++)
    axes << QString::number (Qt::AlignHCenter);

  for (i = 0; i < 4; i++)
    {
      if (d_plot->axisEnabled (i))
	axes[i] = QString::number (d_plot->axisTitle (i).renderFlags ());
    }

  s += axes.join ("\t") + "\n";
  return s;
}

QString
Graph::savePieCurveLayout ()
{
  QString s = "PieCurve\t";

  QwtPieCurve *pieCurve = (QwtPieCurve *) curve (0);
  s += pieCurve->title ().text () + "\t";
  QPen pen = pieCurve->pen ();

  s += QString::number (pen.width ()) + "\t";
  s += pen.color ().name () + "\t";
  s += penStyleName (pen.style ()) + "\t";

  Qt::BrushStyle pattern = pieCurve->pattern ();
  int index;
  if (pattern == Qt::SolidPattern)
    index = 0;
  else if (pattern == Qt::HorPattern)
    index = 1;
  else if (pattern == Qt::VerPattern)
    index = 2;
  else if (pattern == Qt::CrossPattern)
    index = 3;
  else if (pattern == Qt::BDiagPattern)
    index = 4;
  else if (pattern == Qt::FDiagPattern)
    index = 5;
  else if (pattern == Qt::DiagCrossPattern)
    index = 6;
  else if (pattern == Qt::Dense1Pattern)
    index = 7;
  else if (pattern == Qt::Dense2Pattern)
    index = 8;
  else if (pattern == Qt::Dense3Pattern)
    index = 9;
  else if (pattern == Qt::Dense4Pattern)
    index = 10;
  else if (pattern == Qt::Dense5Pattern)
    index = 11;
  else if (pattern == Qt::Dense6Pattern)
    index = 12;
  else if (pattern == Qt::Dense7Pattern)
    index = 13;

  s += QString::number (index) + "\t";
  //s += QString::number (pieRay) + "\t";
  s += QString::number (pieCurve->ray()) + "\t";
  s += QString::number (pieCurve->first ()) + "\n";

  return s;
}

QString
Graph::saveCurveLayout (int index)
{
  QString s = QString::null;
  int style = curveType(index);
  QwtPlotCurve *c = (QwtPlotCurve *) curve (index);
  if (c)
    {
      s += QString::number (style) + "\t";
      if (style == Spline)
	s += "5\t";
      else if (style == VerticalSteps)
	s += "6\t";
      else
	s += QString::number (c->style ()) + "\t";

      s += QString::number (ColorBox::colorIndex (c->pen ().color ())) + "\t";
      s += QString::number (c->pen ().style () - 1) + "\t";
      s += QString::number (c->pen ().width ()) + "\t";

      const QwtSymbol symbol = c->symbol ();
      s += QString::number (symbol.size ().width ()) + "\t";
      s += QString::number (SymbolBox::symbolIndex (symbol.style ())) + "\t";
      s +=
	QString::number (ColorBox::colorIndex (symbol.pen ().color ())) +
	"\t";
      if (symbol.brush ().style () != Qt::NoBrush)
	s +=
	  QString::number (ColorBox::colorIndex (symbol.brush ().color ())) +
	  "\t";
      else
	s += QString::number (-1) + "\t";

      bool filled = c->brush ().style () == Qt::NoBrush ? false : true;
      s += QString::number (filled) + "\t";

      s +=
	QString::number (ColorBox::colorIndex (c->brush ().color ())) + "\t";
      s +=
	QString::number (PatternBox::patternIndex (c->brush ().style ())) +
	"\t";
      if (style <= LineSymbols || style == Box)
	s += QString::number (symbol.pen ().width ()) + "\t";
    }

  if (style == VerticalBars || style == HorizontalBars || style == Histogram)
    {
      QwtBarCurve *b = (QwtBarCurve *) c;
      s += QString::number (b->gap ()) + "\t";
      s += QString::number (b->offset ()) + "\t";
    }

  if (style == Histogram)
    {
      QwtHistogram *h = (QwtHistogram *) c;
      s += QString::number (h->autoBinning ()) + "\t";
      s += QString::number (h->binSize ()) + "\t";
      s += QString::number (h->begin ()) + "\t";
      s += QString::number (h->end ()) + "\t";
    }
  else if (style == VectXYXY || style == VectXYAM)
    {
      VectorCurve *v = (VectorCurve *) c;
      s += v->color ().name () + "\t";
      s += QString::number (v->width ()) + "\t";
      s += QString::number (v->headLength ()) + "\t";
      s += QString::number (v->headAngle ()) + "\t";
      s += QString::number (v->filledArrowHead ()) + "\t";

      QStringList colsList =
	QStringList::split (',', associations[index], false);
      s += colsList[2] + "\t" + colsList[3];
      if (style == VectXYAM)
	s += "\t" + QString::number (v->position ());
      s += "\t";
    }
  else if (style == Box)
    {
      BoxCurve *b = (BoxCurve *) c;
      s += QString::number (SymbolBox::symbolIndex (b->maxStyle ())) + "\t";
      s += QString::number (SymbolBox::symbolIndex (b->p99Style ())) + "\t";
      s += QString::number (SymbolBox::symbolIndex (b->meanStyle ())) + "\t";
      s += QString::number (SymbolBox::symbolIndex (b->p1Style ())) + "\t";
      s += QString::number (SymbolBox::symbolIndex (b->minStyle ())) + "\t";
      s += QString::number (b->boxStyle ()) + "\t";
      s += QString::number (b->boxWidth ()) + "\t";
      s += QString::number (b->boxRangeType ()) + "\t";
      s += QString::number (b->boxRange ()) + "\t";
      s += QString::number (b->whiskersRangeType ()) + "\t";
      s += QString::number (b->whiskersRange ()) + "\t";
    }
  return s;
}

QString
Graph::saveCurves ()
{
  QString s;
  if (piePlot)
    s += savePieCurveLayout ();
  else
    {
      for (int i = 0; i < n_curves; i++)
	{
	  QwtPlotCurve *c = this->curve (i);
	  if (!c)
	    continue;
	  int ct = curveType(i);
	  if (ct != XErrorBars && ct != YErrorBars)
	    {
	      if (c->rtti () == FunctionCurve::RTTI)
		s += ((FunctionCurve *) c)->saveToString ();
	      else if (c->rtti () == QwtPlotItem::Rtti_PlotSpectrogram)
		{
		  s += ((Spectrogram *) c)->saveToString ();
		  continue;
		}
	      else if (ct == Box)
		s +=
		  "curve\t" + QString::number (c->x (0)) + "\t" +
		  c->title ().text () + "\t";
	      else
		{
		  QStringList as =
		    QStringList::split (',', associations[i], FALSE);
		  s += "curve\t" + as[0] + "\t" + as[1] + "\t";
		}
	      s += saveCurveLayout (i);
	      s +=
		QString::number (c->xAxis ()) + "\t" +
		QString::number (c->yAxis ()) + "\n";
	    }
	  else if (ct == XErrorBars || ct == YErrorBars)
	    {
	      QwtErrorPlotCurve *er = (QwtErrorPlotCurve *) c;
	      s += "ErrorBars\t";
	      s += QString::number (er->direction ()) + "\t";
	      QStringList cvs =
		QStringList::split (',', associations[i], FALSE);

	      s += cvs[0] + "\t" + cvs[1] + "\t" + cvs[2] + "\t";
	      s += QString::number (er->width ()) + "\t";
	      s += QString::number (er->capLength ()) + "\t";
	      s += er->color ().name () + "\t";
	      s += QString::number (er->throughSymbol ()) + "\t";
	      s += QString::number (er->plusSide ()) + "\t";
	      s += QString::number (er->minusSide ()) + "\t";
	      s += QString::number (er->xDataOffset ()) + "\t";
	      s += QString::number (er->yDataOffset ()) + "\n";
	    }
	}
    }
  return s;
}

QString
Graph::saveGridOptions ()
{
  QString s = "grid\t";
  s += QString::number (grid.majorOnX) + "\t";
  s += QString::number (grid.minorOnX) + "\t";
  s += QString::number (grid.majorOnY) + "\t";
  s += QString::number (grid.minorOnY) + "\t";
  s += QString::number (grid.majorCol) + "\t";
  s += QString::number (grid.majorStyle) + "\t";
  s += QString::number (grid.majorWidth) + "\t";
  s += QString::number (grid.minorCol) + "\t";
  s += QString::number (grid.minorStyle) + "\t";
  s += QString::number (grid.minorWidth) + "\t";
  s += QString::number (grid.xZeroOn) + "\t";
  s += QString::number (grid.yZeroOn) + "\t";
  s += QString::number (grid.xAxis) + "\t";
  s += QString::number (grid.yAxis) + "\n";
  return s;
}

void
Graph::newLegend ()
{
  LegendMarker *mrk = new LegendMarker (d_plot);
  mrk->setOrigin (QPoint (10, 10));

  if (piePlot)
    mrk->setText (pieLegendText ());
  else
    mrk->setText (legendText ());

  mrk->setBackground (defaultMarkerFrame);
  mrk->setFont (defaultMarkerFont, this->frameGeometry());
  mrk->setTextColor (defaultTextMarkerColor);
  mrk->setBackgroundColor (defaultTextMarkerBackground);

  legendMarkerID = d_plot->insertMarker (mrk);
  int texts = d_texts.size ();
  d_texts.resize (++texts);
  d_texts[texts - 1] = legendMarkerID;

  d_plot->replot ();
  emit modifiedGraph (); //071115
}

void
Graph::addTimeStamp ()
{
  LegendMarker *mrk =
    newLegend (QDateTime::currentDateTime ().toString (Qt::LocalDate));
  mrk->setOrigin (QPoint (d_plot->canvas ()->width () / 2, 10));
  d_plot->replot ();
}

LegendMarker *
Graph::newLegend (const QString & text)
{
  LegendMarker *mrk = new LegendMarker (d_plot);
  int key = d_plot->insertMarker (mrk);

  int texts = d_texts.size ();
  d_texts.resize (++texts);
  d_texts[texts - 1] = key;

  mrk->setOrigin (QPoint (10, 10));
  mrk->setText (text);
  mrk->setBackground (defaultMarkerFrame);
  mrk->setFont (defaultMarkerFont, this->frameGeometry());
  mrk->setTextColor (defaultTextMarkerColor);
  mrk->setBackgroundColor (defaultTextMarkerBackground);

  selectedMarker = key; // necessary to edit new textMarker on new layer.

  return mrk;
}

void
Graph::insertLegend (const QStringList & lst, int fileVersion)
{
  legendMarkerID = insertTextMarker (lst, fileVersion);
}

int
Graph::insertTextMarker (const QStringList & list, int fileVersion)
{
  QStringList fList = list;
  LegendMarker *mrk = new LegendMarker (d_plot);
  int key = d_plot->insertMarker (mrk);

  int texts = d_texts.size ();
  d_texts.resize (++texts);
  d_texts[texts - 1] = key;

  if (fileVersion < 86)
    mrk->setOrigin (QPoint (fList[1].toInt (), fList[2].toInt ()));
  else
    mrk->setOriginCoord (fList[1].toDouble (), fList[2].toDouble ());

  QFont fnt =
    QFont (fList[3], fList[4].toInt (), fList[5].toInt (), fList[6].toInt ());

  fnt.setUnderline (fList[7].toInt ());
  fnt.setStrikeOut (fList[8].toInt ());
  mrk->setFont (fnt, this->frameGeometry());
  mrk->setAngle (fList[11].toInt ());

  if (fileVersion < 71)
    {
      int bkg = fList[10].toInt ();
      if (bkg <= 2)
	mrk->setBackground (bkg);
      else if (bkg == 3)
	{
	  mrk->setBackground (0);
	  mrk->setBackgroundColor (QColor (255, 255, 255));
	}
      else if (bkg == 4)
	{
	  mrk->setBackground (0);
	  mrk->setBackgroundColor (QColor (black));
	}

      QString text;
      int n = (int) fList.count ();
      for (int i = 0; i < n - 12; i++)
	text += fList[12 + i] + "\n";
      mrk->setText (text);
    }
  else
    {
      mrk->setTextColor (QColor (fList[9]));
      mrk->setBackground (fList[10].toInt ());
      mrk->setBackgroundColor (QColor (fList[12]));

      QString text;
      int n = (int) fList.count ();
      for (int i = 0; i < n - 13; i++)
	text += fList[13 + i] + "\n";
      mrk->setText (text);
    }
  return key;
}

void
Graph::insertLineMarker( double x1, double y1, double x2, double y2, bool startArrow, bool endArrow, const QColor &color )
{
  LineMarker *mrk = new LineMarker();
  mrk->setStartPoint( x1,y1 );
  mrk->setEndPoint( x2,y2 );
  mrk->drawStartArrow( startArrow );
  mrk->drawEndArrow( endArrow );
  mrk->setColor( color );
  int linesOnPlot = (int) d_lines.size ();
  d_lines.resize(++linesOnPlot);
  d_lines[linesOnPlot - 1] = d_plot->insertMarker (mrk);
}

void
Graph::insertLineMarker (const QStringList & list, int fileVersion)
{
  LineMarker *mrk = new LineMarker ();
  int mrkID = d_plot->insertMarker (mrk);
  int linesOnPlot = (int) d_lines.size ();
  d_lines.resize (++linesOnPlot);
  d_lines[linesOnPlot - 1] = mrkID;

  if (fileVersion < 86)
    {
      mrk->setStartPoint (QPoint (list[1].toInt (), list[2].toInt ()));
      mrk->setEndPoint (QPoint (list[3].toInt (), list[4].toInt ()));
    }
  else
    {
      mrk->setStartPoint (list[1].toDouble (), list[2].toDouble ());
      mrk->setEndPoint (list[3].toDouble (), list[4].toDouble ());
    }

  mrk->setWidth (list[5].toInt ());
  mrk->setColor (QColor (list[6]));
  mrk->setStyle (getPenStyle (list[7]));
  mrk->drawEndArrow (list[8] == "1");
  mrk->drawStartArrow (list[9] == "1");
  if (list.count () > 10)
    {
      mrk->setHeadLength (list[10].toInt ());
      mrk->setHeadAngle (list[11].toInt ());
      mrk->fillArrowHead (list[12] == "1");
    }
}

void
Graph::insertLineMarker (LineMarker * mrk)
{
  LineMarker *aux = new LineMarker ();
  int linesOnPlot = (int) d_lines.size ();
  d_lines.resize (++linesOnPlot);
  d_lines[linesOnPlot - 1] = d_plot->insertMarker (aux);

  aux->setStartPoint (mrk->startPointCoord ().x (),
		      mrk->startPointCoord ().y ());
  aux->setEndPoint (mrk->endPointCoord ().x (), mrk->endPointCoord ().y ());
  aux->setWidth (mrk->width ());
  aux->setColor (mrk->color ());
  aux->setStyle (mrk->style ());
  aux->drawEndArrow (mrk->hasEndArrow ());
  aux->drawStartArrow (mrk->hasStartArrow ());
  aux->setHeadLength (mrk->headLength ());
  aux->setHeadAngle (mrk->headAngle ());
  aux->fillArrowHead (mrk->filledArrowHead ());
}

LineMarker *
Graph::lineMarker (int id)
{
  return (LineMarker *) d_plot->marker (id);
}

ImageMarker *
Graph::imageMarker (int id)
{
  return (ImageMarker *) d_plot->marker (id);
}

LegendMarker *
Graph::textMarker (int id)
{
  return (LegendMarker *) d_plot->marker (id);
}

int
Graph::insertTextMarker (LegendMarker * mrk)
{
  LegendMarker *aux = new LegendMarker (d_plot);
  int key = d_plot->insertMarker (aux);

  int texts = d_texts.size ();
  d_texts.resize (++texts);
  d_texts[texts - 1] = key;

  aux->setTextColor (mrk->getTextColor ());
  aux->setBackgroundColor (mrk->backgroundColor ());
  aux->setOrigin (mrk->rect ().topLeft ());
  aux->setFont (mrk->getFont (), this->frameGeometry());
  aux->setBackground (mrk->getBkgType ());
  aux->setAngle (mrk->getAngle ());
  aux->setText (mrk->getText ());

  selectedMarker = key; // necessary for textDialog
  return key;
}

QString
Graph::saveMarkers ()
{
  QString s;
  int i;
  for (i = 0; i < (int) d_images.size (); i++)
    {
      ImageMarker *mrkI = (ImageMarker *) d_plot->marker (d_images[i]);
      s += "<image>\t";
      s += mrkI->getFileName () + "\t";

      QwtDoubleRect rect = mrkI->boundingRect ();
      s += QString::number (rect.left (), 'g', 15) + "\t";
      s += QString::number (rect.top (), 'g', 15) + "\t";
      s += QString::number (rect.width (), 'g', 15) + "\t";
      s += QString::number (rect.height (), 'g', 15) + "</image>\n";
    }
  for (i = 0; i < (int) d_lines.size (); i++)
    {
      LineMarker *mrkL = (LineMarker *) d_plot->marker (d_lines[i]);
      s += "<line>\t";

      QwtDoublePoint sp = mrkL->startPointCoord ();
      s += (QString::number (sp.x (), 'g', 15)) + "\t";
      s += (QString::number (sp.y (), 'g', 15)) + "\t";

      QwtDoublePoint ep = mrkL->endPointCoord ();
      s += (QString::number (ep.x (), 'g', 15)) + "\t";
      s += (QString::number (ep.y (), 'g', 15)) + "\t";

      s += QString::number (mrkL->width ()) + "\t";
      s += mrkL->color ().name () + "\t";
      s += penStyleName (mrkL->style ()) + "\t";
      s += QString::number (mrkL->hasEndArrow ()) + "\t";
      s += QString::number (mrkL->hasStartArrow ()) + "\t";
      s += QString::number (mrkL->headLength ()) + "\t";
      s += QString::number (mrkL->headAngle ()) + "\t";
      s += QString::number (mrkL->filledArrowHead ()) + "</line>\n";
    }

  for (i = 0; i < (int) d_texts.size (); i++)
    {
      LegendMarker *mrk = (LegendMarker *) d_plot->marker (d_texts[i]);
      if (d_texts[i] != legendMarkerID)
	s += "<text>\t";
      else
	s += "<legend>\t";

      s += QString::number (mrk->xValue (), 'g', 15) + "\t";
      s += QString::number (mrk->yValue (), 'g', 15) + "\t";

      QFont f = mrk->getFont ();
      s += f.family () + "\t";
      s += QString::number (f.pointSize ()) + "\t";
      s += QString::number (f.weight ()) + "\t";
      s += QString::number (f.italic ()) + "\t";
      s += QString::number (f.underline ()) + "\t";
      s += QString::number (f.strikeOut ()) + "\t";
      s += mrk->getTextColor ().name () + "\t";
      s += QString::number (mrk->getBkgType ()) + "\t";
      s += QString::number (mrk->getAngle ()) + "\t";
      s += mrk->backgroundColor ().name () + "\t";

      QStringList textList =
	QStringList::split ('\n', mrk->getText (), FALSE);
      s += textList.join ("\t");
      if (d_texts[i] != legendMarkerID)
	s += "</text>\n";
      else
	s += "</legend>\n";
    }
  return s;
}

double
Graph::selectedXStartValue ()
{
  QwtPlotCurve *c = curve (selectedCurve);
  return c->x (startPoint);
}

double
Graph::selectedXEndValue ()
{
  QwtPlotCurve *c = curve (selectedCurve);
  return c->x (endPoint);
}

int
Graph::curveIndex (int key)
{
  for(int i=0; i<n_curves; i++)
    {
      if( curveKey(i) == key )
	return i;
    }
  return -1;
}

int
Graph::curveIndex(const QString & title)
{
  for(int i=0;i<n_curves;i++)
    {
      QwtPlotItem *c = d_plot->curve( curveKey(i) );
      if(c->title().text() == title) return i;
    }
  return -1;
}

int
Graph::curveKey (int curve)
{
  return c_attr.at(curve)->c_key;
}

QwtPlotCurve *
Graph::curve (int index)
{
  if (!n_curves || index >= n_curves || index < 0)
    return 0;

  return (QwtPlotCurve *) d_plot->curve( curveKey(index) );
}

bool
Graph::getDisplayRange(int index, int *from, int *to)
{
  const QwtPlotCurve *c = (QwtPlotCurve *) curve(index);
  if (!c)
    {
      fprintf(stderr,"curve %d is invalid\n",index);
      return false;
    }

  const QwtScaleDiv *scDivX = d_plot->axisScaleDiv (c->xAxis());
  double xmin = scDivX->lBound();
  double xmax = scDivX->hBound();
  const QwtScaleDiv *scDivY = d_plot->axisScaleDiv (c->yAxis());
  double ymin = scDivY->lBound();
  double ymax = scDivY->hBound();

  bool first = true;
  for(int i=0; i<c->dataSize(); i++)
    {
      double xval = c->x(i);
      double yval = c->y(i);
      if( xmin <= xval && xval <= xmax && ymin <= yval && yval <= ymax)
	{
	  if(first)
	    {
	      *from = i;
	      *to = i;
	      first = false;
	    }
	  else
	    *to = i;
	}
    }
  return !first;
}

void
Graph::range (int index, double *start, double *end)
{
  if (rangeSelectorsEnabled && selectedCurve == index)
    {
      *start = selectedXStartValue ();
      *end = selectedXEndValue ();
    }
  else
    {
      int from,to;
      QwtPlotCurve *c = curve (index);
      if(getDisplayRange(index,&from,&to))
	{
	  *start = c->x(from);
	  *end = c->x(to);
	}
      else
	{
	  *start = c->minXValue ();
	  *end = c->maxXValue ();
	}
    }
  if(*start>*end)
    {
      SWAP( double, *start, *end );
    }
}

double
Graph::maxPositiveValue(int axis)
{
  double value=1.0;
  bool first=true;
  for(int i=0;i<n_curves;i++)
    {
       QwtPlotCurve *c=curve(i);
       QwtPlotItem *item = (QwtPlotItem *)c;
       if(item->xAxis()==axis)
	 {
	   double v = c->maxXValue();
	   if(first && v>0)
	     {
	       first=false;
	       value = v;
	     }
	   else if (v > value)
	     value = v;
	 }
       else if(item->yAxis()==axis)
	 {
	   double v = c->maxYValue();
	   if(first && v>0)
	     {
	       first=false;
	       value = v;
	     }
	   else if (v > value)
	     value = v;
	 }
    }
  return value;
}

double
Graph::minPositiveValue(int axis)
{
  double value=1e-100;
  bool first=true;
  for(int i=0;i<n_curves;i++)
    {
       QwtPlotCurve *c=curve(i);
       QwtPlotItem *item = (QwtPlotItem *)c;
       if(item->xAxis()==axis)
	 {
	   double v = c->minXValue();
	   if(v<=0){
	     v=c->maxXValue();
	     for(int j=0;j<c->dataSize();j++)
	       if (0<c->x(j) && c->x(j)<v) v=c->x(j);
	   }
	   if( first && v>0 )
	     {
	       first = false;
	       value = v;
	     }
	   else if (v>0 && v < value)
	     value = v;
	 }
       else if(item->yAxis()==axis)
	 {
	   double v = c->minYValue();
	   if(v<=0){
	     v=c->maxYValue();
	     for(int j=0;j<c->dataSize();j++)
	       if (0<c->y(j) && c->y(j)<v) v=c->y(j);
	   }
	   if( first && v>0 )
	     {
	       first = false;
	       value = v;
	     }
	   else if (v>0 && v < value)
	     value = v;
	 }
    }
  return value;
}

int
Graph::numPoints (int index, double start, double end)
{
  QwtPlotCurve *c = curve (index);
  if (!c || c->rtti () != QwtPlotItem::Rtti_PlotCurve)
    return 0;

  int points = 0;
  for (int i = 0; i < c->dataSize (); i++)
    if (c->x (i) >= start && c->x (i) <= end)
      points++;
  return points;

  /*
  bool first = true;
  int from,to;
  for(int i=0; i<c->dataSize(); i++)
    {
      double xval = c->x(i);
      if( (end - xval)*(xval - start) >= 0 )
        {
          if(first)
            {
              from = i;
              to = i;
              first = false;
            }
          else
            to = i;
        }
    }
  if (first) 
    return 0;
  else
    return to-from+1;
  */
}

int
Graph::curveData (int index, double start, double end,
		  QMemArray <double> &x, QMemArray <double> &y)
{
  int n = numPoints (index, start, end);
  if(n==0) return 0;
  //qDebug("numPoints = %d",n);
  x.resize(n);
  y.resize(n);
  QwtPlotCurve *c = curve (index);
  if (!c || c->rtti () != QwtPlotItem::Rtti_PlotCurve)
    return 0;

  double pr_x=0;
  int j = 0;
  for (int i = 0; i < c->dataSize (); i++)
    if (c->x (i) >= start && c->x (i) <= end && j < n)
      {
	x[j] = c->x (i);
	if ( j>0 && x[j] == pr_x )
	  return -1;
	pr_x = x[j];
	y[j++] = c->y (i);
      }
  return n;
}

int
Graph::sortedCurveData (int index, QValueList <double> &ranges,
			QMemArray <double> &x, QMemArray <double> &y)
{
  QwtPlotCurve *c = curve (index);
  if (!c || c->rtti () != QwtPlotItem::Rtti_PlotCurve)
    return 0;

  int n0 = c->dataSize();
  QMemArray <double> xtemp(n0), ytemp(n0);
  
  int n = 0;
  for (int i = 0; i < n0; i++)
    for (int k = 0; k < (int)ranges.count()/2; k++)
      if (ranges[k*2] <= c->x(i) && c->x(i) <= ranges[k*2+1])
	{
	  xtemp[n] = c->x (i);
	  ytemp[n] = c->y (i);
	  n++;
	  break;
	}

  x.resize(n);
  y.resize(n);
  QMemArray <size_t> p(n);
  gsl_sort_index (p.data(), xtemp.data(), 1, n);
  double pr_x=0;
  for (int i = 0; i < n; i++)
    {
      if(i>0 && xtemp[p[i]] == pr_x)
	return -1;
      x[i] = xtemp[p[i]];
      y[i] = ytemp[p[i]];
      pr_x = xtemp[p[i]];
    }
  return n;
}

int
Graph::sortedCurveData (int index, double start, double end,
			QMemArray <double> &x, QMemArray <double> &y)
{
  QValueList <double> ranges;
  ranges.append(start);
  ranges.append(end);
  return sortedCurveData( index, ranges, x, y );
}

curveLayout
Graph::initCurveLayout ()
{
  curveLayout cl;
  cl.connectType = 1;
  cl.lStyle = 0;
  cl.lWidth = 1;
  cl.sSize = 7;
  cl.sType = 0;
  cl.filledArea = 0;
  cl.aCol = 0;
  cl.aStyle = 0;
  cl.lCol = 0;
  cl.penWidth = 1;
  cl.symCol = 0;
  cl.fillCol = 0;
  return cl;
}


curveLayout
Graph::initCurveLayout (int i, int curves, int style)
{
  curveLayout cl = initCurveLayout ();

  int color;
  guessUniqueCurveLayout (color, cl.sType);
  int ct = curveType(i);
  if(ct==XErrorBars || ct==YErrorBars)
    color = 0;

  cl.lCol = color;
  cl.symCol = color;
  cl.fillCol = color;

  if (style == Graph::Line)
    cl.sType = 0;
  else if (style == Graph::VerticalDropLines)
    cl.connectType = 2;
  else if (style == Graph::HorizontalSteps || style == Graph::VerticalSteps)
    {
      cl.connectType = 3;
      cl.sType = 0;
    }
  else if (style == Graph::Spline)
    cl.connectType = 5;
  else if (style == Graph::VerticalBars || style == Graph::HorizontalBars)
    {
      cl.filledArea = 1;
      cl.lCol = 0;		//black color pen
      cl.aCol = i + 1;
      cl.sType = 0;
      if (ct == Graph::VerticalBars || style == Graph::HorizontalBars)
	{
	  QwtBarCurve *b = (QwtBarCurve *) curve (i);
	  if (b)
	    {
	      b->setGap (qRound (100 * (1 - 1.0 / (double) curves)));
	      b->setOffset (-50 * (curves - 1) + i * 100);
	    }
	}
    }
  else if (style == Graph::Histogram)
    {
      cl.filledArea = 1;
      cl.lCol = i + 1;		//start with red color pen
      cl.aCol = i + 1;		//start with red fill color
      cl.aStyle = 4;
      cl.sType = 0;
    }
  else if (style == Graph::Area)
    {
      cl.filledArea = 1;
      cl.aCol = color;
      cl.sType = 0;
    }
 
  return cl;
}

void
Graph::setCurveType (int index, int style)
{
  c_attr.at(index)->c_type = style;
}

void
Graph::updateCurveLayout (int index, const curveLayout * cL)
{
  QwtPlotCurve *c = this->curve (index);
  if (!c)
    return;

  QPen pen = QPen (ColorBox::color (cL->symCol),
		   cL->penWidth,
		   Qt::SolidLine,
		   Qt::FlatCap,
		   Qt::BevelJoin);
  if (cL->fillCol != -1)
    c->setSymbol (QwtSymbol (SymbolBox::style (cL->sType),
			     QBrush (ColorBox::color (cL->fillCol)), pen,
			     QSize (cL->sSize, cL->sSize)));
  else
    c->
      setSymbol (QwtSymbol
		 (SymbolBox::style (cL->sType), QBrush (), pen,
		  QSize (cL->sSize, cL->sSize)));


  c->
    setPen (QPen
	    (ColorBox::color (cL->lCol), cL->lWidth,
	     getPenStyle (cL->lStyle),
	     Qt::FlatCap,
	     Qt::BevelJoin));

  // add by S.Yamamoto
  double factor = scaleFactor();
  CurveAttr *attr = c_attr.at(index);
  attr->line_width = cL->lWidth / factor;
  attr->symbol_size = cL->sSize / factor;
  attr->symbol_edge_width = cL->penWidth / factor;

  int style = curveType(index);
  if (style == Scatter)
    c->setStyle (QwtPlotCurve::NoCurve);
  else if (style == Spline)
    {
      c->setStyle (QwtPlotCurve::Lines);
      c->setCurveAttribute (QwtPlotCurve::Fitted, true);
    }
  else if (style == VerticalSteps)
    {
      c->setStyle (QwtPlotCurve::Steps);
      c->setCurveAttribute (QwtPlotCurve::Inverted, true);
    }
  else
    c->setStyle ((QwtPlotCurve::CurveStyle) cL->connectType);

  QBrush brush = QBrush (ColorBox::color (cL->aCol), QBrush::NoBrush);
  if (cL->filledArea)
    brush.setStyle (getBrushStyle (cL->aStyle));
  c->setBrush (brush);

  QString yColName = c->title().text ();
  for (int i = 0; i < n_curves; i++)
    {
      QStringList lst = QStringList::split(',',associations[i],false);
      int ct = curveType(i);
      if (lst.contains (yColName) && i != index
	  && (ct == XErrorBars || ct == YErrorBars))
	{
	  QwtErrorPlotCurve *er = (QwtErrorPlotCurve *) this->curve (i);
	  if (er)
	    er->setSymbolSize (QSize (cL->sSize, cL->sSize));
	}
    }
}

void
Graph::updateErrorBars (int curve, bool xErr, int width, int cap,
			const QColor & c, bool plus, bool minus, bool through)
{
  QwtErrorPlotCurve *er = (QwtErrorPlotCurve *) this->curve (curve);
  if (!er)
    return;

  if (er->width () == width &&
      er->capLength () == cap &&
      er->color () == c &&
      er->plusSide () == plus &&
      er->minusSide () == minus &&
      er->throughSymbol () == through && er->xErrors () == xErr)
    return;

  er->setWidth (width);
  er->setCapLength (cap);
  er->setColor (c);

  er->setXErrors (xErr);
  if(xErr)
    setCurveType(curve, XErrorBars);
  else
    setCurveType(curve, YErrorBars);

  er->drawThroughSymbol (through);
  er->drawPlusSide (plus);
  er->drawMinusSide (minus);
}

bool
Graph::addErrorBars (Table * w, const QString & yColName,
		     Table * errTable, const QString & errColName,
		     int type, int width, int cap, const QColor & color,
		     bool through, bool minus, bool plus)
{
  //QValueList < int >keys = d_plot->curveKeys ();
  for (int i = 0; i < n_curves; i++)
    {
      QwtPlotCurve *c = (QwtPlotCurve *) d_plot->curve( curveKey(i) );
      int ct = curveType(i);
      if (c && c->title ().text () == yColName 
	  && ct != XErrorBars && ct != YErrorBars)
	{
	  //QStringList lst = QStringList::split (',', associations[i], false);
	  QString xColName = associations[i].section(',',0,0);
	  return addErrorBars (w, xColName, yColName, errTable,
			       errColName, type, width, cap, color, through,
			       minus, plus);
	}
    }
  return false;
}

bool
Graph::addErrorBars (Table * w, const QString & xColName,
		     const QString & yColName, Table * errTable,
		     const QString & errColName, int type, int width, int cap,
		     const QColor & color, bool through, bool minus,
		     bool plus, double xOffset, double yOffset)
{
  int xcol = w->colIndex (xColName);
  int ycol = w->colIndex (yColName);
  int errcol = errTable->colIndex (errColName);
  if (xcol < 0 || ycol < 0 || errcol < 0)
    return false;

  int i, it = 0;
  int xColType = w->columnType (xcol);
  int yColType = w->columnType (ycol);
  QStringList xLabels, yLabels;	// store text labels
  QwtArray <double> X, Y, err;
  for (i = 0; i < w->tableRows (); i++)
    {
      QString xval = w->text (i, xcol);
      QString yval = w->text (i, ycol);
      QString errval = errTable->text (i, errcol);
      if (!xval.isEmpty () && !yval.isEmpty () && !errval.isEmpty ())
	{
	  X.resize (++it);
	  Y.resize (it);
	  err.resize (it);
	  if (xColType == Table::Text)
	    {
	      xLabels << xval;
	      X[it - 1] = (double) it;
	    }
	  else
	    X[it - 1] = xval.toDouble ();

	  if (yColType == Table::Text)
	    {
	      yLabels << yval;
	      Y[it - 1] = (double) it;
	    }
	  else
	    Y[it - 1] = yval.toDouble ();
	  err[it - 1] = errval.toDouble ();
	}
    }
  if (!it)
    return false;

  QSize size;
  QwtPlotCurve *c;
  for (i = 0; i < n_curves; i++)
    {
      QStringList lst = QStringList::split(',',associations[i],false);
      int ct = curveType(i);
      if (lst.contains (yColName) && ct != XErrorBars && ct != YErrorBars)
	{
	  c = curve(i);
	  size = c->symbol ().size ();
	  if (ct == VerticalBars)
	    {
	      QwtBarCurve *bc = (QwtBarCurve *) c;
	      xOffset = bc->dataOffset ();
	    }
	  else if (ct == HorizontalBars)
	    {
	      QwtBarCurve *bc = (QwtBarCurve *) c;
	      yOffset = bc->dataOffset ();
	    }
	}
    }

  QwtErrorPlotCurve *er = new QwtErrorPlotCurve (type, errColName);
  er->setData (X, Y, it);
  er->setErrors (err);
  er->setAxis( c->xAxis(), c->yAxis() );
  er->setCapLength (cap);
  er->setSymbolSize (size);
  //qDebug("Error Bar color = %d",ColorBox::colorIndex(color));
  er->setColor (color);
  er->setWidth (width);
  er->drawPlusSide (plus);
  er->drawMinusSide (minus);
  er->drawThroughSymbol (through);
  er->setXDataOffset (xOffset);
  er->setYDataOffset (yOffset);

  n_curves++;
  int tp;
  if(type==QwtErrorPlotCurve::Vertical)
    tp = YErrorBars;
  else
    tp = XErrorBars;
  c_attr.append( new CurveAttr( d_plot->insertCurve(er), tp ) );  

  QString errLabel = xColName + "," + yColName + "," + errColName;

  associations << errLabel;

  //updatePlot ();
  d_plot->replot();
  return true;
}

void
Graph::updatePie (const QPen & pen, const Qt::BrushStyle & brushStyle,
		  int size, int firstColor)
{
  if (curves () > 0)
    {
      QwtPieCurve *pieCurve = (QwtPieCurve *) curve (0);
      if (pieCurve)
	{
	  pieCurve->setPen (pen);
	  pieCurve->setRay (size);
	  //pieRay = size;
	  pieSize = (double) size / scaleFactor(); 
	  pieCurve->setBrushStyle (brushStyle);
	  pieCurve->setFirstColor (firstColor);
	  d_plot->replot ();
	  emit modifiedGraph ();
	}
    }
}

void
Graph::plotPie (QwtPieCurve * curve)
{
  int n = curve->dataSize ();
  QMemArray <double> dat(n);
  for (int i = 0; i < n; i++)
    dat[i] = curve->y (i);

  QwtPieCurve *pieCurve = new QwtPieCurve (curve->title ().text ());
  pieCurve->setData (dat.data(), dat.data(), n);

  n_curves++;
  c_attr.append( new CurveAttr( d_plot->insertCurve (pieCurve), Pie ) );

  pieCurve->setPen (curve->pen ());
  int size = curve->ray ();
  pieCurve->setRay (size);

  pieSize = (double) size / scaleFactor();

  pieCurve->setBrushStyle (curve->pattern ());
  pieCurve->setFirstColor (curve->first ());
  piePlot = TRUE;

  associations << curve->title ().text ();
}

void
Graph::plotPie (Table * w, const QString & name, const QPen & pen, int brush,
		int size, int firstColor)
{
  associations << name;
  int ycol = w->colIndex (name);
  QMemArray < double >Y (1);
  int it = 0;
  for (int i = 0; i < w->tableRows (); i++)
    {
      QString yval = w->text (i, ycol);
      if (!yval.isEmpty ())
	{
	  it++;
	  Y.resize (it);
	  Y[it - 1] = yval.toDouble ();
	}
    }
  QwtPieCurve *pieCurve = new QwtPieCurve (name);
  pieCurve->setData (Y, Y, it);

  n_curves++;
  c_attr.append( new CurveAttr( d_plot->insertCurve (pieCurve), Pie ) );

  pieCurve->setPen (pen);
  pieCurve->setRay (size);
  //pieRay = size;
  pieSize = (double) size / scaleFactor(); 

  pieCurve->setFirstColor (firstColor);
  pieCurve->setBrushStyle (getBrushStyle (brush));
  piePlot = TRUE;
}

void
Graph::plotPie (Table * w, const QString & name)
{
  int ycol = w->colIndex (name);
  int i, it = 0;
  double sum = 0.0;
  QMemArray < double >Y (1);
  for (i = 0; i < w->tableRows (); i++)
    {
      QString yval = w->text (i, ycol);
      if (!yval.isEmpty ())
	{
	  it++;
	  Y.resize (it);
	  Y[it - 1] = yval.toDouble ();
	  sum += Y[it - 1];
	}
    }
  if (!it)
    return;

  associations << name;

  QwtPlotLayout *pLayout = d_plot->plotLayout ();
  pLayout->activate (d_plot, d_plot->rect (), 0);
  const QRect rect = pLayout->canvasRect ();

  QwtPieCurve *pieCurve = new QwtPieCurve (name);
  pieCurve->setData (Y, Y, it);

  n_curves++;
  c_attr.append( new CurveAttr( d_plot->insertCurve (pieCurve), Pie ) );

  pieSize = (double) pieCurve->ray() / scaleFactor();

  const int ray = 125;
  int xc = int (rect.width () / 2 + 10);
  int yc =
    int (rect.y () + rect.height () / 2 + pLayout->titleRect ().height () +
	 15);

  double PI = 4 * atan (1.0);
  double angle = 90;

  for (i = 0; i < it; i++)
    {
      const double value = Y[i] / sum * 360;
      double alabel = (angle - value * 0.5) * PI / 180.0;

      const int x = int (xc + ray * cos (alabel));
      const int y = int (yc - ray * sin (alabel));

      LegendMarker *aux = new LegendMarker (d_plot);
      aux->setOrigin (QPoint (x, y));
      aux->setBackground (0);
      aux->setText (QString::number (Y[i] / sum * 100, 'g', 2) + "%");
      int key = d_plot->insertMarker (aux);

      int texts = d_texts.size ();
      d_texts.resize (++texts);
      d_texts[texts - 1] = key;

      angle -= value;
    }

  piePlot = TRUE;
  if (legendMarkerID >= 0)
    {
      LegendMarker *mrk = (LegendMarker *) d_plot->marker (legendMarkerID);
      if (mrk)
	{
	  QString text = "";
	  for (i = 0; i < it; i++)
	    {
	      text += "\\p{";
	      text += QString::number (i + 1);
	      text += "} ";
	      text += QString::number (i + 1);
	      text += "\n";
	    }
	  mrk->setText (text);
	}
    }

  for (i = 0; i < QwtPlot::axisCnt; i++)
    d_plot->enableAxis (i, FALSE);

  d_plot->setTitle (QString::null);

  QwtPlotCanvas *canvas = (QwtPlotCanvas *) d_plot->canvas ();
  canvas->setLineWidth (1);

  scalePicker->refresh ();

  updateScale ();
}

bool
Graph::insertCurvesList (Table * w, const QStringList & names, int style,
                         int lWidth, int sSize)
{
  //fprintf(stderr,"graph.cpp : Graph::insertCurvesList\n");

  if (style == Graph::Pie)
    plotPie (w, names[0]);
  else if (style == Box)
    plotBoxDiagram (w, names);
  else if (style == Graph::VectXYXY || style == Graph::VectXYAM)
    plotVectorCurve (w, names, style);
  else
  {
    int curves = (int) names.count ();
    int errCurves = 0;
    for (int j = 0; j < curves; j++)
    {
      int k = w->colIndex (names[j]);
      if (w->colPlotDesignation (k) == Table::xErr
        || w->colPlotDesignation (k) == Table::yErr)
        errCurves++;
    }

    for (int i = 0; i < curves; i++)
    {
      //qDebug("adding curve %d",i);
      if (insertCurve (w, names[i], style))
      {
        curveLayout cl = initCurveLayout (i, curves - errCurves, style);
        cl.sSize = sSize;
        cl.lWidth = lWidth;
        // if the curve is ErrorBar then the color should be the same
        // to the data curve
        updateCurveLayout (i, &cl); // here!
      }
      else
        return false;
    }
  }
  updatePlot ();
  return true;
}

void
Graph::insertPlotItem (QwtPlotItem * i, const QString & plotAssociation,
		       int style)
{
  n_curves++;
  c_attr.append( new CurveAttr( d_plot->insertCurve(i), style ) );

  if (i->rtti () != QwtPlotItem::Rtti_PlotSpectrogram)
    addLegendItem (i->title ().text ());

  associations << plotAssociation;
}

bool
Graph::insertCurve (Table * w, const QString & name, int style)
{				//provided for convenience
  int ycol = w->colIndex (name);
  int xcol = w->colX (ycol);

  bool succes = insertCurve (w, w->colName (xcol), w->colName (ycol), style);
  if (succes)
    emit modifiedGraph ();

  return succes;
}

bool
Graph::insertCurve (Table * w, int xcol, const QString & name, int style)
{				//provided for convenience
  return insertCurve (w, w->colName (xcol), w->colName (w->colIndex (name)),
		      style);
}

bool
Graph::insertCurve (Table * w, const QString & xColName,
		    const QString & yColName, int style)
{
  //qDebug("graph.cpp : Graph::insertCurve");

  int xcol = w->colIndex (xColName);
  int ycol = w->colIndex (yColName);
  if (xcol < 0 || ycol < 0)
    return false;

  int colPlotDesignation = w->colPlotDesignation (ycol);
  if (colPlotDesignation == Table::xErr || colPlotDesignation == Table::yErr)
    {				//add error bars
      ycol = w->colY (ycol);
      if (colPlotDesignation == Table::xErr)
	return addErrorBars (w, xColName, w->colName (ycol), w, yColName,
			     (int) QwtErrorPlotCurve::Horizontal);
      else
	return addErrorBars (w, xColName, w->colName (ycol), w, yColName);
    }

  int xColType = w->columnType (xcol);
  int yColType = w->columnType (ycol);

  QMemArray < double >X (1), Y (1);
  int i, it = 0;
  QStringList xLabels, yLabels;	// store text labels
  QTime time0;
  QDate date;
  int r = w->tableRows ();
  if (xColType == Table::Time)
    {
      for (i = 0; i < r; i++)
	{
	  QString xval = w->text (i, xcol);
	  if (!xval.isEmpty ())
	    {
	      time0 = QTime::fromString (xval, Qt::TextDate);
	      if (time0.isValid ())
		break;
	    }
	}
    }
  else if (xColType == Table::Date)
    {
      for (i = 0; i < r; i++)
	{
	  QString xval = w->text (i, xcol);
	  if (!xval.isEmpty ())
	    {
	      date = QDate::fromString (xval, Qt::ISODate);
	      if (date.isValid ())
		break;
	    }
	}
    }

  for (i = 0; i < r; i++)
    {
      QString xval = w->text (i, xcol);
      QString yval = w->text (i, ycol);
      if (!xval.isEmpty () && !yval.isEmpty ())
	{
	  it++;
	  X.resize (it);
	  Y.resize (it);
	  if (xColType == Table::Text)
	    {
	      if (xLabels.contains (xval) == 0)
		xLabels << xval;
	      X[it - 1] = (double) (xLabels.findIndex (xval) + 1);
	    }
	  else if (xColType == Table::Time)
	    {
	      QTime time = QTime::fromString (xval, Qt::TextDate);
	      if (time.isValid ())
		X[it - 1] = time0.msecsTo (time);
	      else
		X[it - 1] = 0;
	    }
	  else if (xColType == Table::Date)
	    {
	      QDate d = QDate::fromString (xval, Qt::ISODate);
	      if (d.isValid ())
		X[it - 1] = (double) date.daysTo (d);
	    }
	  else
	    X[it - 1] = xval.toDouble ();

	  if (yColType == Table::Text)
	    {
	      yLabels << yval;
	      Y[it - 1] = (double) it;
	    }
	  else
	    Y[it - 1] = yval.toDouble ();
	}
    }

  if (!it)
    return false;

  int curveID = -1;
  QwtPlotCurve *c = 0;
  if (style == VerticalBars)
    {
      c = new QwtBarCurve (QwtBarCurve::Vertical, w->colName (ycol));
      curveID = d_plot->insertCurve (c);
      c->setStyle (QwtPlotCurve::UserCurve);
    }
  else if (style == HorizontalBars)
    {
      c = new QwtBarCurve (QwtBarCurve::Horizontal, w->colName (ycol));
      curveID = d_plot->insertCurve (c);
      c->setStyle (QwtPlotCurve::UserCurve);
    }
  else if (style == Histogram)
    {
      c = new QwtHistogram (w->colName (ycol));
      curveID = d_plot->insertCurve (c);
      c->setStyle (QwtPlotCurve::UserCurve);
    }
  else
    {
      c = new QwtPlotCurve (w->colName (ycol));
      curveID = d_plot->insertCurve (c);
    }

  associations << w->colName (xcol) + "," + w->colName (ycol);

  n_curves++;
  c_attr.append( new CurveAttr( curveID, style ) );

  c->setPen (QPen (black, 
		   widthLine,
		   Qt::SolidLine,
		   Qt::FlatCap,
		   Qt::BevelJoin));

  if (style == Histogram)
    initHistogram (n_curves-1, Y, it);
  else if (style == HorizontalBars)
    c->setData (Y, X, it);
  else
    c->setData (X, Y, it);

  if (xColType == Table::Text)
    {
      if (style == HorizontalBars)
	{
	  axesFormatInfo[QwtPlot::yLeft] = w->colName (xcol);
	  axesFormatInfo[QwtPlot::yRight] = w->colName (xcol);
	  axisType[QwtPlot::yLeft] = Txt;
	  d_plot->setAxisScaleDraw (QwtPlot::yLeft,
				    new QwtTextScaleDraw (xLabels));
	}
      else
	{
	  axesFormatInfo[QwtPlot::xBottom] = w->colName (xcol);
	  axesFormatInfo[QwtPlot::xTop] = w->colName (xcol);
	  axisType[QwtPlot::xBottom] = Txt;
	  d_plot->setAxisScaleDraw (QwtPlot::xBottom,
				    new QwtTextScaleDraw (xLabels));
	}
    }
  else if (xColType == Table::Time)
    {
      QString fmtInfo =
	time0.toString (Qt::TextDate) + ";" + w->columnFormat (xcol);
      if (style == HorizontalBars)
	setLabelsDateTimeFormat (QwtPlot::yLeft, Time, fmtInfo);
      else
	setLabelsDateTimeFormat (QwtPlot::xBottom, Time, fmtInfo);
    }
  else if (xColType == Table::Date)
    {
      QString fmtInfo = date.toString (Qt::ISODate) + ";yyyy-MM-dd";
      if (style == HorizontalBars)
	setLabelsDateTimeFormat (QwtPlot::yLeft, Date, fmtInfo);
      else
	setLabelsDateTimeFormat (QwtPlot::xBottom, Date, fmtInfo);
    }

  if (yColType == Table::Text)
    {
      axesFormatInfo[QwtPlot::yLeft] = w->colName (ycol);
      axesFormatInfo[QwtPlot::yRight] = w->colName (ycol);
      axisType[QwtPlot::yLeft] = Txt;
      d_plot->setAxisScaleDraw (QwtPlot::yLeft,
				new QwtTextScaleDraw (yLabels));
    }

  addLegendItem (w->colName (ycol));
  //updatePlot ();
  d_plot->replot();
  return true;
}

void
Graph::plotVectorCurve (Table * w, const QStringList & colList, int style)
{
  int xcol = w->colIndex (colList[0]);
  int ycol = w->colIndex (colList[1]);
  int endXCol = w->colIndex (colList[2]);
  int endYCol = w->colIndex (colList[3]);
  QMemArray < double >X (1), Y (1), X2 (1), Y2 (1);
  int it = 0;
  for (int i = 0; i < w->tableRows (); i++)
    {
      QString xval = w->text (i, xcol);
      QString yval = w->text (i, ycol);
      QString xend = w->text (i, endXCol);
      QString yend = w->text (i, endYCol);
      if (!xval.isEmpty () && !yval.isEmpty () && !xend.isEmpty ()
	  && !yend.isEmpty ())
	{
	  it++;
	  X.resize (it);
	  Y.resize (it);
	  X2.resize (it);
	  Y2.resize (it);
	  Y[it - 1] = yval.toDouble ();
	  X[it - 1] = xval.toDouble ();
	  Y2[it - 1] = yend.toDouble ();
	  X2[it - 1] = xend.toDouble ();
	}
    }

  if (!it)
    return;

  QString label = colList[1];
  associations << colList[0] + "," + colList[1] + "," +
    colList[2] + "," + colList[3] ;

  VectorCurve *v = 0;
  if (style == VectXYAM)
    {
      v = new VectorCurve (VectorCurve::XYAM, label);
      v->setPosition(VectorCurve::Middle);
    }
  else
    v = new VectorCurve (VectorCurve::XYXY, label);

  if (!v)
    return;

  v->setVectorEnd (X2, Y2);

  n_curves++;
  c_attr.append( new CurveAttr( d_plot->insertCurve(v), style ) );

  v->setStyle (QwtPlotCurve::NoCurve);
  v->setData (X, Y, it);

  addLegendItem (label);
  //updatePlot ();
  d_plot->replot();
}

void
Graph::updateVectorsLayout (Table * w, int curve, int colorIndex, int width,
			    int arrowLength, int arrowAngle, bool filled,
			    int position, const QString & xEndColName,
			    const QString & yEndColName)
{
  VectorCurve *vect = (VectorCurve *) this->curve (curve);
  if (!vect)
    return;

  QColor c = ColorBox::color (colorIndex);
  QStringList cols = QStringList::split (',', associations[curve], false);
  if (vect->width () == width && vect->color () == c &&
      vect->headLength () == arrowLength && vect->headAngle () == arrowAngle
      && vect->position () == position && vect->filledArrowHead () == filled
      && cols[2] == xEndColName && cols[3] == yEndColName)
    return;

  vect->setColor (c);
  vect->setWidth (width);
  vect->setHeadLength (arrowLength);
  vect->setHeadAngle (arrowAngle);
  vect->fillArrowHead (filled);
  vect->setPosition (position);

  if (cols[2] != xEndColName || cols[3] != yEndColName)
    {
      int r = w->tableRows ();
      int xcol = w->colIndex (cols[0]);
      int ycol = w->colIndex (cols[1]);
      int endXCol = w->colIndex (xEndColName);
      int endYCol = w->colIndex (yEndColName);
      int size = vect->dataSize ();
      QMemArray < double >X (size), Y (size);
      int i, it = 0;
      for (i = 0; i < r; i++)
	{
	  QString xval = w->text (i, xcol);
	  QString yval = w->text (i, ycol);
	  QString xend = w->text (i, endXCol);
	  QString yend = w->text (i, endYCol);
	  if (!xval.isEmpty () && !yval.isEmpty () && !xend.isEmpty ()
	      && !yend.isEmpty ())
	    {
	      it++;
	      X[it - 1] = xend.toDouble ();
	      Y[it - 1] = yend.toDouble ();
	    }
	}
      vect->setVectorEnd (X, Y);
      cols[2] = xEndColName;
      cols[3] = yEndColName;
      associations[curve] = cols.join (",");
    }

  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::setVectorsLook (int curve, const QColor & c, int width,
		       int arrowLength, int arrowAngle, bool filled,
		       int position)
{
  VectorCurve *vect = (VectorCurve *) this->curve (curve);
  if (!vect)
    return;

  vect->setColor (c);
  vect->setWidth (width);
  vect->setHeadLength (arrowLength);
  vect->setHeadAngle (arrowAngle);
  vect->fillArrowHead (filled);
  vect->setPosition (position);
}

void
Graph::updatePlot ()
{
  //fprintf(stderr,"updatePlot\n");
  if (autoscale && !removePointsEnabled && !movePointsEnabled &&
      !zoomOn () && !rangeSelectorsEnabled)
    {
      setAutoScale();
      d_plot->replot();
    }
  else
    {
      //fprintf(stderr,"No Autoscale because\n");
      //if(!autoscale) fprintf(stderr,"autoscale is false\n");
      //if(removePointsEnabled) fprintf(stderr,"removePointsEnabled\n");
      d_plot->replot();
      //we need these here because d_plot->replot() sometimes change scales
      d_zoomer[0]->setZoomBase ();
      d_zoomer[1]->setZoomBase ();
      updateMarkersBoundingRect (); //keep markers on canvas area
      d_plot->replot();
    }
}

void
Graph::activateLayout()
{
  int lw = d_plot->lineWidth ();
  QRect rect = QRect (d_plot->x () + lw, 
                      d_plot->y () + lw, 
                      d_plot->width () - 2 * lw,
	              d_plot->height () - 2 * lw);
  QwtPlotLayout *plotLayout = d_plot->plotLayout ();
  plotLayout->activate (d_plot, rect, 0);
}

void
Graph::updateScale ()
{
  //fprintf(stderr,"updateScale\n");
  d_plot->replot ();
  updateMarkersBoundingRect ();

  updateSecondaryAxis (QwtPlot::xTop);
  updateSecondaryAxis (QwtPlot::yRight);

  d_plot->replot ();		//TO DO: avoid 2nd replot!
}

void
Graph::setBarsGap (int curve, int gapPercent, int offset)
{
  QwtBarCurve *bars = (QwtBarCurve *) this->curve (curve);
  if (!bars || (bars->gap () == gapPercent && bars->offset () == offset))
    return;

  bars->setGap (gapPercent);
  bars->setOffset (offset);

  QString yColName = bars->title ().text ();
  for (int i = 0; i < n_curves; i++)
    {
      QStringList lst = QStringList::split(',',associations[i],false);
      int ct = curveType(i);
      if (lst.contains (yColName) && i != curve
	  && (ct == XErrorBars || ct == YErrorBars))
	{
	  QwtErrorPlotCurve *er = (QwtErrorPlotCurve *) this->curve (i);
	  if (!er)
	    continue;

	  if (bars->orientation () == QwtBarCurve::Vertical)
	    er->setXDataOffset (bars->dataOffset ());
	  else
	    er->setYDataOffset (bars->dataOffset ());
	}
    }
}

void
Graph::removePie ()
{
  if (legendMarkerID >= 0)
    {
      LegendMarker *mrk = (LegendMarker *) d_plot->marker (legendMarkerID);
      if (mrk)
	mrk->setText (QString::null);
    }

  d_plot->removeCurve( curveKey(0) );
  //d_plot->replot (); //commented out by S.Yamamoto

  n_curves = 0;
  c_attr.clear();

  associations.clear ();
  piePlot = FALSE;
  emit modifiedGraph ();
}

void
Graph::removeCurves (const QStringList & titles)
{
  bool modified = false;
  for(int i = 0; i<(int)titles.count(); i++)
    {
      QString curveTitle = titles[i];

      //qDebug("removing : %s",curveTitle.ascii());

      //First look into the titles list...
      int index = curveIndex( curveTitle );

      //...then look into the plot associations list
      if (index < 0)
	index = associations.findIndex (curveTitle);	

      if (index < 0)
	{
	  QMessageBox::critical (0, tr ("QtiPlot - Error"),
				 tr("There is no curve called '%1' on this layer.").arg (curveTitle));
	}
      else
	{
	  removeCurve (index, false);
	  modified = true;
	}
    }

  if(modified)
    {
      updatePlot();
      emit modifiedGraph ();
    }
}

// remove curve without replot
void
Graph::removeCurve(int index, bool update)
{
  if (index < 0 || index >= n_curves)
    {
      QMessageBox::critical (0, tr ("QtiPlot - Error"),
			     tr
			     ("There is no curve with index %1 on this layer.").
			     arg (index) + "\n" +
			     tr
			     ("Valid indexes must have values between 0 and %1").
			     arg (n_curves - 1));
      return;
    }

  resetErrorBarsOffset (index);

  QStringList::Iterator it = associations.at (index);
  associations.remove (it);

  removeLegendItem (index);

  d_plot->removeCurve( curveKey(index) );
  n_curves--;

  if (piePlot)
    {
      piePlot = FALSE;
      c_attr.clear();
    }
  else
    {
      if (rangeSelectorsEnabled && index == selectedCurve)
	{
	  if (n_curves > 0)
	    shiftCurveSelector (1);
	  else
	    {
	      d_plot->removeMarker (startID);
	      d_plot->removeMarker (endID);
	      startID = -1;
	      endID = -1;
	      endPoint = -1;
	      selectedCurve = -1;
	      selectedPoint = -1;
	      selectedCursor = -1;
	    }
	}
      c_attr.remove(index);
    }
  if(update)
    {
      updatePlot();
      emit modifiedGraph ();
    }
}

void
Graph::resetErrorBarsOffset (int index)
{
  int ct0 = curveType(index);
  if (ct0 != VerticalBars && ct0 != Histogram && ct0 != HorizontalBars)
    return;

  QwtBarCurve *bars = (QwtBarCurve *) this->curve (index);
  if (!bars || !bars->offset ())
    return;

  QString yColName = bars->title ().text ();
  for (int i = 0; i < n_curves; i++)
    {
      QStringList lst = QStringList::split(',',associations[i],false);
      int ct = curveType(i);
      if (lst.contains (yColName) && i != index
	  && (ct == XErrorBars || ct == YErrorBars))
	{
	  QwtErrorPlotCurve *er = (QwtErrorPlotCurve *) this->curve (i);
	  if (!er)
	    continue;

	  if (bars->orientation () == QwtBarCurve::Vertical)
	    er->setXDataOffset (0);
	  else
	    er->setYDataOffset (0);
	}
    }
}

void
Graph::removeLegendItem (int index)
{
  int ct = curveType(index);
  if (legendMarkerID < 0 || ct == XErrorBars || ct == YErrorBars)
    return;

  LegendMarker *mrk = (LegendMarker *) d_plot->marker (legendMarkerID);
  if (!mrk)
    return;

  if (piePlot)
    {
      mrk->setText (QString::null);
      return;
    }

  QString text = mrk->getText ();
  QStringList items = QStringList::split ('\n', text, FALSE);

  QStringList l = items.grep ("\\c{" + QString::number (index + 1) + "}");
  if(l.count()==0) return;

  for(int i = 0; i < (int)l.count(); i++)
    items.remove (l[i]);		//remove the corresponding legend string 
  QRegExp rx("\\\\c\\{(\\d+)\\}(.*)",true,false);
  for (int i = 0; i < (int) items.count (); i++)
    { //set new curves indexes in legend text
      int pos = rx.search( items[i] );
      if(pos != -1)
	{
	  QStringList list = rx.capturedTexts();
	  int old_index = list[1].toInt();
	  if(index+1 < old_index)
	    items[i] = "\\c{" +  QString::number(old_index - 1) +"}"+list[2];
	}
    }
  text = items.join ("\n") + "\n";
  mrk->setText (text);
}

void
Graph::addLegendItem (const QString & colName)
{
  if (legendMarkerID >= 0)
    {
      LegendMarker *mrk = (LegendMarker *) d_plot->marker (legendMarkerID);
      if (mrk)
	{
	  QString text = mrk->getText ();
	  if (text.endsWith ("\n", true))
	    text.append ("\\c{" + QString::number (curves ()) + "}" +
			 colName + "\n");
	  else
	    text.append ("\n\\c{" + QString::number (curves ()) + "}" +
			 colName + "\n");

	  mrk->setText (text);
	}
    }
}

void
Graph::contextMenuEvent (QContextMenuEvent * e)
{
  if (selectedMarker >= 0)
    return;

  QPoint pos = d_plot->canvas ()->mapFrom (d_plot, e->pos ());
  int dist, point;
  int key = d_plot->closestCurve (pos.x (), pos.y (), dist, point);
  int index = curveIndex(key);

  if (index>=0 && dist < 5)		//5 pixels tolerance
    emit showCurveContextMenu (index);
  else
    emit showContextMenu ();

  e->accept ();
}

void
Graph::closeEvent (QCloseEvent * e)
{
  emit closedGraph ();
  e->accept ();
}

bool
Graph::zoomOn ()
{
  return (d_zoomer[0]->isEnabled () || d_zoomer[1]->isEnabled ());
}

void
Graph::zoomed (const QwtDoubleRect &rect)
{
  updateScale();
  emit modifiedGraph ();
}

void
Graph::zoom (bool on)
{
  d_zoomer[0]->setEnabled (on);
  d_zoomer[1]->setEnabled (on);

  for (int i = 0; i < n_curves; i++)
    {
      Spectrogram *sp = (Spectrogram *) this->curve (i);
      if (sp && sp->rtti () == QwtPlotItem::Rtti_PlotSpectrogram)
	{
	  if (sp->colorScaleAxis () == QwtPlot::xBottom
	      || sp->colorScaleAxis () == QwtPlot::yLeft)
	    d_zoomer[0]->setEnabled (false);
	  else
	    d_zoomer[1]->setEnabled (false);
	}
    }

  QCursor cursor = QCursor (QPixmap (lens_xpm), -1, -1);
  if (on)
    d_plot->canvas ()->setCursor (cursor);
  else
    d_plot->canvas ()->setCursor (Qt::arrowCursor);
}

void
Graph::zoomOut ()
{
  //fprintf(stderr,"zoomOut\n");
  //fprintf(stderr,"zoomRectIndex = %d\n\n",d_zoomer[0]->zoomRectIndex());
  if(d_zoomer[0]->zoomRectIndex()>0)
    {
      d_zoomer[0]->zoom (-1);
      d_zoomer[1]->zoom (-1);
    }
  // Are these necessary? because it is done in zoomed().
  //updateSecondaryAxis (QwtPlot::xTop);
  //updateSecondaryAxis (QwtPlot::yRight);
}

void
Graph::drawText (bool on)
{
  QCursor c = QCursor (IbeamCursor);
  if (on)
    d_plot->canvas ()->setCursor (c);
  else
    d_plot->canvas ()->setCursor (Qt::arrowCursor);

  drawTextOn = on;
}

void
Graph::insertImageMarker (ImageMarker * mrk)
{
  QPixmap photo = mrk->image ();
  ImageMarker *mrk2 = new ImageMarker (photo);

  int imagesOnPlot = d_images.size ();
  d_images.resize (++imagesOnPlot);
  d_images[imagesOnPlot - 1] = d_plot->insertMarker (mrk2);

  mrk2->setFileName (mrk->getFileName ());
  mrk2->setBoundingRect (mrk->boundingRect ());
}

void
Graph::insertImageMarker (const QPixmap & photo, const QString & fileName)
{
  ImageMarker *mrk = new ImageMarker (photo);
  int mrkID = d_plot->insertMarker (mrk);

  int imagesOnPlot = d_images.size ();
  d_images.resize (++imagesOnPlot);
  d_images[imagesOnPlot - 1] = mrkID;

  mrk->setFileName (fileName);
  mrk->setOrigin (QPoint (10, 10));

  QSize picSize = photo.size ();
  int w = d_plot->canvas ()->width ();
  if (picSize.width () > w)
    picSize.setWidth (w);
  int h = d_plot->canvas ()->height ();
  if (picSize.height () > h)
    picSize.setHeight (h);

  mrk->setSize (picSize);
  d_plot->replot ();

  emit modifiedGraph ();
}

void
Graph::insertImageMarker (const QStringList & lst, int fileVersion,
			  const QString & imageDir)
{
  QString fn = lst[1];
  QFile f (fn);
  if (!f.exists ())
    {
      if(QMessageBox::question(0, tr ("QtiPlot - File open error"),
			       tr("Image file: <p><b> %1 </b><p>does not exist anymore! Do you want to indicate the file position?").arg (fn),
			       tr ("&Yes"), tr ("&No"), QString::null, 0, 1))
	return;

      QStringList list = QImage::inputFormatList ();
      QString filter = "Images (*.jpg *.JPG ", aux;
      for (int i = 0; i < (int) list.count (); i++)
	{
	  aux = "*." + (list[i]).lower () + " *." + list[i] + " ";
	  filter += aux;
	}
      filter += ");;";
	  
      aux = "*.jpg *.JPG;; ";
      filter += aux;
      for (int i = 0; i < (int) list.count (); i++)
	{
	  aux = "*." + (list[i]).lower () + " *." + list[i] + ";;";
	  filter += aux;
	}

      fn = QFileDialog::getOpenFileName(imageDir, filter, this, 0,
					tr("QtiPlot - Indicate image file"),
					0, TRUE);
      if (fn.isEmpty ()) return;
    }

  QPixmap photo;
  if (fn.contains (".jpg", false))
    photo.load (fn, "JPEG", QPixmap::Auto);
  else
    {
      QFileInfo fi (fn);
      QString baseName = fi.fileName ();
      QString type = baseName.section('.',-1);
      photo.load (fn, type.upper (), QPixmap::Auto);
    }
  
  ImageMarker *mrk = new ImageMarker (photo);
  int mrkID = d_plot->insertMarker (mrk);
  mrk->setFileName (fn);
  
  if (fileVersion < 86)
    {
      mrk->setOrigin (QPoint (lst[2].toInt (), lst[3].toInt ()));
      mrk->setSize (QSize (lst[4].toInt (), lst[5].toInt ()));
    }
  else
    {
      mrk->
	setBoundingRect (QwtDoubleRect
			 (lst[2].toDouble (), lst[3].toDouble (),
			  lst[4].toDouble (), lst[5].toDouble ()));
    }
  
  int imagesOnPlot = d_images.size ();
  d_images.resize (++imagesOnPlot);
  d_images[imagesOnPlot - 1] = mrkID;
}

void
Graph::drawLine (bool on, bool arrow)
{
  drawLineOn = on;
  drawArrowOn = arrow;
  if (!on)
    emit drawLineEnded (true);
}

void
Graph::setSubtractLine(const QString & curveTitle)
{
  //QStringList cl = curvesList ();
  //int index = cl.findIndex (curveTitle);
  int index = curveIndex( curveTitle );
  if (index < 0)
    return;

  selectedCurve = index;
  subtractLineOn = true;
}


void
Graph::setSubtractLine(bool on)
{
  subtractLineOn = on;
  if(on)
    {
      QStringList clist = analyzableCurvesList();

      int crvs = clist.count();

      if(crvs==1)
	selectedCurve = curveIndex(clist[0]);
      else if(crvs>1)
	{
	  subtractLineOn = false;
	  analysisDialog *ad =
	    new analysisDialog (0, tr ("Curve:"), "analysisDialog", TRUE,
				WDestructiveClose);
	  connect (ad, SIGNAL (options (const QString &)),
		   this, SLOT (setSubtractLine(const QString &)));
	  ad->setCurveNames ( clist, -1 );
	  ad->showNormal ();
	  ad->setActiveWindow ();	  
	}
      else
	{
	  QMessageBox::warning (this, tr ("QtiPlot - Warning"),
				tr
				("<h4>There are no curve available in this layer.</h4>"
				 "<p><h4>Please add a curve and try again!</h4>"));
	  subtractLineOn = false;
	}
    }
}

void
Graph::subtractLine(const QPoint & p1, const QPoint & p2)
{
  subtractLineOn = false;

  const QwtPlotCurve *c = curve (selectedCurve);

  if (!c)
    {
      fprintf(stderr,"invalid curve\n");
      return;
    }

  // qDebug("curve = %s\n",c->title().text().ascii());

  double x1 = d_plot->invTransform (c->xAxis(), p1.x ());
  double y1 = d_plot->invTransform (c->yAxis(), p1.y ());
  double x2 = d_plot->invTransform (c->xAxis(), p2.x ());
  double y2 = d_plot->invTransform (c->yAxis(), p2.y ());

  if(x1==x2) return;

  int n = c->dataSize ();
  QMemArray <double> dat(n);
  const QwtScaleEngine *se = d_plot->axisScaleEngine( c->yAxis() );
  if(se->transformation()->type() == QwtScaleTransformation::Log10)
    {
      if (y1<=0 || y2<=0)
	{
	  qDebug("y value is negative. this should not happen.");
	  return;
	}
      double b = log(y1/y2)/(x1-x2);
      double a = y1 * exp( - b * x1 );
      for (int i = 0; i < n; i++)
	dat[i] = c->y (i) - a * exp( b * c->x(i) );
    }
  else
    {
      double a = (y2-y1)/(x2-x1);
      double b = y1 - a * x1;

      for (int i = 0; i < n; i++)
	dat[i] = c->y(i) - ( c->x(i) * a + b );
    }
  emit updateTableColumn (c->title().text(), dat);
}

void
Graph::modifyFunctionCurve (int curve, int type, const QStringList & formulas,
			    const QString & var, QValueList < double >&ranges,
			    int points)
{
  FunctionCurve *c = (FunctionCurve *) this->curve (curve);
  if (!c)
    return;

  if (c->functionType () == type &&
      c->variable () == var &&
      c->formulas () == formulas &&
      c->startRange () == ranges[0] &&
      c->endRange () == ranges[1] && c->dataSize () == points)
    return;

  bool error = FALSE;
  QMemArray < double >X (points), Y (points);
  double step = (ranges[1] - ranges[0]) / (double) (points - 1);
  QString oldLegend = c->legend ();
  if (type == FunctionCurve::Normal)
    {
      myParser parser;
      double x;
      try
      {
	parser.DefineVar (var.ascii (), &x);
	parser.SetExpr (formulas[0].ascii ());
	X[0] = ranges[0];
	x = ranges[0];
	Y[0] = parser.Eval ();
	for (int i = 1; i < points; i++)
	  {
	    x += step;
	    X[i] = x;
	    Y[i] = parser.Eval ();
	  }
      }
      catch (mu::ParserError & e)
      {
	QMessageBox::critical (this, tr ("QtiPlot - Input function error"),
			       e.GetMsg ().c_str());
	error = TRUE;
      }
    }
  else if (type == FunctionCurve::Parametric || type == FunctionCurve::Polar)
    {
      QStringList aux = formulas;
      if (type == FunctionCurve::Polar)
	{
	  QString swap = aux[0];
	  aux[0] = "(" + swap + ")*cos(" + aux[1] + ")";
	  aux[1] = "(" + swap + ")*sin(" + aux[1] + ")";
	}
      myParser xparser;
      myParser yparser;
      double par;
      try
      {
	xparser.DefineVar (var.ascii (), &par);
	yparser.DefineVar (var.ascii (), &par);
	xparser.SetExpr (aux[0].ascii ());
	yparser.SetExpr (aux[1].ascii ());
	par = ranges[0];
	for (int i = 0; i < points; i++)
	  {
	    X[i] = xparser.Eval ();
	    Y[i] = yparser.Eval ();
	    par += step;
	  }
      }
      catch (mu::ParserError & e)
      {
	QMessageBox::critical (0, tr ("QtiPlot - Input function error"),
			       e.GetMsg ().c_str());
	error = TRUE;
      }
    }

  if (error)
    return;

  c->setData (X, Y, points);
  c->setFunctionType ((FunctionCurve::FunctionType) type);
  c->setRange (ranges[0], ranges[1]);
  c->setFormulas (formulas);
  c->setVariable (var);

  if (legendMarkerID >= 0)
    {				//update the legend marker     
      LegendMarker *mrk = (LegendMarker *) d_plot->marker (legendMarkerID);
      if (mrk)
	{
	  QString text = (mrk->getText ()).replace (oldLegend, c->legend ());
	  mrk->setText (text);
	}
    }
  updatePlot ();
  emit modifiedGraph ();
  emit modifiedFunction ();
}

QString
Graph::generateFunctionName ()
{
  QString newName = "F1";
  int index = 1;
  QStringList lst;
  for (int i = 0; i < n_curves; i++)
    {
      const QwtPlotCurve *c = this->curve (i);
      if (c && c->rtti () == FunctionCurve::RTTI
	  && c->title ().text ().startsWith ("F"))
	lst << c->title ().text ();
    }

  while (lst.contains (newName))
    {
      newName = "F" + QString::number (++index);
    }
  return newName;
}

void
Graph::addFunctionCurve (int type, const QStringList & formulas,
			 const QString & var, QValueList < double >&ranges,
			 int points, const QString & title)
{
  bool error = FALSE;
  QMemArray < double >X (points), Y (points);
  QString name;
  if (!title.isEmpty ())
    name = title;
  else
    name = generateFunctionName ();

  double step = (ranges[1] - ranges[0]) / (double) (points - 1);

  if (type == FunctionCurve::Normal)
    {
      myParser parser;
      double x;
      try
      {
        //qDebug("formula = %s",formulas[0].ascii());
	parser.DefineVar (var.ascii (), &x);
	parser.SetExpr (formulas[0].ascii ());

	X[0] = ranges[0];
	x = ranges[0];
	Y[0] = parser.Eval ();
	for (int i = 1; i < points; i++)
	  {
	    x += step;
	    X[i] = x;
	    Y[i] = parser.Eval ();
	  }
      }
      catch (mu::ParserError & e)
      {
	QMessageBox::critical (this, tr ("QtiPlot - Input function error"),
			       e.GetMsg ().c_str());
	error = TRUE;
      }
    }
  else if (type == FunctionCurve::Parametric || type == FunctionCurve::Polar)
    {
      QStringList aux = formulas;
      myParser xparser;
      myParser yparser;
      double par;
      if (type == FunctionCurve::Polar)
	{
	  QString swap = aux[0];
	  aux[0] = "(" + swap + ")*cos(" + aux[1] + ")";
	  aux[1] = "(" + swap + ")*sin(" + aux[1] + ")";
	}
      try
      {
	xparser.DefineVar (var.ascii (), &par);
	yparser.DefineVar (var.ascii (), &par);
	xparser.SetExpr (aux[0].ascii ());
	yparser.SetExpr (aux[1].ascii ());
	par = ranges[0];
	for (int i = 0; i < points; i++)
	  {
	    X[i] = xparser.Eval ();
	    Y[i] = yparser.Eval ();
	    par += step;
	  }
      }
      catch (mu::ParserError & e)
      {
	QMessageBox::critical (this, tr ("QtiPlot - Input function error"),
			       e.GetMsg ().c_str());
	error = TRUE;
      }
    }

  if (error)
    return;

  associations << name;

  FunctionCurve *c =
    new FunctionCurve ((const FunctionCurve::FunctionType) type, name);
  c->setPen (QPen (black, 
		   widthLine,
		   Qt::SolidLine,
		   Qt::FlatCap,
		   Qt::BevelJoin));
  c->setData (X, Y, points);
  c->setRange (ranges[0], ranges[1]);
  c->setFormulas (formulas);
  c->setVariable (var);

  n_curves++;
  c_attr.append( new CurveAttr( d_plot->insertCurve(c), Line ) );

  addLegendItem (c->legend ());
  //updatePlot ();
  d_plot->replot();
  emit modifiedGraph ();
}

void
Graph::insertFunctionCurve (const QString & formula, int points,
			    int fileVersion)
{
  int type;
  QStringList formulas;
  QString var, name = QString::null;
  QValueList < double >ranges;

  //formula may contain ','!!
  //QStringList curve = QStringList::split (',', formula, TRUE);
  QStringList curve;
  int p0 = 0;
  int p = 0;
  int level = 0;
  QChar ch;
  while((ch=formula[p])!=QChar::null)
  {
    if(ch=='(') level++;
    else if(ch==')') level--;
    else if(ch==',' && level==0) 
    {
      curve.append(formula.mid(p0,p-p0));
      p0 = p+1;
    }
    p++;
  }
  curve.append(formula.mid(p0,p-p0));

  if (fileVersion < 87)
    {
      if (curve[0][0] == 'f')
	{
	  type = FunctionCurve::Normal;
	  formulas += curve[0].section ('=', 1, 1);
	  var = curve[1];
	  ranges += curve[2].toDouble ();
	  ranges += curve[3].toDouble ();
	}
      else if (curve[0][0] == 'X')
	{
	  type = FunctionCurve::Parametric;
	  formulas += curve[0].section ('=', 1, 1);
	  formulas += curve[1].section ('=', 1, 1);
	  var = curve[2];
	  ranges += curve[3].toDouble ();
	  ranges += curve[4].toDouble ();
	}
      else if (curve[0][0] == 'R')
	{
	  type = FunctionCurve::Polar;
	  formulas += curve[0].section ('=', 1, 1);
	  formulas += curve[1].section ('=', 1, 1);
	  var = curve[2];
	  ranges += curve[3].toDouble ();
	  ranges += curve[4].toDouble ();
	}
    }
  else
    {
      type = curve[0].toInt ();
      name = curve[1];
      if (type == FunctionCurve::Normal)
	{
	  formulas << curve[2];
	  var = curve[3];
	  ranges += curve[4].toDouble ();
	  ranges += curve[5].toDouble ();
	}
      else if (type == FunctionCurve::Polar
	       || type == FunctionCurve::Parametric)
	{
	  formulas << curve[2];
	  formulas << curve[3];
	  var = curve[4];
	  ranges += curve[5].toDouble ();
	  ranges += curve[6].toDouble ();
	}
    }
  addFunctionCurve (type, formulas, var, ranges, points, name);
}

void
Graph::createWorksheet (const QString & name)
{
  //QValueList < int >keys = d_plot->curveKeys ();
  for (int i = 0; i < n_curves; i++)
    {
      const QwtPlotCurve *cv =
	(QwtPlotCurve *) d_plot->curve( curveKey(i) );
      if (cv->title ().text () == name)
	{
	  int size = cv->dataSize ();
	  QString text = "1\t2\n";
	  for (int j = 0; j < size; j++)
	    {
	      text += QString::number (cv->x (j)) + "\t";
	      text += QString::number (cv->y (j)) + "\n";
	    }
	  emit createTable ("Table1", size, 2, text);
	  return;
	}
    }
}

bool
Graph::lineProfile ()
{
  return lineProfileOn;
}

QString
Graph::saveToString (bool saveAsTemplate)
{
  QString s = "<graph>\n";
  s += "ggeometry\t";
  QPoint p = this->pos ();
  s += QString::number (p.x ()) + "\t";
  s += QString::number (p.y ()) + "\t";
  s += QString::number (this->frameGeometry ().width ()) + "\t";
  s += QString::number (this->frameGeometry ().height ()) + "\n";
  s += saveTitle ();
  s += "Background\t" + d_plot->paletteBackgroundColor ().name () + "\n";
  s += "Margin\t" + QString::number (d_plot->margin ()) + "\n";
  s += "Spacing\t" + QString::number (d_spacing) + "\n";
  s +=
    "Border\t" + QString::number (d_plot->lineWidth ()) + "\t" +
    d_plot->frameColor ().name () + "\n";
  s += saveGridOptions ();
  s += saveEnabledAxes ();
  s += "AxesTitles\t" + saveScaleTitles ();
  s += saveAxesTitleColors ();
  s += saveAxesTitleAlignement ();
  s += saveFonts ();
  //  s += saveEnabledTickLabels (); // moved to later
  s += saveAxesColors ();
  s += saveAxesBaseline ();
  s += saveCanvas ();

  if (!saveAsTemplate)
    s += saveCurves ();

  s += saveScale ();
  s += saveAxesFormulas ();
  s += saveLabelsFormat ();
  s += saveAxesLabelsType ();
  s += saveTicksType ();
  s += saveEnabledTickLabels ();
  s +=
    "TicksLength\t" + QString::number (minorTickLength ()) + "\t" +
    QString::number (majorTickLength ()) + "\n";
  s += "DrawAxesBackbone\t" + QString::number (drawAxesBackbone) + "\n";
  s += "AxesLineWidth\t" + QString::number (d_plot->axesLinewidth ()) + "\n";
  s += saveLabelsRotation ();
  s += saveMarkers ();
  s += "</graph>\n";
  return s;
}

void
Graph::showIntensityTable ()
{
  ImageMarker *mrk = (ImageMarker *) d_plot->marker (selectedMarker);
  if (!mrk)
    return;

  QPixmap pic = mrk->image ();
  emit createIntensityTable (pic);
}

void
Graph::updateMarkersBoundingRect ()
{
  //fprintf(stderr,"updateMarkersBoundingRect\n");
  int i = 0;
  /*
  for (i = 0; i < (int) d_lines.size (); i++)
    {
      LineMarker *mrkL = (LineMarker *) d_plot->marker (d_lines[i]);
      if (mrkL)
	mrkL->updateBoundingRect ();
    }
  */
  for (i = 0; i < (int) d_texts.size (); i++)
    {
      LegendMarker *mrkT = (LegendMarker *) d_plot->marker (d_texts[i]);
      if (mrkT)
	mrkT->updateOrigin ();
    }
  /*
  for (i = 0; i < (int) d_images.size (); i++)
    {
      ImageMarker *mrk = (ImageMarker *) d_plot->marker (d_images[i]);
      if (mrk)
	mrk->updateBoundingRect ();
    }
  */
}

void
Graph::updateMarkerPos()
{
  for (int i = 0; i < (int) d_texts.size (); i++)
    {
      LegendMarker *mrk = (LegendMarker *) d_plot->marker (d_texts[i]);
      mrk->updatePixelPos();
    }
}


void
Graph::resizeEvent (QResizeEvent * e)
{

  if(d_first_resize){
    d_first_resize=false;
    return;
  }

  if (ignoreResize){
    return;
  }

  this->hide();

  double scale = scaleFactor();
  if (autoScaleFonts) scaleFonts( scale );

  if(!d_textlayer)
    {
      scaleSymbols( scale );
      scaleSpaces( scale );
      scaleTickLength( scale );
      if(autoScaleFrameWidth) scaleFrame( scale );
      if (autoScaleGridWidth) scaleGrid( scale );

      if (this->isPiePlot() && curves () > 0)
	{
	  QwtPieCurve *pieCurve = (QwtPieCurve *) curve (0);
	  if (pieCurve)
	    {
	      int size = int(pieSize*scale);
	      pieCurve->setRay(size);
	    }
	}
    }

  d_plot->replot();

  this->show();  
}

void
Graph::scaleFonts (double factor)
{
  //qDebug("scaleFonts");
  
  int i;
  for (i = 0; i < (int) d_texts.size (); i++)
    {
      LegendMarker *mrk = (LegendMarker *) d_plot->marker (d_texts[i]);
      mrk->scaleFont(factor);
    }

  if(!d_textlayer)
    {
      QFont font;
      for (i = 0; i < QwtPlot::axisCnt; i++)
	{
	  font = axisFont (i);
	  font.setPointSizeFloat ((float)axisFontSize[i] * factor );
	  d_plot->setAxisFont (i, font);

	  font = axisTitleFont (i);
	  font.setPointSizeFloat ((float)axisTitleFontSize[i] * factor );
	  QwtText t = d_plot->axisTitle (i);
	  t.setFont (font);
	  d_plot->setAxisTitle (i, t);
	}

      font = d_plot->title ().font ();
      font.setPointSizeFloat ((float)titleFontSize * factor);
      QwtText t = d_plot->title ();
      t.setFont (font);
      d_plot->setTitle (t);
    }
}

void
Graph::scaleSymbols(double factor)
{
  //qDebug("scaleSymbols");

  for (int i = 0; i < n_curves; i++)
    {
      QwtPlotCurve *c = curve (i);
      if(c){
	int ct = curveType(i);
	if(autoScaleLineWidth &&
	   (ct==Graph::Line || ct==Graph::LineSymbols))
	  {
	    QPen pen = c->pen();
	    int penwidth = QMAX( qRound(c_attr.at(i)->line_width * factor), 1);
	    pen.setWidth(penwidth);
	    c->setPen(pen);
	  }
	if(ct==Graph::Scatter || ct==Graph::LineSymbols)
	  {
	    QwtSymbol s = c->symbol();
	    QPen pen = s.pen();
	    if(autoScaleLineWidth)
	      {
		int penwidth = QMAX( qRound( c_attr.at(i)->symbol_edge_width * factor), 1);
		pen.setWidth(penwidth);
	      }
	    int size = QMAX( qRound( c_attr.at(i)->symbol_size * factor), 1);
	    QwtSymbol ss = QwtSymbol (s.style(), s.brush(), pen, 
				      QSize (size, size));
	    c->setSymbol (ss);
	  }
      }
    }
}

void
Graph::scaleSpaces(double factor)
{
  //qDebug("scaleSpaces");

  int d = qRound(d_base_margin * factor);
  d_plot->setMargin(d);

  int s = qRound(d_base_spacing * factor);
  for (int i = 0; i < QwtPlot::axisCnt; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale) scale->setSpacing(s);
    }
  d_spacing = s;
}

void
Graph::scaleTickLength(double factor)
{
  //qDebug("scaleTickLength");

  int minor = QMAX( qRound( d_minor_ticklength * factor ), 1);
  int major = QMAX( qRound( d_major_ticklength * factor ), 1);

  QValueList < int >majTicksType = d_plot->getMajorTicksType ();
  QValueList < int >minTicksType = d_plot->getMinorTicksType ();
  for (int i = 0; i < 4; i++)
    setAxisTicksLength (i, majTicksType[i], minTicksType[i], minor, major );
}

void
Graph::scaleFrame(double factor)
{
  //qDebug("scaleFrame");

  QwtPlotCanvas *canvas = (QwtPlotCanvas *) d_plot->canvas ();
  int width = canvas->lineWidth();
  if (width>0)
    {
      width = qRound( d_frame_width * factor );
      if (width<1) width = 1;
      canvas->setLineWidth (width);
    }

  width = QMAX( qRound( d_axes_width * factor ), 1);
  d_plot->setAxesLinewidth (width);
}

void
Graph::scaleGrid(double factor)
{
  //qDebug("scaleGrid");

  int maj = QMAX( qRound(d_grid_major * factor), 1 );
  int min = QMAX( qRound(d_grid_minor * factor), 1 );

  QPen majPen = QPen (ColorBox::color(grid.majorCol), maj,
		      getPenStyle(grid.majorStyle));
  grid.majorWidth = maj;
  d_plot->grid ()->setMajPen (majPen);

  QPen minPen = QPen (ColorBox::color(grid.minorCol), min,
		      getPenStyle(grid.minorStyle));
  d_plot->grid ()->setMinPen (minPen);
  grid.minorWidth = min;
}

void
Graph::setAxisTitleInverted(int axis, bool inv)
{
  QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (axis);
  if (scale) scale->setTitleInverted(inv);
}

void
Graph::changeMargin (int d)
{
  if (d_plot->margin () == d)
    return;

  d_plot->setMargin (d);
  d_base_margin = d / scaleFactor();

  emit modifiedGraph ();
}

void
Graph::changeSpacing (int s)
{
  for (int i = 0; i < QwtPlot::axisCnt; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	scale->setSpacing(s);
    }
  d_spacing = s;
  d_base_spacing = s / scaleFactor();

  emit modifiedGraph ();
}

void
Graph::drawBorder (int width, const QColor & color)
{
  if (d_plot->frameColor () == color && width == d_plot->lineWidth ())
    return;

  QPalette pal = d_plot->palette ();
  pal.setColor (QColorGroup::Foreground, color);

  d_plot->setPalette (pal);
  d_plot->setLineWidth (width);

  int margin = d_plot->margin ();
  d_plot->setMargin (margin + 5);
  d_plot->setMargin (margin);

  emit modifiedGraph ();
}

void
Graph::setBorder (int width, const QColor & color)
{
  if (d_plot->frameColor () == color && width == d_plot->lineWidth ())
    return;

  QPalette pal = d_plot->palette ();
  pal.setColor (QColorGroup::Foreground, color);

  d_plot->setPalette (pal);
  d_plot->setLineWidth (width);
}

void
Graph::setBackgroundColor (const QColor & color)
{
  d_plot->setPaletteBackgroundColor (color);

  QwtTextLabel *title = d_plot->titleLabel ();
  title->setPaletteBackgroundColor (color);

  for (int i = 0; i < QwtPlot::axisCnt; i++)
    {
      QwtScaleWidget *scale = (QwtScaleWidget *) d_plot->axisWidget (i);
      if (scale)
	scale->setPaletteBackgroundColor (color);
    }

  emit modifiedGraph ();
}

void
Graph::setCanvasBackground (const QColor & color)
{
  d_plot->setCanvasBackground (color);
  emit modifiedGraph ();
}

Qt::BrushStyle Graph::getBrushStyle (int style)
{
  Qt::BrushStyle brushStyle;
  switch (style)
    {
    case 0:
      brushStyle = Qt::SolidPattern;
      break;
    case 1:
      brushStyle = Qt::HorPattern;
      break;
    case 2:
      brushStyle = Qt::VerPattern;
      break;
    case 3:
      brushStyle = Qt::CrossPattern;
      break;
    case 4:
      brushStyle = Qt::BDiagPattern;
      break;
    case 5:
      brushStyle = Qt::FDiagPattern;
      break;
    case 6:
      brushStyle = Qt::DiagCrossPattern;
      break;
    case 7:
      brushStyle = Qt::Dense1Pattern;
      break;
    case 8:
      brushStyle = Qt::Dense2Pattern;
      break;
    case 9:
      brushStyle = Qt::Dense3Pattern;
      break;
    case 10:
      brushStyle = Qt::Dense4Pattern;
      break;
    case 11:
      brushStyle = Qt::Dense5Pattern;
      break;
    case 12:
      brushStyle = Qt::Dense6Pattern;
      break;
    case 13:
      brushStyle = Qt::Dense7Pattern;
      break;
    }
  return brushStyle;
}

QString
Graph::penStyleName (Qt::PenStyle style)
{
  if (style == Qt::SolidLine)
    return "SolidLine";
  else if (style == Qt::DashLine)
    return "DashLine";
  else if (style == Qt::DotLine)
    return "DotLine";
  else if (style == Qt::DashDotLine)
    return "DashDotLine";
  else if (style == Qt::DashDotDotLine)
    return "DashDotDotLine";
  else
    return "SolidLine";
}

Qt::PenStyle Graph::getPenStyle (int style)
{
  Qt::PenStyle linePen;
  switch (style)
    {
    case 0:
      linePen = Qt::SolidLine;
      break;
    case 1:
      linePen = Qt::DashLine;
      break;
    case 2:
      linePen = Qt::DotLine;
      break;
    case 3:
      linePen = Qt::DashDotLine;
      break;
    case 4:
      linePen = Qt::DashDotDotLine;
      break;
    }
  return linePen;
}

Qt::PenStyle Graph::getPenStyle (const QString & s)
{
  Qt::PenStyle style;
  if (s == "SolidLine")
    style = Qt::SolidLine;
  else if (s == "DashLine")
    style = Qt::DashLine;
  else if (s == "DotLine")
    style = Qt::DotLine;
  else if (s == "DashDotLine")
    style = Qt::DashDotLine;
  else if (s == "DashDotDotLine")
    style = Qt::DashDotDotLine;
  return style;
}

int
Graph::obsoleteSymbolStyle (int type)
{
  if (type <= 4)
    return type + 1;
  else
    return type + 2;
}

int
Graph::curveType (int index)
{
  if(index >= 0 && index < n_curves)
    return c_attr.at(index)->c_type;
  else
    return -1;
}

void
Graph::showPlotErrorMessage (QWidget * parent,
			     const QStringList & emptyColumns)
{
  QApplication::restoreOverrideCursor ();

  int n = (int) emptyColumns.count ();
  if (n > 1)
    {
      QString columns;
      for (int i = 0; i < n; i++)
	columns += "<p><b>" + emptyColumns[i] + "</b></p>";

      QMessageBox::warning (parent, tr ("QtiPlot - Warning"),
			    tr ("The columns") + ": " + columns +
			    tr
			    ("are empty and will not be added to the plot!"));
    }
  else if (n == 1)
    QMessageBox::warning (parent, tr ("QtiPlot - Warning"),
			  tr ("The column") + ": <p><b>" + emptyColumns[0] +
			  "</b></p>" +
			  tr ("is empty and will not be added to the plot!"));
}

void
Graph::moveMarkerBy (int dx, int dy)
{
  bool line = false, image = false;
  if (d_lines.contains (selectedMarker))
    {
      LineMarker *mrk = (LineMarker *) d_plot->marker (selectedMarker);

      QPoint point = mrk->startPoint ();
      mrk->setStartPoint (QPoint (point.x () + dx, point.y () + dy));

      point = mrk->endPoint ();
      mrk->setEndPoint (QPoint (point.x () + dx, point.y () + dy));

      line = true;
    }
  else if (d_images.contains (selectedMarker))
    {
      ImageMarker *mrk = (ImageMarker *) d_plot->marker (selectedMarker);
      QPoint point = mrk->getOrigin ();
      mrk->setOrigin (QPoint (point.x () + dx, point.y () + dy));

      image = true;
    }
  else if (d_texts.contains (selectedMarker))
    {
      LegendMarker *mrk = (LegendMarker *) d_plot->marker (selectedMarker);
      QPoint point = mrk->rect ().topLeft ();
      mrk->setOrigin (QPoint (point.x () + dx, point.y () + dy));
    }

  d_plot->replot ();

  if (line)
    highlightLineMarker (selectedMarker);
  else if (image)
    highlightImageMarker (selectedMarker);
  else
    highlightTextMarker (selectedMarker);

  emit modifiedGraph ();
}

void
Graph::showTitleContextMenu ()
{
  QPopupMenu titleMenu (this);
  titleMenu.insertItem (QPixmap (cut_xpm), tr ("&Cut"), this,
			SLOT (cutTitle ()));
  titleMenu.insertItem (QPixmap (copy_xpm), tr ("&Copy"), this,
			SLOT (copyTitle ()));
  titleMenu.insertItem (tr ("&Delete"), this, SLOT (removeTitle ()));
  titleMenu.insertSeparator ();
  titleMenu.insertItem (tr ("&Properties..."), this,
			SIGNAL (viewTitleDialog ()));
  titleMenu.exec (QCursor::pos ());
}

void
Graph::cutTitle ()
{
  QApplication::clipboard ()->
    setData (new
	     QTextDrag (d_plot->title ().text (), d_plot->titleLabel (), 0));
  removeTitle ();
}

void
Graph::copyTitle ()
{
  QApplication::clipboard ()->
    setData (new
	     QTextDrag (d_plot->title ().text (), d_plot->titleLabel (), 0));
}

void
Graph::removeAxisTitle ()
{
  int axis = (selectedAxis + 2) % 4;	//unconsistent notation in Qwt enumerations between 
  //QwtScaleDraw::alignement and QwtPlot::Axis
  d_plot->setAxisTitle (axis, QString::null);
  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::cutAxisTitle ()
{
  copyAxisTitle ();
  removeAxisTitle ();
}

void
Graph::copyAxisTitle ()
{

  int axis = (selectedAxis + 2) % 4;	//unconsistent notation in Qwt enumerations between 
  //QwtScaleDraw::alignement and QwtPlot::Axis
  QApplication::clipboard ()->
    setData (new
	     QTextDrag (d_plot->axisTitle (axis).text (),
			(QWidget *) d_plot->axisWidget (axis), 0));
}

void
Graph::showAxisTitleMenu (int axis)
{
  selectedAxis = axis;

  QPopupMenu titleMenu (this);
  titleMenu.insertItem (QPixmap (cut_xpm), tr ("&Cut"), this,
			SLOT (cutAxisTitle ()));
  titleMenu.insertItem (QPixmap (copy_xpm), tr ("&Copy"), this,
			SLOT (copyAxisTitle ()));
  titleMenu.insertItem (tr ("&Delete"), this, SLOT (removeAxisTitle ()));
  titleMenu.insertSeparator ();
  switch (axis)
    {
    case QwtScaleDraw::BottomScale:
      titleMenu.insertItem (tr ("&Properties..."), this,
			    SIGNAL (xAxisTitleDblClicked ()));
      break;

    case QwtScaleDraw::LeftScale:
      titleMenu.insertItem (tr ("&Properties..."), this,
			    SIGNAL (yAxisTitleDblClicked ()));
      break;

    case QwtScaleDraw::TopScale:
      titleMenu.insertItem (tr ("&Properties..."), this,
			    SIGNAL (topAxisTitleDblClicked ()));
      break;

    case QwtScaleDraw::RightScale:
      titleMenu.insertItem (tr ("&Properties..."), this,
			    SIGNAL (rightAxisTitleDblClicked ()));
      break;
    }

  titleMenu.exec (QCursor::pos ());
}

void
Graph::showAxisContextMenu (int axis)
{
  selectedAxis = axis;

  QPopupMenu menu (this);
  menu.setCheckable (true);
  menu.insertItem (QPixmap (unzoom_xpm), tr ("&Rescale to show all"), this,
		   SLOT (setAutoScale ()), tr ("Ctrl+Shift+R"));
  menu.insertSeparator ();
  menu.insertItem (tr ("&Hide axis"), this, SLOT (hideSelectedAxis ()));

  int gridsID =
    menu.insertItem (tr ("&Show grids"), this, SLOT (showGrids ()));
  if (axis == QwtScaleDraw::LeftScale || axis == QwtScaleDraw::RightScale)
    {
      if (grid.majorOnY)
	menu.setItemChecked (gridsID, true);
    }
  else
    {
      if (grid.majorOnX)
	menu.setItemChecked (gridsID, true);
    }

  menu.insertSeparator ();
  menu.insertItem (tr ("&Scale..."), this, SLOT (showScaleDialog ()));
  menu.insertItem (tr ("&Properties..."), this, SLOT (showAxisDialog ()));
  menu.exec (QCursor::pos ());
}

void
Graph::showAxisDialog ()
{
  emit showAxisDialog (selectedAxis);
}

void
Graph::showScaleDialog ()
{
  emit axisDblClicked (selectedAxis);
}

void
Graph::hideSelectedAxis ()
{
  int axis = -1;
  if (selectedAxis == QwtScaleDraw::LeftScale
      || selectedAxis == QwtScaleDraw::RightScale)
    axis = selectedAxis - 2;
  else
    axis = selectedAxis + 2;

  d_plot->enableAxis (axis, false);
  scalePicker->refresh ();
  emit modifiedGraph ();
}

void
Graph::showGrids ()
{
  if (selectedAxis == QwtScaleDraw::LeftScale
      || selectedAxis == QwtScaleDraw::RightScale)
    {
      grid.majorOnY = 1 - grid.majorOnY;
      d_plot->grid ()->enableY (grid.majorOnY);

      grid.minorOnY = 1 - grid.minorOnY;
      d_plot->grid ()->enableYMin (grid.minorOnY);
    }
  else
    {
      grid.majorOnX = 1 - grid.majorOnX;
      d_plot->grid ()->enableX (grid.majorOnX);

      grid.minorOnX = 1 - grid.minorOnX;
      d_plot->grid ()->enableXMin (grid.minorOnX);
    }
  d_plot->replot ();
  emit modifiedGraph ();
}

void
Graph::copy (Graph * g)
{
  int i;
  Plot *plot = g->plotWidget ();
  d_plot->setMargin (plot->margin ());
  setBackgroundColor (plot->paletteBackgroundColor ());
  setBorder (plot->lineWidth (), plot->frameColor ());
  setCanvasBackground (plot->canvasBackground ());

  enableAxes (g->enabledAxes ());
  setAxesColors (g->axesColors ());
  setAxesNumColors (g->axesNumColors ());
  setAxesBaseline (g->axesBaseline ());

  setGridOptions (g->grid);

  d_plot->setTitle (plot->title ());
  setTitleFont( plot->title().font() );

  drawCanvasFrame (g->framed (), g->canvasFrameWidth (),
		   g->canvasFrameColor ());

  QStringList lst = g->scalesTitles ();
  for (i = 0; i < (int) lst.count (); i++)
    setAxisTitle (i, lst[i]);

  for (i = 0; i < 4; i++)
    setAxisFont (i, g->axisFont (i));

  setXAxisTitleColor (g->axisTitleColor (2));
  setXAxisTitleFont (g->axisTitleFont (2));
  setXAxisTitleAlignment (g->axisTitleAlignment (2));

  setYAxisTitleColor (g->axisTitleColor (0));
  setYAxisTitleFont (g->axisTitleFont (0));
  setYAxisTitleAlignment (g->axisTitleAlignment (0));

  setTopAxisTitleColor (g->axisTitleColor (3));
  setTopAxisTitleFont (g->axisTitleFont (3));
  setTopAxisTitleAlignment (g->axisTitleAlignment (3));

  setRightAxisTitleColor (g->axisTitleColor (1));
  setRightAxisTitleFont (g->axisTitleFont (1));
  setRightAxisTitleAlignment (g->axisTitleAlignment (1));

  setAxesLinewidth (plot->axesLinewidth ());
  removeLegend ();

  associations = g->associations;
  if (g->isPiePlot ())
    plotPie ((QwtPieCurve *) g->curve (0));
  else
    {
      for (i = 0; i < g->curves (); i++)
	{
	  QwtPlotItem *it = (QwtPlotItem *) g->curve (i);
	  if (it->rtti () == QwtPlotItem::Rtti_PlotCurve
	      || it->rtti () == FunctionCurve::RTTI)
	    {
	      const QwtPlotCurve *cv = (QwtPlotCurve *) it;
	      int n = cv->dataSize ();
	      int style = g->curveType(i);
	      QMemArray <double> x(n), y(n);
	      for (int j = 0; j < n; j++)
		{
		  x[j] = cv->x (j);
		  y[j] = cv->y (j);
		}

	      QwtPlotCurve *c = 0;
	      if (cv->rtti () == FunctionCurve::RTTI)
		{
		  c = new FunctionCurve (cv->title ().text ());
		  ((FunctionCurve *) c)->copy ((FunctionCurve *) cv);
		}
	      else if (style == VerticalBars || style == HorizontalBars)
		{
		  c =
		    new QwtBarCurve (((QwtBarCurve *) cv)->orientation (),
				     cv->title ().text ());
		  ((QwtBarCurve *) c)->copy ((const QwtBarCurve *) cv);
		}
	      else if (style == XErrorBars || style == YErrorBars)
		{
		  c = new QwtErrorPlotCurve (cv->title ().text ());
		  ((QwtErrorPlotCurve *) c)->
		    copy ((const QwtErrorPlotCurve *) cv);
		}
	      else if (style == Histogram)
		{
		  c = new QwtHistogram (cv->title ().text ());
		  ((QwtHistogram *) c)->copy ((const QwtHistogram *) cv);
		}
	      else if (style == VectXYXY || style == VectXYAM)
		{
		  VectorCurve::VectorStyle vs = VectorCurve::XYXY;
		  if (style == VectXYAM)
		    vs = VectorCurve::XYAM;
		  c = new VectorCurve (vs, cv->title ().text ());
		  ((VectorCurve *) c)->copy ((const VectorCurve *) cv);
		}
	      else if (style == Box)
		{
		  c = new BoxCurve (cv->title ().text ());
		  ((BoxCurve *) c)->copy ((const BoxCurve *) cv);
		  QwtSingleArrayData dat (x[0], y, n);
		  c->setData (dat);
		}
	      else
		c = new QwtPlotCurve (cv->title ());

	      n_curves++;
	      CurveAttr *new_attr = new CurveAttr( g->c_attr.at(i) );
	      new_attr->c_key = d_plot->insertCurve(c);
	      c_attr.append( new_attr );

	      if (new_attr->c_type != Box)
		c->setData (x.data(), y.data(), n);

	      c->setPen (cv->pen ());
	      c->setBrush (cv->brush ());
	      c->setStyle (cv->style ());
	      c->setSymbol (cv->symbol ());

	      if (cv->testCurveAttribute (QwtPlotCurve::Fitted))
		c->setCurveAttribute (QwtPlotCurve::Fitted, true);
	      else if (cv->testCurveAttribute (QwtPlotCurve::Inverted))
		c->setCurveAttribute (QwtPlotCurve::Inverted, true);

	      c->setAxis (cv->xAxis (), cv->yAxis ());
	    }
	  else if (it->rtti () == QwtPlotItem::Rtti_PlotSpectrogram)
	    {
	      Spectrogram *sp = ((Spectrogram *) it)->copy ();
	      n_curves++;
	      CurveAttr *new_attr = new CurveAttr( g->c_attr.at(i) );
	      new_attr->c_key = d_plot->insertCurve(sp);
	      c_attr.append( new_attr );

	      sp->showColorScale (((Spectrogram *) it)->colorScaleAxis (),
				  ((Spectrogram *) it)->hasColorScale ());
	      sp->setColorBarWidth (((Spectrogram *) it)->colorBarWidth ());
	    }
	}
    }
  axesFormulas = g->axesFormulas;
  axisType = g->axisType;
  axesFormatInfo = g->axesFormatInfo;
  axisType = g->axisType;

  for (i = 0; i < QwtPlot::axisCnt; i++)
    {
      QwtScaleWidget *sc = plot->axisWidget (i);
      if (!sc)
	continue;

      QwtScaleDraw *sd = plot->axisScaleDraw (i);
      if (sd->hasComponent (QwtAbstractScaleDraw::Labels))
	{
	  if (axisType[i] == Graph::Numeric)
	    setLabelsNumericFormat (i, plot->axisLabelFormat (i),
				    plot->axisLabelPrecision (i),
				    axesFormulas[i]);
	  else if (axisType[i] == Graph::Day)
	    setLabelsDayFormat (i, axesFormatInfo[i].toInt ());
	  else if (axisType[i] == Graph::Month)
	    setLabelsMonthFormat (i, axesFormatInfo[i].toInt ());
	  else if (axisType[i] == Graph::Time || axisType[i] == Graph::Date)
	    setLabelsDateTimeFormat (i, axisType[i], axesFormatInfo[i]);
	  else
	    {
	      QwtTextScaleDraw *sd =
		(QwtTextScaleDraw *) plot->axisScaleDraw (i);
	      d_plot->setAxisScaleDraw (i,
					new QwtTextScaleDraw (sd->
							      labelsList ()));
	    }
	}
      else
	{
	  sd = d_plot->axisScaleDraw (i);
	  sd->enableComponent (QwtAbstractScaleDraw::Labels, false);
	}
    }

  for (i = 0; i < QwtPlot::axisCnt; i++)
    {				//set same scales
      const QwtScaleEngine *se = plot->axisScaleEngine (i);
      if (!se)
	continue;

      QwtScaleEngine *sc_engine = 0;
      if (se->transformation ()->type () == QwtScaleTransformation::Log10)
	sc_engine = new QwtLog10ScaleEngine ();
      else if (se->transformation ()->type () ==
	       QwtScaleTransformation::Linear)
	sc_engine = new QwtLinearScaleEngine ();

      int majorTicks = plot->axisMaxMajor (i);
      int minorTicks = plot->axisMaxMinor (i);
      d_plot->setAxisMaxMajor (i, majorTicks);
      d_plot->setAxisMaxMinor (i, minorTicks);

      const QwtScaleDiv *sd = plot->axisScaleDiv (i);
      QwtValueList lst = sd->ticks (QwtScaleDiv::MajorTick);

      d_user_step[i] = g->axisStep (i);

      QwtScaleDiv div =
	sc_engine->divideScale (QMIN (sd->lBound (), sd->hBound ()),
				QMAX (sd->lBound (), sd->hBound ()),
				majorTicks, minorTicks, d_user_step[i]);

      if (se->testAttribute (QwtScaleEngine::Inverted))
	{
	  sc_engine->setAttribute (QwtScaleEngine::Inverted);
	  div.invert ();
	}

      d_plot->setAxisScaleEngine (i, sc_engine);
      d_plot->setAxisScaleDiv (i, div);
    }

  drawAxesBackbones (g->drawAxesBackbone);
  setMajorTicksType (g->plotWidget ()->getMajorTicksType ());
  setMinorTicksType (g->plotWidget ()->getMinorTicksType ());
  setTicksLength (g->minorTickLength (), g->majorTickLength ());

  setAxisLabelRotation (QwtPlot::xBottom,
			g->labelsRotation (QwtPlot::xBottom));
  setAxisLabelRotation (QwtPlot::xTop, g->labelsRotation (QwtPlot::xTop));

  QwtArray < int >imag = g->imageMarkerKeys ();
  for (i = 0; i < (int) imag.size (); i++)
    {
      ImageMarker *imrk = (ImageMarker *) g->imageMarker (imag[i]);
      if (imrk)
	insertImageMarker (imrk);
    }

  QwtArray < int >txtMrkKeys = g->textMarkerKeys ();
  for (i = 0; i < (int) txtMrkKeys.size (); i++)
    {
      LegendMarker *mrk = (LegendMarker *) g->textMarker (txtMrkKeys[i]);
      if (!mrk)
	continue;

      if (txtMrkKeys[i] == g->legendMarkerID)
	legendMarkerID = insertTextMarker (mrk);
      else
	insertTextMarker (mrk);
    }

  QwtArray < int >l = g->lineMarkerKeys ();
  for (i = 0; i < (int) l.size (); i++)
    {
      LineMarker *lmrk = (LineMarker *) g->lineMarker (l[i]);
      if (lmrk)
	insertLineMarker (lmrk);
    }
  d_plot->replot ();
}

void
Graph::plotBoxDiagram (Table * w, const QStringList & names)
{
  for (int j = 0; j < (int) names.count (); j++)
    {
      const QString name = names[j];
      associations << name;
      int ycol = w->colIndex (name);
      int i, size = 0;
      for (i = 0; i < w->tableRows (); i++)
	{
	  if (!w->text (i, ycol).isEmpty ())
	    size++;
	}
      if (size > 0)
	{
	  QMemArray <double> y(size);
	  int it = 0;
	  for (i = 0; i < w->tableRows (); i++)
	    {
	      QString s = w->text (i, ycol);
	      if (!s.isEmpty ())
		y[it++] = s.toDouble ();
	    }
	  gsl_sort (y.data(), 1, size);

	  QwtSingleArrayData data (double (j + 1), y.data(), size);
	  BoxCurve *c = new BoxCurve (name);
	  c->setData (data);

	  n_curves++;
	  c_attr.append( new CurveAttr( d_plot->insertCurve(c), Box ) );

	  c->setPen(QPen (ColorBox::color (j),1,Qt::SolidLine));
	  c->setSymbol(QwtSymbol(QwtSymbol::NoSymbol, QBrush (),
				 QPen(ColorBox::color (j), 1), 
				 QSize (7, 7)));
	}
    }

  LegendMarker *mrk = (LegendMarker *) d_plot->marker (legendMarkerID);
  if (mrk)
    mrk->setText (legendText ());

  axisType[QwtPlot::xBottom] = ColHeader;
  d_plot->setAxisScaleDraw (QwtPlot::xBottom,
			    new QwtTextScaleDraw (w->selectedYLabels ()));
  d_plot->setAxisMaxMajor (QwtPlot::xBottom, names.count () + 1);
  d_plot->setAxisMaxMinor (QwtPlot::xBottom, 0);

  axisType[QwtPlot::xTop] = ColHeader;
  d_plot->setAxisScaleDraw (QwtPlot::xTop,
			    new QwtTextScaleDraw (w->selectedYLabels ()));
  d_plot->setAxisMaxMajor (QwtPlot::xTop, names.count () + 1);
  d_plot->setAxisMaxMinor (QwtPlot::xTop, 0);

  axesFormatInfo[QwtPlot::xBottom] = w->getName();
  axesFormatInfo[QwtPlot::xTop] = w->getName();
}

void
Graph::updateBoxData (Table * w, const QString & yColName, int curve)
{
  int ycol = w->colIndex (yColName);
  int i, size = 0;
  for (i = 0; i < w->tableRows (); i++)
    {
      if (!w->text (i, ycol).isEmpty ())
	size++;
    }

  if (size > 0)
    {
      QMemArray <double> y(size);
      int it = 0;
      for (i = 0; i < w->tableRows (); i++)
	{
	  QString s = w->text (i, ycol);
	  if (!s.isEmpty ())
	    y[it++] = s.toDouble ();
	}
      gsl_sort (y.data(), 1, size);
      BoxCurve *c = (BoxCurve *) this->curve (curve);
      if (c)
	{
	  QwtSingleArrayData dat (c->x (0), y.data(), size);
	  c->setData (dat);
	}
    }
  else
    removeCurve (curve,true);
  d_plot->replot ();
}

void
Graph::setCurveStyle (int index, int s)
{
  QwtPlotCurve *c = curve (index);
  if (!c)
    return;

  if (s == 5)			//ancient spline style in Qwt 4.2.0
    {
      s = QwtPlotCurve::Lines;
      c->setCurveAttribute (QwtPlotCurve::Fitted, true);
      setCurveType(index, Spline);
    }
  else if (s == QwtPlotCurve::Lines)
    c->setCurveAttribute (QwtPlotCurve::Fitted, false);
  else if (s == 6)		// Vertical Steps
    {
      s = QwtPlotCurve::Steps;
      c->setCurveAttribute (QwtPlotCurve::Inverted, true);
      setCurveType(index, VerticalSteps);
    }
  else if (s == QwtPlotCurve::Steps)	// Horizontal Steps
    {
      c->setCurveAttribute (QwtPlotCurve::Inverted, false);
      setCurveType(index, HorizontalSteps);
    }
  c->setStyle ((QwtPlotCurve::CurveStyle) s);
}

void
Graph::setCurveSymbol (int index, const QwtSymbol & s)
{
  QwtPlotCurve *c = curve (index);
  if (!c || c->symbol () == s)
    return;

  c->setSymbol (s);
  double factor = scaleFactor();
  c_attr.at(index)->symbol_size = s.size().width() / factor;
  c_attr.at(index)->symbol_edge_width = s.pen().width() / factor;

  QString yColName = c->title ().text ();
  for (int i = 0; i < n_curves; i++)
    {
      QStringList lst = QStringList::split(',',associations[i],false);
      int ct = curveType(i);
      if (lst.contains (yColName) && i != index
	  && (ct == XErrorBars || ct == YErrorBars))
	{
	  QwtErrorPlotCurve *er = (QwtErrorPlotCurve *) this->curve (i);
	  if (er)
	    {
	      er->setSymbolSize (s.size ());
	    }
	}
    }
}

void
Graph::setCurvePen (int index, const QPen & p)
{
  QwtPlotCurve *c = curve (index);
  if (!c || c->pen () == p)
    return;

  c_attr.at(index)->line_width = p.width() / scaleFactor();

  c->setPen (p);
}

void
Graph::setCurveBrush (int index, const QBrush & b)
{
  QwtPlotCurve *c = curve (index);
  if (!c || c->brush () == b)
    return;

  c->setBrush (b);
}

void
Graph::openBoxDiagram (Table * w, const QStringList & l)
{
  associations << l[2];
  int ycol = w->colIndex (l[2]);
  int i, size = 0;
  for (i = 0; i < w->tableRows (); i++)
    {
      if (!w->text (i, ycol).isEmpty ())
	size++;
    }
  if (!size)
    return;

  QMemArray <double> y(size);
  int it = 0;
  for (i = 0; i < w->tableRows (); i++)
    {
      QString s = w->text (i, ycol);
      if (!s.isEmpty ())
	y[it++] = s.toDouble ();
    }

  gsl_sort (y.data(), 1, size);	//the data must be sorted first!

  BoxCurve *c = new BoxCurve (l[2]);
  QwtSingleArrayData dat (l[1].toDouble (), y.data(), size);
  c->setData (dat);

  n_curves++;
  c_attr.append( new CurveAttr( d_plot->insertCurve(c), Box ) );

  c->setMaxStyle (SymbolBox::style (l[16].toInt ()));
  c->setP99Style (SymbolBox::style (l[17].toInt ()));
  c->setMeanStyle (SymbolBox::style (l[18].toInt ()));
  c->setP1Style (SymbolBox::style (l[19].toInt ()));
  c->setMinStyle (SymbolBox::style (l[20].toInt ()));

  c->setBoxStyle (l[21].toInt ());
  c->setBoxWidth (l[22].toInt ());
  c->setBoxRange (l[23].toInt (), l[24].toDouble ());
  c->setWhiskersRange (l[25].toInt (), l[26].toDouble ());
}

QString
Graph::curveXColName (const QString & curveTitle)
{
  int index = curveIndex( curveTitle );
  if (index < 0)
    return QString::null;

  return associations[index].section(',',0,0);
}

void
Graph::disableTools ()
{
  if (selectorsEnabled ())
    {
      disableRangeSelectors ();
      return;
    }
  else if (enabledCursor ())
    {
      enableCursor (false);
      replot ();
      return;
    }
  else if (pickerActivated ())
    {
      showPlotPicker (false);
      return;
    }
  else if (movePointsActivated ())
    {
      movePoints (false);
      return;
    }
  else if (removePointActivated ())
    {
      removePoints (false);
      return;
    }
  else if (zoomOn ())
    {
      zoom (false);
      return;
    }
  else if (drawLineActive ())
    {
      drawLine (false);
      return;
    }
  else if (subtractLineActive())
    {
      setSubtractLine(false);
      return;
    }
}

void
Graph::setTextMarkerDefaults (int f, const QFont & font,
			      const QColor & textCol,
			      const QColor & backgroundCol)
{
  defaultMarkerFont = font;
  defaultMarkerFrame = f;
  defaultTextMarkerColor = textCol;
  defaultTextMarkerBackground = backgroundCol;
}

void
Graph::setArrowDefaults (int lineWidth, const QColor & c, Qt::PenStyle style,
			 int headLength, int headAngle, bool fillHead)
{
  defaultArrowLineWidth = lineWidth;
  defaultArrowColor = c;
  defaultArrowLineStyle = style;
  defaultArrowHeadLength = headLength;
  defaultArrowHeadAngle = headAngle;
  defaultArrowHeadFill = fillHead;
}

const QString &
Graph::parentPlotName ()
{
  QWidget *w = (QWidget *) parent ()->parent ();
  return ((myWidget *)w)->getName();
}

void
Graph::guessUniqueCurveLayout (int &colorIndex, int &symbolIndex)
{
  colorIndex = 0;
  symbolIndex = 0;
  for (int i = 0; i < n_curves; i++)
    {
      const QwtPlotCurve *c = curve (i);
      if (c && c->rtti () == QwtPlotItem::Rtti_PlotCurve)
	{
	  int index = ColorBox::colorIndex (c->pen ().color ());
	  if (index > colorIndex) // find maximum color index
	    colorIndex = index;

	  QwtSymbol symb = c->symbol ();
	  index = SymbolBox::symbolIndex (symb.style ());
	  if (index > symbolIndex) // find maximum symbol index
	    symbolIndex = index;
	}
    }

  if(n_curves > 1)
    colorIndex = (++colorIndex) % 16;  // succeed index

  if (colorIndex == 13)		//avoid white invisible curves
    colorIndex = 0;

  symbolIndex = (++symbolIndex) % 15; // succeed index
  if (!symbolIndex)
    symbolIndex = 1;
}

void
Graph::deleteFitCurves ()
{
  QRegExp rx("^(.+Fit|Peak)\\d+$");
  bool modified = false;
  for (int i = 0; i < n_curves; i++)
    {
      QwtPlotCurve *c = curve(i);
      if (c && c->title().text().contains( rx ) )
	{
	  removeCurve(i, false);
	  modified = true;
	}
    }
  if(modified)
    {
      updatePlot();
      emit modifiedGraph ();
    }
}

void
Graph::plotSpectrogram (Matrix * m, int type)
{
  if (type != GrayMap && type != ColorMap && type != ContourMap)
    return;

  Spectrogram *d_spectrogram = new Spectrogram (m);
  if (type == GrayMap)
    d_spectrogram->setGrayScale ();
  else if (type == ContourMap)
    {
      d_spectrogram->setDisplayMode (QwtPlotSpectrogram::ImageMode, false);
      d_spectrogram->setDisplayMode (QwtPlotSpectrogram::ContourMode, true);
    }
  else if (type == ColorMap)
    {
      d_spectrogram->setDefaultColorMap ();
      d_spectrogram->setDisplayMode (QwtPlotSpectrogram::ContourMode, true);
    }

  n_curves++;
  c_attr.append( new CurveAttr( d_plot->insertCurve(d_spectrogram), type ) );

  associations << m->getName();

  QwtScaleWidget *rightAxis = d_plot->axisWidget (QwtPlot::yRight);
  rightAxis->setColorBarEnabled (type != ContourMap);
  rightAxis->setColorMap (d_spectrogram->data ().range (),
			  d_spectrogram->colorMap ());

  d_plot->setAxisScale (QwtPlot::xBottom, m->xStart(), m->xEnd());
  if(m->xStart() > m->xEnd())
    {
      QwtScaleEngine *se = d_plot->axisScaleEngine (QwtPlot::xBottom);
      se->setAttribute (QwtScaleEngine::Inverted);
    }
  d_plot->setAxisScale (QwtPlot::yLeft, m->yStart(), m->yEnd());
  if(m->yStart() > m->yEnd())
    {
      QwtScaleEngine *se = d_plot->axisScaleEngine (QwtPlot::yLeft);
      se->setAttribute (QwtScaleEngine::Inverted);
    }

  d_plot->setAxisScale (QwtPlot::yRight,
			d_spectrogram->data ().range ().minValue (),
			d_spectrogram->data ().range ().maxValue ());
  d_plot->enableAxis (QwtPlot::yRight, type != ContourMap);

  d_plot->replot ();
}

void
Graph::restoreSpectrogram (ApplicationWindow * app, const QStringList & lst)
{
  QStringList::const_iterator line = lst.begin ();
  QString s = (*line).stripWhiteSpace ();
  QString matrixName = s.remove ("<matrix>").remove ("</matrix>");
  Matrix *m = app->matrix (matrixName);
  if (!m)
    return;

  Spectrogram *sp = new Spectrogram (m);
  insertPlotItem (sp, matrixName, Graph::ColorMap);

  for (line++; line != lst.end (); line++)
    {
      QString s = *line;
      if (s.contains ("<ColorPolicy>"))
	{
	  int color_policy =
	    s.remove ("<ColorPolicy>").remove ("</ColorPolicy>").
	    stripWhiteSpace ().toInt ();
	  if (color_policy == Spectrogram::GrayScale)
	    sp->setGrayScale ();
	  else if (color_policy == Spectrogram::Default)
	    sp->setDefaultColorMap ();
	}
      else if (s.contains ("<ColorMap>"))
	{
	  s = *(++line);
	  int mode =
	    s.remove ("<Mode>").remove ("</Mode>").stripWhiteSpace ().
	    toInt ();
	  s = *(++line);
	  QColor color1 =
	    QColor (s.remove ("<MinColor>").remove ("</MinColor>").
		    stripWhiteSpace ());
	  s = *(++line);
	  QColor color2 =
	    QColor (s.remove ("<MaxColor>").remove ("</MaxColor>").
		    stripWhiteSpace ());

	  QwtLinearColorMap colorMap = QwtLinearColorMap (color1, color2);
	  colorMap.setMode ((QwtLinearColorMap::Mode) mode);

	  s = *(++line);
	  int stops =
	    s.remove ("<ColorStops>").remove ("</ColorStops>").
	    stripWhiteSpace ().toInt ();
	  for (int i = 0; i < stops; i++)
	    {
	      s = (*(++line)).stripWhiteSpace ();
	      QStringList l =
		QStringList::split ('\t',
				    s.remove ("<Stop>").remove ("</Stop>"));
	      colorMap.addColorStop (l[0].toDouble (), QColor (l[1]));
	    }
	  sp->setCustomColorMap (colorMap);
	  line++;
	}
      else if (s.contains ("<Image>"))
	{
	  int mode =
	    s.remove ("<Image>").remove ("</Image>").stripWhiteSpace ().
	    toInt ();
	  sp->setDisplayMode (QwtPlotSpectrogram::ImageMode, mode);
	}
      else if (s.contains ("<ContourLines>"))
	{
	  int contours =
	    s.remove ("<ContourLines>").remove ("</ContourLines>").
	    stripWhiteSpace ().toInt ();
	  sp->setDisplayMode (QwtPlotSpectrogram::ContourMode, contours);
	  if (contours)
	    {
	      s = (*(++line)).stripWhiteSpace ();
	      int levels =
		s.remove ("<Levels>").remove ("</Levels>").toInt ();
	      sp->setLevelsNumber (levels);

	      s = (*(++line)).stripWhiteSpace ();
	      int defaultPen =
		s.remove ("<DefaultPen>").remove ("</DefaultPen>").toInt ();
	      if (!defaultPen)
		sp->setDefaultContourPen (Qt::NoPen);
	      else
		{
		  s = (*(++line)).stripWhiteSpace ();
		  QColor c =
		    QColor (s.remove ("<PenColor>").remove ("</PenColor>"));
		  s = (*(++line)).stripWhiteSpace ();
		  int width =
		    s.remove ("<PenWidth>").remove ("</PenWidth>").toInt ();
		  s = (*(++line)).stripWhiteSpace ();
		  int style =
		    s.remove ("<PenStyle>").remove ("</PenStyle>").toInt ();
		  sp->
		    setDefaultContourPen (QPen(c,
					       width,
					       Graph::getPenStyle (style)));
		}
	    }
	}
      else if (s.contains ("<ColorBar>"))
	{
	  s = *(++line);
	  int color_axis =
	    s.remove ("<axis>").remove ("</axis>").stripWhiteSpace ().
	    toInt ();
	  s = *(++line);
	  int width =
	    s.remove ("<width>").remove ("</width>").stripWhiteSpace ().
	    toInt ();

	  QwtScaleWidget *colorAxis = d_plot->axisWidget (color_axis);
	  if (colorAxis)
	    {
	      colorAxis->setColorBarWidth (width);
	      colorAxis->setColorBarEnabled (true);
	    }
	  line++;
	}
    }
}

QString
Graph::generateUniqueName( const QString & basename )
{
  QStringList curves = plotItemsList ();
  int index = 1;
  QString name = basename + QString::number(index);
  while( curves.contains( name ) )
    {
      index++;
      name = basename + QString::number(index);
    }
  return name;
}


Graph::~Graph ()
{
  delete titlePicker;
  delete scalePicker;
  delete cp;
  delete d_plot;
  c_attr.clear();
}


void
Graph::swapCurves(int c1, int c2)
{
  CurveAttr *dummy1 = c_attr.at(c1);
  CurveAttr *dummy2 = c_attr.at(c2);
  c_attr.setAutoDelete(false);
  c_attr.replace( c1, dummy2 );
  c_attr.replace( c2, dummy1 );
  c_attr.setAutoDelete(true);
  SWAP(QString, associations[c1], associations[c2]);
}

void
Graph::resetColors(bool black)
{
  int n = 0;
  for (int i = 0; i < curves(); i++)
    {
      QwtPlotCurve *c = curve(i);
      int ct = curveType(i);
      if (c && c->rtti () != QwtPlotItem::Rtti_PlotSpectrogram
	  && ct != XErrorBars && ct != YErrorBars)
	{
	  QColor color;
	  if(black)
	    color = ColorBox::color(0);
	  else
	    color = ColorBox::color(n);

	  QPen pen = c->pen();
	  pen.setColor(color);
	  c->setPen(pen);
	  QwtSymbol symbol = c->symbol();
	  QPen spen = symbol.pen();
	  spen.setColor(color);
	  symbol.setPen(spen);
	  QBrush br = symbol.brush();
	  br.setColor(color);
	  symbol.setBrush(br);
	  c->setSymbol(symbol);
      	  n++;
	}
    }
  d_plot->replot ();
}

void
Graph::resetSymbols()
{
  int n = 0;
  for (int i = 0; i < curves(); i++)
    {
      QwtPlotCurve *c = curve(i);
      int ct = curveType(i);
      if (c && c->rtti () != QwtPlotItem::Rtti_PlotSpectrogram
	  && (ct == Scatter || ct == LineSymbols))
	{
	  QwtSymbol symbol = c->symbol();
	  symbol.setStyle(SymbolBox::style(n+1));
	  c->setSymbol(symbol);
	  n++;
	}
    }
  d_plot->replot ();
}

void
Graph::setFonts(const QFont & titleFont, const QFont & axesFont, 
		const QFont & numbersFont, const QFont & legendFont)
{
  //qDebug("setFonts");
  setTitleFont( titleFont );
  initFonts( axesFont, numbersFont );
  for (int i = 0; i < (int) d_texts.size (); i++)
    {
      LegendMarker *mrk = (LegendMarker *) d_plot->marker (d_texts[i]);
      mrk->setFont( legendFont, this->frameGeometry() );
    }
  d_plot->replot();
}

void
Graph::updateTableName(const QString &oldName, const QString &newName)
{
  //update plotted curves list and associations
  //if the curve title contains the table name, so does the association
  bool contain = false;
  for (int i = 0; i < n_curves; i++)
  {
    QwtPlotItem *c = d_plot->curve ( curveKey(i) );
    if (c){
      QString title = c->title ().text ();
      if(title.startsWith(oldName+"_"))
      {
        title.replace(oldName,newName,true);
        c->setTitle( title );
	associations[i].replace(oldName,newName,true);
	contain = true;
	//qDebug("curve title and association is changed from %s to %s",
	//       oldName.ascii(),newName.ascii());
      }
    }
  }

  //update legend
  if (contain && legendMarkerID >= 0)
  {
    LegendMarker *legendMrk = (LegendMarker *) d_plot->marker (legendMarkerID);
    QString text = legendMrk->getText();
    text.replace("}"+oldName+"_","}"+newName+"_",true);
    legendMrk->setText(text);
    d_plot->replot();
  }
}

void
Graph::updateColName(const QString &oldName, const QString &newName )
{
  //qDebug("updateColName %s -> %s",oldName.ascii(),newName.ascii());

  //update plotted curves list
  bool contain = false;
  for (int i = 0; i < n_curves; i++)
  {
    QwtPlotItem *c = d_plot->curve ( curveKey(i) );
    if (c && c->title ().text () == oldName)
      {
	c->setTitle( newName );
	contain = true;
      }
  }
  
  //update plot associations
  //
  // associations.gres(QRegExp("\\b"+oldName+"\\b"),newName); 
  // does not work because \b matches '-' 
  //
  QString tableName = oldName.section('_',0,0,QString::SectionIncludeTrailingSep);
  //qDebug("tableName = %s",tableName.ascii());
  QRegExp rex("^"+oldName+"$");
  for (int k = 0; k < (int) associations.count (); k++)
    if(associations[k].startsWith(tableName))
      {
	QStringList cols = QStringList::split (',', associations[k], false);
	cols.gres(rex,newName);
	associations[k] = cols.join(",");
      }

  //update legend
  //be cafeful that column names may include '-'
  QRegExp rex2("\\}"+oldName+"$");
  if (contain && legendMarkerID >= 0)
  {
    LegendMarker *legendMrk = (LegendMarker *) d_plot->marker (legendMarkerID);
    QStringList lst = QStringList::split('\n',legendMrk->getText(),true);
    lst.gres(rex2,"}"+newName);
    legendMrk->setText(lst.join("\n"));
    d_plot->replot();
  }
}

void
Graph::saveComponentSizes()
{
  int i;
  sizeStore.clear();
  //Fonts
  for (i = 0; i < (int) d_texts.size (); i++)
    {
      //MarkerFonts
      LegendMarker *mrk = (LegendMarker *) d_plot->marker(d_texts[i]);
      sizeStore.append( mrk->getFont().pointSize() );
    }
  for (i = 0; i < QwtPlot::axisCnt; i++)
    {
      //AxisFonts
      sizeStore.append( axisFont(i).pointSize() );
      sizeStore.append( axisTitleFont(i).pointSize() );
    }
  //TitleFont
  sizeStore.append( d_plot->title().font().pointSize() );
  //FrameWidth
  sizeStore.append( canvasFrameWidth() );
  sizeStore.append( d_plot->axesLinewidth() );
  //GridWidth
  sizeStore.append( grid.majorWidth );
  sizeStore.append( grid.minorWidth );
  //LineWidth and symbol sizes
  for (int i = 0; i < n_curves; i++)
    {
      QwtPlotCurve *c = curve (i);
      int ct = curveType(i);
      if(ct==Graph::Line || ct==Graph::LineSymbols)
	{
	  //LineWidth
	  QPen pen = c->pen();
	  sizeStore.append( pen.width() );
	}
      if(ct==Graph::Scatter || ct==Graph::LineSymbols)
	{
	  //SymbolEdgeWidth
	  QwtSymbol s = c->symbol();
	  sizeStore.append( s.pen().width() );
	  //SymbolSize
	  sizeStore.append( s.size().width() );
	}
    }
  //TickLength does not need saving
  //Spacing does not need saving
  //qDebug( "buffer size = %d", sizeStore.size() );
}

void
Graph::restoreComponentSizes()
{
  QFont font;
  int i, n = 0;

  //MarkerFonts
  for (i = 0; i < (int) d_texts.size (); i++)
    {
      LegendMarker *mrk = (LegendMarker *) d_plot->marker(d_texts[i]);
      font = mrk->getFont();
      font.setPointSize( sizeStore[n++] );
      mrk->setFont( font, frameGeometry() );
    }
  //AxisFonts
  for (i = 0; i < QwtPlot::axisCnt; i++)
    {
      //AxisFonts
      font = axisFont(i);
      font.setPointSize( sizeStore[n++] );
      d_plot->setAxisFont(i,font);

      //AxisTitleFonts
      font = axisTitleFont (i);
      font.setPointSize( sizeStore[n++] );
      QwtText t = d_plot->axisTitle (i);
      t.setFont (font);
      d_plot->setAxisTitle (i, t);
    }

  //TitleFont
  font = d_plot->title ().font ();
  font.setPointSize( sizeStore[n++] );
  QwtText t = d_plot->title ();
  t.setFont (font);
  d_plot->setTitle (t);

  //FrameWidth
  int width = sizeStore[n++];
  QwtPlotCanvas *canvas = (QwtPlotCanvas *) d_plot->canvas ();
  if (canvas->lineWidth()>0) canvas->setLineWidth( width );
  d_plot->setAxesLinewidth( sizeStore[n++] );

  //GridWidth
  int maj = sizeStore[n++];
  int min = sizeStore[n++];
  QPen majPen = QPen (ColorBox::color(grid.majorCol), maj,
                      getPenStyle(grid.majorStyle));
  grid.majorWidth = maj;
  d_plot->grid ()->setMajPen (majPen);
  QPen minPen = QPen (ColorBox::color(grid.minorCol), min,
                      getPenStyle(grid.minorStyle));
  d_plot->grid ()->setMinPen (minPen);
  grid.minorWidth = min;

  //LineWidth and symbol sizes
  for (int i = 0; i < n_curves; i++)
    {
      QwtPlotCurve *c = curve (i);
      int ct = curveType(i);
      if(ct==Graph::Line || ct==Graph::LineSymbols)
	{
	  //LineWidth
	  QPen pen = c->pen(); 
	  pen.setWidth( sizeStore[n++] );
	  c->setPen(pen);
	}
      if(ct==Graph::Scatter || ct==Graph::LineSymbols)
	{
	  //SymbolEdgeWidth
	  QwtSymbol s = c->symbol();
	  QPen pen = s.pen();
	  pen.setWidth( sizeStore[n++] );
	  //SymbolSize
	  int size = sizeStore[n++];
	  QwtSymbol ss = QwtSymbol (s.style(), s.brush(), pen,
				    QSize (size, size));
	  c->setSymbol (ss);
	}
    }

  //TickLength does not need restoring
  //Spacing does not need restoring
  //qDebug("restored size = %d",n);
}

